@@ -1,6 +1,6 @@
 #! /bin/sh
 
-# $Id: xenofarm.sh,v 1.27 2005/03/18 19:11:01 grubba Exp $
+# $Id: xenofarm.sh,v 1.33 2008/07/27 09:32:40 grubba Exp $
 # This file scripts the xenofarm actions and creates a result package
 # to send back.
 
@@ -25,20 +25,27 @@ log_end() {
 
 xenofarm_build() {
   log_start compile
-  $MAKE "CONFIGUREARGS=$CONFIGUREARGS" >xenofarm_result/compilelog.txt 2>&1
+  $MAKE > xenofarm_result/compilelog.txt 2>&1
   log_end $?
   [ $LASTERR = 0 ] || return 1
 }
 
 xenofarm_post_build() {
   POST_RESULT=0
+  log_start features
+  bin/pike -x features > xenofarm_result/features.txt 2>&1
+  log_end $?
+  [ $LASTERR = 0 ] || POST_RESULT=$LASTERR
+
   log_start benchmark
   $MAKE benchmark > xenofarm_result/benchmark.txt 2>&1
   log_end $?
-  POST_RESULT=$LASTERR
+  [ $LASTERR = 0 ] || POST_RESULT=$LASTERR
 
   log_start verify
-  $MAKE METATARGET=verify TESTARGS="-a -T -F" > \
+  # Note: verify and valgrind_verify perform the same actions
+  #       if not compiled --with-valgrind.
+  $MAKE METATARGET=valgrind_verify TESTARGS="-a -T -F" > \
     xenofarm_result/verifylog.txt 2>&1
   log_end $?
   [ $LASTERR = 0 ] || return 1
@@ -74,8 +81,7 @@ log_start response_assembly
   # Configuration
   (
     cd "$BUILDDIR"
-    test -f config.log && cat config.log
-    for f in `find modules post_modules -name config.log -type f`; do
+    for f in `find . -name config.log -type f | sort`; do
       echo
       echo '###################################################'
       echo '##' `dirname "$f"`
