@@ -1,7 +1,7 @@
 /*
  * Created by
- * $Id: make_ci.pike,v 1.11 2004/04/12 01:11:15 nilsson Exp $
- * on Mon Jan 29 00:56:52 2007
+ * $Id: make_ci.pike,v 1.12 2008/05/29 20:08:43 nilsson Exp $
+ * on Sun Aug  9 19:47:01 2009
  *
  * Table used for looking up the case of
  * Unicode characters.
@@ -31,7 +31,7 @@
 { 0x000178, CIM_UPPERDELTA, -0x0079, },
 { 0x000179, CIM_CASEBITOFF, 0x0001, },
 { 0x00017f, CIM_LOWERDELTA, 0x012c, },
-{ 0x000180, CIM_NONE, 0x0000, },
+{ 0x000180, CIM_LOWERDELTA, -0x00c3, },
 { 0x000181, CIM_UPPERDELTA, 0x00d2, },
 { 0x000182, CIM_CASEBIT, 0x0001, },
 { 0x000186, CIM_UPPERDELTA, 0x00ce, },
@@ -49,7 +49,8 @@
 { 0x000196, CIM_UPPERDELTA, 0x00d3, },
 { 0x000197, CIM_UPPERDELTA, 0x00d1, },
 { 0x000198, CIM_CASEBIT, 0x0001, },
-{ 0x00019a, CIM_NONE, 0x0000, },
+{ 0x00019a, CIM_LOWERDELTA, -0x00a3, },
+{ 0x00019b, CIM_NONE, 0x0000, },
 { 0x00019c, CIM_UPPERDELTA, 0x00d3, },
 { 0x00019d, CIM_UPPERDELTA, 0x00d5, },
 { 0x00019e, CIM_LOWERDELTA, -0x0082, },
@@ -89,6 +90,19 @@
 { 0x000221, CIM_NONE, 0x0000, },
 { 0x000222, CIM_CASEBIT, 0x0001, },
 { 0x000234, CIM_NONE, 0x0000, },
+{ 0x00023a, CIM_UPPERDELTA, 0x2a2b, },
+{ 0x00023b, CIM_CASEBITOFF, 0x0001, },
+{ 0x00023d, CIM_UPPERDELTA, -0x00a3, },
+{ 0x00023e, CIM_UPPERDELTA, 0x2a28, },
+{ 0x00023f, CIM_NONE, 0x0000, },
+{ 0x000241, CIM_CASEBITOFF, 0x0001, },
+{ 0x000243, CIM_UPPERDELTA, -0x00c3, },
+{ 0x000244, CIM_UPPERDELTA, 0x0045, },
+{ 0x000245, CIM_UPPERDELTA, 0x0047, },
+{ 0x000246, CIM_CASEBIT, 0x0001, },
+{ 0x000250, CIM_LOWERDELTA, -0x2a1f, },
+{ 0x000251, CIM_LOWERDELTA, -0x2a1c, },
+{ 0x000252, CIM_NONE, 0x0000, },
 { 0x000253, CIM_LOWERDELTA, 0x00d2, },
 { 0x000254, CIM_LOWERDELTA, 0x00ce, },
 { 0x000255, CIM_NONE, 0x0000, },
@@ -105,24 +119,36 @@
 { 0x000268, CIM_LOWERDELTA, 0x00d1, },
 { 0x000269, CIM_LOWERDELTA, 0x00d3, },
 { 0x00026a, CIM_NONE, 0x0000, },
+{ 0x00026b, CIM_LOWERDELTA, -0x29f7, },
+{ 0x00026c, CIM_NONE, 0x0000, },
 { 0x00026f, CIM_LOWERDELTA, 0x00d3, },
 { 0x000270, CIM_NONE, 0x0000, },
+{ 0x000271, CIM_LOWERDELTA, -0x29fd, },
 { 0x000272, CIM_LOWERDELTA, 0x00d5, },
 { 0x000273, CIM_NONE, 0x0000, },
 { 0x000275, CIM_LOWERDELTA, 0x00d6, },
 { 0x000276, CIM_NONE, 0x0000, },
+{ 0x00027d, CIM_LOWERDELTA, -0x29e7, },
+{ 0x00027e, CIM_NONE, 0x0000, },
 { 0x000280, CIM_LOWERDELTA, 0x00da, },
 { 0x000281, CIM_NONE, 0x0000, },
 { 0x000283, CIM_LOWERDELTA, 0x00da, },
 { 0x000284, CIM_NONE, 0x0000, },
 { 0x000288, CIM_LOWERDELTA, 0x00da, },
-{ 0x000289, CIM_NONE, 0x0000, },
+{ 0x000289, CIM_LOWERDELTA, 0x0045, },
 { 0x00028a, CIM_LOWERDELTA, 0x00d9, },
-{ 0x00028c, CIM_NONE, 0x0000, },
+{ 0x00028c, CIM_LOWERDELTA, 0x0047, },
+{ 0x00028d, CIM_NONE, 0x0000, },
 { 0x000292, CIM_LOWERDELTA, 0x00db, },
 { 0x000293, CIM_NONE, 0x0000, },
 { 0x000345, CIM_LOWERDELTA, -0x0054, },
 { 0x000346, CIM_NONE, 0x0000, },
+{ 0x000370, CIM_CASEBIT, 0x0001, },
+{ 0x000374, CIM_NONE, 0x0000, },
+{ 0x000376, CIM_CASEBIT, 0x0001, },
+{ 0x000378, CIM_NONE, 0x0000, },
+{ 0x00037b, CIM_LOWERDELTA, -0x0082, },
+{ 0x00037e, CIM_NONE, 0x0000, },
 { 0x000386, CIM_UPPERDELTA, 0x0026, },
 { 0x000387, CIM_NONE, 0x0000, },
 { 0x000388, CIM_UPPERDELTA, 0x0025, },
@@ -144,13 +170,13 @@
 { 0x0003c3, CIM_CASEBITOFF, 0x0020, },
 { 0x0003cc, CIM_CASEBIT, 0x0040, },
 { 0x0003cd, CIM_LOWERDELTA, 0x003f, },
-{ 0x0003cf, CIM_NONE, 0x0000, },
+{ 0x0003cf, CIM_CASEBITOFF, 0x0008, },
 { 0x0003d0, CIM_LOWERDELTA, 0x003e, },
 { 0x0003d1, CIM_LOWERDELTA, 0x0039, },
 { 0x0003d2, CIM_NONE, 0x0000, },
 { 0x0003d5, CIM_LOWERDELTA, 0x002f, },
 { 0x0003d6, CIM_LOWERDELTA, 0x0036, },
-{ 0x0003d7, CIM_NONE, 0x0000, },
+{ 0x0003d7, CIM_CASEBITOFF, 0x0008, },
 { 0x0003d8, CIM_CASEBIT, 0x0001, },
 { 0x0003f0, CIM_LOWERDELTA, 0x0056, },
 { 0x0003f1, CIM_LOWERDELTA, 0x0050, },
@@ -163,6 +189,7 @@
 { 0x0003f9, CIM_UPPERDELTA, -0x0007, },
 { 0x0003fa, CIM_CASEBIT, 0x0001, },
 { 0x0003fc, CIM_NONE, 0x0000, },
+{ 0x0003fd, CIM_UPPERDELTA, -0x0082, },
 { 0x000400, CIM_UPPERDELTA, 0x0050, },
 { 0x000410, CIM_CASEBIT, 0x0020, },
 { 0x000420, CIM_CASEBITOFF, 0x0020, },
@@ -172,25 +199,28 @@
 { 0x000460, CIM_CASEBIT, 0x0001, },
 { 0x000482, CIM_NONE, 0x0000, },
 { 0x00048a, CIM_CASEBIT, 0x0001, },
-{ 0x0004c0, CIM_NONE, 0x0000, },
+{ 0x0004c0, CIM_UPPERDELTA, 0x000f, },
 { 0x0004c1, CIM_CASEBITOFF, 0x0001, },
-{ 0x0004cf, CIM_NONE, 0x0000, },
+{ 0x0004cf, CIM_LOWERDELTA, 0x000f, },
 { 0x0004d0, CIM_CASEBIT, 0x0001, },
-{ 0x0004f6, CIM_NONE, 0x0000, },
-{ 0x0004f8, CIM_CASEBIT, 0x0001, },
-{ 0x0004fa, CIM_NONE, 0x0000, },
-{ 0x000500, CIM_CASEBIT, 0x0001, },
-{ 0x000510, CIM_NONE, 0x0000, },
+{ 0x000524, CIM_NONE, 0x0000, },
 { 0x000531, CIM_UPPERDELTA, 0x0030, },
 { 0x000557, CIM_NONE, 0x0000, },
 { 0x000561, CIM_LOWERDELTA, 0x0030, },
 { 0x000587, CIM_NONE, 0x0000, },
+{ 0x0010a0, CIM_UPPERDELTA, 0x1c60, },
+{ 0x0010c6, CIM_NONE, 0x0000, },
+{ 0x001d79, CIM_LONGLOWERDELTA, -0x0a04, },
+{ 0x001d7a, CIM_NONE, 0x0000, },
+{ 0x001d7d, CIM_LOWERDELTA, -0x0ee6, },
+{ 0x001d7e, CIM_NONE, 0x0000, },
 { 0x001e00, CIM_CASEBIT, 0x0001, },
 { 0x001e96, CIM_NONE, 0x0000, },
 { 0x001e9b, CIM_LOWERDELTA, 0x003b, },
 { 0x001e9c, CIM_NONE, 0x0000, },
+{ 0x001e9e, CIM_UPPERDELTA, -0x1dbf, },
+{ 0x001e9f, CIM_NONE, 0x0000, },
 { 0x001ea0, CIM_CASEBIT, 0x0001, },
-{ 0x001efa, CIM_NONE, 0x0000, },
 { 0x001f00, CIM_LOWERDELTA, -0x0008, },
 { 0x001f08, CIM_UPPERDELTA, -0x0008, },
 { 0x001f10, CIM_LOWERDELTA, -0x0008, },
@@ -274,11 +304,56 @@
 { 0x00212a, CIM_UPPERDELTA, -0x20bf, },
 { 0x00212b, CIM_UPPERDELTA, -0x2046, },
 { 0x00212c, CIM_NONE, 0x0000, },
+{ 0x002132, CIM_UPPERDELTA, 0x001c, },
+{ 0x002133, CIM_NONE, 0x0000, },
+{ 0x00214e, CIM_LOWERDELTA, 0x001c, },
+{ 0x00214f, CIM_NONE, 0x0000, },
 { 0x002160, CIM_CASEBIT, 0x0010, },
 { 0x002180, CIM_NONE, 0x0000, },
+{ 0x002183, CIM_CASEBITOFF, 0x0001, },
+{ 0x002185, CIM_NONE, 0x0000, },
 { 0x0024b6, CIM_UPPERDELTA, 0x001a, },
 { 0x0024d0, CIM_LOWERDELTA, 0x001a, },
 { 0x0024ea, CIM_NONE, 0x0000, },
+{ 0x002c00, CIM_UPPERDELTA, 0x0030, },
+{ 0x002c2f, CIM_NONE, 0x0000, },
+{ 0x002c30, CIM_LOWERDELTA, 0x0030, },
+{ 0x002c5f, CIM_NONE, 0x0000, },
+{ 0x002c60, CIM_CASEBIT, 0x0001, },
+{ 0x002c62, CIM_UPPERDELTA, -0x29f7, },
+{ 0x002c63, CIM_UPPERDELTA, -0x0ee6, },
+{ 0x002c64, CIM_UPPERDELTA, -0x29e7, },
+{ 0x002c65, CIM_LOWERDELTA, 0x2a2b, },
+{ 0x002c66, CIM_LOWERDELTA, 0x2a28, },
+{ 0x002c67, CIM_CASEBITOFF, 0x0001, },
+{ 0x002c6d, CIM_UPPERDELTA, -0x2a1c, },
+{ 0x002c6e, CIM_UPPERDELTA, -0x29fd, },
+{ 0x002c6f, CIM_UPPERDELTA, -0x2a1f, },
+{ 0x002c70, CIM_NONE, 0x0000, },
+{ 0x002c72, CIM_CASEBIT, 0x0001, },
+{ 0x002c74, CIM_NONE, 0x0000, },
+{ 0x002c75, CIM_CASEBITOFF, 0x0001, },
+{ 0x002c77, CIM_NONE, 0x0000, },
+{ 0x002c80, CIM_CASEBIT, 0x0001, },
+{ 0x002ce4, CIM_NONE, 0x0000, },
+{ 0x002d00, CIM_LOWERDELTA, 0x1c60, },
+{ 0x002d26, CIM_NONE, 0x0000, },
+{ 0x00a640, CIM_CASEBIT, 0x0001, },
+{ 0x00a660, CIM_NONE, 0x0000, },
+{ 0x00a662, CIM_CASEBIT, 0x0001, },
+{ 0x00a66e, CIM_NONE, 0x0000, },
+{ 0x00a680, CIM_CASEBIT, 0x0001, },
+{ 0x00a698, CIM_NONE, 0x0000, },
+{ 0x00a722, CIM_CASEBIT, 0x0001, },
+{ 0x00a730, CIM_NONE, 0x0000, },
+{ 0x00a732, CIM_CASEBIT, 0x0001, },
+{ 0x00a770, CIM_NONE, 0x0000, },
+{ 0x00a779, CIM_CASEBITOFF, 0x0001, },
+{ 0x00a77d, CIM_LONGUPPERDELTA, -0x0a04, },
+{ 0x00a77e, CIM_CASEBIT, 0x0001, },
+{ 0x00a788, CIM_NONE, 0x0000, },
+{ 0x00a78b, CIM_CASEBITOFF, 0x0001, },
+{ 0x00a78d, CIM_NONE, 0x0000, },
 { 0x00ff21, CIM_CASEBITOFF, 0x0020, },
 { 0x00ff3b, CIM_NONE, 0x0000, },
 { 0x00ff41, CIM_CASEBITOFF, 0x0020, },
