@@ -2,7 +2,7 @@
 || This file is part of Pike. For copyright information see COPYRIGHT.
 || Pike is distributed under GPL, LGPL and MPL. See the file COPYING
 || for more information.
-|| $Id: ia32.c,v 1.41 2004/05/25 07:56:48 grubba Exp $
+|| $Id: ia32.c,v 1.50 2008/11/19 21:37:34 mast Exp $
 */
 
 /*
@@ -323,6 +323,11 @@ static void update_arg2(INT32 value)
   MOV_VAL_TO_RELSTACK (value, 4);
 }
 
+#define PIKE_CPU_VENDOR_UNKNOWN	0
+#define PIKE_CPU_VENDOR_INTEL	1
+#define PIKE_CPU_VENDOR_AMD	2
+static int cpu_vendor = PIKE_CPU_VENDOR_UNKNOWN;
+
 static enum ia32_reg next_reg;
 static enum ia32_reg sp_reg, fp_reg, mark_sp_reg;
 ptrdiff_t ia32_prev_stored_pc; /* PROG_PC at the last point Pike_fp->pc was updated. */
@@ -521,7 +526,8 @@ static void ia32_push_global (INT32 arg)
   enum ia32_reg tmp_reg = alloc_reg ((1 << fp_reg) | (1 << sp_reg));
   load_fp_reg ((1 << tmp_reg) | (1 << sp_reg));
 
-  MOV_RELADDR16_TO_REG (OFFSETOF (pike_frame, context.identifier_level), fp_reg, tmp_reg);
+  MOV_RELADDR_TO_REG (OFFSETOF (pike_frame, context), fp_reg, tmp_reg);
+  MOV_RELADDR16_TO_REG (OFFSETOF (inherit, identifier_level), tmp_reg, tmp_reg);
   ADD_VAL_TO_REG (arg, tmp_reg);
   MOV_REG_TO_RELSTACK (tmp_reg, 8);
 
@@ -564,19 +570,16 @@ static void ia32_push_int(INT32 x)
 static void ia32_push_string (INT32 x, int subtype)
 {
   size_t e;
-  struct svalue tmp = {PIKE_T_STRING, subtype,
-#ifdef HAVE_UNION_INIT
-		       {0}
-#endif
-		      };
+  struct svalue tmp = SVALUE_INIT (PIKE_T_STRING, subtype, 0);
 
   enum ia32_reg tmp_reg = alloc_reg ((1 << fp_reg) | (1 << sp_reg));
   load_fp_reg ((1 << tmp_reg) | (1 << sp_reg));
 
-  MOV_RELADDR_TO_REG (OFFSETOF (pike_frame, context.prog), fp_reg, tmp_reg);
+  MOV_RELADDR_TO_REG (OFFSETOF (pike_frame, context), fp_reg, tmp_reg);
+  MOV_RELADDR_TO_REG (OFFSETOF (inherit, prog), tmp_reg, tmp_reg);
   MOV_RELADDR_TO_REG (OFFSETOF (program, strings), tmp_reg, tmp_reg);
   MOV_RELADDR_TO_REG (x * sizeof (struct pike_string *), tmp_reg, tmp_reg);
-  /* tmp_reg is now Pike_fp->context.prog->strings[x] */
+  /* tmp_reg is now Pike_fp->context->prog->strings[x] */
 
   load_sp_reg (1 << tmp_reg);
 
@@ -643,10 +646,11 @@ static void maybe_update_pc(void)
   if(
 #ifdef PIKE_DEBUG
     /* Update the pc more often for the sake of the opcode level trace. */
-    d_flag ||
+     d_flag ||
 #endif
-    last_prog_id != Pike_compiler->new_program->id ||
-    last_num_linenumbers != Pike_compiler->new_program->num_linenumbers
+     (ia32_prev_stored_pc == -1) ||
+     last_prog_id != Pike_compiler->new_program->id ||
+     last_num_linenumbers != Pike_compiler->new_program->num_linenumbers
   ) {
     last_prog_id=Pike_compiler->new_program->id;
     last_num_linenumbers = Pike_compiler->new_program->num_linenumbers;
@@ -659,6 +663,8 @@ static void ins_debug_instr_prologue (PIKE_INSTR_T instr, INT32 arg1, INT32 arg2
 {
   int flags = instrs[instr].flags;
 
+  maybe_update_pc();
+
   if (flags & I_HASARG2)
     MOV_VAL_TO_RELSTACK (arg2, 8);
   if (flags & I_HASARG)
@@ -759,7 +765,7 @@ void ins_f_byte(unsigned int b)
 #endif
 }
 
-void ins_f_byte_with_arg(unsigned int a,unsigned INT32 b)
+void ins_f_byte_with_arg(unsigned int a, INT32 b)
 {
   maybe_update_pc();
 
@@ -819,7 +825,7 @@ void ins_f_byte_with_arg(unsigned int a,unsigned INT32 b)
 
     case F_NEG_NUMBER:
       ins_debug_instr_prologue (a - F_OFFSET, b, 0);
-      ia32_push_int(-b);
+      ia32_push_int(-(INT32) b);
       return;
 
     case F_CONSTANT:
@@ -850,6 +856,7 @@ void ins_f_byte_with_arg(unsigned int a,unsigned INT32 b)
       if(Pike_compiler->new_program->constants[b].sval.u.efun->internal_flags & CALLABLE_DYNAMIC)
 	break;
       ins_debug_instr_prologue (a - F_OFFSET, b, 0);
+      ia32_call_c_function (call_check_threads_etc);
       update_arg1(0);
       ia32_call_c_function(Pike_compiler->new_program->constants[b].sval.u.efun->function);
       return;
@@ -858,6 +865,7 @@ void ins_f_byte_with_arg(unsigned int a,unsigned INT32 b)
       if(Pike_compiler->new_program->constants[b].sval.u.efun->internal_flags & CALLABLE_DYNAMIC)
 	break;
       ins_debug_instr_prologue (a - F_OFFSET, b, 0);
+      ia32_call_c_function (call_check_threads_etc);
       update_arg1(1);
       ia32_call_c_function(Pike_compiler->new_program->constants[b].sval.u.efun->function);
       return;
@@ -869,8 +877,8 @@ void ins_f_byte_with_arg(unsigned int a,unsigned INT32 b)
 }
 
 void ins_f_byte_with_2_args(unsigned int a,
-			    unsigned INT32 b,
-			    unsigned INT32 c)
+			    INT32 b,
+			    INT32 c)
 {
   maybe_update_pc();
 
@@ -922,11 +930,11 @@ static INT32 do_ins_jump (unsigned int op, int backward_jump)
       PUSH_INT (0);		/* 4 bytes */
     }
     else {
-#if 0
-      /* Slows down Intel PIII by 7%, speeds up Athlon XP by 22%. :P */
-      if (op == F_LOOP || op == F_FOREACH)
+      if ((cpu_vendor == PIKE_CPU_VENDOR_AMD) &&
+	  (op == F_LOOP || op == F_FOREACH)) {
+	/* Slows down Intel PIII by 7%, speeds up Athlon XP by 22%. :P */
 	add_to_program (0x3e);	/* Branch taken hint. */
-#endif
+      }
       add_to_program (0x0f);	/* jnz rel32 */
       add_to_program (0x85);
       ret = DO_NOT_WARN ((INT32) PIKE_PC);
@@ -1012,3 +1020,145 @@ void ia32_decode_program(struct program *p)
     *(INT32*)(prog + p->relocations[rel]) -= delta;
   }
 }
+
+static size_t ia32_clflush_size = 0;
+
+void ia32_flush_instruction_cache(void *start, size_t len)
+{
+  if (ia32_clflush_size) {
+    char *addr;
+    char *end_addr;
+    /* fprintf (stderr, "flush %p+%"PRINTSIZET"d\n", start, len); */
+    /* We assume even multiple of 2. */
+    len += ((size_t)start) & (ia32_clflush_size-1);
+    addr = (char *)(((size_t)start) & ~(ia32_clflush_size-1));
+    end_addr = addr + len;
+    while (addr < end_addr) {
+#ifdef USE_CL_IA32_ASM_STYLE
+      __asm {
+	mov eax, addr;
+	clflush [eax];
+      }
+#else  /* USE_GCC_IA32_ASM_STYLE */
+#ifdef HAVE_ASM_CLFLUSH
+      __asm__ __volatile__("clflush %0" :: "m" (*addr));
+#else
+      /* clflush (%eax) */
+      __asm__ __volatile__(".byte 0x0f\n"
+			   ".byte 0xae\n"
+			   ".byte 0x38" :: "eax" (addr));
+#endif
+#endif
+      addr += ia32_clflush_size;
+    }
+  }
+}
+
+static void ia32_get_cpuid(int oper, INT32 *cpuid_ptr)
+{
+  static int cpuid_supported = 0;
+  if (!cpuid_supported) {
+    int fbits=0;
+#ifdef USE_CL_IA32_ASM_STYLE
+    __asm {
+      pushf
+      pop  eax
+      mov  ecx, eax
+      xor  eax, 00200000h
+      push eax
+      popf
+      pushf
+      pop  eax
+      xor  ecx, eax
+      mov  fbits, ecx
+    };
+#else  /* USE_GCC_IA32_ASM_STYLE */
+    /* Note: gcc swaps the argument order... */
+    __asm__("pushf\n\t"
+	    "pop  %%eax\n\t"
+	    "movl %%eax, %%ecx\n\t"
+	    "xorl $0x00200000, %%eax\n\t"
+	    "push %%eax\n\t"
+	    "popf\n\t"
+	    "pushf\n\t"
+	    "pop  %%eax\n\t"
+	    "xorl %%eax, %%ecx\n\t"
+	    "movl %%ecx, %0"
+	    : "=m" (fbits)
+	    :
+	    : "cc", "eax", "ecx");
+#endif
+    if (fbits & 0x00200000) {
+      cpuid_supported = 1;
+    } else {
+      cpuid_supported = -1;
+    }
+  }
+
+  if (cpuid_supported > 0) {
+#ifdef USE_CL_IA32_ASM_STYLE
+    __asm {
+      mov eax, oper;
+      mov edi, cpuid_ptr;
+      cpuid;
+      mov [edi], eax;
+      mov [edi+4], ebx;
+      mov [edi+8], edx;
+      mov [edi+12], ecx;
+    };
+#else  /* USE_GCC_IA32_ASM_STYLE */
+    __asm__ __volatile__("cpuid"
+			 : "=a" (cpuid_ptr[0]),
+			   "=b" (cpuid_ptr[1]),
+			   "=d" (cpuid_ptr[2]),
+			   "=c" (cpuid_ptr[3])
+			 : "0" (oper));
+#endif
+  } else {
+    cpuid_ptr[0] = cpuid_ptr[1] = cpuid_ptr[2] = cpuid_ptr[3] = 0;
+  }
+}
+
+void ia32_init_interpreter_state(void)
+{
+  /* Note: One extra zero for nul-termination. */
+  INT32 cpuid[5] = { 0, 0, 0, 0, 0 };
+
+  /* fprintf(stderr, "Calling ia32_get_cpuid()...\n"); */
+
+  ia32_get_cpuid(0, cpuid);
+
+  /* fprintf(stderr, "CPUID: %d, \"%.12s\"\n", cpuid[0], (char *)(cpuid+1)); */
+
+  if (!memcmp (cpuid + 1, "AuthenticAMD", 12))
+    cpu_vendor = PIKE_CPU_VENDOR_AMD;
+  else if (!memcmp (cpuid + 1, "GenuineIntel", 12))
+    cpu_vendor = PIKE_CPU_VENDOR_INTEL;
+
+  if (cpuid[0] > 0) {
+    /* http://www.sandpile.org/ia32/cpuid.htm */
+    struct cpu_info {
+      int signature;
+      unsigned char brand_id;
+      unsigned char clflush_size;
+      unsigned char cpu_count;
+      unsigned char apic_id;
+      int feature_flags_edx;
+      int feature_flags_ecx;
+    } cpu_info;
+    ia32_get_cpuid(1, (INT32 *) &cpu_info);
+#if 0
+    fprintf(stderr,
+	    "features: 0x%08x\n"
+	    "CLFLUSH size: %d\n",
+	    cpu_info.feature_flags_edx,
+	    cpu_info.clflush_size);
+#endif /* 0 */
+
+    if (cpu_info.feature_flags_edx & 0x00080000) {
+      /* CLFLUSH present. */
+      /* fprintf (stderr, "Enabling clflush, size %d\n", cpu_info.clflush_size); */
+      ia32_clflush_size = cpu_info.clflush_size * 8;
+    }
+  }
+}
