@@ -2,7 +2,7 @@
 || This file is part of Pike. For copyright information see COPYRIGHT.
 || Pike is distributed under GPL, LGPL and MPL. See the file COPYING
 || for more information.
-|| $Id: ppc32.c,v 1.38 2005/08/27 20:30:57 jonasw Exp $
+|| $Id: ppc32.c,v 1.44 2008/11/19 21:38:15 mast Exp $
 */
 
 /*
@@ -52,7 +52,6 @@
     /* blrl */								\
     BCLRL(20, 0);							\
   } while(0)
-#define ADD_LABEL_CALL(X) ADD_CALL(&(X))
 #else
 #define ADD_CALL(X) do {				\
     INT32 func_=(INT32)(void*)(X);			\
@@ -69,7 +68,6 @@
       BLA(func_);					\
     }							\
   } while(0)
-#define ADD_LABEL_CALL(X) ADD_CALL(X)
 #endif
 
 int ppc32_codegen_state = 0, ppc32_codegen_last_pc = 0;
@@ -196,8 +194,10 @@ void ppc32_push_constant(INT32 arg)
   }
 
   LOAD_FP_REG();
-  /* lwz r3,context.prog(pike_fp) */
-  LWZ(PPC_REG_ARG1, PPC_REG_PIKE_FP, OFFSETOF(pike_frame, context.prog));
+  /* lwz r3,context(pike_fp) */
+  LWZ(PPC_REG_ARG1, PPC_REG_PIKE_FP, OFFSETOF(pike_frame, context));
+  /* lwz r3,prog(r3) */
+  LWZ(PPC_REG_ARG1, PPC_REG_ARG1, OFFSETOF(inherit, prog));
   /* lwz r3,constants(r3) */
   LWZ(PPC_REG_ARG1, PPC_REG_ARG1, OFFSETOF(program, constants));
 
@@ -254,12 +254,13 @@ void ppc32_local_lvalue(INT32 arg)
 void ppc32_push_global(INT32 arg)
 {
   LOAD_FP_REG();
-  /* lha r5,context.identifier_level(pike_fp) */
-  LHA(PPC_REG_ARG3, PPC_REG_PIKE_FP,
-      OFFSETOF(pike_frame, context.identifier_level));
+  /* lwz r5,context(pike_fp) */
+  LWZ(PPC_REG_ARG3, PPC_REG_PIKE_FP, OFFSETOF(pike_frame, context));
   /* lwz r4,current_object(pike_fp) */
   LWZ(PPC_REG_ARG2, PPC_REG_PIKE_FP, 
       OFFSETOF(pike_frame, current_object));
+  /* lha r5,identifier_level(r5) */
+  LHA(PPC_REG_ARG3, PPC_REG_ARG3, OFFSETOF(inherit, identifier_level));
   if(arg > 32767) {
     /* addis r5,r5,%hi(arg) */
     ADDIS(PPC_REG_ARG3, PPC_REG_ARG3, (arg+32768)>>16);
@@ -315,8 +316,10 @@ void ppc32_push_string(INT32 arg, int st)
   LOAD_FP_REG();
   LOAD_SP_REG();
   
-  /* lwz r11,context.prog(pike_fp) */
-  LWZ(11, PPC_REG_PIKE_FP, OFFSETOF(pike_frame, context.prog));
+  /* lwz r11,context(pike_fp) */
+  LWZ(11, PPC_REG_PIKE_FP, OFFSETOF(pike_frame, context));
+  /* lwz r11,prog(r11) */
+  LWZ(11, 11, OFFSETOF(inherit, prog));
   /* lwz r11,strings(r11) */
   LWZ(11, 11, OFFSETOF(program, strings));
 
@@ -371,24 +374,31 @@ void ppc32_mark(void)
 
 static void ppc32_escape_catch(void)
 {
-  extern void *do_escape_catch_label;
-  void *toc_;
-  INT32 pos_;
-
   LOAD_FP_REG();
+  /* lwz r3,catch_ctx(pike_interpreter) */
+  LWZ(PPC_REG_ARG1, PPC_REG_PIKE_INTERP,
+      OFFSETOF(Pike_interpreter, catch_ctx));
+  /* lwz r4,recovery.previous(r3) */
+  LWZ(PPC_REG_ARG2, PPC_REG_ARG1,
+      OFFSETOF(catch_context, recovery.previous));
+  /* stw r4,recoveries(pike_interpreter) */
+  STW(PPC_REG_ARG2, PPC_REG_PIKE_INTERP,
+      OFFSETOF(Pike_interpreter, recoveries));
+  /* lwz r4,save_expendible(r3) */
+  LWZ(PPC_REG_ARG2, PPC_REG_ARG1,
+      OFFSETOF(catch_context, save_expendible));
+  /* stw r4,expendible(pike_fp) */
+  STW(PPC_REG_ARG2, PPC_REG_PIKE_FP,
+      OFFSETOF(pike_frame, expendible));
+  /* lwz r4,prev(r3) */
+  LWZ(PPC_REG_ARG2, PPC_REG_ARG1,
+      OFFSETOF(catch_context, prev));
+  /* stw r4,catch_ctx(pike_interpreter) */
+  STW(PPC_REG_ARG2, PPC_REG_PIKE_INTERP,
+      OFFSETOF(Pike_interpreter, catch_ctx));
+
   FLUSH_CODE_GENERATOR_STATE();
-  /* bl .+4 */
-  BL(4);
-  pos_ = PIKE_PC;
-  /* mflr pike_pc */
-  MFSPR(PPC_REG_PIKE_PC, PPC_SPREG_LR);
-  /* addi pike_pc,pike_pc,N */
-  ADDI(PPC_REG_PIKE_PC, PPC_REG_PIKE_PC, 0);
-  /* stw pike_pc,return_addr(pike_fp) */
-  STW(PPC_REG_PIKE_PC, PPC_REG_PIKE_FP, OFFSETOF(pike_frame, return_addr));
-  ADD_LABEL_CALL(do_escape_catch_label);
-  Pike_compiler->new_program->program[pos_+1] +=
-    sizeof(PIKE_OPCODE_T) * (PIKE_PC-pos_);
+  ADD_CALL(really_free_catch_context);
 }
 
 static void maybe_update_pc(void)
@@ -471,7 +481,7 @@ void ins_f_byte(unsigned int b)
 #endif
 }
 
-void ins_f_byte_with_arg(unsigned int a,unsigned INT32 b)
+void ins_f_byte_with_arg(unsigned int a, INT32 b)
 {
   maybe_update_pc();
 
@@ -530,6 +540,7 @@ void ins_f_byte_with_arg(unsigned int a,unsigned INT32 b)
    case F_CALL_BUILTIN1:
      if(Pike_compiler->new_program->constants[b].sval.u.efun->internal_flags & CALLABLE_DYNAMIC)
        break;
+     /* FIXME: Add call to call_check_threads_etc. */
      if(a == F_CALL_BUILTIN1)
        SET_REG(PPC_REG_ARG1, 1);
      else
@@ -544,8 +555,8 @@ void ins_f_byte_with_arg(unsigned int a,unsigned INT32 b)
 }
 
 void ins_f_byte_with_2_args(unsigned int a,
-			    unsigned INT32 b,
-			    unsigned INT32 c)
+			    INT32 b,
+			    INT32 c)
 {
   maybe_update_pc();
 
@@ -940,6 +951,9 @@ void ppc32_disassemble_code(void *addr, size_t bytes)
 	else if(opcd == 10 || opcd == 11)
 	  fprintf(stderr, "%s cr%d,%d,r%d,%s\n", instr_name,
 		  (instr>>23)&7, (instr>>21)&1, (instr>>16)&31, immtext);
+	else if(opcd >= 48 && opcd < 56)
+	  fprintf(stderr, "%s fr%d,%s(r%d)\n", instr_name,
+		  (instr>>21)&31, immtext, (instr>>16)&31);
 	else if(opcd >= 32)
 	  fprintf(stderr, "%s r%d,%s(r%d)\n", instr_name,
 		  (instr>>21)&31, immtext, (instr>>16)&31);
@@ -971,7 +985,7 @@ void ppc32_disassemble_code(void *addr, size_t bytes)
 	else
 	  fprintf(stderr, "%s crf%d,fr%d,fr%d\n", instr_name,
 		  (instr>>23)&7, (instr>>16)&31, (instr>>11)&31);
-      } else if((xo & 135)==135) {
+      } else if((xo & 543)==535) {
 	fprintf(stderr, "%s fr%d,r%d,r%d\n", instr_name,
 		(instr>>21)&31, (instr>>16)&31, (instr>>11)&31);
       } else {
@@ -995,27 +1009,28 @@ void ppc32_disassemble_code(void *addr, size_t bytes)
 	  fprintf(stderr, "%s%s r%d,r%d\n", instr_name, ((instr&1)? ".":""),
 		  (instr>>16)&31, (instr>>21)&31);
 	else if(!(xo & 772)) {
-	  if(xo & 32)
+	  if((xo & 96)==96)
 	    fprintf(stderr, "%s r%d,r%d\n", instr_name,
 		    (instr>>21)&31, (instr>>11)&31);
 	  else if((xo & 192)==192)
-	    fprintf(stderr, "%s r%d\n", instr_name, (instr>>21)&31);
-	  else
 	    fprintf(stderr, "%s %d,r%d\n", instr_name,
 		    (instr>>16)&15, (instr>>21)&31);
-	} else if((xo & 127)==85)
+	  else
+	    fprintf(stderr, "%s r%d\n", instr_name, (instr>>21)&31);
+	} else if((xo & 639)==597)
 	  fprintf(stderr, "%s r%d,r%d,%d\n", instr_name,
 		  (instr>>21)&31, (instr>>16)&31, (instr>>11)&31);
-	else if(xo == 370 || xo == 566 || xo == 598 || xo == 854)
+	else if(xo == 370 || xo == 498 || xo == 566 || xo == 598 || xo == 854)
 	  fprintf(stderr, "%s\n", instr_name);
-	else if(!(xo & 8)) {
-	  if(xo & 128)
-	    fprintf(stderr, "%s r%d,r%d\n", instr_name,
-		    (instr>>21)%31, (instr>>11)&31);
-	  else if(xo & 512)
-	    fprintf(stderr, "%s r%d,%d\n", instr_name,
-		    (instr>>21)%31, (instr>>16)&15);
+	else if(!(xo & 4)) {
+	  if(xo & 32)
 	    fprintf(stderr, "%s r%d\n", instr_name, (instr>>11)&31);
+	  else if(xo == 595)
+	    fprintf(stderr, "%s r%d,%d\n", instr_name,
+		    (instr>>21)&31, (instr>>16)&15);
+	  else
+	    fprintf(stderr, "%s r%d,r%d\n", instr_name,
+		    (instr>>21)&31, (instr>>11)&31);
 	} else if(instr_name[0]=='d' || instr_name[0]=='i')
 	  fprintf(stderr, "%s r%d,r%d\n", instr_name,
 		  (instr>>16)&31, (instr>>11)&31);
