@@ -2,7 +2,7 @@
 || This file is part of Pike. For copyright information see COPYRIGHT.
 || Pike is distributed under GPL, LGPL and MPL. See the file COPYING
 || for more information.
-|| $Id: cpp.c,v 1.141 2004/11/01 03:05:23 mast Exp $
+|| $Id: cpp.c,v 1.177 2009/03/08 22:08:27 grubba Exp $
 */
 
 #include "global.h"
@@ -25,19 +25,17 @@
 #include "version.h"
 #include "pike_types.h"
 #include "cpp.h"
+#include "lex.h"
 
 #include <ctype.h>
 
 #define sp Pike_sp
 
-#undef ATTRIBUTE
-#define ATTRIBUTE(X)
-
-#define CPP_NO_OUTPUT 1
-#define CPP_EXPECT_ELSE 2
-#define CPP_EXPECT_ENDIF 4
-#define CPP_REALLY_NO_OUTPUT 8
-#define CPP_END_AT_NEWLINE 16
+#define CPP_NO_OUTPUT 1		/* Inside false section of #if/#else */
+#define CPP_EXPECT_ELSE 2	/* Expect #else/#elif/#elseif. */
+#define CPP_EXPECT_ENDIF 4	/* Expect #endif */
+#define CPP_REALLY_NO_OUTPUT 8	/* Entire preprocessor is in false section. */
+#define CPP_END_AT_NEWLINE 16	/* Halt at end of line. */
 #define CPP_DO_IF 32
 #define CPP_NO_EXPAND 64
 
@@ -53,6 +51,27 @@
 #define DEF_ARG_NOPOSTSPACE 0x400000
 #define DEF_ARG_MASK        0x0fffff
 
+#if 0
+#define CALC_DUMPPOS(X)	DUMPPOS(X)
+#else /* !0 */
+#define CALC_DUMPPOS(X)
+#endif /* 0 */
+
+static struct pike_string *efun_str;
+static struct pike_string *constant_str;
+static struct pike_string *defined_str;
+
+/* Some string builder debug. */
+#if 0
+
+#define string_builder_putchar(X, Y)	do {				\
+    int Y_Y_ = Y;							\
+    fprintf(stderr, "%s:%d string_builder_putchar(%s, %s, '%c')\n",	\
+	    __FILE__,__LINE__, #X, #Y, Y_Y_);				\
+    string_builder_putchar(X, Y_Y_);					\
+  } while(0)
+
+#endif /* 0 */
 
 struct pike_predef_s
 {
@@ -110,7 +129,8 @@ struct cpp
   struct object *compat_handler;
   int compat_major;
   int compat_minor;
-  int warn_if_constant_throws;
+  struct pike_string *data;
+  int picky_cpp;
 };
 
 struct define *defined_macro =0;
@@ -118,7 +138,7 @@ struct define *defined_macro =0;
 static struct pike_string *binary_findstring1(p_wchar1 *str, ptrdiff_t len);
 static struct pike_string *binary_findstring2(p_wchar2 *str, ptrdiff_t len);
 
-void cpp_error(struct cpp *this, const char *err)
+static void cpp_error(struct cpp *this, const char *err)
 {
   this->compile_errors++;
   if(this->compile_errors > 10) return;
@@ -139,15 +159,49 @@ void cpp_error(struct cpp *this, const char *err)
   }
 }
 
-void cpp_error_vsprintf (struct cpp *this, const char *fmt, va_list args)
+static void cpp_error_vsprintf (struct cpp *this, const char *fmt,
+				va_list args)
 {
-  char buf[8192];
-  VSNPRINTF (buf, sizeof (buf), fmt, args);
-  cpp_error(this, buf);
+  struct string_builder s;
+  struct pike_string *msg;
+
+  this->compile_errors++;
+  if (this->compile_errors > 10) return;
+
+  init_string_builder(&s, 0);
+
+  string_builder_vsprintf(&s, fmt, args);
+
+  msg = finish_string_builder(&s);
+
+  if((this->handler && this->handler->prog) || get_master())
+  {
+    ref_push_string(this->current_file);
+    push_int(this->current_line);
+    push_string(msg);
+    low_safe_apply_handler("compile_error", this->handler,
+			   this->compat_handler, 3);
+    pop_stack();
+    return;
+  }
+
+  if (this->current_file->size_shift) {
+    fprintf(stderr, "WIDE:%ld: ", (long)this->current_line);
+  } else {
+    fprintf(stderr, "%s:%ld: ",
+	    this->current_file->str, (long)this->current_line);
+  }
+
+  if (!msg->size_shift) {
+    fprintf(stderr, "%s\n", msg->str);
+  } else {
+    fprintf(stderr, "WIDE (fmt: %s)\n", fmt);
+  }
+  free_string(msg);
+  fflush(stderr);
 }
 
-void cpp_error_sprintf(struct cpp *this, const char *fmt, ...)
-  ATTRIBUTE((format(printf,2,3)))
+static void cpp_error_sprintf(struct cpp *this, const char *fmt, ...)
 {
   va_list args;
   va_start(args,fmt);
@@ -155,12 +209,12 @@ void cpp_error_sprintf(struct cpp *this, const char *fmt, ...)
   va_end(args);
 }
 
-void cpp_handle_exception(struct cpp *this, const char *cpp_error_fmt, ...)
-  ATTRIBUTE((format(printf,2,3)))
+static void cpp_handle_exception(struct cpp *this,
+				 const char *cpp_error_fmt, ...)
 {
   struct svalue thrown;
   move_svalue (&thrown, &throw_value);
-  throw_value.type = T_INT;
+  mark_free_svalue (&throw_value);
 
   if (cpp_error_fmt) {
     va_list args;
@@ -170,13 +224,13 @@ void cpp_handle_exception(struct cpp *this, const char *cpp_error_fmt, ...)
   }
 
   push_svalue(&thrown);
-  low_safe_apply_handler("compile_exception", error_handler, compat_handler, 1);
+  low_safe_apply_handler("compile_exception",
+			 this->handler, this->compat_handler, 1);
 
   if (SAFE_IS_ZERO(sp-1)) {
-    /* FIXME: Doesn't handle wide string error messages. */
     struct pike_string *s = format_exception_for_error_msg (&thrown);
     if (s) {
-      if (!s->size_shift) cpp_error (this, s->str);
+      cpp_error_sprintf(this, "%S", s);
       free_string (s);
     }
   }
@@ -185,13 +239,13 @@ void cpp_handle_exception(struct cpp *this, const char *cpp_error_fmt, ...)
   free_svalue(&thrown);
 }
 
-static void cpp_warning(struct cpp *this, const char *warn_fmt, ...)
+static void cpp_warning(struct cpp *this, const char *cpp_warn_fmt, ...)
 {
   char msg[8192];
   va_list args;
 
-  va_start(args, warn_fmt);
-  VSNPRINTF(msg, sizeof(msg), warn_fmt, args);
+  va_start(args, cpp_warn_fmt);
+  Pike_vsnprintf(msg, sizeof(msg), cpp_warn_fmt, args);
   va_end(args);
 
   if((this->handler && this->handler->prog) || get_master())
@@ -261,6 +315,45 @@ static void cpp_warning(struct cpp *this, const char *warn_fmt, ...)
  */
 
 /*! @class CompilationHandler
+ *!
+ *!   Objects used by the compiler to handle references to global symbols,
+ *!   modules, external files, etc.
+ *!
+ *!   There can be up to three compilation handlers active at the same
+ *!   time during a compilation. They are in order of precedence:
+ *!
+ *!   @ol
+ *!     @item
+ *!       The error handler
+ *!
+ *!       This is the object passed to @[compile()] as
+ *!       the second argument (if any). This object is returned by
+ *!       @[get_active_error_handler()] during a compilation.
+ *!
+ *!     @item
+ *!       The compatibility handler
+ *!
+ *!       This is the object returned by
+ *!       @[master()->get_compilation_handler()] (if any), which
+ *!       the compiler calls when it sees @tt{#pike@}-directives,
+ *!       or expressions using the version scope
+ *!       (eg @expr{7.4::rusage@}). This object is returned by
+ *!       @[get_active_compilation_handler()] during a compilation.
+ *!
+ *!     @item
+ *!       The master object.
+ *!
+ *!       This is returned by @[master()] at any time.
+ *!   @endol
+ *!
+ *!   Any of the objects may implement a subset of the @[CompilationHandler]
+ *!   functions, and the first object that implements a function will be
+ *!   used. The error handler object can thus be used to block certain
+ *!   functionality (eg to restrict the number of available functions).
+ *!
+ *! @seealso
+ *!   @[master()->get_compilation_handler()], @[get_active_error_handler()],
+ *!   @[get_active_compilation_handler()], @[compile()]
  */
 
 /*! @decl void compile_error(string filename, int line, string msg)
@@ -369,6 +462,8 @@ static void cpp_warning(struct cpp *this, const char *warn_fmt, ...)
 /*! @endclass
  */
 
+/* #pike handling. */
+
 void cpp_change_compat(struct cpp *this, int major, int minor)
 {
   if(this->compat_major == major &&
@@ -392,6 +487,10 @@ void cpp_change_compat(struct cpp *this, int major, int minor)
   SAFE_APPLY_MASTER("get_compilation_handler",2);
   if(sp[-1].type == T_OBJECT)
   {
+    if (sp[-1].subtype) {
+      cpp_error(this,
+		"#pike: Subtyped compilation handlers are not supported yet.");
+    }
     this->compat_handler=sp[-1].u.object;
     dmalloc_touch_svalue(Pike_sp-1);
     sp--;
@@ -418,9 +517,11 @@ static void check_defined(struct cpp *this,
   case 2:
     s = binary_findstring2((p_wchar2 *)args[0].arg.ptr, args[0].len);
     break;
+#ifdef PIKE_DEBUG
   default:
     Pike_fatal("cpp(): Symbol has unsupported shift: %d\n", args[0].arg.shift);
     break;
+#endif
   }
   if(s && find_define(s))
   {
@@ -439,13 +540,10 @@ static int do_safe_index_call(struct cpp *this, struct pike_string *s)
   if (SETJMP_SP(recovery, 1)) {
     if (CPP_TEST_COMPAT (this, 7, 4)) {
       free_svalue (&throw_value);
-      throw_value.type = T_INT;
+      mark_free_svalue (&throw_value);
     }
-    else {
-      if (!s->size_shift)
-	cpp_handle_exception (this, "Error indexing module with \"%s\".", s->str);
-      else
-	cpp_handle_exception (this, "Error indexing module in '.' operator.");
+    else if(this->picky_cpp) {
+      cpp_warning (this, "Error indexing module with %S.", s);
     }
     res = 0;
     push_undefined();
@@ -531,30 +629,23 @@ void cpp_func_constant(struct cpp *this, INT32 args)
 	     Pike_sp[-1].u.object == placeholder_object) ||
 	    (Pike_sp[-1].type == T_PROGRAM &&
 	     Pike_sp[-1].u.program == placeholder_program)) {
-	  if (!str->size_shift) {
-	    if(this->warn_if_constant_throws)
-	      cpp_warning (this, "Got placeholder %s (resolver problem) "
-			   "when resolving '%s'.",
-			   get_name_of_type (Pike_sp[-1].type), str->str);
-	  }
-	  else if(this->warn_if_constant_throws)
-	    cpp_warning (this, "Got placeholder %s (resolver problem).",
-			 get_name_of_type (Pike_sp[-1].type));
+	  cpp_error_sprintf (this, "Got placeholder %s (resolver problem) "
+			     "when resolving %S.",
+			     get_name_of_type(Pike_sp[-1].type),
+			     str);
 	}
 	else
 	  res = !(SAFE_IS_ZERO(sp-1) && sp[-1].subtype == NUMBER_UNDEFINED);
       }
       else if (throw_value.type == T_STRING &&
 	       !throw_value.u.string->size_shift) {
-	if(this->warn_if_constant_throws)
-	  cpp_warning(this, throw_value.u.string->str);
+	cpp_error(this, throw_value.u.string->str);
 	free_svalue(&throw_value);
-	throw_value.type = T_INT;
+	mark_free_svalue (&throw_value);
+	res = 0;
+      } else if(this->picky_cpp) {
+	cpp_warning (this, "Error resolving %S.", str);
 	res = 0;
-      } else if (!str->size_shift) {
-	cpp_handle_exception (this, "Error resolving '%s'.", str->str);
-      } else {
-	cpp_handle_exception (this, "Error resolving identifier.");
       }
     }
   } else {
@@ -587,7 +678,7 @@ void cpp_func_constant(struct cpp *this, INT32 args)
 
 /* Macro handling. */
 
-static struct mapping *initial_predefs_mapping()
+static struct mapping *initial_predefs_mapping(void)
 {
   struct pike_predef_s *def;
   struct mapping *map = allocate_mapping (0);
@@ -788,10 +879,25 @@ static void simple_add_define(struct cpp *this,
 		  fprintf(stderr,"\n");					\
 		  fflush(stderr)
 
-#define FIND_EOL() do {				\
-    while(pos < len && data[pos]!='\n') pos++;	\
-  } while(0)
+#define FIND_EOL() do {						\
+    while(pos < len && data[pos]!='\n') pos++;			\
+    if (data[pos] == '\\') {					\
+      if (data[pos+1] == '\n') {				\
+	pos+=2;							\
+      } else if ((data[pos+1] == '\r') &&			\
+		 (data[pos+2] == '\n')) {			\
+	pos+=3;							\
+      } else {							\
+	break;							\
+      }								\
+    } else {							\
+      break;							\
+    }								\
+    PUTNL();							\
+    this->current_line++;					\
+  } while (1)
 
+/* Skips horizontal whitespace and newlines. */
 #define SKIPWHITE() do {					\
     if(!WC_ISSPACE(data[pos])) {				\
       if (data[pos] == '\\') {					\
@@ -814,6 +920,7 @@ static void simple_add_define(struct cpp *this,
     pos++;							\
   } while(1)
 
+/* Skips horizontal whitespace and escaped newlines. */
 #define SKIPSPACE()						\
   do {								\
     while (WC_ISSPACE(data[pos]) && data[pos]!='\n') {		\
@@ -835,6 +942,7 @@ static void simple_add_define(struct cpp *this,
     this->current_line++;					\
   } while (1)
 
+/* The current char is assumed to be '*', the previous '/'. */
 #define SKIPCOMMENT()	do{				\
   	pos++;						\
 	while(data[pos]!='*' || data[pos+1]!='/')	\
@@ -856,39 +964,42 @@ static void simple_add_define(struct cpp *this,
 	pos+=2;						\
   }while(0)
 
-#define READCHAR(C) do {			\
-  switch(data[++pos])				\
-  {						\
-  case 'n': C='\n'; break;			\
-  case 'r': C='\r'; break;			\
-  case 'b': C='\b'; break;			\
-  case 't': C='\t'; break;			\
-    						\
-  case '0': case '1': case '2': case '3':	\
-  case '4': case '5': case '6': case '7':	\
-    C=data[pos]-'0';				\
-    if(data[pos+1]>='0' && data[pos+1]<='8')	\
-    {						\
-      C*=8;					\
-      C+=data[++pos]-'0';			\
-      						\
-      if(data[pos+1]>='0' && data[pos+1]<='8')	\
-      {						\
-	C*=8;					\
-	C+=data[++pos]-'0';			\
-      }						\
-    }						\
-    break;					\
-    						\
-  case '\n':					\
-    this->current_line++;			\
-    C='\n';					\
-    break;					\
-    						\
-  default:					\
-    C = data[pos];				\
-  }						\
-}while (0)
+/* pos is assumed to be at the backslash. pos it at the last char in
+ * the escape afterwards. */
+#define READCHAR(C) do {						\
+    ptrdiff_t l;							\
+    switch (parse_esc_seq (data + pos + 1, &C, &l)) {			\
+      case 0:								\
+	pos += l;							\
+	break;								\
+      case 1:								\
+	C = '\r';							\
+	pos++;								\
+	break;								\
+      case 3:								\
+	/* The eof will get caught in the next round. */		\
+	C = 0;								\
+	pos++;								\
+	break;								\
+      case 4: case 5: case 6:						\
+	cpp_warning (this, "Too large character value in escape.");	\
+	C = (int) MAX_UINT32;						\
+	pos += l;							\
+	break;								\
+      case 7:								\
+	cpp_warning (this, "Too few hex digits in \\u escape.");	\
+	C = '\\';							\
+	break;								\
+      case 8:								\
+	cpp_warning (this, "Too few hex digits in \\U escape.");	\
+	C = '\\';							\
+	break;								\
+      DO_IF_DEBUG (							\
+	case 2: Pike_fatal ("Not supposed to happen.\n");		\
+	default: Pike_fatal ("Unknown error from parse_esc_seq.\n");	\
+      );								\
+    }									\
+  } while (0)
 
 /* At entry pos points to the start-quote.
  * At end pos points past the end-quote.
@@ -912,7 +1023,7 @@ while(1)					\
   case '"':  break;				\
   case '\\':					\
   {						\
-    int tmp;					\
+    p_wchar2 tmp;				\
     if(data[pos+1]=='\n')			\
     {						\
       pos++;					\
@@ -976,6 +1087,7 @@ while(1)						\
     }							\
     if(outp) string_builder_putchar(&nf, '\\');	        \
     pos++;                                              \
+    /* Fall through. */					\
 							\
   default:						\
     if(outp) string_builder_putchar(&nf, data[pos-1]);	\
@@ -1002,7 +1114,7 @@ while(1)					\
   case '"':  break;				\
   case '\\':					\
   {						\
-    int tmp;					\
+    p_wchar2 tmp;				\
     if(data[pos+1]=='\n')			\
     {						\
       pos++;					\
@@ -1026,6 +1138,7 @@ while(1)					\
   case '\n':					\
     PUTNL();					\
     this->current_line++;			\
+    /* Fall through. */				\
   default:					\
     string_builder_putchar(&nf, data[pos]);	\
     continue;					\
@@ -1034,41 +1147,6 @@ while(1)					\
   break;					\
 }
 
-#define FINDTOK() 				\
-  do {						\
-  SKIPWHITE();					\
-  if(data[pos]=='/')				\
-  {						\
-    ptrdiff_t tmp;				\
-    switch(data[pos+1])				\
-    {						\
-    case '/':					\
-      FIND_EOL();				\
-      break;					\
-      						\
-    case '*':					\
-      tmp=pos+2;				\
-      while(1)					\
-      {						\
-        if(data[tmp]=='*')			\
-	{					\
-	  if(data[tmp+1] == '/')		\
-	  {					\
-	    pos=tmp+2;				\
-	    break;				\
-	  }					\
-	  break;				\
-	}					\
-						\
-	if(data[tmp]=='\n')			\
-	  break;				\
-        tmp++;					\
-      }						\
-    }						\
-  }						\
-  break;					\
-  }while(1)
-
 static struct pike_string *recode_string(struct cpp *this, struct pike_string *data)
 {
   /* Observations:
@@ -1275,7 +1353,7 @@ static struct pike_string *recode_string(struct cpp *this, struct pike_string *d
 
     if (!safe_apply_handler ("decode_charset", this->handler, this->compat_handler,
 			     2, BIT_STRING)) {
-      cpp_handle_exception (this, "Error decoding with charset '%s'", new_str->str);
+      cpp_handle_exception (this, "Error decoding with charset %S", new_str);
       free_string (new_str);
       return data;
     }
@@ -1316,14 +1394,14 @@ static struct pike_string *filter_bom(struct pike_string *data)
     for(i = 0; i<len; i++) {
       if (ptr[i] == 0xfeff) {
 	if (i != j) {
-	  string_builder_append(&buf, MKPCHARP(ptr + j, 1), i - j);
+	  string_builder_binary_strcat1 (&buf, ptr + j, i - j);
 	  j = i+1;
 	}
       }
     }
     if ((j) && (i != j)) {
       /* Add the trailing string */
-      string_builder_append(&buf, MKPCHARP(ptr + j, 1), i - j);
+      string_builder_binary_strcat1 (&buf, ptr + j, i - j);
       free_string(data);
       data = finish_string_builder(&buf);
     } else {
@@ -1336,14 +1414,14 @@ static struct pike_string *filter_bom(struct pike_string *data)
     for(i = 0; i<len; i++) {
       if (ptr[i] == 0xfeff) {
 	if (i != j) {
-	  string_builder_append(&buf, MKPCHARP(ptr + j, 2), i - j);
+	  string_builder_binary_strcat2 (&buf, ptr + j, i - j);
 	  j = i+1;
 	}
       }
     }
     if ((j) && (i != j)) {
       /* Add the trailing string */
-      string_builder_append(&buf, MKPCHARP(ptr + j, 2), i - j);
+      string_builder_binary_strcat2 (&buf, ptr + j, i - j);
       free_string(data);
       data = finish_string_builder(&buf);
     } else {
@@ -1396,8 +1474,10 @@ static ptrdiff_t low_cpp(struct cpp *this, void *data, ptrdiff_t len,
   case 2:
     return lower_cpp2(this, (p_wchar2 *)data, len,
 		      flags, auto_convert, charset);
+#ifdef PIKE_DEBUG
   default:
     Pike_fatal("low_cpp(): Bad shift: %d\n", shift);
+#endif
   }
   /* NOT_REACHED */
   return 0;
@@ -1409,9 +1489,7 @@ static void insert_current_line(struct cpp *this,
 				struct define_argument *args,
 				struct string_builder *tmp)
 {
-  char buf[20];
-  sprintf(buf," %ld ",(long)this->current_line);
-  string_builder_binary_strcat(tmp, buf, strlen(buf));
+  string_builder_sprintf(tmp, " %ld ", (long)this->current_line);
 }
 
 static void insert_current_file_as_string(struct cpp *this,
@@ -1423,6 +1501,19 @@ static void insert_current_file_as_string(struct cpp *this,
 		    this->current_file->size_shift, tmp);
 }
 
+static void insert_current_dir_as_string(struct cpp *this,
+                                         struct define *def,
+                                         struct define_argument *args,
+                                         struct string_builder *tmp)
+{
+  ref_push_string(this->current_file);
+  /* FIXME: This isn't safe if the master hasn't been compiled yet. */
+  SAFE_APPLY_MASTER("dirname",1);
+  PUSH_STRING_SHIFT(Pike_sp[-1].u.string->str, Pike_sp[-1].u.string->len,
+                    Pike_sp[-1].u.string->size_shift, tmp);
+  pop_stack();
+}
+
 static void insert_current_time_as_string(struct cpp *this,
 					  struct define *def,
 					  struct define_argument *args,
@@ -1457,9 +1548,8 @@ static void insert_current_version(struct cpp *this,
 				   struct define_argument *args,
 				   struct string_builder *tmp)
 {
-  char buf[50];
-  sprintf(buf," %d.%d ",this->compat_major, this->compat_minor);
-  string_builder_binary_strcat(tmp, buf, strlen(buf));
+  string_builder_sprintf(tmp, " %d.%d ", this->compat_major,
+			 this->compat_minor);
 }
 
 
@@ -1468,9 +1558,7 @@ static void insert_current_minor(struct cpp *this,
 				 struct define_argument *args,
 				 struct string_builder *tmp)
 {
-  char buf[20];
-  sprintf(buf," %d ",this->compat_minor);
-  string_builder_binary_strcat(tmp, buf, strlen(buf));
+  string_builder_sprintf(tmp, " %d ", this->compat_minor);
 }
 
 
@@ -1479,9 +1567,7 @@ static void insert_current_major(struct cpp *this,
 				 struct define_argument *args,
 				 struct string_builder *tmp)
 {
-  char buf[20];
-  sprintf(buf," %d ",this->compat_major);
-  string_builder_binary_strcat(tmp, buf, strlen(buf));
+  string_builder_sprintf(tmp, " %d ", this->compat_major);
 }
 
 
@@ -1576,6 +1662,11 @@ static void insert_current_major(struct cpp *this,
  *! This define contains the file path and name of the source file.
  */
 
+/*! @decl constant __DIR__
+ *!
+ *! This define contains the directory path of the source file.
+ */
+
 /*! @decl constant __DATE__
  *!
  *! This define contains the current date at the time of compilation,
@@ -1618,7 +1709,7 @@ static void insert_current_major(struct cpp *this,
 /*! @decl string cpp(string data, string|void current_file, @
  *!                  int|string|void charset, object|void handler, @
  *!                  void|int compat_major, void|int compat_minor, @
- *!                  void|int warn_if_constant_throws)
+ *!                  void|int picky_cpp)
  *!
  *! Run a string through the preprocessor.
  *!
@@ -1629,101 +1720,117 @@ static void insert_current_major(struct cpp *this,
  *! @seealso
  *!   @[compile()]
  */
+
+/* Doesn't free string_builder buf! */
+static void free_cpp(struct cpp *this)
+{
+  if(this->defines)
+    free_hashtable(this->defines, free_one_define);
+
+  if(this->current_file)
+    free_string(this->current_file);
+
+  if(this->handler) {
+    free_object(this->handler);
+    this->handler = 0;
+  }
+
+  if(this->compat_handler) {
+    free_object(this->compat_handler);
+    this->compat_handler=0;
+  }
+
+  if(this->data)
+    free_string(this->data);
+}
+
 void f_cpp(INT32 args)
 {
-  struct pike_string *data;
-  struct mapping *predefs = NULL;
-  struct svalue *save_sp = sp - args;
   struct cpp this;
+  struct svalue *save_sp = sp - args;
+  struct mapping *predefs = NULL;
+
+  struct pike_string *data;
+
+  struct pike_string *current_file = 0;
+
+  struct svalue *charset_sv = 0;
   int auto_convert = 0;
   struct pike_string *charset = NULL;
 
+  struct object *handler = 0;
+
+  int compat_major, compat_minor, picky_cpp;
+
+  ONERROR err;
 #ifdef PIKE_DEBUG
   ONERROR tmp;
 #endif /* PIKE_DEBUG */
 
-  if(args<1)
-    SIMPLE_TOO_FEW_ARGS_ERROR("cpp", 1);
+  get_all_args("cpp", args, "%t.%T%*%O%d%d%d", &data, &current_file,
+	       &charset_sv, &handler, &compat_major, &compat_minor,
+	       &picky_cpp);
 
-  if(sp[-args].type != T_STRING)
-    SIMPLE_BAD_ARG_ERROR("cpp", 1, "string");
-
-  data = sp[-args].u.string;
+  this.current_line=1;
+  this.compile_errors=0;
+  this.defines=0;
 
+  this.data = data;
   add_ref(data);
 
-  this.handler = NULL;
-  this.compat_handler = NULL;
-
-  if(args>1)
-  {
-    if(args > 5)
-    {
-      if(sp[4-args].type != T_INT)
-	SIMPLE_BAD_ARG_ERROR("cpp", 5, "int");
-      if(sp[5-args].type != T_INT)
-	SIMPLE_BAD_ARG_ERROR("cpp", 6, "int");
-    }
-    if(sp[1-args].type != T_STRING) {
-      free_string(data);
-      SIMPLE_BAD_ARG_ERROR("cpp", 2, "string");
-    }
-    copy_shared_string(this.current_file, sp[1-args].u.string);
-
-    if (args > 2) {
-      if (sp[2-args].type == T_STRING) {
-	charset = sp[2 - args].u.string;
-	push_string(data);
-	ref_push_string(charset);
-	if (!safe_apply_handler ("decode_charset", this.handler, this.compat_handler,
-				 2, BIT_STRING)) {
-	  free_string(this.current_file);
-	  cpp_handle_exception (&this, "Error decoding with charset '%s'",
-				charset->str);
-	  Pike_error("Unknown charset.\n");
-	}
-	data = sp[-1].u.string;
-	sp--;
-	dmalloc_touch_svalue(sp);
-      } else if (sp[2-args].type == T_INT) {
-	auto_convert = sp[2-args].u.integer;
-      } else {
-	free_string(data);
-	SIMPLE_BAD_ARG_ERROR("cpp", 3, "string|int");
-      }
-      if (args > 3) {
-	if (sp[3-args].type == T_OBJECT) {
-	  if ((this.handler = sp[3-args].u.object)) {
-	    add_ref(this.handler);
-	  }
-	} else if (sp[3-args].type != T_INT) {
-	  free_string(data);
-	  free_string(this.current_file);
-	  SIMPLE_BAD_ARG_ERROR("cpp", 4, "object");
-	}
-      }
-    }
-  }else{
-    this.current_file=make_shared_string("-");
-  }
+  if(current_file)
+    add_ref(current_file);
+  else
+    current_file = make_shared_string("-");
+  this.current_file = current_file;
 
   this.compat_major=PIKE_MAJOR_VERSION;
   this.compat_minor=PIKE_MINOR_VERSION;
-  if(args > 5)
-  {
-    cpp_change_compat(&this, sp[4-args].u.integer, sp[5-args].u.integer);
+  this.compat_handler = 0;
+  this.handler = handler;
+  if(handler)
+    add_ref(handler);
+
+  /* Don't call free_cpp before all variables are cleared or set. */
+  SET_ONERROR(err, free_cpp, &this);
+
+  if(charset_sv) {
+    if(charset_sv->type == T_STRING) {
+      charset = charset_sv->u.string;
+      push_string(data);
+      ref_push_string(charset);
+      if (!safe_apply_handler ("decode_charset", this.handler,
+			       this.compat_handler, 2, BIT_STRING)) {
+	cpp_handle_exception (&this, "Error decoding with charset %S",
+			      charset);
+	Pike_error("Unknown charset.\n");
+      }
+      free(data);
+      this.data = data = sp[-1].u.string;
+      sp--;
+      dmalloc_touch_svalue(sp);
+    }
+    else if(charset_sv->type == T_INT)
+      auto_convert = charset_sv->u.integer;
+    else {
+      SIMPLE_BAD_ARG_ERROR("cpp", 3, "string|int");
+    }
   }
 
+  if(args > 5)
+    cpp_change_compat(&this, compat_major, compat_minor);
+
   if(args > 6)
-    this.warn_if_constant_throws = sp[6-args].u.integer;
+    this.picky_cpp = picky_cpp;
   else
-    this.warn_if_constant_throws = 0;
+    this.picky_cpp = 0;
 
   if (use_initial_predefs)
     /* Typically compiling the master here. */
     predefs = initial_predefs_mapping();
   else {
-    low_unsafe_apply_handler ("get_predefines", this.handler, this.compat_handler, 0);
+    low_unsafe_apply_handler ("get_predefines", this.handler,
+			      this.compat_handler, 0);
     if (!UNSAFE_IS_ZERO (sp - 1)) {
       struct keypair *k;
       int e, sprintf_args = 0;
@@ -1759,15 +1866,8 @@ void f_cpp(INT32 args)
       }
       if (!predefs) {
       predef_map_error:
-	free_string (data);
-	free_string (this.current_file);
-	if (this.handler) free_object (this.handler);
-	if (this.compat_handler) free_object (this.compat_handler);
 	f_sprintf (sprintf_args);
-	if (!sp[-1].u.string->size_shift)
-	  Pike_error ("%s", sp[-1].u.string->str);
-	else
-	  Pike_error ("%s", sp[-2].u.string->str);
+	Pike_error("%S", sp[-1].u.string);
       }
     }
     pop_stack();
@@ -1777,22 +1877,20 @@ void f_cpp(INT32 args)
     /* Try to determine if we need to recode the string */
     struct pike_string *new_data = recode_string(&this, data);
     free_string(data);
-    data = new_data;
+    this.data = data = new_data;
   }
   if (data->size_shift) {
     /* Get rid of any byte order marks (0xfeff) */
     struct pike_string *new_data = filter_bom(data);
     free_string(data);
-    data = new_data;
+    this.data = data = new_data;
   }
 
   init_string_builder(&this.buf, 0);
-  this.current_line=1;
-  this.compile_errors=0;
-  this.defines=0;
 
   do_magic_define(&this,"__LINE__",insert_current_line);
   do_magic_define(&this,"__FILE__",insert_current_file_as_string);
+  do_magic_define(&this,"__DIR__",insert_current_dir_as_string);
   do_magic_define(&this,"__DATE__",insert_current_date_as_string);
   do_magic_define(&this,"__TIME__",insert_current_time_as_string);
   do_magic_define(&this,"__VERSION__",insert_current_version);
@@ -1800,6 +1898,17 @@ void f_cpp(INT32 args)
   do_magic_define(&this,"__MINOR__",insert_current_minor);
 
   {
+#if 0
+    /* Left in place for documentation reference purposes. */
+    struct define *def =
+      alloc_empty_define(make_shared_string("__deprecated__"), 1);
+    def->args = 1;
+    REF_MAKE_CONST_STRING(def->first, "__attribute__(\"deprecated\", ");
+    def->parts[0].argument = 0;
+    REF_MAKE_CONST_STRING(def->parts[0].postfix, ")");
+    this.defines = hash_insert(this.defines, &def->link);
+#endif /* 0 */
+
     simple_add_define(&this, "__PIKE__", " 1 ");
 
     simple_add_define(&this, "__REAL_VERSION__",
@@ -1821,6 +1930,9 @@ void f_cpp(INT32 args)
 #endif
 #ifdef __amigaos__
     simple_add_define(&this, "__amigaos__", " 1 ");
+#endif
+#ifdef __APPLE__
+    simple_add_define(&this, "__APPLE__", " 1 ");
 #endif
     simple_add_define(&this, "SIZEOF_INT",
 		      " " DEFINETOSTR(SIZEOF_INT) " ");
@@ -1857,29 +1969,13 @@ void f_cpp(INT32 args)
   UNSET_ONERROR(tmp);
 #endif /* PIKE_DEBUG */
 
-  if(this.defines)
-    free_hashtable(this.defines, free_one_define);
-
-  if(this.compat_handler)
-  {
-    free_object(this.compat_handler);
-    this.compat_handler=0;
-  }
-
-  if (this.handler)
-  {
-    free_object(this.handler);
-    this.handler = 0;
-  }
-
-  free_string(this.current_file);
-
-  free_string(data);
+  UNSET_ONERROR(err);
+  free_cpp(&this);
 
   if(this.compile_errors)
   {
     free_string_builder(&this.buf);
-    throw_error_object(low_clone(cpp_error_program), 0, 0, 0,
+    throw_error_object(fast_clone_object(cpp_error_program), 0, 0, 0,
 		       "Cpp() failed\n");
   }else{
     pop_n_elems(sp - save_sp);
@@ -1915,6 +2011,10 @@ void init_cpp()
   defined_macro->magic=check_defined;
   defined_macro->args=1;
 
+  efun_str = make_shared_string ("efun");
+  constant_str = make_shared_string ("constant");
+  defined_str = make_shared_string ("defined");
+
   use_initial_predefs = 1;
 
 /* function(string, string|void, int|string|void, object|void, int|void, int|void:string) */
@@ -1923,6 +2023,7 @@ void init_cpp()
 			       tOr(tObj,tVoid)
 			       tOr(tInt,tVoid)
 			       tOr(tInt,tVoid)
+			       tOr(tInt,tVoid)
 			       , tStr),
 	   /* OPT_SIDE_EFFECT since we might instantiate modules etc. */
 	   OPT_EXTERNAL_DEPEND|OPT_SIDE_EFFECT);
@@ -1979,5 +2080,9 @@ void exit_cpp(void)
   }
   free_string(defined_macro->link.s);
   free((char *)defined_macro);
+
+  free_string (efun_str);
+  free_string (constant_str);
+  free_string (defined_str);
 #endif
 }
