@@ -2,7 +2,7 @@
 || This file is part of Pike. For copyright information see COPYRIGHT.
 || Pike is distributed under GPL, LGPL and MPL. See the file COPYING
 || for more information.
-|| $Id: dmalloc.h,v 1.53 2006/08/06 16:58:07 mast Exp $
+|| $Id: dmalloc.h,v 1.68 2008/09/04 20:31:07 mast Exp $
 */
 
 #ifndef DMALLOC_H
@@ -13,15 +13,41 @@ PMOD_EXPORT extern void debug_xfree(void *);
 PMOD_EXPORT extern void *debug_xmalloc(size_t);
 PMOD_EXPORT extern void *debug_xcalloc(size_t,size_t);
 PMOD_EXPORT extern void *debug_xrealloc(void *,size_t);
+PMOD_EXPORT char *debug_xstrdup(const char *src);
+
+#if defined (HAVE_EXECINFO_H) && defined (HAVE_BACKTRACE)
+/* GNU libc provides some tools to inspect the stack. */
+#include <execinfo.h>
+typedef void *c_stack_frame;
+
+#define DUMP_C_STACK_TRACE() do {					\
+    c_stack_frame bt[100];						\
+    int n = backtrace (bt, 100);					\
+    backtrace_symbols_fd (bt, n, 2);					\
+  } while (0)
+
+#else
+#undef DMALLOC_C_STACK_TRACE
+#define DUMP_C_STACK_TRACE() do {} while (0)
+#endif
 
 #define DMALLOC_NAMED_LOCATION(NAME)	\
     (("NS" __FILE__ ":" DEFINETOSTR(__LINE__) NAME )+1)
 
 #define DMALLOC_LOCATION() DMALLOC_NAMED_LOCATION("")
 
+/* Location types in loc[0]:
+ * 'S': static (DMALLOC_NAMED_LOCATION)
+ * 'D': dynamic (dynamic_location)
+ * 'B': dynamic with backtrace (debug_malloc_update_location_bt)
+ * 'T': memory map template
+ * 'M': memory map
+ */
 typedef char *LOCATION;
+#define LOCATION_TYPE(X) ((X)[0])
 #define LOCATION_NAME(X) ((X)+1)
-#define LOCATION_IS_DYNAMIC(X) ((X)[0]=='D')
+#define LOCATION_IS_DYNAMIC(X)						\
+  (LOCATION_TYPE (X)=='D' || LOCATION_TYPE (X) == 'B')
 
 #ifdef DMALLOC_TRACE
 #define DMALLOC_TRACELOGSIZE 131072
@@ -34,12 +60,13 @@ extern size_t dmalloc_tracelogptr;
 #endif /* DMALLOC_TRACE */
 
 #ifdef PIKE_DEBUG
-extern int gc_external_refs_zapped;
-void gc_check_zapped (void *a, TYPE_T type, const char *file, int line);
+PMOD_EXPORT extern int gc_external_refs_zapped;
+PMOD_EXPORT void gc_check_zapped (void *a, TYPE_T type, const char *file, int line);
 #endif
 
 #ifdef DO_PIKE_CLEANUP
 extern int exit_with_cleanup;
+extern int exit_cleanup_in_progress;
 #define DO_IF_PIKE_CLEANUP(X) X
 #else
 #define DO_IF_PIKE_CLEANUP(X)
@@ -58,31 +85,35 @@ void really_free_memhdr(struct memhdr *mh);
 void add_marks_to_memhdr(struct memhdr *to,void *ptr);
 
 extern int verbose_debug_malloc;
-extern void dmalloc_trace(void *);
-extern void dmalloc_register(void *, int, LOCATION);
-extern int dmalloc_unregister(void *, int);
-extern void *debug_malloc(size_t, LOCATION);
-extern void *debug_calloc(size_t, size_t, LOCATION);
-extern void *debug_realloc(void *, size_t, LOCATION);
-extern void debug_free(void *, LOCATION, int);
-extern char *debug_strdup(const char *, LOCATION);
-extern void reset_debug_malloc(void);
+PMOD_EXPORT void dmalloc_trace(void *);
+PMOD_EXPORT void dmalloc_register(void *, int, LOCATION);
+PMOD_EXPORT int dmalloc_unregister(void *, int);
+PMOD_EXPORT void *debug_malloc(size_t, LOCATION);
+PMOD_EXPORT void *debug_calloc(size_t, size_t, LOCATION);
+PMOD_EXPORT void *debug_realloc(void *, size_t, LOCATION);
+PMOD_EXPORT void debug_free(void *, LOCATION, int);
+PMOD_EXPORT char *debug_strdup(const char *, LOCATION);
+void reset_debug_malloc(void);
+PMOD_EXPORT int dmalloc_check_allocated (void *p, int must_be_freed);
+#if 0
 void dmalloc_check_block_free(void *, LOCATION, char *, describe_block_fn *);
-extern void dmalloc_free(void *p);
-extern int debug_malloc_touch_fd(int, LOCATION);
-extern int debug_malloc_register_fd(int, LOCATION);
-extern int debug_malloc_close_fd(int, LOCATION);
-extern int dmalloc_mark_as_free(void*,int);
+#endif
+PMOD_EXPORT void dmalloc_free(void *p);
+PMOD_EXPORT int debug_malloc_touch_fd(int, LOCATION);
+PMOD_EXPORT int debug_malloc_register_fd(int, LOCATION);
+PMOD_EXPORT void debug_malloc_accept_leak_fd(int);
+PMOD_EXPORT int debug_malloc_close_fd(int, LOCATION);
+PMOD_EXPORT int dmalloc_mark_as_free(void*,int);
 
-void *debug_malloc_update_location(void *, LOCATION);
-void *debug_malloc_update_location_ptr(void *, ptrdiff_t, LOCATION);
+PMOD_EXPORT void *debug_malloc_update_location(void *, LOCATION);
+PMOD_EXPORT void *debug_malloc_update_location_ptr(void *, ptrdiff_t, LOCATION);
+PMOD_EXPORT void *debug_malloc_update_location_bt (void *p, const char *file,
+						   int line, const char *name);
 void search_all_memheaders_for_references(void);
-void cleanup_memhdrs(void);
-void cleanup_debug_malloc(void);
 
 /* Beware! names of named memory regions are never ever freed!! /Hubbe */
-void *debug_malloc_name(void *p, const char *fn, int line);
-int debug_malloc_copy_names(void *p, void *p2);
+PMOD_EXPORT void *debug_malloc_name(void *p, const char *fn, int line);
+PMOD_EXPORT int debug_malloc_copy_names(void *p, void *p2);
 char *dmalloc_find_name(void *p);
 
 /* glibc 2.1 defines this as a macro. */
@@ -99,22 +130,32 @@ char *dmalloc_find_name(void *p);
 #define DO_IF_DMALLOC(X) X
 #define DO_IF_NOT_DMALLOC(X)
 #define debug_malloc_touch(X) debug_malloc_update_location((void *)(X),DMALLOC_LOCATION())
+#define debug_malloc_touch_named(X,NAME) debug_malloc_update_location((void *)(X),DMALLOC_NAMED_LOCATION(" " NAME))
 #define debug_malloc_pass(X) debug_malloc_update_location((void *)(X),DMALLOC_LOCATION())
+#define debug_malloc_pass_named(X,NAME) debug_malloc_update_location((void *)(X),DMALLOC_NAMED_LOCATION(" " NAME))
 #define dmalloc_touch_struct_ptr(TYPE,X,MEMBER) ((TYPE)debug_malloc_update_location_ptr((void *)(X), ((ptrdiff_t)& (((TYPE)0)->MEMBER)), DMALLOC_LOCATION()))
 
+/* These also save the call stack if support for that exists. Beware
+ * that this can take a lot of extra memory - for temporary use only. */
+#define debug_malloc_touch_bt(X) debug_malloc_update_location_bt ((void *)(X), __FILE__, __LINE__, NULL)
+#define debug_malloc_touch_named_bt(X,NAME) debug_malloc_update_location_bt ((void *)(X), __FILE__, __LINE__, (NAME))
+
 #define xalloc(X) ((void *)debug_malloc_update_location((void *)debug_xalloc(X), DMALLOC_NAMED_LOCATION(" xalloc")))
 #define xfree(X) debug_xfree(debug_malloc_update_location((X), DMALLOC_NAMED_LOCATION(" free")))
-void debug_malloc_dump_references(void *x, int indent, int depth, int flags);
-#define dmalloc_touch(TYPE,X) ((TYPE)debug_malloc_update_location((void *)(X),DMALLOC_LOCATION()))
+PMOD_EXPORT void debug_malloc_dump_references(void *x, int indent, int depth, int flags);
+#define dmalloc_touch(TYPE,X) ((TYPE) debug_malloc_pass (X))
+#define dmalloc_touch_named(TYPE,X,NAME) ((TYPE) debug_malloc_pass_named (X, NAME))
 void debug_malloc_dump_fd(int fd);
 #define dmalloc_touch_svalue(X) do { struct svalue *_tmp = (X); if (_tmp->type <= MAX_REF_TYPE) { debug_malloc_touch(_tmp->u.refs); } } while(0)
+#define dmalloc_touch_svalue_named(X,NAME) do { struct svalue *_tmp = (X); if (_tmp->type <= MAX_REF_TYPE) { debug_malloc_touch_named(_tmp->u.refs,NAME); } } while(0)
 
 #define DMALLOC_LINE_ARGS ,char * dmalloc_location
 #define DMALLOC_POS ,DMALLOC_LOCATION()
 #define DMALLOC_PROXY_ARGS ,dmalloc_location
-void dmalloc_accept_leak(void *);
+PMOD_EXPORT void dmalloc_accept_leak(void *);
 #define dmalloc_touch_fd(X) debug_malloc_touch_fd((X),DMALLOC_LOCATION())
 #define dmalloc_register_fd(X) debug_malloc_register_fd((X),DMALLOC_LOCATION())
+#define dmalloc_accept_leak_fd(X) debug_malloc_accept_leak_fd(X)
 #define dmalloc_close_fd(X) debug_malloc_close_fd((X),DMALLOC_LOCATION())
 
 /* Beware, these do not exist without DMALLOC */
@@ -159,6 +200,7 @@ PMOD_EXPORT void* dlpvalloc(size_t);
 
 #define dmalloc_touch_fd(X) (X)
 #define dmalloc_register_fd(X) (X)
+#define dmalloc_accept_leak_fd(X)
 #define dmalloc_close_fd(X) (X)
 #define dmfree(X) free((X))
 #define dmalloc_accept_leak(X) (void)(X)
@@ -169,7 +211,7 @@ PMOD_EXPORT void* dlpvalloc(size_t);
 #define debug_malloc_dump_fd(fd)
 #define xalloc debug_xalloc
 
-#if defined(DYNAMIC_MODULE) && defined(__NT__)
+#if defined(DYNAMIC_MODULE) && defined(__NT__) && !defined(USE_DLL)
 #define xmalloc debug_xmalloc
 #define xcalloc debug_xcalloc
 #define xrealloc debug_xrealloc
@@ -188,7 +230,7 @@ PMOD_EXPORT void* dlpvalloc(size_t);
 #define DO_IF_NOT_DMALLOC(X) X
 #define dmalloc_trace(X)
 #define dmalloc_register(X,Y,Z)
-#define dmalloc_unregister(X,Y)
+#define dmalloc_unregister(X,Y) 1
 #define debug_free(X,Y,Z) free((X))
 #define debug_malloc_name(P,FN,LINE)
 #define debug_malloc_copy_names(p,p2) 0
@@ -199,15 +241,27 @@ PMOD_EXPORT void* dlpvalloc(size_t);
 #ifdef DMALLOC_TRACE
 #define debug_malloc_update_location(X,Y) (DMALLOC_TRACE_LOG(DMALLOC_LOCATION()),(X))
 #define dmalloc_touch_svalue(X) DMALLOC_TRACE_LOG(DMALLOC_LOCATION())
+#define dmalloc_touch_svalue_named(X,NAME) DMALLOC_TRACE_LOG(DMALLOC_NAMED_LOCATION(" " NAME))
 #define debug_malloc_touch(X) DMALLOC_TRACE_LOG(DMALLOC_LOCATION())
+#define debug_malloc_touch_named(X,NAME) DMALLOC_TRACE_LOG(DMALLOC_NAMED_LOCATION(" " NAME))
 #define debug_malloc_pass(X) (DMALLOC_TRACE_LOG(DMALLOC_LOCATION()),(X))
-#define dmalloc_touch(TYPE,X) (DMALLOC_TRACE_LOG(DMALLOC_LOCATION()),(X))
+#define debug_malloc_pass_named(X,NAME) (DMALLOC_TRACE_LOG(DMALLOC_NAMED_LOCATION(" " NAME)), (X))
+#define debug_malloc_touch_bt(X) DMALLOC_TRACE_LOG(DMALLOC_LOCATION())
+#define debug_malloc_touch_named_bt(X,NAME) DMALLOC_TRACE_LOG(DMALLOC_NAMED_LOCATION(" " NAME))
+#define dmalloc_touch(TYPE,X) debug_malloc_pass (X)
+#define dmalloc_touch_named(TYPE,X,NAME) debug_malloc_pass_named (X, NAME)
 #else /* DMALLOC_TRACE */
 #define debug_malloc_update_location(X,Y) (X)
 #define dmalloc_touch_svalue(X)
+#define dmalloc_touch_svalue_named(X,NAME)
 #define debug_malloc_touch(X)
+#define debug_malloc_touch_named(X,NAME)
 #define debug_malloc_pass(X) (X)
+#define debug_malloc_pass_named(X,NAME) (X)
+#define debug_malloc_touch_bt(X)
+#define debug_malloc_touch_named_bt(X,NAME)
 #define dmalloc_touch(TYPE,X) (X)
+#define dmalloc_touch_named(TYPE,X,NAME) (X)
 #endif /* !MALLOC_TRACE */
 
 #endif /* !DEBUG_MALLOC */
