@@ -2,11 +2,10 @@
 || This file is part of Pike. For copyright information see COPYRIGHT.
 || Pike is distributed under GPL, LGPL and MPL. See the file COPYING
 || for more information.
-|| $Id: docode.c,v 1.174 2004/01/22 23:17:27 nilsson Exp $
+|| $Id: docode.c,v 1.204 2008/08/28 15:13:19 grubba Exp $
 */
 
 #include "global.h"
-RCSID("$Id: docode.c,v 1.174 2004/01/22 23:17:27 nilsson Exp $");
 #include "las.h"
 #include "program.h"
 #include "pike_types.h"
@@ -18,17 +17,17 @@ RCSID("$Id: docode.c,v 1.174 2004/01/22 23:17:27 nilsson Exp $");
 #include "pike_error.h"
 #include "pike_memory.h"
 #include "svalue.h"
-#include "main.h"
+#include "pike_embed.h"
 #include "builtin_functions.h"
 #include "peep.h"
 #include "docode.h"
 #include "operators.h"
 #include "object.h"
 #include "opcodes.h"
-#include "language.h"
 #include "lex.h"
 #include "mapping.h"
 #include "multiset.h"
+#include "pike_compiler.h"
 
 static int do_docode2(node *n, int flags);
 
@@ -172,7 +171,6 @@ struct switch_data
 };
 
 static struct switch_data current_switch = {0, 0, 0, 0, 0, NULL, NULL};
-static int in_catch=0;
 
 void upd_int(int offset, INT32 tmp)
 {
@@ -190,6 +188,7 @@ int alloc_label(void) { return ++label_no; }
 
 int do_jump(int token,INT32 lbl)
 {
+  struct compilation *c = THIS_COMPILATION;
   if(lbl==-1) lbl=alloc_label();
   emit1(token, lbl);
   return lbl;
@@ -202,8 +201,9 @@ int do_jump(int token,INT32 lbl)
 
 static int lbl_cache[LBLCACHESIZE];
 
-int do_branch(INT32 lbl)
+static int do_branch(INT32 lbl)
 {
+  struct compilation *c = THIS_COMPILATION;
   if(lbl==-1)
   {
     lbl=alloc_label();
@@ -233,21 +233,33 @@ int do_branch(INT32 lbl)
   return lbl;
 }
 
-void low_insert_label(int lbl)
+static void low_insert_label(int lbl)
 {
+  struct compilation *c = THIS_COMPILATION;
   lbl_cache[ lbl % LBLCACHESIZE ] = CURRENT_INSTR;
   emit1(F_LABEL, lbl);
 }
 
-int ins_label(int lbl)
+static int ins_label(int lbl)
 {
   if(lbl==-1) lbl=alloc_label();
   low_insert_label(lbl);
   return lbl;
 }
 
+void modify_stack_depth(int delta)
+{
+  current_stack_depth += delta;
+#ifdef PIKE_DEBUG
+  if (current_stack_depth < 0) {
+    Pike_fatal("Popped out of virtual stack.\n");
+  }
+#endif
+}
+
 void do_pop(int x)
 {
+  struct compilation *c = THIS_COMPILATION;
 #ifdef PIKE_DEBUG
   if (x < 0) Pike_fatal("Cannot do pop of %d args.\n", x);
 #endif
@@ -260,24 +272,30 @@ void do_pop(int x)
   current_stack_depth -= x;
 }
 
-void do_pop_mark(void *ignored)
+static void do_pop_mark(void *ignored)
 {
+  struct compilation *c = THIS_COMPILATION;
   emit0(F_POP_MARK);
 }
 
-void do_pop_to_mark(void *ignored)
+static void do_pop_to_mark(void *ignored)
 {
+  struct compilation *c = THIS_COMPILATION;
   emit0(F_POP_TO_MARK);
 }
 
-void do_cleanup_synch_mark(void)
+#ifdef PIKE_DEBUG
+static void do_cleanup_synch_mark(void)
 {
+  struct compilation *c = THIS_COMPILATION;
   if (d_flag > 2)
     emit0(F_CLEANUP_SYNCH_MARK);
 }
+#endif
 
-void do_escape_catch(void)
+static void do_escape_catch(void)
 {
+  struct compilation *c = THIS_COMPILATION;
   emit0(F_ESCAPE_CATCH);
 }
 
@@ -287,16 +305,17 @@ int do_docode(node *n, int flags)
 {
   int i;
   int stack_depth_save = current_stack_depth;
-  int save_current_line=lex.current_line;
+  struct compilation *c = THIS_COMPILATION;
+  int save_current_line = c->lex.current_line;
   if(!n) return 0;
-  lex.current_line=n->line_number;
+  c->lex.current_line=n->line_number;
 #ifdef PIKE_DEBUG
   if (current_stack_depth == -4711) Pike_fatal("do_docode() used outside docode().\n");
 #endif
-  i=do_docode2(check_node_hash(n), flags);
+  i=do_docode2(n, flags);
   current_stack_depth = stack_depth_save + i;
 
-  lex.current_line=save_current_line;
+  c->lex.current_line=save_current_line;
   return i;
 }
 
@@ -309,15 +328,15 @@ static int is_efun(node *n, c_fun fun)
 
 static void code_expression(node *n, int flags, char *err)
 {
-  switch(do_docode(check_node_hash(n), flags & ~DO_POP))
+  switch(do_docode(n, flags & ~DO_POP))
   {
   case 0: my_yyerror("Void expression for %s",err);
   case 1: return;
   case 2:
     Pike_fatal("Internal compiler error (%s), line %ld, file %s\n",
-	  err,
-	  (long)lex.current_line,
-	  lex.current_file?lex.current_file->str:"Unknown");
+	       err,
+	       (long)THIS_COMPILATION->lex.current_line,
+	       THIS_COMPILATION->lex.current_file->str);
   }
 }
 
@@ -415,6 +434,7 @@ static INT32 count_cases(node *n)
 
 int generate_call_function(node *n)
 {
+  struct compilation *c = THIS_COMPILATION;
   emit0(F_MARK);
   PUSH_CLEANUP_FRAME(do_pop_mark, 0);
   do_docode(CDR(n),DO_NOT_COPY);
@@ -423,7 +443,7 @@ int generate_call_function(node *n)
   return 1;
 }
 
-static inline struct compiler_frame *find_local_frame(INT32 depth)
+static INLINE struct compiler_frame *find_local_frame(INT32 depth)
 {
   struct compiler_frame *f=Pike_compiler->compiler_frame;
   while(--depth>=0) f=f->previous;
@@ -433,9 +453,10 @@ static inline struct compiler_frame *find_local_frame(INT32 depth)
 /* Emit code for a function call to the identifier reference #id,
  * with the arguments specified by args.
  */
-int do_lfun_call(int id, node *args)
+static int do_lfun_call(int id, node *args)
 {
 #if 1
+  struct compilation *c = THIS_COMPILATION;
   struct reference *ref =
     Pike_compiler->new_program->identifier_references + id;
 
@@ -447,7 +468,9 @@ int do_lfun_call(int id, node *args)
    *
    * * Check if id is an alternate reference to the current function.
    *
-   * * Check that the function isn't varargs.
+   * * Check that the function isn't varargs or contains scoped functions.
+   *
+   * * Check that the current function doesn't contain scoped functions.
    */
   if((Pike_compiler->compiler_frame->current_function_number >= 0) &&
      ((id == Pike_compiler->compiler_frame->current_function_number) ||
@@ -456,9 +479,10 @@ int do_lfun_call(int id, node *args)
 	Pike_compiler->new_program->
 	identifier_references[Pike_compiler->compiler_frame->
 			      current_function_number].identifier_offset))) &&
-     (!(Pike_compiler->new_program->
-	identifiers[ref->identifier_offset].identifier_flags &
-	IDENTIFIER_VARARGS)))
+     !(Pike_compiler->new_program->
+       identifiers[ref->identifier_offset].identifier_flags &
+       (IDENTIFIER_VARARGS|IDENTIFIER_SCOPE_USED)) &&
+     !(Pike_compiler->compiler_frame->lexical_scope & SCOPE_SCOPE_USED))
   {
     int n=count_args(args);
     if(n == Pike_compiler->compiler_frame->num_args)
@@ -512,6 +536,7 @@ int do_lfun_call(int id, node *args)
 static void emit_apply_builtin(char *func)
 {
   INT32 tmp1;
+  struct compilation *c = THIS_COMPILATION;
   struct pike_string *n1=make_shared_string(func);
   node *n=find_module_identifier(n1,0);
   free_string(n1);
@@ -538,6 +563,7 @@ static void emit_apply_builtin(char *func)
 static int do_encode_automap_arg_list(node *n,
 				      int flags)
 {
+  struct compilation *c = THIS_COMPILATION;
   int stack_depth_save = current_stack_depth;
   if(!n) return 0;
   switch(n->token)
@@ -575,6 +601,7 @@ static int do_encode_automap_arg_list(node *n,
 static void emit_builtin_svalue(char *func)
 {
   INT32 tmp1;
+  struct compilation *c = THIS_COMPILATION;
   struct pike_string *n1=make_shared_string(func);
   node *n=find_module_identifier(n1,0);
   free_string(n1);
@@ -595,8 +622,179 @@ static void emit_builtin_svalue(char *func)
   free_node(n);
 }
 
+static void emit_range (node *n DO_IF_DEBUG (COMMA int num_args))
+{
+  struct compilation *c = THIS_COMPILATION;
+  node *low = CADR (n), *high = CDDR (n);
+  int bound_types;		/* Got bogus gcc warning here. */
+
+  switch (low->token) {
+    case F_RANGE_FROM_BEG: bound_types = RANGE_LOW_FROM_BEG; break;
+    case F_RANGE_FROM_END: bound_types = RANGE_LOW_FROM_END; break;
+    case F_RANGE_OPEN:     bound_types = RANGE_LOW_OPEN; break;
+#ifdef PIKE_DEBUG
+    default:
+      Pike_fatal ("Unexpected node %d as range lower bound.\n", low->token);
+#endif
+  }
+
+  switch (high->token) {
+    case F_RANGE_FROM_BEG: bound_types |= RANGE_HIGH_FROM_BEG; break;
+    case F_RANGE_FROM_END: bound_types |= RANGE_HIGH_FROM_END; break;
+    case F_RANGE_OPEN:     bound_types |= RANGE_HIGH_OPEN; break;
+#ifdef PIKE_DEBUG
+    default:
+      Pike_fatal ("Unexpected node %d as range upper bound.\n", high->token);
+#endif
+  }
+
+#ifdef PIKE_DEBUG
+  {
+    int expected_args = 0;
+    switch (bound_types & (RANGE_LOW_OPEN|RANGE_HIGH_OPEN)) {
+      case 0:
+	expected_args = 2; break;
+      case RANGE_LOW_OPEN:
+      case RANGE_HIGH_OPEN:
+	expected_args = 1; break;
+      case RANGE_LOW_OPEN|RANGE_HIGH_OPEN:
+	expected_args = 0; break;
+    }
+    if (num_args != expected_args)
+      Pike_fatal ("Wrong number of args to o_range opcode. Expected %d, got %d.\n",
+		  expected_args, num_args);
+  }
+#endif
+
+  emit1 (F_RANGE, bound_types);
+}
+
+static void emit_multi_assign(node *vals, node *vars, int no)
+{
+  struct compilation *c = THIS_COMPILATION;
+  node *var;
+  node *val;
+  node **valp = my_get_arg(&vals, no);
+
+  if (!vars && (!valp || !*valp)) return;
+  if (!(vars && valp && (val = *valp))) {
+    yyerror("Argument count mismatch for multi-assignment.\n");
+    return;
+  }
+  
+  if (vars->token == F_LVALUE_LIST) {
+    var = CAR(vars);
+    vars = CDR(vars);
+  } else {
+    var = vars;
+    vars = NULL;
+  }
+
+  switch(var->token) {
+  case F_LOCAL:
+    if(var->u.integer.a >= 
+       find_local_frame(var->u.integer.b)->max_number_of_locals)
+      yyerror("Illegal to use local variable here.");
+
+    if(var->u.integer.b) goto normal_assign;
+
+    if (var->node_info & OPT_ASSIGNMENT) {
+      /* Initialize. */
+      emit0(F_CONST0);
+      emit1(F_ASSIGN_LOCAL_AND_POP, var->u.integer.a);
+    }
+    code_expression(val, 0, "RHS");
+    emit_multi_assign(vals, vars, no+1);
+    emit1(F_ASSIGN_LOCAL_AND_POP, var->u.integer.a );
+    break;
+
+    /* FIXME: Make special case for F_EXTERNAL */
+  case F_IDENTIFIER:
+    if(!IDENTIFIER_IS_VARIABLE( ID_FROM_INT(Pike_compiler->new_program,
+					    var->u.id.number)->identifier_flags))
+    {
+      yyerror("Cannot assign functions or constants.\n");
+    }else{
+      code_expression(val, 0, "RHS");
+      emit_multi_assign(vals, vars, no+1);
+      emit1(F_ASSIGN_GLOBAL_AND_POP, var->u.id.number);
+    }
+    break;
+
+  case F_GET_SET:
+    {
+      /* Check for the setter function. */
+      struct program_state *state = Pike_compiler;
+      int program_id = var->u.integer.a;
+      int level = 0;
+      while (state && (state->new_program->id != program_id)) {
+	state = state->previous;
+	level++;
+      }
+      if (!state) {
+	yyerror("Lost parent.");
+      } else {
+	struct reference *ref =
+	  PTR_FROM_INT(state->new_program, var->u.integer.b);
+	struct identifier *id =
+	  ID_FROM_PTR(state->new_program, ref);
+	struct inherit *inh =
+	  INHERIT_FROM_PTR(state->new_program, ref);
+	int f;
+#ifdef PIKE_DEBUG
+	if (!IDENTIFIER_IS_VARIABLE(id->identifier_flags) ||
+	    (id->run_time_type != PIKE_T_GET_SET)) {
+	  Pike_fatal("Not a getter/setter in a F_GET_SET node!\n"
+		     "  identifier_flags: 0x%08x\n"
+		     "  run_time_type; %s (%d)\n",
+		     id->identifier_flags,
+		     get_name_of_type(id->run_time_type),
+		     id->run_time_type);
+	}
+#endif /* PIKE_DEBUG */
+	f = id->func.gs_info.setter;
+	if (f == -1) {
+	  yywarning("Variable %S lacks a setter.", id->name);
+	} else if (!level) {
+	  f += inh->identifier_level;
+	  emit0(F_MARK);
+	  code_expression(val, 0, "RHS");
+	  emit_multi_assign(vals, vars, no+1);
+	  emit1(F_CALL_LFUN, f);
+	  emit0(F_POP_VALUE);
+	}
+      }
+    }
+    /* FALL_THROUGH */
+  case F_EXTERNAL:
+    /* Check that it is in this context */
+    if(Pike_compiler ->new_program->id == var->u.integer.a)
+    {
+      /* Check that it is a variable */
+      if(var->u.integer.b != IDREF_MAGIC_THIS &&
+	 IDENTIFIER_IS_VARIABLE( ID_FROM_INT(Pike_compiler->new_program, var->u.integer.b)->identifier_flags))
+      {
+	code_expression(val, 0, "RHS");
+	emit_multi_assign(vals, vars, no+1);
+	emit1(F_ASSIGN_GLOBAL_AND_POP, var->u.integer.b);
+	break;
+      }
+    }
+    /* fall through */
+
+  default:
+  normal_assign:
+    do_docode(var, DO_LVALUE);
+    if(do_docode(val, 0) != 1) yyerror("RHS is void!");
+    emit_multi_assign(vals, vars, no+1);
+    emit0(F_ASSIGN_AND_POP);
+    break;
+  }
+}
+
 static int do_docode2(node *n, int flags)
 {
+  struct compilation *c = THIS_COMPILATION;
   ptrdiff_t tmp1,tmp2,tmp3;
   int ret;
 
@@ -622,6 +820,7 @@ static int do_docode2(node *n, int flags)
       case F_ARG_LIST:
       case F_COMMA_EXPR:
       case F_EXTERNAL:
+      case F_GET_SET:
       case F_AUTO_MAP_MARKER:
 	  break;
       }
@@ -666,6 +865,7 @@ static int do_docode2(node *n, int flags)
     return 1;
       
   case F_EXTERNAL:
+  case F_GET_SET:
     {
       int level = 0;
       struct program_state *state = Pike_compiler;
@@ -676,45 +876,86 @@ static int do_docode2(node *n, int flags)
       if (!state) {
 	my_yyerror("Program parent %d lost during compiling.", n->u.integer.a);
 	emit1(F_NUMBER,0);
-	return 1;
-      }
-      if(level)
-      {
-	if(flags & WANT_LVALUE)
-	{
-	  if (n->u.integer.b == IDREF_MAGIC_THIS)
-	    yyerror ("Cannot assign to this.");
+      } else if (flags & WANT_LVALUE) {
+	if (n->u.integer.b == IDREF_MAGIC_THIS) {
+	  my_yyerror("this is not an lvalue.");
+	}
+	if (level) {
 	  emit2(F_EXTERNAL_LVALUE, n->u.integer.b, level);
-	  return 2;
-	}else{
-	  emit2(F_EXTERNAL, n->u.integer.b, level);
-	  return 1;
-	}
-      }else{
-	if(flags & WANT_LVALUE)
-	{
+	} else {
 	  emit1(F_GLOBAL_LVALUE, n->u.integer.b);
-	  return 2;
-	}else{
-	  if (n->u.integer.b == IDREF_MAGIC_THIS)
-	    emit1(F_THIS_OBJECT, 0);
-	  else {
-	    struct identifier *id = ID_FROM_INT(state->new_program,n->u.integer.b);
-	    if(IDENTIFIER_IS_FUNCTION(id->identifier_flags) &&
-	       id->identifier_flags & IDENTIFIER_HAS_BODY)
-	    {
-	      /* Only use this opcode when it's certain that the result
-	       * can't zero, i.e. when we know the function isn't just a
-	       * prototype. */
-	      emit1(F_LFUN, n->u.integer.b);
-	    }else{
-	      emit1(F_GLOBAL, n->u.integer.b);
-	    }
+	}
+	return 2;
+      } else {
+	struct reference *ref =
+	  PTR_FROM_INT(state->new_program, n->u.integer.b);
+	struct identifier *id =
+	  ID_FROM_PTR(state->new_program, ref);
+	if (n->token == F_GET_SET) {
+	  struct inherit *inh =
+	    INHERIT_FROM_PTR(state->new_program, ref);
+	  int f;
+#ifdef PIKE_DEBUG
+	  if (!IDENTIFIER_IS_VARIABLE(id->identifier_flags) ||
+	      (id->run_time_type != PIKE_T_GET_SET)) {
+	    Pike_fatal("Not a getter/setter in a F_GET_SET node!\n"
+		       "  identifier_flags: 0x%08x\n"
+		       "  run_time_type; %s (%d)\n",
+		     id->identifier_flags,
+		       get_name_of_type(id->run_time_type),
+		       id->run_time_type);
 	  }
-	  return 1;
+#endif /* PIKE_DEBUG */
+	  f = id->func.gs_info.getter;
+	  if (f == -1) {
+	    yywarning("Variable %S lacks a getter.", id->name);
+	  } else if (!level) {
+	    return do_lfun_call(f + inh->identifier_level, NULL);
+	  } else {
+	    /* FIXME: Support inlining for the parent case.
+	     *
+	     * do_call_external(n->u.integer.a, f + inh->identifier_level,
+	     *                  NULL);
+	     */
+	    emit2(F_EXTERNAL, n->u.integer.b, level);
+	  }
+	} else if (level) {
+	  emit2(F_EXTERNAL, n->u.integer.b, level);
+	} else if (n->u.integer.b == IDREF_MAGIC_THIS) {
+	  emit1(F_THIS_OBJECT, 0);
+	} else if(IDENTIFIER_IS_FUNCTION(id->identifier_flags) &&
+		  id->identifier_flags & IDENTIFIER_HAS_BODY)
+	{
+	  /* Only use this opcode when it's certain that the result
+	   * can't zero, i.e. when we know the function isn't just a
+	   * prototype. */
+	  emit1(F_LFUN, n->u.integer.b);
+	}else{
+	  emit1(F_GLOBAL, n->u.integer.b);
 	}
       }
     }
+    return 1;
+
+  case F_THIS:
+    {
+      int level = 0;
+      struct program_state *state = Pike_compiler;
+      int inh = n->u.integer.b;
+      while (state && (state->new_program->id != n->u.integer.a)) {
+	state = state->previous;
+	level++;
+      }
+      if (!state) {
+	my_yyerror("Program parent %d lost during compiling.", n->u.integer.a);
+	emit1(F_NUMBER,0);
+      } else if (!level && (inh < 0)) {
+	emit1(F_THIS_OBJECT, 0);
+      } else {
+	emit2(F_THIS, level, inh);
+      }
+      return 1;
+    }
     break;
 
   case F_UNDEFINED:
@@ -739,6 +980,17 @@ static int do_docode2(node *n, int flags)
     return 0;
   }
 
+  case F_APPEND_ARRAY: {
+    emit0(F_MARK);
+    PUSH_CLEANUP_FRAME(do_pop_mark, 0);
+    do_docode(CAR(n),DO_LVALUE);
+    emit0(F_CONST0);	/* Reserved for svalue. */
+    do_docode(CDR(n),0);
+    emit0(F_APPEND_ARRAY);
+    POP_AND_DONT_CLEANUP;
+    return 1;
+  }
+
   case '?':
   {
     INT32 *prev_switch_jumptable = current_switch.jumptable;
@@ -922,6 +1174,27 @@ static int do_docode2(node *n, int flags)
       return 1;
     }
 
+  case F_MULTI_ASSIGN:
+    if (flags & DO_POP) {
+      emit_multi_assign(CAR(n), CDR(n), 0);
+      return 0;
+    } else {
+      /* Fall back to the normal assign case. */
+      tmp1=do_docode(CDR(n),DO_LVALUE);
+#ifdef PIKE_DEBUG
+      if(tmp1 & 1)
+	Pike_fatal("Very internal compiler error.\n");
+#endif
+      emit1(F_ARRAY_LVALUE, DO_NOT_WARN((INT32)(tmp1>>1)));
+      emit0(F_MARK);
+      PUSH_CLEANUP_FRAME(do_pop_mark, 0);
+      do_docode(CAR(n), 0);
+      emit_apply_builtin("aggregate");
+      POP_AND_DONT_CLEANUP;
+      emit0(F_ASSIGN);
+      return 1;
+    }
+
   case F_ASSIGN:
     switch(CAR(n)->token)
     {
@@ -938,6 +1211,7 @@ static int do_docode2(node *n, int flags)
     case F_MULTIPLY:
       if(node_is_eq(CDR(n),CAAR(n)))
       {
+	int num_args;
 	tmp1=do_docode(CDR(n),DO_LVALUE);
 	if(match_types(CDR(n)->type, array_type_string) ||
 	   match_types(CDR(n)->type, string_type_string) ||
@@ -945,8 +1219,10 @@ static int do_docode2(node *n, int flags)
 	   match_types(CDR(n)->type, multiset_type_string) ||
 	   match_types(CDR(n)->type, mapping_type_string))
 	{
-	  switch(do_docode(check_node_hash(CDAR(n)), 0))
+	  num_args = do_docode(CDAR(n), 0);
+	  switch (num_args)
 	  {
+	    case 0: emit0(F_LTOSVAL1); break;
 	    case 1: emit0(F_LTOSVAL2); break;
 	    case 2: emit0(F_LTOSVAL3); break;
 #ifdef PIKE_DEBUG
@@ -956,15 +1232,50 @@ static int do_docode2(node *n, int flags)
 	  }
 	}else{
 	  emit0(F_LTOSVAL);
-	  do_docode(check_node_hash(CDAR(n)), 0);
+	  num_args = do_docode(CDAR(n), 0);
 	}
 
-	emit0(CAR(n)->token);
+	if (CAR (n)->token == F_RANGE)
+	  emit_range (CAR (n) DO_IF_DEBUG (COMMA num_args));
+	else
+	  emit0(CAR(n)->token);
 
 	emit0(n->token);
 	return n->token==F_ASSIGN;
       }
-
+    case F_APPLY:
+      if ((CAAR(n)->token == F_CONSTANT) &&
+	  (CAAR(n)->u.sval.type == T_FUNCTION) &&
+	  (CAAR(n)->u.sval.subtype == FUNCTION_BUILTIN) &&
+	  (CAAR(n)->u.sval.u.efun->function != f_map) &&
+	  (CAAR(n)->u.sval.u.efun->function != f_filter)) {
+	/* efuns typically don't access object variables. */
+	node *args = CDAR(n);
+	if (args) {
+	  node **arg = my_get_arg(&args, 0);
+	  if (arg && node_is_eq(CDR(n), *arg) &&
+	      !(args->tree_info & OPT_ASSIGNMENT)) {
+	    /* First arg is the lvalue.
+	     *
+	     * We optimize this to allow for destructive operations.
+	     */
+	    int no = 0;
+	    tmp1 = do_docode(CDR(n), DO_LVALUE);
+	    emit0(F_MARK_AND_CONST0);
+	    PUSH_CLEANUP_FRAME(do_pop_mark, 0);
+	    while ((arg = my_get_arg(&args, ++no)) && *arg) {
+	      do_docode(*arg, 0);
+	    }
+	    tmp1=store_constant(&CAAR(n)->u.sval,
+				!(CAAR(n)->tree_info & OPT_EXTERNAL_DEPEND),
+				CAAR(n)->name);
+	    emit1(F_LTOSVAL_CALL_BUILTIN_AND_ASSIGN, DO_NOT_WARN((INT32)tmp1));
+	    POP_AND_DONT_CLEANUP;
+	    return 1;
+	  }
+	}
+      }
+      /* FALL_THROUGH */
     default:
       switch(CDR(n)->token)
       {
@@ -975,6 +1286,11 @@ static int do_docode2(node *n, int flags)
 
 	if(CDR(n)->u.integer.b) goto normal_assign;
 
+	if (CDR(n)->node_info & OPT_ASSIGNMENT) {
+	  /* Initialize. */
+	  emit0(F_CONST0);
+	  emit1(F_ASSIGN_LOCAL_AND_POP, CDR(n)->u.integer.a);
+	}
 	code_expression(CAR(n), 0, "RHS");
 	emit1(flags & DO_POP ? F_ASSIGN_LOCAL_AND_POP:F_ASSIGN_LOCAL,
 	     CDR(n)->u.integer.a );
@@ -988,10 +1304,61 @@ static int do_docode2(node *n, int flags)
 	}else{
 	  code_expression(CAR(n), 0, "RHS");
 	  emit1(flags & DO_POP ? F_ASSIGN_GLOBAL_AND_POP:F_ASSIGN_GLOBAL,
-	       CDR(n)->u.id.number);
+		CDR(n)->u.id.number);
 	}
 	break;
 
+        case F_GET_SET:
+	  {
+	    /* Check for the setter function. */
+	    struct program_state *state = Pike_compiler;
+	    int program_id = CDR(n)->u.integer.a;
+	    int level = 0;
+	    while (state && (state->new_program->id != program_id)) {
+	      state = state->previous;
+	      level++;
+	    }
+	    if (!state) {
+	      yyerror("Lost parent.");
+	    } else {
+	      struct reference *ref =
+		PTR_FROM_INT(state->new_program, CDR(n)->u.integer.b);
+	      struct identifier *id =
+		ID_FROM_PTR(state->new_program, ref);
+	      struct inherit *inh =
+		INHERIT_FROM_PTR(state->new_program, ref);
+	      int f;
+#ifdef PIKE_DEBUG
+	      if (!IDENTIFIER_IS_VARIABLE(id->identifier_flags) ||
+		  (id->run_time_type != PIKE_T_GET_SET)) {
+		Pike_fatal("Not a getter/setter in a F_GET_SET node!\n"
+			   "  identifier_flags: 0x%08x\n"
+			   "  run_time_type; %s (%d)\n",
+			   id->identifier_flags,
+			   get_name_of_type(id->run_time_type),
+			   id->run_time_type);
+	      }
+#endif /* PIKE_DEBUG */
+	      f = id->func.gs_info.setter;
+	      if (f == -1) {
+		yywarning("Variable %S lacks a setter.", id->name);
+	      } else if (!level) {
+		f += inh->identifier_level;
+		if (flags & DO_POP) {
+		  emit0(F_MARK);
+		  code_expression(CAR(n), 0, "RHS");
+		} else {
+		  code_expression(CAR(n), 0, "RHS");
+		  emit0(F_MARK);
+		  emit0(F_DUP);
+		}
+		emit1(F_CALL_LFUN, f);
+		emit0(F_POP_VALUE);
+		return !(flags & DO_POP);
+	      }
+	    }
+	  }
+	  /* FALL_THROUGH */
 	case F_EXTERNAL:
 	  /* Check that it is in this context */
 	  if(Pike_compiler ->new_program->id == CDR(n)->u.integer.a)
@@ -1057,10 +1424,23 @@ static int do_docode2(node *n, int flags)
 
   case F_RANGE:
     tmp1=do_docode(CAR(n),DO_NOT_COPY_TOPLEVEL);
-    if(do_docode(CDR(n),DO_NOT_COPY)!=2)
-      Pike_fatal("Compiler internal error (at %ld).\n",(long)lex.current_line);
-    emit0(n->token);
-    return DO_NOT_WARN((INT32)tmp1);
+    {
+#ifdef PIKE_DEBUG
+      int num_args =
+#endif
+	do_docode (CDR (n), DO_NOT_COPY);
+      emit_range (n DO_IF_DEBUG (COMMA num_args));
+      return DO_NOT_WARN((INT32)tmp1);
+    }
+
+    /* The special bound type nodes are simply ignored when the
+     * arglist to the range operator is coded. emit_range looks at
+     * them later on instead. */
+  case F_RANGE_FROM_BEG:
+  case F_RANGE_FROM_END:
+    return do_docode (CAR (n), flags);
+  case F_RANGE_OPEN:
+    return 0;
 
   case F_INC:
   case F_POST_INC:
@@ -1279,20 +1659,21 @@ static int do_docode2(node *n, int flags)
 
     if(CAR(arr) && CAR(arr)->token==F_RANGE)
     {
-      node **a1=my_get_arg(&_CAR(arr),0);
-      node **a2=my_get_arg(&_CAR(arr),1);
-      if(a1 && a2 && a2[0]->token==F_CONSTANT &&
-	 a2[0]->u.sval.type==T_INT &&
-	 a2[0]->u.sval.u.integer==MAX_INT_TYPE &&
-	a1[0]->type == int_type_string)
+      node *range = CAR(arr);
+      node *low = CADR(range);
+      node *high = CDDR(range);
+      if(high->token == F_RANGE_OPEN &&
+	 low->token == F_RANGE_FROM_BEG &&
+	 match_types (low->type, int_type_string))
       {
 	/* Optimize foreach(x[start..],y). */
-	do_docode(CAR(arr),DO_NOT_COPY_TOPLEVEL);
-	do_docode(*a1,DO_NOT_COPY);
+	do_docode (CAR(range), DO_NOT_COPY_TOPLEVEL);
+	do_docode (CDR(arr), DO_NOT_COPY|DO_LVALUE);
+	do_docode (CAR(low), DO_NOT_COPY);
 	goto foreach_arg_pushed;
       }
     }
-    do_docode(CAR(n),DO_NOT_COPY);
+    do_docode(arr,DO_NOT_COPY);
     emit0(F_CONST0);
     current_stack_depth++;
   foreach_arg_pushed:
@@ -1397,11 +1778,12 @@ static int do_docode2(node *n, int flags)
     }
 
   case F_CAST:
-    if(n->type==void_type_string)
-    {
+    switch(n->type->type) {
+    case T_VOID:
       DO_CODE_BLOCK(CAR(n));
       return 0;
-    } else if (n->type == int_type_string) {
+    case T_INT:
+      /* FIXME: Integer range? */
       tmp1 = do_docode(CAR(n), 0);
       if(!tmp1)
 	emit0(F_CONST0);
@@ -1411,7 +1793,8 @@ static int do_docode2(node *n, int flags)
 	emit0(F_CAST_TO_INT);
       }
       return 1;
-    } else if (n->type == string_type_string) {
+    case T_STRING:
+      /* FIXME: String width? */
       tmp1 = do_docode(CAR(n), 0);
       if(!tmp1)
 	emit0(F_CONST0);
@@ -1419,13 +1802,15 @@ static int do_docode2(node *n, int flags)
 	do_pop(DO_NOT_WARN((INT32)(tmp1-1)));
       emit0(F_CAST_TO_STRING);
       return 1;
-    } else if (compile_type_to_runtime_type(n->type) == PIKE_T_MIXED) {
-      tmp1 = do_docode(CAR(n), 0);
-      if(!tmp1) 
-	emit0(F_CONST0);
-      else if(tmp1>1)
-	do_pop(DO_NOT_WARN((INT32)(tmp1-1)));
-      return 1;
+    default:
+      if (compile_type_to_runtime_type(n->type) == PIKE_T_MIXED) {
+	tmp1 = do_docode(CAR(n), 0);
+	if(!tmp1) 
+	  emit0(F_CONST0);
+	else if(tmp1>1)
+	  do_pop(DO_NOT_WARN((INT32)(tmp1-1)));
+	return 1;
+      }
     }
     {
       struct svalue sv;
@@ -1522,6 +1907,12 @@ static int do_docode2(node *n, int flags)
     {
       return do_lfun_call(CAR(n)->u.integer.b, CDR(n));
     }
+    else if(CAR(n)->token == F_GET_SET &&
+	    CAR(n)->u.integer.a == Pike_compiler->new_program->id &&
+	    CAR(n)->u.integer.b != IDREF_MAGIC_THIS)
+    {
+      return do_lfun_call(CAR(n)->u.integer.b, CDR(n));
+    }
     else if(CAR(n)->token == F_ARROW)
     {
       emit0(F_MARK);
@@ -1567,6 +1958,7 @@ static int do_docode2(node *n, int flags)
 
   case F_ARG_LIST:
   case F_COMMA_EXPR:
+  case ':':
     {
       node *root = n;
       node *parent = n->parent;
@@ -1769,7 +2161,7 @@ static int do_docode2(node *n, int flags)
 	      if (!match_types(case_val->type, current_switch.type)) {
 		yytype_error("Type mismatch in case.",
 			     current_switch.type, case_val->type, 0);
-	      } else if (lex.pragmas & ID_STRICT_TYPES) {
+	      } else if (c->lex.pragmas & ID_STRICT_TYPES) {
 		yytype_error("Type mismatch in case.",
 			     current_switch.type, case_val->type, YYTE_IS_WARNING);
 	      }
@@ -1840,7 +2232,7 @@ static int do_docode2(node *n, int flags)
 	for (lbl_name = label->name; lbl_name; lbl_name = lbl_name->next)
 	  if (lbl_name->str == name)
 	    goto label_found_1;
-      my_yyerror("No surrounding statement labeled '%s'.", name->str);
+      my_yyerror("No surrounding statement labeled %S.", name);
       return 0;
 
     label_found_1:
@@ -1902,11 +2294,11 @@ static int do_docode2(node *n, int flags)
       struct statement_label_name *lbl_name;
       for (lbl_name = label->name; lbl_name; lbl_name = lbl_name->next)
 	if (lbl_name->str == name.str) {
-	  INT32 save_line = lex.current_line;
-	  lex.current_line = name.line_number;
+	  INT32 save_line = c->lex.current_line;
+	  c->lex.current_line = name.line_number;
 	  my_yyerror("Duplicate nested labels, previous one on line %d.",
 		     lbl_name->line_number);
-	  lex.current_line = save_line;
+	  c->lex.current_line = save_line;
 	  goto label_check_done;
 	}
     }
@@ -1932,15 +2324,45 @@ static int do_docode2(node *n, int flags)
     return 0;
   }
 
-  case F_RETURN:
+  case F_RETURN: {
+    struct statement_label *p;
+    int in_catch = 0;
     do_docode(CAR(n),0);
+
+    /* Insert the appropriate number of F_ESCAPE_CATCH. The rest of
+     * the cleanup is handled wholesale in low_return et al.
+     * Alternatively we could handle this too in low_return and
+     * then allow tail recursion of these kind of returns too. */
+    for (p = current_label; p; p = p->prev) {
+      struct cleanup_frame *q;
+      for (q = p->cleanups; q; q = q->prev) {
+	if (q->cleanup == (cleanup_func) do_escape_catch) {
+	  in_catch = 1;
+	  do_escape_catch();
+	}
+#ifdef PIKE_DEBUG
+	/* Have to pop marks from F_SYNCH_MARK too if the debug level
+	 * is high enough to get them inserted, otherwise we'll get
+	 * false alarms from debug checks in e.g. POP_CATCH_CONTEXT. */
+	else if (d_flag > 2 &&
+		 q->cleanup == (cleanup_func) do_cleanup_synch_mark) {
+	  /* Use the ordinary pop mark instruction here since we know
+	   * the stack isn't in synch and we don't want debug checks
+	   * for that. */
+	  do_pop_mark (NULL);
+	}
+#endif
+      }
+    }
+
     emit0(in_catch ? F_VOLATILE_RETURN : F_RETURN);
     return 0;
+  }
 
   case F_SSCANF:
-    tmp1=do_docode(CAR(n),DO_NOT_COPY);
+    tmp1=do_docode(CDAR(n),DO_NOT_COPY);
     tmp2=do_docode(CDR(n),DO_NOT_COPY | DO_LVALUE);
-    emit1(F_SSCANF, DO_NOT_WARN((INT32)(tmp1+tmp2)));
+    emit2(F_SSCANF, DO_NOT_WARN((INT32)(tmp1+tmp2)), CAAR(n)->u.sval.u.integer);
     return 1;
 
   case F_CATCH: {
@@ -1949,7 +2371,7 @@ static int do_docode2(node *n, int flags)
     tmp1=do_jump(F_CATCH,-1);
     PUSH_CLEANUP_FRAME(do_escape_catch, 0);
 
-    /* Entry point called by eval_instruction() via o_catch(). */
+    /* Entry point called via catching_eval_instruction(). */
     emit0(F_ENTRY);
 
     PUSH_STATEMENT_LABEL;
@@ -1958,17 +2380,25 @@ static int do_docode2(node *n, int flags)
     if (TEST_COMPAT(7,0))
       current_label->continue_label = current_label->break_label;
 
-    in_catch++;
     DO_CODE_BLOCK(CAR(n));
-    in_catch--;
 
     ins_label(current_label->break_label);
     emit0(F_EXIT_CATCH);
     POP_STATEMENT_LABEL;
     current_switch.jumptable = prev_switch_jumptable;
+    do_branch (tmp1);
 
-    ins_label(DO_NOT_WARN((INT32)tmp1));
     current_stack_depth++;
+    /* Entry point called via catching_eval_instruction() after
+     * catching an error.
+     *
+     * NB: This is reached by subtracting ENTRY_PROLOGUE_SIZE
+     *     from the label below.
+     * NB: The label must be after the entry, since it may expand to code
+     *     that requires the entry code to have run.
+     */
+    emit0(F_ENTRY);
+    ins_label(DO_NOT_WARN((INT32)tmp1));
 
     POP_AND_DONT_CLEANUP;
     return 1;
@@ -2146,8 +2576,7 @@ static int do_docode2(node *n, int flags)
 	  switch(n->u.sval.type)
 	  {
 	    case T_ARRAY:
-	      array_fix_type_field(n->u.sval.u.array);
-	      if(n->u.sval.u.array -> type_field & BIT_COMPLEX)
+	      if(array_fix_type_field(n->u.sval.u.array) & BIT_COMPLEX)
 		emit0(F_COPY_VALUE);
 	      break;
 
@@ -2181,19 +2610,30 @@ static int do_docode2(node *n, int flags)
     {
       if(flags & WANT_LVALUE)
       {
-	emit2(F_LEXICAL_LOCAL_LVALUE,n->u.id.number,n->u.integer.b);
+	emit2(F_LEXICAL_LOCAL_LVALUE, n->u.integer.a, n->u.integer.b);
 	return 2;
       }else{
-	emit2(F_LEXICAL_LOCAL,n->u.id.number,n->u.integer.b);
+	emit2(F_LEXICAL_LOCAL, n->u.integer.a, n->u.integer.b);
 	return 1;
       }
     }else{
       if(flags & WANT_LVALUE)
       {
-	emit1(F_LOCAL_LVALUE,n->u.id.number);
+	if (n->node_info & OPT_ASSIGNMENT) {
+	  /* Initialize the variable. */
+	  emit0(F_CONST0);
+	  emit1(F_ASSIGN_LOCAL_AND_POP, n->u.integer.a);
+	}
+	emit1(F_LOCAL_LVALUE, n->u.integer.a);
 	return 2;
       }else{
-	emit1(F_LOCAL,n->u.id.number);
+	if (n->node_info & OPT_ASSIGNMENT) {
+	  /* Initialize the variable. */
+	  emit0(F_CONST0);
+	  emit1(F_ASSIGN_LOCAL, n->u.integer.a);
+	} else {
+	  emit1(F_LOCAL, n->u.integer.a);
+	}
 	return 1;
       }
     }
@@ -2263,6 +2703,7 @@ static int do_docode2(node *n, int flags)
 /* Used to generate code for functions. */
 INT32 do_code_block(node *n)
 {
+  struct compilation *c = THIS_COMPILATION;
   INT32 entry_point;
 #ifdef PIKE_DEBUG
   if (current_stack_depth != -4711) Pike_fatal("Reentrance in do_code_block().\n");
