@@ -2,7 +2,7 @@
 || This file is part of Pike. For copyright information see COPYRIGHT.
 || Pike is distributed under GPL, LGPL and MPL. See the file COPYING
 || for more information.
-|| $Id: error.c,v 1.125 2006/05/10 19:43:53 mast Exp $
+|| $Id: error.c,v 1.168 2009/07/17 14:32:34 grubba Exp $
 */
 
 #define NO_PIKE_SHORTHAND
@@ -14,6 +14,7 @@
 #include "stralloc.h"
 #include "builtin_functions.h"
 #include "array.h"
+#include "mapping.h"
 #include "object.h"
 #include "main.h"
 #include "builtin_functions.h"
@@ -23,11 +24,13 @@
 #include "threads.h"
 #include "gc.h"
 
-RCSID("$Id: error.c,v 1.125 2006/05/10 19:43:53 mast Exp $");
-
+/* __attribute__ only applies to function declarations, not
+   definitions, so we disable them here. */
 #undef ATTRIBUTE
 #define ATTRIBUTE(X)
 
+/* #define ERROR_DEBUG */
+
 PMOD_EXPORT const char msg_fatal_error[] =
   "%s:%d: Fatal error:\n";
 #ifdef PIKE_DEBUG
@@ -60,20 +63,28 @@ PMOD_EXPORT const char msg_div_by_zero[] =
  * Recoveries handling.
  */
 
-JMP_BUF *recoveries=0;
-
 #ifdef PIKE_DEBUG
+/* struct mapping *recovery_lookup = NULL; */
+
 PMOD_EXPORT void check_recovery_context(void)
 {
   char foo;
 #define TESTILITEST ((((char *)Pike_interpreter.recoveries)-((char *)&foo))*STACK_DIRECTION)
-  if(Pike_interpreter.recoveries && TESTILITEST > 0) {
+  if(Pike_interpreter.recoveries &&
+     Pike_interpreter.recoveries->on_stack &&
+     TESTILITEST > 0) {
     fprintf(stderr, "Recoveries is out biking (Pike_interpreter.recoveries=%p, Pike_sp=%p, %ld)!\n",
 	    Pike_interpreter.recoveries, &foo,
 	    DO_NOT_WARN((long)TESTILITEST));
-    fprintf(stderr, "Last recovery was added at %s\n",
-	    Pike_interpreter.recoveries->file);
-    Pike_fatal("Recoveries is out biking (Pike_interpreter.recoveries=%p, Pike_sp=%p, %ld)!\n",
+    fprintf(stderr, "Last recovery was added at %s:%d\n",
+	    Pike_interpreter.recoveries->file,
+	    Pike_interpreter.recoveries->line);
+    push_int((ptrdiff_t)Pike_interpreter.recoveries);
+/*     if ((s = low_mapping_lookup(recovery_lookup, Pike_sp-1))) { */
+/*       fprintf(stderr, "Try also looking at %s.\n", (char*)s->u.integer); */
+/*     } */
+    pop_stack();
+    Pike_fatal("Recoveries is out biking (Pike_interpreter.recoveries=%p, C sp=%p, %ld)!\n",
 	  Pike_interpreter.recoveries, &foo,
 	  DO_NOT_WARN((long)TESTILITEST));
   }
@@ -87,12 +98,13 @@ PMOD_EXPORT void pike_gdb_breakpoint(INT32 args)
   pop_n_elems(args);
 }
 
-PMOD_EXPORT JMP_BUF *init_recovery(JMP_BUF *r, size_t stack_pop_levels DEBUG_LINE_ARGS)
+PMOD_EXPORT JMP_BUF *init_recovery(JMP_BUF *r, size_t stack_pop_levels DEBUG_INIT_REC_ARGS)
 {
   check_recovery_context();
 #ifdef PIKE_DEBUG
   r->file=location;
   OED_FPRINTF((stderr, "init_recovery(%p) %s\n", r, location));
+  r->on_stack = on_stack;
 #endif
   r->frame_pointer=Pike_fp;
   r->stack_pointer=Pike_sp - stack_pop_levels - Pike_interpreter.evaluator_stack;
@@ -102,6 +114,14 @@ PMOD_EXPORT JMP_BUF *init_recovery(JMP_BUF *r, size_t stack_pop_levels DEBUG_LIN
   r->severity=THROW_ERROR;
   Pike_interpreter.recoveries=r;
   check_recovery_context();
+#ifdef PIKE_DEBUG
+/*   if (recovery_lookup && Pike_sp) { */
+/*     push_int((ptrdiff_t)r); */
+/*     push_int((ptrdiff_t)location); */
+/*     mapping_insert(recovery_lookup, Pike_sp-2, Pike_sp-1); */
+/*     pop_n_elems(2); */
+/*   } */
+#endif
   return r;
 }
 
@@ -115,20 +135,52 @@ PMOD_EXPORT DECLSPEC(noreturn) void pike_throw(void) ATTRIBUTE((noreturn))
   {
     while(Pike_interpreter.recoveries->onerror)
     {
+      while(Pike_fp != Pike_interpreter.recoveries->onerror->frame_pointer)
+      {
+#ifdef PIKE_DEBUG
+	if(!Pike_fp)
+	  Pike_fatal("Popped out of stack frames.\n");
+#endif
+	POP_PIKE_FRAME();
+      }
+
       (*Pike_interpreter.recoveries->onerror->func)(Pike_interpreter.recoveries->onerror->arg);
       Pike_interpreter.recoveries->onerror=Pike_interpreter.recoveries->onerror->previous;
     }
-    
-    Pike_interpreter.recoveries=Pike_interpreter.recoveries->previous;
+
+    {
+      JMP_BUF *prev_rec = Pike_interpreter.recoveries->previous;
+      if (Pike_interpreter.catch_ctx &&
+	  (&Pike_interpreter.catch_ctx->recovery ==
+	   Pike_interpreter.recoveries)) {
+	struct catch_context *cc = Pike_interpreter.catch_ctx;
+	Pike_interpreter.catch_ctx = cc->prev;
+	really_free_catch_context (cc);
+      }
+      Pike_interpreter.recoveries = prev_rec;
+    }
   }
 
   if(!Pike_interpreter.recoveries)
     Pike_fatal("No error recovery context.\n");
 
+  /* Don't pop the stack before the onerrors have run; they might need
+   * items from the stack.
+   */
+
+  while(Pike_interpreter.recoveries->onerror)
+  {
+    while(Pike_fp != Pike_interpreter.recoveries->onerror->frame_pointer)
+    {
 #ifdef PIKE_DEBUG
-  if(Pike_sp - Pike_interpreter.evaluator_stack < Pike_interpreter.recoveries->stack_pointer)
-    Pike_fatal("Stack error in error.\n");
+      if(!Pike_fp)
+	Pike_fatal("Popped out of stack frames.\n");
 #endif
+      POP_PIKE_FRAME();
+    }
+    (*Pike_interpreter.recoveries->onerror->func)(Pike_interpreter.recoveries->onerror->arg);
+    Pike_interpreter.recoveries->onerror=Pike_interpreter.recoveries->onerror->previous;
+  }
 
   while(Pike_fp != Pike_interpreter.recoveries->frame_pointer)
   {
@@ -139,15 +191,14 @@ PMOD_EXPORT DECLSPEC(noreturn) void pike_throw(void) ATTRIBUTE((noreturn))
     POP_PIKE_FRAME();
   }
 
+#ifdef PIKE_DEBUG
+  if(Pike_sp - Pike_interpreter.evaluator_stack < Pike_interpreter.recoveries->stack_pointer)
+    Pike_fatal("Stack error in error.\n");
+#endif
+
   pop_n_elems(Pike_sp - Pike_interpreter.evaluator_stack - Pike_interpreter.recoveries->stack_pointer);
   Pike_mark_sp = Pike_interpreter.mark_stack + Pike_interpreter.recoveries->mark_sp;
 
-  while(Pike_interpreter.recoveries->onerror)
-  {
-    (*Pike_interpreter.recoveries->onerror->func)(Pike_interpreter.recoveries->onerror->arg);
-    Pike_interpreter.recoveries->onerror=Pike_interpreter.recoveries->onerror->previous;
-  }
-
 #if defined(DEBUG_MALLOC) && defined(PIKE_DEBUG)
   /* This will tell us where the value was caught (I hope) */
   if(throw_value.type <= MAX_REF_TYPE)
@@ -162,7 +213,18 @@ PMOD_EXPORT DECLSPEC(noreturn) void pike_throw(void) ATTRIBUTE((noreturn))
   accept_unfinished_type_fields--;
 #endif
 
-  LOW_LONGJMP(Pike_interpreter.recoveries->recovery,1);
+  if (Pike_interpreter.catch_ctx &&
+      &Pike_interpreter.catch_ctx->recovery == Pike_interpreter.recoveries) {
+    /* This is a phony recovery made in F_CATCH that hasn't been passed
+     * to LOW_SETJMP. The real jmpbuf is in catching_eval_instruction. */
+#ifdef PIKE_DEBUG
+    if (!Pike_interpreter.catching_eval_jmpbuf)
+      Pike_fatal ("Got phony F_CATCH recovery but no catching_eval_jmpbuf.\n");
+#endif
+    LOW_LONGJMP (*Pike_interpreter.catching_eval_jmpbuf, 1);
+  }
+  else
+    LOW_LONGJMP(Pike_interpreter.recoveries->recovery,1);
 }
 
 PMOD_EXPORT void push_error(const char *description)
@@ -172,12 +234,7 @@ PMOD_EXPORT void push_error(const char *description)
   f_aggregate(2);
 }
 
-PMOD_EXPORT struct svalue throw_value = {
-  PIKE_T_INT, 0,
-#ifdef HAVE_UNION_INIT
-  {0}, /* Only to avoid warnings. */
-#endif
-};
+PMOD_EXPORT struct svalue throw_value = SVALUE_INIT_FREE;
 int throw_severity;
 static const char *in_error;
 
@@ -191,41 +248,187 @@ PMOD_EXPORT DECLSPEC(noreturn) void low_error(const char *buf) ATTRIBUTE((noretu
   pike_throw();  /* Hope someone is catching, or we will be out of balls. */
 }
 
-/* FIXME: NOTE: This function uses a static buffer.
- * Check sizes of arguments passed!
- */
-void DECLSPEC(noreturn) va_error(const char *fmt, va_list args) ATTRIBUTE((noreturn))
+PMOD_EXPORT void Pike_vsnprintf(char *str, size_t size,
+				const char *fmt, va_list args)
+{
+  size--;
+
+  do {
+    if(fmt[0]=='%') {
+
+      fmt++;
+
+      switch( fmt++[0] ) {
+
+      case 'O':
+	{
+	  dynamic_buffer buf;
+	  dynbuf_string s;
+	  char *ostr;
+	  struct svalue t = va_arg(args, struct svalue);
+	  init_buf(&buf);
+	  describe_svalue(&t,0,0);
+	  s=complex_free_buf(&buf);
+	  ostr=s.str;
+
+	  while( --size>0 && (s.len--)>0 )
+	    str++[0]=s.str++[0];
+
+	  free(ostr);
+	}
+	break;
+
+      case 'S':
+	{
+	  dynamic_buffer buf;
+	  dynbuf_string s;
+	  char *ostr;
+	  struct svalue t;
+
+	  t.type = PIKE_T_STRING;
+	  t.u.string = va_arg(args, struct pike_string *);
+
+	  init_buf(&buf);
+	  describe_svalue(&t,0,0);
+	  s=complex_free_buf(&buf);
+	  ostr=s.str;
+
+	  while( --size>0 && (s.len--)>0 )
+	    str++[0]=s.str++[0];
+
+	  free(ostr);
+	}
+	break;
+
+      case 's':
+	{
+	  char *from = va_arg(args, char *);
+	  while( --size>0 && from[0]!=0 )
+	    str++[0]=from++[0];
+	}
+	break;
+
+      case 'c':
+	{
+	  int c = (char)va_arg(args, int);
+	  sprintf(str, "%c", c);
+	  str++;
+	  size--;
+	}
+	break;
+
+      case 'd':
+	{
+	  char buf[12];
+	  int pos=0;
+	  sprintf(buf, "%d", va_arg(args, int));
+	  while( --size>0 && buf[pos]!=0 )
+	    str++[0]=buf[pos++];
+	}
+	break;
+
+      case 'x':
+	{
+	  char buf[12];
+	  int pos=0;
+	  sprintf(buf, "%x", va_arg(args, int));
+	  while( --size>0 && buf[pos]!=0 )
+	    str++[0]=buf[pos++];
+	}
+	break;
+
+      case '%':
+	str++[0]='%';
+	size--;
+	break;
+
+      default:
+	Pike_fatal("Unknown Pike_vsnprintf formatting char '%c'.\n",
+		   (fmt-1)[0]);
+      }
+    }
+    else {
+      str++[0]=fmt++[0];
+      size--;
+    }
+
+  } while(fmt[0] && size>0);
+
+  str[0]=0;
+  va_end(args);
+}
+
+
+PMOD_EXPORT void va_make_error (const char *fmt, va_list args)
+{
+  struct string_builder s;
+  init_string_builder(&s, 0);
+  string_builder_vsprintf(&s, fmt, args);
+  push_string(finish_string_builder(&s));
+  f_backtrace(0);
+  f_aggregate(2);
+}
+
+PMOD_EXPORT void DECLSPEC(noreturn) va_error(const char *fmt, va_list args)
+  ATTRIBUTE((noreturn))
 {
-  char buf[4096];
   SWAP_IN_THREAD_IF_REQUIRED();
   if(in_error)
   {
     const char *tmp=in_error;
-    in_error=0;
-    Pike_fatal("Recursive error() calls, original error: %s",tmp);
+    in_error = NULL;
+    Pike_fatal("Recursive Pike_error() calls, "
+	       "original error: %s, new error: %s", tmp, fmt);
   }
 
-  in_error=buf;
-  VSNPRINTF (buf, sizeof (buf), fmt, args);
+  in_error=fmt;
 
   if(!Pike_interpreter.recoveries)
   {
+    /* FIXME: Why not use Pike_fatal() here? */
 #ifdef PIKE_DEBUG
     if (d_flag) {
-      fprintf(stderr,"No error recovery context!\n%s",buf);
+      fprintf(stderr,"No error recovery context!\n%s", fmt);
       dump_backlog();
     }
 #endif
 
-    fprintf(stderr,"No error recovery context!\n%s",buf);
+    fprintf(stderr,"No error recovery context!\n%s", fmt);
     exit(99);
   }
 
-  low_error(buf);
+  va_make_error (fmt, args);
+  free_svalue(&throw_value);
+  move_svalue(&throw_value, --Pike_sp);
+  throw_severity = THROW_ERROR;
+  in_error = NULL;
+  pike_throw();	/* Hope someone is catching, or we will be out of balls. */
 }
 
-PMOD_EXPORT DECLSPEC(noreturn) void new_error(const char *name, const char *text, struct svalue *oldsp,
-	       INT32 args, const char *file, int line) ATTRIBUTE((noreturn))
+PMOD_EXPORT void make_error (const char *fmt, ...)
+/* Creates a general error container like Pike_error but doesn't throw
+ * it, instead it is left on the stack. */
+{
+  va_list args;
+  va_start (args,fmt);
+  va_make_error (fmt,args);
+  va_end (args);
+}
+
+PMOD_EXPORT DECLSPEC(noreturn) void Pike_error(const char *fmt,...) ATTRIBUTE((noreturn))
+{
+  va_list args;
+  va_start(args,fmt);
+  va_error(fmt,args);
+  va_end(args);
+}
+
+PMOD_EXPORT DECLSPEC(noreturn) void new_error(const char *name,
+					      const char *text,
+					      struct svalue *oldsp,
+					      INT32 args,
+					      const char *file,
+					      int line) ATTRIBUTE((noreturn))
 {
   int i;
 
@@ -292,6 +495,7 @@ PMOD_EXPORT DECLSPEC(noreturn) void new_error(const char *name, const char *text
 
   f_aggregate(2);
 
+  free_svalue(&throw_value);
   move_svalue (&throw_value, --Pike_sp);
   throw_severity=THROW_ERROR;
 
@@ -334,7 +538,7 @@ PMOD_EXPORT void exit_on_error(const void *msg)
     fprintf(stderr,"Attempting to dump raw error: (may fail)\n");
     init_buf(&save_buf);
     move_svalue (&thrown, &throw_value);
-    throw_value.type = PIKE_T_INT;
+    mark_free_svalue (&throw_value);
     describe_svalue(&thrown,0,0);
     free_svalue (&thrown);
     s=simple_free_buf(&save_buf);
@@ -374,52 +578,52 @@ PMOD_EXPORT void fatal_on_error(const void *msg)
   do_abort();
 }
 
-PMOD_EXPORT DECLSPEC(noreturn) void Pike_error(const char *fmt,...) ATTRIBUTE((noreturn,format (printf, 1, 2)))
+PMOD_EXPORT DECLSPEC(noreturn) void debug_va_fatal(const char *fmt, va_list args) ATTRIBUTE((noreturn))
 {
-  va_list args;
-  va_start(args,fmt);
-  va_error(fmt,args);
-  va_end(args);
-}
-
-PMOD_EXPORT DECLSPEC(noreturn) void debug_fatal(const char *fmt, ...) ATTRIBUTE((noreturn,format (printf, 1, 2)))
-{
-  va_list args;
   static int in_fatal = 0;
 
   /* fprintf(stderr, "Raw error: %s\n", fmt); */
 
-  va_start(args,fmt);
   /* Prevent double fatal. */
   if (in_fatal)
   {
-    (void)VFPRINTF(stderr, fmt, args);
+    if (fmt) (void)VFPRINTF(stderr, fmt, args);
     do_abort();
   }
 
   in_fatal = 1;
 #ifdef PIKE_DEBUG
   if (d_flag) {
-    (void)VFPRINTF(stderr, fmt, args);
+#ifdef HAVE_VA_COPY
+    if (fmt) {
+      va_list a;
+      va_copy (a, args);
+      (void)VFPRINTF(stderr, fmt, a);
+      va_end (a);
+    }
+#endif
     dump_backlog();
   }
 #endif
 
+  if (fmt) (void)VFPRINTF(stderr, fmt, args);
+
   if(Pike_in_gc)
-    fprintf(stderr,"Pike was in GC stage %d when this fatal occured:\n",Pike_in_gc);
+    fprintf(stderr,"Pike was in GC stage %d when this fatal occured.\n",Pike_in_gc);
   Pike_in_gc = GC_PASS_DISABLED;
 
-  (void)VFPRINTF(stderr, fmt, args);
-
   d_flag=Pike_interpreter.trace_level=0;
+
   if(Pike_sp && Pike_interpreter.evaluator_stack &&
      master_object && master_object->prog)
   {
     JMP_BUF jmp;
     struct callback_list saved_eval_cbs = evaluator_callbacks;
-    /* Don't want thread switches or any other evaluator stuff while
-     * we let the master describe the backtrace below. */
-    low_init_threads_disable();
+    /* Simulate threads_disabled to avoid thread switches or any other
+     * evaluator stuff while we let the master describe the backtrace
+     * below. Doing it the naughty way without going through
+     * init_threads_disable etc to avoid hanging on runaway locks. */
+    threads_disabled++;
     MEMSET (&evaluator_callbacks, 0, sizeof (evaluator_callbacks));
     if (SETJMP (jmp))
       fprintf(stderr,"Got exception when trying to describe backtrace.\n");
@@ -431,7 +635,7 @@ PMOD_EXPORT DECLSPEC(noreturn) void debug_fatal(const char *fmt, ...) ATTRIBUTE(
 	write_to_stderr(Pike_sp[-1].u.string->str, Pike_sp[-1].u.string->len);
     }
     UNSETJMP (jmp);
-    exit_threads_disable (NULL);
+    threads_disabled--;
     evaluator_callbacks = saved_eval_cbs;
   }else{
     fprintf(stderr,"No stack - no backtrace.\n");
@@ -440,6 +644,14 @@ PMOD_EXPORT DECLSPEC(noreturn) void debug_fatal(const char *fmt, ...) ATTRIBUTE(
   do_abort();
 }
 
+PMOD_EXPORT DECLSPEC(noreturn) void debug_fatal(const char *fmt, ...) ATTRIBUTE((noreturn))
+{
+  va_list args;
+  va_start(args,fmt);
+  debug_va_fatal (fmt, args);
+  va_end (args);
+}
+
 #if 1
 
 /*! @class MasterObject
@@ -481,6 +693,26 @@ PMOD_EXPORT DECLSPEC(noreturn) void debug_fatal(const char *fmt, ...) ATTRIBUTE(
  *! Class for exception objects for errors of unspecified type.
  */
 
+/*! @decl string error_message
+ *!
+ *! The error message. It always ends with a newline (@expr{'\n'@})
+ *! character and it might be more than one line.
+ *!
+ *! Code that catches and rethrows errors may extend this with more
+ *! error information.
+ */
+
+/*! @decl array(backtrace_frame|array(mixed)) error_backtrace
+ *!
+ *! The backtrace as returned by @[backtrace] where the error
+ *! occurred.
+ *!
+ *! Code that catches and rethrows errors should ensure that this
+ *! remains the same in the rethrown error.
+ */
+
+#define GENERIC_ERROR_THIS ((struct generic_error_struct *)CURRENT_STORAGE)
+
 #define ERR_DECLARE
 #include "errors.h"
 
@@ -490,7 +722,8 @@ PMOD_EXPORT DECLSPEC(noreturn) void debug_fatal(const char *fmt, ...) ATTRIBUTE(
  *!
  *! @note
  *!   The only supported type to cast to is @expr{"array"@}, which
- *!   generates and old-style error.
+ *!   generates an old-style error @expr{({@[message](),
+ *!   @[backtrace]()})@}.
  */
 static void f_error_cast(INT32 args)
 {
@@ -499,14 +732,8 @@ static void f_error_cast(INT32 args)
   if(!strncmp(s,"array",5))
   {
     pop_n_elems(args);
-    if(GENERIC_ERROR_THIS->error_message)
-      ref_push_string(GENERIC_ERROR_THIS->error_message);
-    else
-      push_int(0);
-    if(GENERIC_ERROR_THIS->error_backtrace)
-      ref_push_array(GENERIC_ERROR_THIS->error_backtrace);
-    else
-      push_int(0);
+    apply_current (generic_err_message_fun, 0);
+    apply_current (generic_err_backtrace_fun, 0);
     f_aggregate(2);
   }else{
     SIMPLE_BAD_ARG_ERROR("error->cast", 1, "the value \"array\"");
@@ -520,10 +747,9 @@ static void f_error_cast(INT32 args)
  *! Simulates an array
  *! @array
  *!   @elem string msg
- *!     Error message.
+ *!     Error message as returned by @[message].
  *!   @elem array backtrace
- *!     Backtrace as returned by @[backtrace()] from where
- *!     the error occurred.
+ *!     Backtrace as returned by @[backtrace].
  *! @endarray
  *!
  *! @note
@@ -541,17 +767,11 @@ static void f_error_index(INT32 args)
   {
     case 0:
       pop_n_elems(args);
-      if(GENERIC_ERROR_THIS->error_message)
-	ref_push_string(GENERIC_ERROR_THIS->error_message);
-      else
-	push_int(0);
+      apply_current (generic_err_message_fun, 0);
       break;
     case 1:
       pop_n_elems(args);
-      if(GENERIC_ERROR_THIS->error_backtrace)
-	ref_push_array(GENERIC_ERROR_THIS->error_backtrace);
-      else
-	push_int(0);
+      apply_current (generic_err_backtrace_fun, 0);
       break;
     default:
       index_error("error->`[]", Pike_sp-args, args, NULL, Pike_sp-args,
@@ -573,7 +793,13 @@ static void f_error_describe(INT32 args)
 
 /*! @decl string message()
  *!
- *! Return a readable message describing the error.
+ *! Return a readable message describing the error. Normally simply
+ *! returns @[error_message].
+ *!
+ *! If you override this function then you should ensure that
+ *! @[error_message] is included in the returned message, since there
+ *! might be code that catches your error objects, extends
+ *! @[error_message] with more info, and rethrows the error.
  */
 static void f_error_message(INT32 args)
 {
@@ -586,7 +812,8 @@ static void f_error_message(INT32 args)
 
 /*! @decl array backtrace()
  *!
- *! Return the backtrace where the error occurred.
+ *! Return the backtrace where the error occurred. Normally simply
+ *! returns @[error_backtrace].
  *!
  *! @seealso
  *!   @[predef::backtrace()]
@@ -604,44 +831,94 @@ static void f_error_backtrace(INT32 args)
  */
 static void f_error__sprintf(INT32 args)
 {
-  struct program *p = Pike_fp->current_object->prog;
-  /* FIXME: What about obscure overloadings? */
-  int i = find_identifier("error_type", p);
-  struct identifier *id = ID_FROM_INT(p, i);
   int mode = 0;
+
   if(args>0 && Pike_sp[-args].type == PIKE_T_INT)
     mode = Pike_sp[-args].u.integer;
   pop_n_elems(args);
+
   if(mode != 'O') {
     push_undefined();
     return;
   }
-  push_constant_text ("Error.");
-  push_svalue(&PROG_FROM_INT(p, i)->constants[id->func.offset].sval);
+
+  {
+    struct program *p = Pike_fp->current_object->prog;
+    int i;
+    struct pike_string *error_name_str;
+    /* error_name is the name the error program wants in printouts
+     * like this. To still let inheriting programs fall back to
+     * describe_program if they don't provide it, we make it private
+     * in the base error classes and allow us to see private
+     * identifiers here. */
+    MAKE_CONST_STRING (error_name_str, "error_name");
+    i = really_low_find_shared_string_identifier (error_name_str, p,
+						  SEE_PRIVATE);
+    if (i != -1)
+      push_svalue (&PROG_FROM_INT (p, i)->
+		   constants[ID_FROM_INT (p, i)->func.offset].sval);
+    else {
+      ref_push_program (p);
+      SAFE_APPLY_MASTER ("describe_program", 1);
+    }
+  }
+
   push_constant_text("(%O)");
   if(GENERIC_ERROR_THIS->error_message)
     ref_push_string(GENERIC_ERROR_THIS->error_message);
   else
     push_int(0);
   f_sprintf(2);
-  f_add(3);
+  f_add(2);
 }
 
-/*! @decl void create(string message)
+/*! @decl int(0..1) _is_type(string t)
+ *!
+ *! Claims that the error object is an array, for compatibility with
+ *! old style error handling code.
+ */
+static void f_error__is_type(INT32 args)
+{
+  struct pike_string *array_string;
+  int ret;
+  MAKE_CONST_STRING(array_string, "array");
+  if (args < 0) SIMPLE_TOO_FEW_ARGS_ERROR("_is_type", 1);
+  if (args > 1) SIMPLE_WRONG_NUM_ARGS_ERROR("_is_type", 1);
+  if (Pike_sp[-args].type != PIKE_T_STRING)
+    SIMPLE_ARG_TYPE_ERROR("_is_type", 1, "string");
+  ret = Pike_sp[-args].u.string == array_string;
+  pop_n_elems(args);
+  push_int(ret);
+}
+
+/*! @decl void create(string message, @
+ *!                   void|array(backtrace_frame|array(mixed)) backtrace)
  */
 static void f_error_create(INT32 args)
 {
   struct pike_string *msg;
-  get_all_args("create", args, "%W", &msg);
+  struct array *bt = NULL;
+
+  get_all_args("create", args, "%W.%A", &msg, &bt);
+
   do_free_string(GENERIC_ERROR_THIS->error_message);
   copy_shared_string(GENERIC_ERROR_THIS->error_message, msg);
-  f_backtrace(0);
-  push_int (0);
-  push_int (Pike_sp[-2].u.array->size-2);
-  o_range ();
-  assign_to_short_svalue ((union anything *)&GENERIC_ERROR_THIS->error_backtrace,
-			  PIKE_T_ARRAY, Pike_sp-1);
-  pop_n_elems(args+1);
+
+  if (bt) {
+    if (GENERIC_ERROR_THIS->error_backtrace)
+      free_array (GENERIC_ERROR_THIS->error_backtrace);
+    add_ref (GENERIC_ERROR_THIS->error_backtrace = bt);
+  }
+  else {
+    f_backtrace(0);
+    push_int (1);
+    o_range2 (RANGE_LOW_OPEN|RANGE_HIGH_FROM_END);
+    assign_to_short_svalue ((union anything *)&GENERIC_ERROR_THIS->error_backtrace,
+			    PIKE_T_ARRAY, Pike_sp-1);
+    pop_stack();
+  }
+
+  pop_n_elems(args);
 }
 
 /*! @endclass
@@ -661,13 +938,13 @@ static void f_error_create(INT32 args)
   struct object *o; \
   va_start(foo,desc); \
   ASSERT_THREAD_SWAPPED_IN(); \
-  o=low_clone(PIKE_CONCAT(FEL,_error_program)); \
-  DWERROR((stderr, "%s(): Throwing a " #FEL " error\n", func))
+  DWERROR((stderr, "%s(): Throwing a " #FEL " error\n", func)); \
+  o=fast_clone_object(PIKE_CONCAT(FEL,_error_program))
 
 #define ERROR_DONE(FOO) \
   PIKE_CONCAT(FOO,_error_va(o,func, \
-			      base_sp,  args, \
-			      desc,foo)); \
+			      base_sp, args, \
+			      desc, &foo)); \
   va_end(foo)
 
 #define ERROR_STRUCT(STRUCT,O) \
@@ -691,37 +968,59 @@ static void f_error_create(INT32 args)
   add_ref( ERROR_STRUCT(STRUCT,o)->X=X )
 
 
-DECLSPEC(noreturn) void generic_error_va(struct object *o,
-					 const char *func,
-					 struct svalue *base_sp,  int args,
-					 const char *fmt,
-					 va_list foo)
-     ATTRIBUTE((noreturn))
+/* This prepares the passed object o, which is assumed to inherit
+ * generic_error_program, and throws it:
+ *
+ * o  A backtrace is assigned to error_backtrace.
+ *
+ * o  If func is specified, a frame is constructed for at the end of
+ *    backtrace using it as function name and base_sp[0..args-1] as
+ *    arguments.
+ *
+ * o  If fmt is specified, an error message is created from it and
+ *    fmt_args using string_builder_vsprintf. (fmt_args is passed as a
+ *    va_list pointer to be able to pass NULL if fmt is NULL.)
+ */
+PMOD_EXPORT DECLSPEC(noreturn) void generic_error_va(
+  struct object *o, const char *func, const struct svalue *base_sp, int args,
+  const char *fmt, va_list *fmt_args)
 {
-  char buf[8192];
-  int i;
+  struct generic_error_struct *err =
+    (struct generic_error_struct *) get_storage (o, generic_error_program);
 
-  VSNPRINTF(buf, sizeof(buf), fmt, foo);
+#ifdef PIKE_DEBUG
+  if (!err)
+    Pike_fatal ("Got an object which doesn't inherit generic_error_program.\n");
+#endif
 
   if(in_error)
   {
     const char *tmp=in_error;
     in_error=0;
-    Pike_fatal("Recursive error() calls, original error: %s",tmp);
+    Pike_fatal("Recursive error() calls, original error: %s", tmp);
   }
-  in_error=buf;
+  in_error = fmt ? fmt : "no error message";
+
+  if (fmt) {
+    struct string_builder s;
+    init_string_builder(&s, 0);
+    string_builder_vsprintf(&s, fmt, *fmt_args);
 
 #if 0
-  if (!master_program) {
-    fprintf(stderr, "ERROR: %s\n", buf);
-  }
+    if (!master_program) {
+      fprintf(stderr, "ERROR: %s\n", s.s->str);
+    }
 #endif
 
-  ERROR_STRUCT(generic,o)->error_message=make_shared_string(buf);
+    if (err->error_message) free_string(err->error_message);
+    err->error_message = finish_string_builder(&s);
+  }
+
   f_backtrace(0);
 
   if(func)
   {
+    int i;
     push_int(0);
     push_int(0);
     push_text(func);
@@ -734,14 +1033,16 @@ DECLSPEC(noreturn) void generic_error_va(struct object *o,
   }
 
   if(Pike_sp[-1].type!=PIKE_T_ARRAY)
-    Pike_fatal("Error failed to generate a backtrace!\n");
+    Pike_fatal("f_backtrace failed to generate a backtrace!\n");
 
-  ERROR_STRUCT(generic,o)->error_backtrace=Pike_sp[-1].u.array;
+  if (err->error_backtrace) free_array(err->error_backtrace);
+  err->error_backtrace=Pike_sp[-1].u.array;
   Pike_sp--;
   dmalloc_touch_svalue(Pike_sp);
 
   free_svalue(& throw_value);
   throw_value.type=PIKE_T_OBJECT;
+  throw_value.subtype = 0;
   throw_value.u.object=o;
   throw_severity = THROW_ERROR;
   in_error=0;
@@ -752,11 +1053,17 @@ DECLSPEC(noreturn) void generic_error_va(struct object *o,
 }
 
 
+/* Throw a preallocated error object.
+ *
+ * NOTE: The object MUST NOT have been cloned by a plain low_clone()!
+ *       At least fast_clone_object() MUST have been used, or the object
+ *       data must have been properly initialized in some other way!
+ */
 PMOD_EXPORT DECLSPEC(noreturn) void throw_error_object(
   struct object *o,
   const char *func,
   struct svalue *base_sp,  int args,
-  const char *desc, ...) ATTRIBUTE((noreturn,format (printf, 5, 6)))
+  const char *desc, ...) ATTRIBUTE((noreturn))
 {
   va_list foo;
   va_start(foo,desc);
@@ -768,7 +1075,7 @@ PMOD_EXPORT DECLSPEC(noreturn) void throw_error_object(
 PMOD_EXPORT DECLSPEC(noreturn) void generic_error(
   const char *func,
   struct svalue *base_sp,  int args,
-  const char *desc, ...) ATTRIBUTE((noreturn,format (printf, 4, 5)))
+  const char *desc, ...) ATTRIBUTE((noreturn))
 {
   INIT_ERROR(generic);
   ERROR_DONE(generic);
@@ -779,7 +1086,7 @@ PMOD_EXPORT DECLSPEC(noreturn) void index_error(
   struct svalue *base_sp,  int args,
   struct svalue *value,
   struct svalue *index,
-  const char *desc, ...) ATTRIBUTE((noreturn,format (printf, 6, 7)))
+  const char *desc, ...) ATTRIBUTE((noreturn))
 {
   INIT_ERROR(index);
   ERROR_COPY_SVALUE(index, value);
@@ -793,7 +1100,7 @@ PMOD_EXPORT DECLSPEC(noreturn) void bad_arg_error(
   int which_argument,
   const char *expected_type,
   struct svalue *got_value,
-  const char *desc, ...)  ATTRIBUTE((noreturn,format (printf, 7, 8)))
+  const char *desc, ...)  ATTRIBUTE((noreturn))
 {
   INIT_ERROR(bad_argument);
   ERROR_COPY(bad_argument, which_argument);
@@ -803,7 +1110,7 @@ PMOD_EXPORT DECLSPEC(noreturn) void bad_arg_error(
     ERROR_STRUCT(bad_argument,o)->expected_type = NULL;
   ERROR_COPY_SVALUE(bad_argument, got_value);
   DWERROR((stderr, "%s():Bad arg %d (expected %s)\n",
-	   func, which_arg, expected_type));
+	   func, which_argument, expected_type));
   ERROR_DONE(generic);
 }
 
@@ -811,7 +1118,7 @@ PMOD_EXPORT DECLSPEC(noreturn) void math_error(
   const char *func,
   struct svalue *base_sp,  int args,
   struct svalue *number,
-  const char *desc, ...) ATTRIBUTE((noreturn,format (printf, 5, 6)))
+  const char *desc, ...) ATTRIBUTE((noreturn))
 {
   INIT_ERROR(math);
   if(number)
@@ -830,7 +1137,7 @@ PMOD_EXPORT DECLSPEC(noreturn) void resource_error(
   struct svalue *base_sp,  int args,
   const char *resource_type,
   size_t howmuch_,
-  const char *desc, ...) ATTRIBUTE((noreturn,format (printf, 6, 7)))
+  const char *desc, ...) ATTRIBUTE((noreturn))
 {
   INT_TYPE howmuch = DO_NOT_WARN((INT_TYPE)howmuch_);
   INIT_ERROR(resource);
@@ -843,7 +1150,7 @@ PMOD_EXPORT DECLSPEC(noreturn) void permission_error(
   const char *func,
   struct svalue *base_sp, int args,
   const char *permission_type,
-  const char *desc, ...) ATTRIBUTE((noreturn, format(printf, 5, 6)))
+  const char *desc, ...) ATTRIBUTE((noreturn))
 {
   INIT_ERROR(permission);
   ERROR_STRUCT(permission,o)->permission_type=
@@ -856,12 +1163,12 @@ PMOD_EXPORT void wrong_number_of_args_error(const char *name, int args, int expe
   if(expected>args)
   {
     bad_arg_error (name, Pike_sp-args, args, expected, NULL, NULL,
-		   "Too few arguments to %s(). Expected at least %d.\n",
-		   name, expected);
+		   "Too few arguments to %s(). Expected at least %d (got %d).\n",
+		   name, expected, args);
   }else {
     bad_arg_error (name, Pike_sp-args, args, expected, NULL, NULL,
-		   "Too many arguments to %s(). Expected at most %d.\n",
-		   name, expected);
+		   "Too many arguments to %s(). Expected at most %d (got %d).\n",
+		   name, expected, args);
   }
 }
 
@@ -879,11 +1186,16 @@ void init_error(void)
 
 #ifdef PIKE_DEBUG
   dmalloc_accept_leak(add_gc_callback(gc_check_throw_value,0,0));
+
+/*   recovery_lookup = allocate_mapping(100); */
 #endif
 }
 
 void cleanup_error(void)
 {
+#ifdef PIKE_DEBUG
+/*   free_mapping(recovery_lookup); */
+#endif
 #define ERR_CLEANUP
 #include "errors.h"
 }
