@@ -2,7 +2,7 @@
 || This file is part of Pike. For copyright information see COPYRIGHT.
 || Pike is distributed under GPL, LGPL and MPL. See the file COPYING
 || for more information.
-|| $Id: gc.h,v 1.117 2007/03/09 12:01:56 mast Exp $
+|| $Id: gc.h,v 1.136 2008/07/24 20:02:53 mast Exp $
 */
 
 #ifndef GC_H
@@ -67,12 +67,14 @@ extern double gc_average_slowness;
 #define PRINT_ALLOC_COUNT_TYPE "%lu"
 #endif
 
-extern INT32 num_objects;
+extern int num_objects, got_unlinked_things;
 extern ALLOC_COUNT_TYPE num_allocs, alloc_threshold;
 PMOD_EXPORT extern int Pike_in_gc;
 extern int gc_generation;
 extern int gc_trace, gc_debug;
+#ifdef CPU_TIME_MIGHT_NOT_BE_THREAD_LOCAL
 extern cpu_time_t auto_gc_time;
+#endif
 
 extern struct callback *gc_evaluator_callback;
 #ifdef PIKE_DEBUG
@@ -164,12 +166,12 @@ extern int gc_keep_markers;
   num_objects-- ;							\
 }while(0)
 
-struct gc_frame;
+struct gc_rec_frame;
 
 struct marker
 {
   struct marker *next;
-  struct gc_frame *frame;	/* Pointer into the cycle check stack. */
+  struct gc_rec_frame *frame;	/* Pointer to the cycle check rec frame. */
   void *data;
   INT32 refs;
   /* Internal references (both weak and nonweak). Increased during
@@ -207,8 +209,6 @@ struct marker
  * check pass. */
 #define GC_LIVE_OBJ		0x0010
 /* The thing is a live object. Set in the cycle check pass. */
-#define GC_LIVE_RECURSE		0x0020
-/* The thing is being recursed a second time to propagate GC_LIVE. */
 #define GC_GOT_DEAD_REF		0x0040
 /* 1.  The thing has lost all references but since it can't be freed
  *     inside the gc, it's been given an extra ref to stay around
@@ -226,13 +226,14 @@ struct marker
  * references but only weak external references. */
 #define GC_USER_1		0x0100
 #define GC_USER_2		0x0200
+#define GC_USER_3		0x0400
 /* Flags free for use in the gc callbacks for the specific data types.
  * E.g. multisets use these flags on the multiset_data blocks. */
 
 #define GC_PRETOUCHED		0x4000
 /* The thing has been visited by debug_gc_touch() in the pretouch pass. */
-#define GC_MIDDLETOUCHED	0x8000
-/* The thing has been visited by debug_gc_touch() in the middletouch pass. */
+#define GC_POSTTOUCHED		0x8000
+/* The thing has been visited by debug_gc_touch() in the posttouch pass. */
 
 #ifdef PIKE_DEBUG
 #define GC_IS_REFERENCED	0x00040000
@@ -251,23 +252,30 @@ struct marker
 /* The thing has been visited by gc_checked_as_weak(). */
 #define GC_WATCHED		0x01000000
 /* The thing has been set under watch by gc_watch(). */
-#define GC_CLEANUP_FREED	0x02000000
-/* The thing was freed by the cleanup code under the assumption that
- * references were lost. */
+#define GC_CLEANUP_LEAKED	0x02000000
+/* The thing was reported as leaked by the cleanup code. The extra
+ * unaccounted refs to it have been/will be freed there too. Thus
+ * later frees of this object are logged since the "legitimate" free
+ * might still take place so the thing runs out of refs too early. */
 #endif
 
-#ifdef PIKE_DEBUG
+PMOD_EXPORT struct marker *pmod_get_marker (void *p);
+PMOD_EXPORT struct marker *pmod_find_marker (void *p);
+
 #define get_marker debug_get_marker
 #define find_marker debug_find_marker
-#endif
 
 #include "block_alloc_h.h"
-PTR_HASH_ALLOC_FIXED_FILL_PAGES(marker, n/a)
+PTR_HASH_ALLOC_FIXED_FILL_PAGES(marker, n/a);
 
-#ifdef PIKE_DEBUG
 #undef get_marker
-#define get_marker(X) ((struct marker *) debug_malloc_pass(debug_get_marker(X)))
 #undef find_marker
+
+#ifdef DYNAMIC_MODULE
+#define get_marker(X) ((struct marker *) debug_malloc_pass(pmod_get_marker(X)))
+#define find_marker(X) ((struct marker *) debug_malloc_pass(pmod_find_marker(X)))
+#else
+#define get_marker(X) ((struct marker *) debug_malloc_pass(debug_get_marker(X)))
 #define find_marker(X) ((struct marker *) debug_malloc_pass(debug_find_marker(X)))
 #endif
 
@@ -275,7 +283,6 @@ extern size_t gc_ext_weak_refs;
 
 typedef void gc_cycle_check_cb (void *data, int weak);
 
-/* Prototypes begin here */
 struct gc_frame;
 struct callback *debug_add_gc_callback(callback_func call,
 				 void *arg,
@@ -301,7 +308,7 @@ void debug_describe_svalue(struct svalue *s);
 void gc_watch(void *a);
 void debug_gc_touch(void *a);
 PMOD_EXPORT int real_gc_check(void *a);
-int real_gc_check_weak(void *a);
+PMOD_EXPORT int real_gc_check_weak(void *a);
 void exit_gc(void);
 void locate_references(void *a);
 void debug_gc_add_extra_ref(void *a);
@@ -313,7 +320,7 @@ int gc_do_weak_free(void *a);
 void gc_delayed_free(void *a, int type);
 void debug_gc_mark_enqueue(queue_call call, void *data);
 int gc_mark(void *a);
-void gc_move_marker (void *old, void *new);
+void gc_move_marker (void *old_thing, void *new_thing);
 PMOD_EXPORT void gc_cycle_enqueue(gc_cycle_check_cb *checkfn, void *data, int weak);
 void gc_cycle_run_queue(void);
 int gc_cycle_push(void *x, struct marker *m, int weak);
@@ -322,6 +329,7 @@ void do_gc_recurse_short_svalue(union anything *u, int type);
 int gc_do_free(void *a);
 size_t do_gc(void *ignored, int explicit_call);
 void f__gc_status(INT32 args);
+void f_count_memory (INT32 args);
 void cleanup_gc(void);
 
 #if defined (PIKE_DEBUG) && defined (DEBUG_MALLOC)
@@ -335,6 +343,16 @@ void cleanup_gc(void);
 #define gc_check_weak(VP) \
   DMALLOC_TOUCH_MARKER(VP, real_gc_check_weak(debug_malloc_pass(VP)))
 
+/* An object is considered live if its program has the flag
+ * PROGRAM_LIVE_OBJ set. That flag gets set if:
+ * o  There's a destroy LFUN,
+ * o  a program event callback is set with pike_set_prog_event_callback,
+ * o  an exit callback is set with set_exit_callback, or
+ * o  any inherited program has PROGRAM_LIVE_OBJ set.
+ * */
+#define gc_object_is_live(OBJ) \
+  ((OBJ)->prog && (OBJ)->prog->flags & PROGRAM_LIVE_OBJ)
+
 #ifdef GC_MARK_DEBUG
 
 void gc_mark_enqueue (queue_call fn, void *data);
@@ -370,7 +388,7 @@ extern const char *gc_found_place;
     gc_found_in_type = orig_gc_found_in_type;				\
   } while (0)
 
-static inline int debug_gc_check (void *a, const char *place)
+static INLINE int debug_gc_check (void *a, const char *place)
 {
   int res;
   const char *orig_gc_found_place = gc_found_place;
@@ -380,7 +398,7 @@ static inline int debug_gc_check (void *a, const char *place)
   return res;
 }
 
-static inline int debug_gc_check_weak (void *a, const char *place)
+static INLINE int debug_gc_check_weak (void *a, const char *place)
 {
   int res;
   const char *orig_gc_found_place = gc_found_place;
@@ -458,9 +476,11 @@ static inline int debug_gc_check_weak (void *a, const char *place)
    gc_cycle_check_weak_short_svalue((U), (T)) : gc_mark_weak_short_svalue((U), (T)))
 
 #define GC_RECURSE_THING(V, T)						\
-  (DMALLOC_TOUCH_MARKER(V, Pike_in_gc == GC_PASS_CYCLE) ?		\
-   PIKE_CONCAT(gc_cycle_check_, T)(V, 0) :				\
-   PIKE_CONCAT3(gc_mark_, T, _as_referenced)(V))
+  (mc_pass ?								\
+   PIKE_CONCAT3 (visit_,T,_ref) (debug_malloc_pass (V), REF_TYPE_NORMAL) : \
+   (DMALLOC_TOUCH_MARKER(V, Pike_in_gc == GC_PASS_CYCLE) ?		\
+    PIKE_CONCAT(gc_cycle_check_, T)(V, 0) :				\
+    PIKE_CONCAT3(gc_mark_, T, _as_referenced)(V)))
 #define gc_recurse_array(V) GC_RECURSE_THING((V), array)
 #define gc_recurse_mapping(V) GC_RECURSE_THING((V), mapping)
 #define gc_recurse_multiset(V) GC_RECURSE_THING((V), multiset)
@@ -491,11 +511,10 @@ static inline int debug_gc_check_weak (void *a, const char *place)
 #define GC_PASS_MARK		200
 #define GC_PASS_CYCLE		250
 #define GC_PASS_ZAP_WEAK	260
-#define GC_PASS_MIDDLETOUCH	270
+#define GC_PASS_POSTTOUCH	270
 #define GC_PASS_FREE		300
 #define GC_PASS_KILL		400
 #define GC_PASS_DESTRUCT	500
-#define GC_PASS_POSTTOUCH	510
 
 #define GC_PASS_LOCATE -1
 #define GC_PASS_DISABLED -2
@@ -533,7 +552,7 @@ extern int gc_in_cycle_check;
   struct object *_thing_ = (X);						\
   struct marker *_m_ = get_marker(_thing_);				\
   if (!(_m_->flags & GC_MARKED)) {					\
-    if (_thing_->prog && FIND_LFUN(_thing_->prog, LFUN_DESTROY) != -1)	\
+    if (gc_object_is_live (_thing_))					\
       _m_->flags |= GC_LIVE|GC_LIVE_OBJ;				\
     DO_IF_DEBUG(							\
       if (gc_in_cycle_check)						\
@@ -548,4 +567,103 @@ extern int gc_in_cycle_check;
   }									\
 } while (0)
 
+/* Generic API to follow refs in arbitrary structures (experimental).
+ *
+ * The check/mark/cycle check callbacks in the gc might be converted
+ * to this. Global state is fair play by design; the intention is to
+ * use thread local storage when the time comes. */
+
+/* A visit_thing_fn is made for every type of refcounted block. An
+ * INT32 refcounter is assumed to be first in every block. The
+ * visit_thing_fn should call visit_ref exactly once for every
+ * refcounted ref inside src_thing, in a stable order. dst_thing is
+ * then the target of the ref, ref_type describes the type of the ref
+ * itself (REF_TYPE_*), visit_dst is the visit_thing_fn for the target
+ * block, and extra is an arbitrary value passed along to visit_dst.
+ *
+ * action identifies some action that the visit_thing_fn should take
+ * (VISIT_*). Also, visit_ref_cb is likely to get a return value that
+ * identifies an action that should be taken on the ref.
+ *
+ * visit_thing_fn's must be reentrant. visit_dst might be called
+ * immediately, queued and called later, or not called at all. */
+
+typedef void visit_thing_fn (void *src_thing, int action, void *extra);
+typedef void visit_ref_cb (void *dst_thing, int ref_type,
+			   visit_thing_fn *visit_dst, void *extra);
+PMOD_EXPORT extern visit_ref_cb *visit_ref;
+
+#define REF_TYPE_STRENGTH 0x03	/* Bits for normal/weak/strong. */
+#define REF_TYPE_NORMAL	0x00	/* Normal (nonweak and nonstrong) ref. */
+#define REF_TYPE_WEAK	0x01	/* Weak ref. */
+#define REF_TYPE_STRONG	0x02	/* Strong ref. Note restrictions above. */
+
+#define REF_TYPE_INTERNAL 0x04	/* "Internal" ref. */
+/* An "internal" ref is one that should be considered internal within
+ * a structure that is treated as one unit on the pike level. E.g. the
+ * refs from mappings and multisets to their data blocks are internal.
+ * In general, if the target block isn't one of the eight referenced
+ * pike types then the ref is internal. Internal refs must not make up
+ * a cycle. They don't count towards the lookahead distance in
+ * f_count_memory. */
+
+#define VISIT_NORMAL 0
+/* Alias for zero which indicates normal (no) action: Visit all refs
+ * to all (refcounted) blocks and do nothing else. */
+
+#define VISIT_COMPLEX_ONLY 0x01
+/* Bit flag. If this is set, only follow refs to blocks that might
+ * contain more refs. */
+
+#define VISIT_COUNT_BYTES 0x02
+/* Add the number of bytes allocated for the block to
+ * mc_counted_bytes, then visit the refs. Never combined with
+ * VISIT_COMPLEX_ONLY. */
+
+/* Map between type and visit function for the standard ref types. */
+PMOD_EXPORT extern visit_thing_fn *const visit_fn_from_type[MAX_REF_TYPE + 1];
+PMOD_EXPORT TYPE_T type_from_visit_fn (visit_thing_fn *fn);
+
+static INLINE void real_visit_short_svalue (const union anything *u, TYPE_T t,
+					    int ref_type)
+{
+  check_short_svalue (u, t);
+  if (t <= MAX_REF_TYPE)
+    visit_ref (u->ptr, ref_type, visit_fn_from_type[t], NULL);
+}
+#define visit_short_svalue(U, T, REF_TYPE) \
+  (real_visit_short_svalue (debug_malloc_pass ((U)->ptr), (T), (REF_TYPE)))
+
+#ifdef DEBUG_MALLOC
+static INLINE void dmalloc_visit_svalue (struct svalue *s,
+					 int ref_type, char *l)
+{
+  int t = s->type;
+  check_svalue (s);
+  dmalloc_check_svalue (s, l);
+  if (t <= MAX_REF_TYPE) {
+    if (t == PIKE_T_FUNCTION) visit_function (s, ref_type);
+    else visit_ref (s->u.ptr, ref_type, visit_fn_from_type[t], NULL);
+  }
+}
+#define visit_svalue(S, REF_TYPE) \
+  dmalloc_visit_svalue ((S), (REF_TYPE), DMALLOC_LOCATION())
+#else
+static INLINE void visit_svalue (struct svalue *s, int ref_type)
+{
+  int t = s->type;
+  check_svalue (s);
+  if (t <= MAX_REF_TYPE) {
+    if (t == PIKE_T_FUNCTION) visit_function (s, ref_type);
+    else visit_ref (s->u.ptr, ref_type, visit_fn_from_type[t], NULL);
+  }
+}
+#endif
+
+/* Memory counting */
+
+PMOD_EXPORT extern int mc_pass;
+PMOD_EXPORT extern size_t mc_counted_bytes;
+PMOD_EXPORT int mc_count_bytes (void *thing);
+
 #endif
