@@ -2,7 +2,7 @@
 || This file is part of Pike. For copyright information see COPYRIGHT.
 || Pike is distributed under GPL, LGPL and MPL. See the file COPYING
 || for more information.
-|| $Id: interpret.h,v 1.153 2007/03/28 21:54:37 marcus Exp $
+|| $Id: interpret.h,v 1.180 2009/04/01 20:26:37 mast Exp $
 */
 
 #ifndef INTERPRET_H
@@ -12,6 +12,19 @@
 #include "program.h"
 #include "pike_error.h"
 #include "object.h"
+#include "pike_rusage.h"
+
+struct catch_context
+{
+  struct catch_context *prev;
+  JMP_BUF recovery;
+  struct svalue *save_expendible;
+  PIKE_OPCODE_T *next_addr;
+  ptrdiff_t continue_reladdr;
+#ifdef PIKE_DEBUG
+  struct pike_frame *frame;
+#endif
+};
 
 struct Pike_interpreter {
   /* Swapped variables */
@@ -20,8 +33,6 @@ struct Pike_interpreter {
   struct svalue **mark_stack_pointer;
   struct svalue **mark_stack;
   struct pike_frame *frame_pointer;
-  int evaluator_stack_malloced;
-  int mark_stack_malloced;
   JMP_BUF *recoveries;
 #ifdef PIKE_THREADS
   struct thread_state *thread_state;
@@ -29,14 +40,18 @@ struct Pike_interpreter {
   char *stack_top;
   DO_IF_SECURITY(struct object *current_creds;)
 
+  struct catch_context *catch_ctx;
+  LOW_JMP_BUF *catching_eval_jmpbuf;
+
   int svalue_stack_margin;
   int c_stack_margin;
 
+  INT16 evaluator_stack_malloced;
+  INT16 mark_stack_malloced;
+
 #ifdef PROFILING
-#ifdef HAVE_GETHRTIME
-  long long accounted_time;
-  long long time_base;
-#endif
+  cpu_time_t accounted_time;	/** Time spent and accounted for so far. */
+  cpu_time_t unlocked_time;	/** Time spent unlocked so far. */
   char *stack_bottom;
 #endif
 
@@ -48,41 +63,42 @@ struct Pike_interpreter {
 #endif
 struct pike_frame
 {
-  INT32 refs; /* must be first */
-  INT32 args;			/* Actual number of arguments. */
-  unsigned INT16 fun;		/* Function number. */
-  INT16 num_locals;		/* Number of local variables. */
-  INT16 num_args;		/* Number of argument variables. */
-  unsigned INT16 flags;		/* PIKE_FRAME_* */
+  INT32 refs;/* must be first */
+  INT32 args;			/** Actual number of arguments. */
+  unsigned INT16 fun;		/** Function number. */
+  INT16 num_locals;		/** Number of local variables. */
+  INT16 num_args;		/** Number of argument variables. */
+  unsigned INT16 flags;		/** PIKE_FRAME_* */
   INT16 ident;
   struct pike_frame *next;
   struct pike_frame *scope;
-  PIKE_OPCODE_T *pc;		/* Address of current opcode. */
-  PIKE_OPCODE_T *return_addr;	/* Address of opcode to continue at after call. */
-  struct svalue *locals;	/* Start of local variables. */
+  PIKE_OPCODE_T *pc;		/** Address of current opcode. */
+  PIKE_OPCODE_T *return_addr;	/** Address of opcode to continue at after call. */
+  struct svalue *locals;	/** Start of local variables. */
 
-  /*  This is <= locals, and this is where the
+  /** This is <= locals, and this is where the
    * return value should go.
    */
   struct svalue *save_sp;
 
-  /* This tells us the current level of
-   * svalues on the stack that can be discarded once the
-   * current function is done with them
+  /**
+   * This tells us the current level of svalues on the stack that can
+   * be discarded once the current function is done with them
    */
   struct svalue *expendible;
   struct svalue **save_mark_sp;
   struct svalue **mark_sp_base;
   struct object *current_object;
+  struct program *current_program;	/* program containing the context. */
+  struct inherit *context;
+  char *current_storage;
 
   DO_IF_SECURITY(struct object *current_creds;)
-#if defined(PROFILING) && defined(HAVE_GETHRTIME)
-  long long children_base;
-  long long start_time;
-  INT32 self_time_base;
+#if defined(PROFILING)
+  cpu_time_t children_base;	/** Accounted time when the frame started. */
+  cpu_time_t start_time;	/** Adjusted time when thr frame started. */
+  cpu_time_t self_time_base;	/* ??? */
 #endif
-  struct inherit context;
-  char *current_storage;
 };
 
 #define PIKE_FRAME_RETURN_INTERNAL 1
@@ -106,9 +122,8 @@ extern struct op_2_f {
 
 #ifdef PIKE_DEBUG
 PMOD_EXPORT extern const char msg_stack_error[];
-#define debug_check_stack() do{if(Pike_sp<Pike_interpreter.evaluator_stack)Pike_fatal(msg_stack_error);}while(0)
+#define debug_check_stack() do{if(Pike_sp<Pike_interpreter.evaluator_stack)Pike_fatal("%s", msg_stack_error);}while(0)
 #define check__positive(X,Y) if((X)<0) Pike_fatal Y
-#include "pike_error.h"
 #else
 #define check__positive(X,Y)
 #define debug_check_stack() 
@@ -138,22 +153,27 @@ PMOD_EXPORT extern const char Pike_check_mark_stack_errmsg[];
 
 PMOD_EXPORT extern const char Pike_check_c_stack_errmsg[];
 
-#define check_c_stack(X) do {						\
-  ptrdiff_t x_= ((char *)&x_) +						\
-    STACK_DIRECTION * (Pike_interpreter.c_stack_margin + (X)) -		\
-    Pike_interpreter.stack_top ;					\
-  x_*=STACK_DIRECTION;							\
-  if(x_>0)								\
-    low_error(Pike_check_c_stack_errmsg);				\
+#define low_check_c_stack(MIN_BYTES, RUN_IF_LOW) do {			\
+    ptrdiff_t x_= (((char *)&x_) - Pike_interpreter.stack_top) +	\
+      STACK_DIRECTION * (MIN_BYTES);					\
+    x_*=STACK_DIRECTION;						\
+    if(x_>0) {RUN_IF_LOW;}						\
+  } while (0)
+
+
+#define check_c_stack(MIN_BYTES) do {					\
+    low_check_c_stack (Pike_interpreter.c_stack_margin + (MIN_BYTES), {	\
+	low_error(Pike_check_c_stack_errmsg);				\
+	/* Pike_fatal("C stack overflow: x_:%p &x_:%p top:%p margin:%p\n", \
+	   x_, &x_, Pike_interpreter.stack_top,				\
+	   Pike_interpreter.c_stack_margin + (MIN_BYTES)); */		\
+      });								\
   }while(0)
 
-#define fatal_check_c_stack(X) do { 			\
-    ptrdiff_t x_= 					\
-      ((char *)&x_) + STACK_DIRECTION * (X) - Pike_interpreter.stack_top ; \
-    x_*=STACK_DIRECTION;						\
-    if(x_>0) {								\
-      ((void (*)(const char*, ...))Pike_fatal)(Pike_check_c_stack_errmsg);	\
-    }									\
+#define fatal_check_c_stack(MIN_BYTES) do {				\
+    low_check_c_stack ((MIN_BYTES), {					\
+	((void (*)(const char*, ...))Pike_fatal)(Pike_check_c_stack_errmsg); \
+      });								\
   }while(0)
 
 
@@ -288,6 +308,8 @@ PMOD_EXPORT extern const char msg_pop_neg[];
     _sp_->type=PIKE_T_ARRAY;						\
   }while(0)
 
+#define push_empty_array() ref_push_array(&empty_array)
+
 #define push_multiset(L) do{						\
     struct multiset *_=(L);						\
     struct svalue *_sp_ = Pike_sp++;					\
@@ -309,6 +331,8 @@ PMOD_EXPORT extern const char msg_pop_neg[];
     _sp_->type=PIKE_T_STRING;						\
   }while(0)
 
+#define push_empty_string() ref_push_string(empty_pike_string)
+
 #define push_type_value(S) do{						\
     struct pike_type *_=(S);						\
     struct svalue *_sp_ = Pike_sp++;					\
@@ -323,6 +347,17 @@ PMOD_EXPORT extern const char msg_pop_neg[];
     debug_malloc_touch(_);						\
     _sp_->u.object=_;							\
     _sp_->type=PIKE_T_OBJECT;						\
+    _sp_->subtype = 0;							\
+  }while(0)
+
+#define push_object_inherit(O, INH_NUM) do {				\
+    struct object *_ = (O);						\
+    struct svalue *_sp_ = Pike_sp++;					\
+    int _inh_ = (INH_NUM);						\
+    debug_malloc_touch(_);						\
+    _sp_->u.object = _;							\
+    _sp_->type = PIKE_T_OBJECT;						\
+    _sp_->subtype = _inh_;						\
   }while(0)
 
 #define push_float(F) do{						\
@@ -348,6 +383,12 @@ PMOD_EXPORT extern const char msg_pop_neg[];
     _sp_->type=PIKE_T_STRING;						\
   }while(0)
 
+#define push_constant_string_code(STR, CODE) do{			\
+    struct pike_string *STR;						\
+    REF_MAKE_CONST_STRING_CODE (STR, CODE);				\
+    push_string (STR);							\
+  }while(0)
+
 #define push_function(OBJ, FUN) do {					\
     struct object *_=(OBJ);						\
     struct svalue *_sp_ = Pike_sp++;					\
@@ -416,6 +457,17 @@ PMOD_EXPORT extern const char msg_pop_neg[];
     add_ref(_);								\
     _sp_->u.object=_;							\
     _sp_->type=PIKE_T_OBJECT;						\
+    _sp_->subtype = 0;							\
+  }while(0)
+
+#define ref_push_object_inherit(O, INH_NUM) do{				\
+    struct object  *_ = (O);						\
+    struct svalue *_sp_ = Pike_sp++;					\
+    int _inh_ = (INH_NUM);						\
+    add_ref(_);								\
+    _sp_->u.object = _;							\
+    _sp_->type = PIKE_T_OBJECT;						\
+    _sp_->subtype = _inh_;						\
   }while(0)
 
 #define ref_push_function(OBJ, FUN) do {				\
@@ -471,59 +523,99 @@ PMOD_EXPORT extern const char msg_pop_neg[];
 /* A scope is any frame which may have malloced locals */
 #define free_pike_scope(F) do{ struct pike_frame *f_=(F); if(!sub_ref(f_)) really_free_pike_scope(f_); }while(0)
 
+/* Without fancy accounting stuff. This one can't assume there is an
+ * identifier corresponding to the frame (i.e. _fp_->ident might be
+ * bogus). */
+#define LOW_POP_PIKE_FRAME(_fp_) do {					\
+    struct pike_frame *tmp_=_fp_->next;					\
+    if(!sub_ref(_fp_))							\
+    {									\
+      really_free_pike_frame(_fp_);					\
+    }else{								\
+      ptrdiff_t num_expendible = _fp_->expendible - _fp_->locals;	\
+      DO_IF_DEBUG(							\
+	if( (_fp_->locals + _fp_->num_locals > Pike_sp) ||		\
+	    (Pike_sp < _fp_->expendible) ||				\
+	    (num_expendible < 0) || (num_expendible > _fp_->num_locals)) \
+	  Pike_fatal("Stack failure in POP_PIKE_FRAME %p+%d=%p %p %p!\n", \
+		     _fp_->locals, _fp_->num_locals,			\
+		     _fp_->locals+_fp_->num_locals,			\
+		     Pike_sp,_fp_->expendible));			\
+      debug_malloc_touch(_fp_);						\
+      if(num_expendible)						\
+      {									\
+	struct svalue *s=(struct svalue *)xalloc(sizeof(struct svalue)*	\
+						 num_expendible);	\
+	_fp_->num_locals = num_expendible;				\
+	assign_svalues_no_free(s, _fp_->locals, num_expendible,		\
+			       BIT_MIXED);				\
+	_fp_->locals=s;							\
+	_fp_->flags|=PIKE_FRAME_MALLOCED_LOCALS;			\
+      }else{								\
+	_fp_->locals=0;							\
+      }									\
+      _fp_->next=0;							\
+    }									\
+    Pike_fp=tmp_;							\
+  } while (0)
+
 #define POP_PIKE_FRAME() do {						\
   struct pike_frame *_fp_ = Pike_fp;					\
-  struct pike_frame *tmp_=_fp_->next;					\
   DO_IF_PROFILING({							\
       /* Time spent in this frame + children. */			\
       cpu_time_t time_passed =						\
-	gethrtime() - Pike_interpreter.time_base;			\
+	get_cpu_time() - Pike_interpreter.unlocked_time;		\
       /* Time spent in children to this frame. */			\
       cpu_time_t time_in_children;					\
       /* Time spent in just this frame. */				\
       cpu_time_t self_time;						\
       struct identifier *function;					\
+      DO_IF_PROFILING_DEBUG({						\
+	  fprintf(stderr, "%p}: Pop got %" PRINT_CPU_TIME		\
+		  " (%" PRINT_CPU_TIME ")"				\
+		  " %" PRINT_CPU_TIME " (%" PRINT_CPU_TIME ")\n",	\
+		  Pike_interpreter.thread_state, time_passed,		\
+		  _fp_->start_time, Pike_interpreter.accounted_time,	\
+		  _fp_->children_base);					\
+	});								\
       time_passed -= _fp_->start_time;					\
+      DO_IF_DEBUG(if (time_passed < 0) {				\
+		    Pike_fatal("Negative time_passed: %" PRINT_CPU_TIME	\
+			       " now: %" PRINT_CPU_TIME			\
+			       " unlocked_time: %" PRINT_CPU_TIME	\
+			       " start_time: %" PRINT_CPU_TIME		\
+			       "\n", time_passed, get_cpu_time(),	\
+			       Pike_interpreter.unlocked_time,		\
+			       _fp_->start_time);			\
+		  });							\
       time_in_children =						\
 	Pike_interpreter.accounted_time - _fp_->children_base;		\
+      DO_IF_DEBUG(if (time_in_children < 0) {				\
+		    Pike_fatal("Negative time_in_children: %"		\
+			       PRINT_CPU_TIME				\
+			       " accounted_time: %" PRINT_CPU_TIME	\
+			       " children_base: %" PRINT_CPU_TIME	\
+			       "\n", time_in_children,			\
+			       Pike_interpreter.accounted_time,		\
+			       _fp_->children_base);			\
+		  });							\
       self_time = time_passed - time_in_children;			\
+      DO_IF_DEBUG(if (self_time < 0) {					\
+		    Pike_fatal("Negative self_time: %" PRINT_CPU_TIME	\
+			       " time_passed: %" PRINT_CPU_TIME		\
+			       " time_in_children: %" PRINT_CPU_TIME	\
+			       "\n", self_time, time_passed,		\
+			       time_in_children);			\
+		  });							\
       Pike_interpreter.accounted_time += self_time;			\
-      /* FIXME: Can context.prog be NULL? */				\
-      function = _fp_->context.prog->identifiers + _fp_->ident;		\
+      /* FIXME: Can context->prog be NULL? */				\
+      function = _fp_->context->prog->identifiers + _fp_->ident;		\
       /* function->total_time =						\
 	 Pike_fp->self_time_base + time_passed; */			\
       function->total_time += time_passed;				\
       function->self_time += self_time;					\
     });									\
-  if(!sub_ref(_fp_))							\
-  {									\
-    really_free_pike_frame(_fp_);					\
-  }else{								\
-    ptrdiff_t num_expendible = _fp_->expendible - _fp_->locals;		\
-    DO_IF_DEBUG(							\
-      if( (_fp_->locals + _fp_->num_locals > Pike_sp) ||		\
-	  (Pike_sp < _fp_->expendible) ||				\
-	  (num_expendible < 0) || (num_expendible > _fp_->num_locals))	\
-	Pike_fatal("Stack failure in POP_PIKE_FRAME %p+%d=%p %p %p!\n",	\
-		   _fp_->locals, _fp_->num_locals,			\
-		   _fp_->locals+_fp_->num_locals,			\
-		   Pike_sp,_fp_->expendible));				\
-    debug_malloc_touch(_fp_);						\
-    if(num_expendible)							\
-    {									\
-      struct svalue *s=(struct svalue *)xalloc(sizeof(struct svalue)*	\
-					       num_expendible);		\
-      _fp_->num_locals = num_expendible;				\
-      assign_svalues_no_free(s, _fp_->locals, num_expendible,		\
-			     BIT_MIXED);				\
-      _fp_->locals=s;							\
-      _fp_->flags|=PIKE_FRAME_MALLOCED_LOCALS;				\
-    }else{								\
-      _fp_->locals=0;							\
-    }									\
-    _fp_->next=0;							\
-  }									\
-  Pike_fp=tmp_;								\
+  LOW_POP_PIKE_FRAME (_fp_);						\
  }while(0)
 
 
@@ -544,7 +636,11 @@ do{ \
     fun_=find_identifier(FUN,master_ob->prog); \
     master_cnt = master_ob->prog->id; \
   } \
-  apply_low(master_ob, fun_, ARGS); \
+  if (fun_ >= 0) { \
+    apply_low(master_ob, fun_, ARGS); \
+  } else { \
+    Pike_error("Cannot call undefined function \"%s\" in master.\n", FUN); \
+  } \
 }while(0)
 
 #define SAFE_APPLY_MASTER(FUN,ARGS) \
@@ -556,7 +652,7 @@ do{ \
     fun_=find_identifier(FUN,master_ob->prog); \
     master_cnt = master_ob->prog->id; \
   } \
-  safe_apply_low2(master_ob, fun_, ARGS, 1); \
+  safe_apply_low2(master_ob, fun_, ARGS, FUN); \
 }while(0)
 
 #define SAFE_APPLY_HANDLER(FUN, HANDLER, COMPAT, ARGS) do {	\
@@ -596,14 +692,36 @@ PMOD_EXPORT extern unsigned long evaluator_callback_calls;
   call_callback(& evaluator_callbacks, NULL); \
 }while(0) 
 
-#ifdef PIKE_DEBUG
-#define fast_check_threads_etc(X) do { \
-  static int div_; if(d_flag || !(div_++& ((1<<(X))-1))) check_threads_etc(); } while(0)
+extern int fast_check_threads_counter;
 
-#else
-#define fast_check_threads_etc(X) do { \
-  static int div_; if(!(div_++& ((1<<(X))-1))) check_threads_etc(); } while(0)
-#endif
+#define fast_check_threads_etc(X) do {					\
+    if (++fast_check_threads_counter >= (1 << (X))) {			\
+      fast_check_threads_counter = 0;					\
+      check_threads_etc();						\
+    }									\
+  } while(0)
+
+/* Used before any sort of pike level function call. This covers most
+ * code paths. An interval with magnitude 6 on a 2 GHz AMD 64 averages
+ * on 1000-3000 check_threads_etc calls per second in mixed code, but
+ * it can vary greatly - from 0 to 30000+ calls/sec. In a test case
+ * with about 22000 calls/sec, it took 0.042% of the cpu. */
+#define FAST_CHECK_THREADS_ON_CALL() do {				\
+    if (++fast_check_threads_counter >= (1 << 6)) {			\
+      fast_check_threads_counter = 0;					\
+      check_threads_etc();						\
+    }									\
+    else if (objects_to_destruct)					\
+      /* De facto pike semantics requires that freed objects are */	\
+      /* destructed before function calls. Otherwise done through */	\
+      /* evaluator_callbacks. */					\
+      destruct_objects_to_destruct_cb();				\
+  } while (0)
+
+/* Used before any sort of backward branch. This is only a safeguard
+ * for some corner cases with loops without calls - not relevant in
+ * ordinary code. */
+#define FAST_CHECK_THREADS_ON_BRANCH() fast_check_threads_etc (8)
 
 #include "block_alloc_h.h"
 /* Prototypes begin here */
@@ -611,6 +729,7 @@ void push_sp_mark(void);
 ptrdiff_t pop_sp_mark(void);
 void gc_mark_stack_external (struct pike_frame *frame,
 			     struct svalue *stack_p, struct svalue *stack);
+PMOD_EXPORT int low_init_interpreter(struct Pike_interpreter *interpreter);
 PMOD_EXPORT void init_interpreter(void);
 void lvalue_to_svalue_no_free(struct svalue *to,struct svalue *lval);
 PMOD_EXPORT void assign_lvalue(struct svalue *lval,struct svalue *from);
@@ -619,9 +738,11 @@ void print_return_value(void);
 void reset_evaluator(void);
 struct backlog;
 void dump_backlog(void);
-BLOCK_ALLOC(pike_frame,128)
+BLOCK_ALLOC (catch_context, 0);
+BLOCK_ALLOC(pike_frame,128);
 
 #ifdef PIKE_USE_MACHINE_CODE
+void call_check_threads_etc();
 #if defined(OPCODE_INLINE_BRANCH) || defined(INS_F_JUMP) || \
     defined(INS_F_JUMP_WITH_ARG) || defined(INS_F_JUMP_WITH_TWO_ARGS)
 void branch_check_threads_etc();
@@ -640,13 +761,14 @@ int low_mega_apply(enum apply_type type, INT32 args, void *arg1, void *arg2);
 void low_return(void);
 void low_return_pop(void);
 void unlink_previous_frame(void);
-void mega_apply(enum apply_type type, INT32 args, void *arg1, void *arg2);
+PMOD_EXPORT void mega_apply(enum apply_type type, INT32 args, void *arg1, void *arg2);
 PMOD_EXPORT void f_call_function(INT32 args);
 PMOD_EXPORT void call_handle_error(void);
 PMOD_EXPORT int apply_low_safe_and_stupid(struct object *o, INT32 offset);
-PMOD_EXPORT void safe_apply_low(struct object *o,int fun,int args);
-PMOD_EXPORT void safe_apply_low2(struct object *o,int fun,int args, int handle_errors);
-PMOD_EXPORT void safe_apply(struct object *o, const char *fun ,INT32 args);
+PMOD_EXPORT int safe_apply_low(struct object *o,int fun,int args);
+PMOD_EXPORT int safe_apply_low2(struct object *o,int fun,int args,
+				 const char *fun_name);
+PMOD_EXPORT int safe_apply(struct object *o, const char *fun ,INT32 args);
 PMOD_EXPORT int low_unsafe_apply_handler(const char *fun,
 					 struct object *handler,
 					 struct object *compat,
@@ -666,11 +788,13 @@ PMOD_EXPORT void apply_shared(struct object *o,
 		  int args);
 PMOD_EXPORT void apply(struct object *o, const char *fun, int args);
 PMOD_EXPORT void apply_svalue(struct svalue *s, INT32 args);
+PMOD_EXPORT void safe_apply_svalue (struct svalue *s, INT32 args, int handle_errors);
 PMOD_EXPORT void apply_external(int depth, int fun, INT32 args);
 PMOD_EXPORT void slow_check_stack(void);
 PMOD_EXPORT void custom_check_stack(ptrdiff_t amount, const char *fmt, ...)
   ATTRIBUTE((format (printf, 2, 3)));
 PMOD_EXPORT void cleanup_interpret(void);
+PMOD_EXPORT void low_cleanup_interpret(struct Pike_interpreter *interpreter);
 void really_clean_up_interpret(void);
 /* Prototypes end here */
 
@@ -678,12 +802,12 @@ void really_clean_up_interpret(void);
  * to avoid implicit declaration of mega_apply().
  */
 #ifdef __ECL
-static inline void apply_low(struct object *o, ptrdiff_t fun, INT32 args)
+static INLINE void apply_low(struct object *o, ptrdiff_t fun, INT32 args)
 {
   mega_apply(APPLY_LOW, args, (void*)o, (void*)fun);
 }
 
-static inline void strict_apply_svalue(struct svalue *sval, INT32 args)
+static INLINE void strict_apply_svalue(struct svalue *sval, INT32 args)
 {
   mega_apply(APPLY_SVALUE_STRICT, args, (void*)sval, 0);
 }
@@ -697,9 +821,19 @@ static inline void strict_apply_svalue(struct svalue *sval, INT32 args)
 
 #define apply_current(FUN, ARGS)			\
   apply_low(Pike_fp->current_object,			\
-	    (FUN) + Pike_fp->context.identifier_level,	\
+	    (FUN) + Pike_fp->context->identifier_level,	\
 	    (ARGS))
 
+#define safe_apply_current(FUN, ARGS)				\
+  safe_apply_low(Pike_fp->current_object,			\
+		 (FUN) + Pike_fp->context->identifier_level,	\
+		 (ARGS))
+
+#define safe_apply_current2(FUN, ARGS, FUNNAME)			\
+  safe_apply_low2(Pike_fp->current_object,			\
+		  (FUN) + Pike_fp->context->identifier_level,	\
+		  (ARGS), (FUNNAME))
+
 PMOD_EXPORT extern int d_flag; /* really in main.c */
 
 PMOD_EXPORT extern int Pike_stack_size;
