@@ -2,11 +2,10 @@
 || This file is part of Pike. For copyright information see COPYRIGHT.
 || Pike is distributed under GPL, LGPL and MPL. See the file COPYING
 || for more information.
-|| $Id: mapping.c,v 1.184 2007/03/09 12:01:56 mast Exp $
+|| $Id: mapping.c,v 1.213 2009/09/22 15:45:09 grubba Exp $
 */
 
 #include "global.h"
-RCSID("$Id: mapping.c,v 1.184 2007/03/09 12:01:56 mast Exp $");
 #include "main.h"
 #include "object.h"
 #include "mapping.h"
@@ -15,6 +14,7 @@ RCSID("$Id: mapping.c,v 1.184 2007/03/09 12:01:56 mast Exp $");
 #include "pike_macros.h"
 #include "pike_error.h"
 #include "pike_memory.h"
+#include "pike_types.h"
 #include "dynamic_buffer.h"
 #include "interpret.h"
 #include "las.h"
@@ -78,30 +78,33 @@ DO_IF_DEBUG(								\
 BLOCK_ALLOC_FILL_PAGES(mapping, 2)
 
 #ifndef PIKE_MAPPING_KEYPAIR_LOOP
+#define IF_ELSE_KEYPAIR_LOOP(X, Y)	Y
 #define FREE_KEYPAIR(md, k) do {	\
     k->next = md->free_list;		\
     md->free_list = k;			\
   } while(0)
 #else /* PIKE_MAPPING_KEYPAIR_LOOP */
+#define IF_ELSE_KEYPAIR_LOOP(X, Y)	X
 #define FREE_KEYPAIR(md, k) do {			\
     md->free_list--;					\
     if (k != md->free_list) {				\
       struct keypair **prev_;				\
       unsigned INT32 h_;				\
+      INT32 e;						\
       /* Move the last keypair to the new hole. */	\
       *k = *(md->free_list);				\
       h_ = k->hval % md->hashsize;			\
       prev_ = md->hash + h_;				\
       DO_IF_DEBUG(					\
 	if (!*prev_) {					\
-	  Pike_fatal("Node to move not found!\n");		\
+	  Pike_fatal("Node to move not found!\n");	\
 	}						\
       );						\
       while (*prev_ != md->free_list) {			\
 	prev_ = &((*prev_)->next);			\
         DO_IF_DEBUG(					\
 	  if (!*prev_) {				\
-	    Pike_fatal("Node to move not found!\n");		\
+	    Pike_fatal("Node to move not found!\n");	\
 	  }						\
 	);						\
       }							\
@@ -111,14 +114,15 @@ BLOCK_ALLOC_FILL_PAGES(mapping, 2)
 #endif /* !PIKE_MAPPING_KEYPAIR_LOOP */
 
 #ifdef PIKE_DEBUG
-/* This function checks that the type field isn't lacking any bits.
+
+/** This function checks that the type field isn't lacking any bits.
  * It is used for debugging purposes only.
  */
-static void check_mapping_type_fields(struct mapping *m)
+static void check_mapping_type_fields(const struct mapping *m)
 {
   INT32 e;
-  struct keypair *k=0,**prev;
-  struct mapping_data *md;
+  const struct keypair *k=0;
+  const struct mapping_data *md;
   TYPE_FIELD ind_types, val_types;
 
   ind_types=val_types=0;
@@ -126,7 +130,13 @@ static void check_mapping_type_fields(struct mapping *m)
   md = m->data;
   NEW_MAPPING_LOOP(md)
     {
+      if (k->val.type > MAX_TYPE)
+	Pike_fatal("Invalid mapping keypair value type: %s\n",
+		   get_name_of_type(k->val.type));
       val_types |= 1 << k->val.type;
+      if (k->ind.type > MAX_TYPE)
+	Pike_fatal("Invalid maping keypair index type: %s\n",
+		   get_name_of_type(k->ind.type));
       ind_types |= 1 << k->ind.type;
     }
 
@@ -139,15 +149,19 @@ static void check_mapping_type_fields(struct mapping *m)
 #endif
 
 static struct mapping_data empty_data =
-  { PIKE_CONSTANT_MEMOBJ_INIT(1), 1, 0,0,0,0,0,0, 0, 0,{0}};
+  { PIKE_CONSTANT_MEMOBJ_INIT(1), 1, 0,0,0,0,0,0, 0,
+    IF_ELSE_KEYPAIR_LOOP((struct keypair *)&empty_data.hash, 0), {0}};
 static struct mapping_data weak_ind_empty_data =
-  { PIKE_CONSTANT_MEMOBJ_INIT(1), 1, 0,0,0,0,0,0, MAPPING_WEAK_INDICES, 0,{0}};
+  { PIKE_CONSTANT_MEMOBJ_INIT(1), 1, 0,0,0,0,0,0, MAPPING_WEAK_INDICES,
+    IF_ELSE_KEYPAIR_LOOP((struct keypair *)&weak_ind_empty_data.hash, 0), {0}};
 static struct mapping_data weak_val_empty_data =
-  { PIKE_CONSTANT_MEMOBJ_INIT(1), 1, 0,0,0,0,0,0, MAPPING_WEAK_VALUES, 0,{0}};
+  { PIKE_CONSTANT_MEMOBJ_INIT(1), 1, 0,0,0,0,0,0, MAPPING_WEAK_VALUES,
+    IF_ELSE_KEYPAIR_LOOP((struct keypair *)&weak_val_empty_data.hash, 0), {0}};
 static struct mapping_data weak_both_empty_data =
-  { PIKE_CONSTANT_MEMOBJ_INIT(1), 1, 0,0,0,0,0,0, MAPPING_WEAK, 0,{0}};
+  { PIKE_CONSTANT_MEMOBJ_INIT(1), 1, 0,0,0,0,0,0, MAPPING_WEAK,
+    IF_ELSE_KEYPAIR_LOOP((struct keypair *)&weak_both_empty_data.hash, 0), {0}};
 
-/* This function allocates the hash table and svalue space for a mapping
+/** This function allocates the hash table and svalue space for a mapping
  * struct. The size is the max number of indices that can fit in the
  * allocated space.
  */
@@ -183,12 +197,12 @@ static void init_mapping(struct mapping *m,
     for(e=1;e<size;e++)
     {
       md->free_list[e-1].next = md->free_list + e;
-      md->free_list[e-1].ind.type=T_INT;
-      md->free_list[e-1].val.type=T_INT;
+      mark_free_svalue (&md->free_list[e-1].ind);
+      mark_free_svalue (&md->free_list[e-1].val);
     }
     md->free_list[e-1].next=0;
-    md->free_list[e-1].ind.type=T_INT;
-    md->free_list[e-1].val.type=T_INT;
+    mark_free_svalue (&md->free_list[e-1].ind);
+    mark_free_svalue (&md->free_list[e-1].val);
 #endif /* !PIKE_MAPPING_KEYPAIR_LOOP */
     md->ind_types = 0;
     md->val_types = 0;
@@ -213,8 +227,13 @@ static void init_mapping(struct mapping *m,
 #endif
 }
 
-/* This function allocates an empty mapping with initial room
+/** This function allocates an empty mapping with initial room
  * for 'size' values.
+ *
+ * @param size initial number of values
+ * @return the newly allocated mapping
+ * @see do_free_mapping
+ * @see free_mapping
  */
 PMOD_EXPORT struct mapping *debug_allocate_mapping(int size)
 {
@@ -262,10 +281,10 @@ PMOD_EXPORT void really_free_mapping_data(struct mapping_data *md)
 PMOD_EXPORT void do_free_mapping(struct mapping *m)
 {
   if (m)
-    free_mapping(m);
+    inl_free_mapping(m);
 }
 
-/* This function is used to rehash a mapping without loosing the internal
+/* This function is used to rehash a mapping without losing the internal
  * order in each hash chain. This is to prevent mappings from becoming
  * inefficient just after being rehashed.
  */
@@ -377,9 +396,13 @@ static void mapping_rehash_backwards_good(struct mapping_data *md,
   }
 }
 
-/* This function re-allocates a mapping. It adjusts the max no. of
+/** This function re-allocates a mapping. It adjusts the max no. of
  * values can be fitted into the mapping. It takes a bit of time to
  * run, but is used seldom enough not to degrade preformance significantly.
+ *
+ * @param m the mapping to be rehashed
+ * @param new_size new mappingsize
+ * @return the rehashed mapping
  */
 static struct mapping *rehash(struct mapping *m, int new_size)
 {
@@ -398,6 +421,8 @@ static struct mapping *rehash(struct mapping *m, int new_size)
   if(d_flag>1)  check_mapping(m);
 #endif
 
+  if (md->hashsize == new_size) return m;
+
   init_mapping(m, new_size, md->flags);
   debug_malloc_touch(m);
   new_md=m->data;
@@ -486,6 +511,8 @@ struct mapping_data *copy_mapping_data(struct mapping_data *md)
   nmd->valrefs=0;
   nmd->hardlinks=0;
 
+  /* FIXME: What about nmd->flags? */
+
   if(md->hardlinks)
   {
 #ifdef PIKE_DEBUG
@@ -621,8 +648,8 @@ struct mapping_data *copy_mapping_data(struct mapping_data *md)
 #define PREPARE_FOR_INDEX_CHANGE() \
  if(md->refs>1) COPYMAP()
 
-/* This function brings the type fields in the mapping up to speed.
- * I only use it when the type fields MUST be correct, which is not
+/** This function brings the type fields in the mapping up to speed.
+ * It should be used only when the type fields MUST be correct, which is not
  * very often.
  */
 PMOD_EXPORT void mapping_fix_type_field(struct mapping *m)
@@ -674,19 +701,10 @@ PMOD_EXPORT void mapping_set_flags(struct mapping *m, int flags)
   md->flags = flags;
 }
 
-
-/* This function inserts key:val into the mapping m.
- * Same as doing m[key]=val; in pike.
- *
- * overwrite:
- *   0: Do not replace the value if the entry exists.
- *   1: Replace the value if the entry exists.
- *   2: Replace both the index and the value if the entry exists.
- */
 PMOD_EXPORT void low_mapping_insert(struct mapping *m,
-			struct svalue *key,
-			struct svalue *val,
-			int overwrite)
+				    const struct svalue *key,
+				    const struct svalue *val,
+				    int overwrite)
 {
   unsigned INT32 h,h2;
   struct keypair *k, **prev;
@@ -795,8 +813,8 @@ PMOD_EXPORT void low_mapping_insert(struct mapping *m,
 }
 
 PMOD_EXPORT void mapping_insert(struct mapping *m,
-		    struct svalue *key,
-		    struct svalue *val)
+				const struct svalue *key,
+				const struct svalue *val)
 {
   low_mapping_insert(m,key,val,1);
 }
@@ -805,7 +823,7 @@ PMOD_EXPORT void mapping_insert(struct mapping *m,
 #define mapping_insert(M, KEY, VAL) low_mapping_insert ((M), (KEY), (VAL), 1)
 
 PMOD_EXPORT union anything *mapping_get_item_ptr(struct mapping *m,
-				     struct svalue *key,
+				     const struct svalue *key,
 				     TYPE_T t)
 {
   unsigned INT32 h, h2;
@@ -921,7 +939,7 @@ PMOD_EXPORT union anything *mapping_get_item_ptr(struct mapping *m,
 }
 
 PMOD_EXPORT void map_delete_no_free(struct mapping *m,
-			struct svalue *key,
+			const struct svalue *key,
 			struct svalue *to)
 {
   unsigned INT32 h,h2;
@@ -974,14 +992,14 @@ PMOD_EXPORT void map_delete_no_free(struct mapping *m,
   *prev=k->next;
   free_svalue(& k->ind);
   if(to)
-    to[0]=k->val;
+    move_svalue (to, &k->val);
   else
     free_svalue(& k->val);
 
   FREE_KEYPAIR(md, k);
 
-  md->free_list->ind.type=T_INT;
-  md->free_list->val.type=T_INT;
+  mark_free_svalue (&md->free_list->ind);
+  mark_free_svalue (&md->free_list->val);
 
   md->size--;
 #ifdef MAPPING_SIZE_DEBUG
@@ -994,7 +1012,7 @@ PMOD_EXPORT void map_delete_no_free(struct mapping *m,
     debug_malloc_touch(m);
     rehash(m, MAP_SLOTS(m->data->size));
   }
-  
+
 #ifdef PIKE_DEBUG
   if(d_flag>1)  check_mapping(m);
 #endif
@@ -1041,9 +1059,10 @@ PMOD_EXPORT void check_mapping_for_destruct(struct mapping *m)
 	  free_svalue(& k->ind);
 	  free_svalue(& k->val);
 	  FREE_KEYPAIR(md, k);
-	  md->free_list->ind.type = T_INT;
-	  md->free_list->val.type = T_INT;
+	  mark_free_svalue (&md->free_list->ind);
+	  mark_free_svalue (&md->free_list->val);
 	  md->size--;
+
 #ifdef MAPPING_SIZE_DEBUG
 	  if(m->data ==md)
 	  {
@@ -1076,7 +1095,7 @@ PMOD_EXPORT void check_mapping_for_destruct(struct mapping *m)
 }
 
 PMOD_EXPORT struct svalue *low_mapping_lookup(struct mapping *m,
-				  struct svalue *key)
+					      const struct svalue *key)
 {
   unsigned INT32 h,h2;
   struct keypair *k=0, **prev=0;
@@ -1102,7 +1121,7 @@ PMOD_EXPORT struct svalue *low_mapping_lookup(struct mapping *m,
 }
 
 PMOD_EXPORT struct svalue *low_mapping_string_lookup(struct mapping *m,
-					 struct pike_string *p)
+                                                     struct pike_string *p)
 {
   struct svalue tmp;
   tmp.type=T_STRING;
@@ -1111,8 +1130,8 @@ PMOD_EXPORT struct svalue *low_mapping_string_lookup(struct mapping *m,
 }
 
 PMOD_EXPORT void mapping_string_insert(struct mapping *m,
-			   struct pike_string *p,
-			   struct svalue *val)
+                                       struct pike_string *p,
+                                       const struct svalue *val)
 {
   struct svalue tmp;
   tmp.type=T_STRING;
@@ -1139,10 +1158,11 @@ PMOD_EXPORT struct svalue *simple_mapping_string_lookup(struct mapping *m,
   return 0;
 }
 
+/* Lookup in a mapping of mappings */
 PMOD_EXPORT struct svalue *mapping_mapping_lookup(struct mapping *m,
-				      struct svalue *key1,
-				      struct svalue *key2,
-				      int create)
+						  const struct svalue *key1,
+						  const struct svalue *key2,
+						  int create)
 {
   struct svalue tmp;
   struct mapping *m2;
@@ -1200,16 +1220,20 @@ PMOD_EXPORT struct svalue *mapping_mapping_string_lookup(struct mapping *m,
 
 PMOD_EXPORT void mapping_index_no_free(struct svalue *dest,
 			   struct mapping *m,
-			   struct svalue *key)
+			   const struct svalue *key)
 {
   struct svalue *p;
 
-  if((p=low_mapping_lookup(m,key)))
+  if(!IS_DESTRUCTED (key) && (p=low_mapping_lookup(m,key)))
   {
-    if(p->type==T_INT)
-      p->subtype=NUMBER_NUMBER;
-
     assign_svalue_no_free(dest, p);
+
+    /* Never return NUMBER_UNDEFINED for existing entries. */
+    /* Note: There is code that counts on storing UNDEFINED in mapping
+     * values (using e.g. low_mapping_lookup to get them), so we have
+     * to fix the subtype here rather than in mapping_insert. */
+    if(p->type==T_INT)
+      dest->subtype=NUMBER_NUMBER;
   }else{
     dest->type=T_INT;
     dest->u.integer=0;
@@ -1404,6 +1428,25 @@ PMOD_EXPORT struct mapping *copy_mapping(struct mapping *m)
 
 #endif
 
+/* copy_mapping() for when destructive operations are ok.
+ *
+ * Note: It destructive operations on the resulting mapping *will*
+ *       affect eg NEW_MAPPING_LOOP() on the original mapping.
+ */
+static struct mapping *destructive_copy_mapping(struct mapping *m)
+{
+  if ((m->refs == 1) && !m->data->hardlinks &&
+      !(m->data->flags & MAPPING_WEAK)) {
+    /* We may perform destructive operations on the mapping. */
+    add_ref(m);
+    return m;
+  }
+  return copy_mapping(m);
+}
+
+/* NOTE: May perform destructive operations on either of the arguments
+ *       if it has only a single reference.
+ */
 static struct mapping *subtract_mappings(struct mapping *a, struct mapping *b)
 {
   struct mapping *res;
@@ -1415,7 +1458,7 @@ static struct mapping *subtract_mappings(struct mapping *a, struct mapping *b)
 
   /* First some special cases. */
   if (!a_md->size || !b_md->size || !a_md->hashsize || !b_md->hashsize) {
-    return copy_mapping(a);
+    return destructive_copy_mapping(a);
   }
   if (a_md == b_md) {
     return allocate_mapping(0);
@@ -1439,7 +1482,7 @@ static struct mapping *subtract_mappings(struct mapping *a, struct mapping *b)
     }
   } else {
     /* Remove the elements in a that are in b. */
-    res = copy_mapping(a);
+    res = destructive_copy_mapping(a);
     SET_ONERROR(err, do_free_mapping, res);
     NEW_MAPPING_LOOP(b_md) {
       map_delete(res, &k->ind);
@@ -1449,6 +1492,9 @@ static struct mapping *subtract_mappings(struct mapping *a, struct mapping *b)
   return res;
 }
 
+/* NOTE: May perform destructive operations on either of the arguments
+ *       if it has only a single reference.
+ */
 static struct mapping *and_mappings(struct mapping *a, struct mapping *b)
 {
   struct mapping *res;
@@ -1460,7 +1506,7 @@ static struct mapping *and_mappings(struct mapping *a, struct mapping *b)
 
   /* First some special cases. */
   if (!a_md->size || !b_md->size) return allocate_mapping(0);
-  if (a_md == b_md) return copy_mapping(a);
+  if (a_md == b_md) return destructive_copy_mapping(a);
 
   /* Copy the second mapping. */
   res = copy_mapping(b);
@@ -1483,6 +1529,9 @@ static struct mapping *and_mappings(struct mapping *a, struct mapping *b)
   return res;
 }
 
+/* NOTE: May perform destructive operations on either of the arguments
+ *       if it has only a single reference.
+ */
 static struct mapping *or_mappings(struct mapping *a, struct mapping *b)
 {
   struct mapping *res;
@@ -1493,31 +1542,51 @@ static struct mapping *or_mappings(struct mapping *a, struct mapping *b)
   ONERROR err;
 
   /* First some special cases. */
-  if (!a_md->size) return copy_mapping(b);
-  if (!b_md->size) return copy_mapping(a);
-  if (a_md == b_md) return copy_mapping(a);
+  if (!a_md->size) return destructive_copy_mapping(b);
+  if (!b_md->size) return destructive_copy_mapping(a);
+  if (a_md == b_md) return destructive_copy_mapping(a);
 
-  /* Copy the second mapping. */
-  res = copy_mapping(b);
-  SET_ONERROR(err, do_free_mapping, res);
+  if (a_md->size <= b_md->size) {
+    /* Copy the second mapping. */
+    res = destructive_copy_mapping(b);
+    SET_ONERROR(err, do_free_mapping, res);
 
-  /* Add elements in a that aren't in b. */
-  NEW_MAPPING_LOOP(a_md) {
-    size_t h = k->hval % b_md->hashsize;
-    struct keypair *k2;
-    for (k2 = b_md->hash[h]; k2; k2 = k2->next) {
-      if ((k2->hval == k->hval) && is_eq(&k2->ind, &k->ind)) {
-	break;
+    if (!b_md->hashsize) {
+      Pike_fatal("Invalid hashsize.\n");
+    }
+
+    /* Add elements in a that aren't in b. */
+    NEW_MAPPING_LOOP(a_md) {
+      size_t h = k->hval % b_md->hashsize;
+      struct keypair *k2;
+      for (k2 = b_md->hash[h]; k2; k2 = k2->next) {
+	if ((k2->hval == k->hval) && is_eq(&k2->ind, &k->ind)) {
+	  break;
+	}
+      }
+      if (!k2) {
+	mapping_insert(res, &k->ind, &k->val);
+	b_md = b->data;
       }
     }
-    if (!k2) {
+    UNSET_ONERROR(err);
+  } else {
+    /* Copy the first mapping. */
+    res = destructive_copy_mapping(a);
+    SET_ONERROR(err, do_free_mapping, res);
+
+    /* Add all elements in b. */
+    NEW_MAPPING_LOOP(b_md) {
       mapping_insert(res, &k->ind, &k->val);
     }
+    UNSET_ONERROR(err);
   }
-  UNSET_ONERROR(err);
   return res;
 }
 
+/* NOTE: May perform destructive operations on either of the arguments
+ *       if it has only a single reference.
+ */
 static struct mapping *xor_mappings(struct mapping *a, struct mapping *b)
 {
   struct mapping *res;
@@ -1528,8 +1597,8 @@ static struct mapping *xor_mappings(struct mapping *a, struct mapping *b)
   ONERROR err;
 
   /* First some special cases. */
-  if (!a_md->size) return copy_mapping(b);
-  if (!b_md->size) return copy_mapping(a);
+  if (!a_md->size) return destructive_copy_mapping(b);
+  if (!b_md->size) return destructive_copy_mapping(a);
   if (a_md == b_md) return allocate_mapping(0);
 
   /* Copy the largest mapping. */
@@ -1540,7 +1609,7 @@ static struct mapping *xor_mappings(struct mapping *a, struct mapping *b)
     a_md = b_md;
     b_md = b->data;
   }    
-  res = copy_mapping(b);
+  res = destructive_copy_mapping(b);
   SET_ONERROR(err, do_free_mapping, res);
 
   /* Add elements in a that aren't in b, and remove those that are. */
@@ -1557,11 +1626,15 @@ static struct mapping *xor_mappings(struct mapping *a, struct mapping *b)
     } else {
       map_delete(res, &k2->ind);
     }
+    b_md = b->data;
   }
   UNSET_ONERROR(err);
   return res;
 }
 
+/* NOTE: May perform destructive operations on either of the arguments
+ *       if it has only a single reference.
+ */
 PMOD_EXPORT struct mapping *merge_mappings(struct mapping *a, struct mapping *b, INT32 op)
 {
   ONERROR r1,r2,r3,r4;
@@ -1714,6 +1787,9 @@ PMOD_EXPORT struct mapping *merge_mapping_array_unordered(struct mapping *a,
   return m;
 }
 
+/* NOTE: May perform destructive operations on either of the arguments
+ *       if it has only a single reference.
+ */
 PMOD_EXPORT struct mapping *add_mappings(struct svalue *argp, INT32 args)
 {
   INT32 e,d;
@@ -1722,10 +1798,11 @@ PMOD_EXPORT struct mapping *add_mappings(struct svalue *argp, INT32 args)
 
   for(e=d=0;d<args;d++)
   {
+    struct mapping *m = argp[d].u.mapping;
 #ifdef PIKE_DEBUG
-    if(d_flag>1) check_mapping(argp[d].u.mapping);
+    if(d_flag>1) check_mapping(m);
 #endif
-    e+=argp[d].u.mapping->data->size;
+    e += m->data->size;
   }
 
   if(!e) return allocate_mapping(0);
@@ -1793,7 +1870,8 @@ PMOD_EXPORT int mapping_equal_p(struct mapping *a, struct mapping *b, struct pro
 
   if (a->data == b->data) return 1;
 
-  if (a->data->flags || b->data->flags) return 0;
+  /* If either is weak, they're different. */
+  if ((a->data->flags | b->data->flags) & MAPPING_WEAK) return 0;
 
   check_mapping_for_destruct(a);
   check_mapping_for_destruct(b);
@@ -1927,6 +2005,7 @@ void describe_mapping(struct mapping *m,struct processing *p,int indent)
   }
   else {
     int save_t_flag = Pike_interpreter.trace_level;
+    dynamic_buffer save_buf;
 
     if (m->data->size == 1) {
       my_strcat("([ /* 1 element */\n");
@@ -1935,15 +2014,17 @@ void describe_mapping(struct mapping *m,struct processing *p,int indent)
       my_strcat(buf);
     }
 
+    save_buffer (&save_buf);
     Pike_interpreter.trace_level = 0;
     if(SETJMP(catch)) {
       free_svalue(&throw_value);
-      throw_value.type = T_INT;
+      mark_free_svalue (&throw_value);
     }
     else
       sort_array_destructively(a);
     UNSETJMP(catch);
     Pike_interpreter.trace_level = save_t_flag;
+    restore_buffer (&save_buf);
 
     for(e = 0; e < a->size; e++)
     {
@@ -1953,7 +2034,7 @@ void describe_mapping(struct mapping *m,struct processing *p,int indent)
     
       for(d = 0; d < indent; d++)
 	my_putchar(' ');
-    
+
       describe_svalue(ITEM(a)+e, indent+2, &doing);
       my_strcat (": ");
 
@@ -2036,7 +2117,11 @@ PMOD_EXPORT void f_aggregate_mapping(INT32 args)
 
   m=allocate_mapping(MAP_SLOTS(args / 2));
 
-  for(e=-args;e<0;e+=2) low_mapping_insert(m, Pike_sp+e, Pike_sp+e+1, 2);
+  for(e=-args;e<0;e+=2) {
+    STACK_LEVEL_START(-e);
+    low_mapping_insert(m, Pike_sp+e, Pike_sp+e+1, 2);
+    STACK_LEVEL_DONE(-e);
+  }
   pop_n_elems(args);
 #ifdef PIKE_DEBUG
   if(d_flag)
@@ -2046,30 +2131,19 @@ PMOD_EXPORT void f_aggregate_mapping(INT32 args)
 }
 
 PMOD_EXPORT struct mapping *copy_mapping_recursively(struct mapping *m,
-						     struct processing *p)
+						     struct mapping *p)
 {
-  struct processing doing;
   struct mapping *ret;
   INT32 e;
   struct keypair *k;
   struct mapping_data  *md;
+  struct svalue aa, bb;
 
 #ifdef PIKE_DEBUG
   if(m->data->refs <=0)
     Pike_fatal("Zero refs in mapping->data\n");
 #endif
 
-  doing.next=p;
-  doing.pointer_a=(void *)m;
-  for(;p;p=p->next)
-  {
-    if(p->pointer_a == (void *)m)
-    {
-      add_ref(ret=(struct mapping *)p->pointer_b);
-      return ret;
-    }
-  }
-
 #ifdef PIKE_DEBUG
   if(d_flag > 1) check_mapping_type_fields(m);
 #endif
@@ -2078,7 +2152,14 @@ PMOD_EXPORT struct mapping *copy_mapping_recursively(struct mapping *m,
     return copy_mapping(m);
 
   ret=allocate_mapping(MAP_SLOTS(m->data->size));
-  doing.pointer_b=ret;
+
+  aa.type = T_MAPPING;
+  aa.subtype = 0;
+  aa.u.mapping = m;
+  bb.type = T_MAPPING;
+  bb.subtype = 0;
+  bb.u.mapping = ret;
+  mapping_insert(p, &aa, &bb);
 
   ret->data->flags = m->data->flags;
 
@@ -2089,10 +2170,16 @@ PMOD_EXPORT struct mapping *copy_mapping_recursively(struct mapping *m,
   add_ref(md);
   NEW_MAPPING_LOOP(md)
   {
+    /* Place holders.
+     *
+     * NOTE: copy_svalues_recursively_no_free() may store stuff in
+     *       the destination svalue, and then call stuff that uses
+     *       the stack (eg itself).
+     */
     push_int(0);
     push_int(0);
-    copy_svalues_recursively_no_free(Pike_sp-2,&k->ind, 1, &doing);
-    copy_svalues_recursively_no_free(Pike_sp-1,&k->val, 1, &doing);
+    copy_svalues_recursively_no_free(Pike_sp-2,&k->ind, 1, p);
+    copy_svalues_recursively_no_free(Pike_sp-1,&k->val, 1, p);
     
     low_mapping_insert(ret, Pike_sp-2, Pike_sp-1, 2);
     pop_n_elems(2);
@@ -2106,8 +2193,8 @@ PMOD_EXPORT struct mapping *copy_mapping_recursively(struct mapping *m,
 
 PMOD_EXPORT void mapping_search_no_free(struct svalue *to,
 			    struct mapping *m,
-			    struct svalue *look_for,
-			    struct svalue *key /* start */)
+			    const struct svalue *look_for,
+			    const struct svalue *key /* start */)
 {
   struct mapping_data *md, *omd;
 
@@ -2178,14 +2265,13 @@ PMOD_EXPORT void mapping_search_no_free(struct svalue *to,
   to->u.integer=0;
 }
 
-
 #ifdef PIKE_DEBUG
 
-void check_mapping(struct mapping *m)
+void check_mapping(const struct mapping *m)
 {
   int e,num;
   struct keypair *k;
-  struct mapping_data *md;
+  const struct mapping_data *md;
 
   static int in_check_mapping;
   if(in_check_mapping) return;
@@ -2269,9 +2355,15 @@ void check_mapping(struct mapping *m)
     {
       num++;
 
+      if (k->ind.type > MAX_TYPE)
+	Pike_fatal("Invalid maping keypair index type: %s\n",
+		   get_name_of_type(k->ind.type));
       if(! ( (1 << k->ind.type) & (md->ind_types) ))
 	Pike_fatal("Mapping indices type field lies.\n");
 
+      if (k->val.type > MAX_TYPE)
+	Pike_fatal("Invalid mapping keypair value type: %s\n",
+		   get_name_of_type(k->val.type));
       if(! ( (1 << k->val.type) & (md->val_types) ))
 	Pike_fatal("Mapping values type field lies.\n");
 
@@ -2298,6 +2390,56 @@ void check_all_mappings(void)
 }
 #endif
 
+static void visit_mapping_data (struct mapping_data *md, int action,
+				struct mapping *m)
+{
+  switch (action) {
+#ifdef PIKE_DEBUG
+    default:
+      Pike_fatal ("Unknown visit action %d.\n", action);
+    case VISIT_NORMAL:
+    case VISIT_COMPLEX_ONLY:
+      break;
+#endif
+    case VISIT_COUNT_BYTES:
+      mc_counted_bytes += MAPPING_DATA_SIZE (md->hashsize, md->num_keypairs);
+      break;
+  }
+
+  if ((md->ind_types | md->val_types) &
+      (action & VISIT_COMPLEX_ONLY ? BIT_COMPLEX : BIT_REF_TYPES)) {
+    int ind_ref_type =
+      md->flags & MAPPING_WEAK_INDICES ? REF_TYPE_WEAK : REF_TYPE_NORMAL;
+    int val_ref_type =
+      md->flags & MAPPING_WEAK_VALUES ? REF_TYPE_WEAK : REF_TYPE_NORMAL;
+    INT32 e;
+    struct keypair *k;
+    NEW_MAPPING_LOOP (md) {
+      visit_svalue (&k->ind, ind_ref_type);
+      visit_svalue (&k->val, val_ref_type);
+    }
+  }
+}
+
+PMOD_EXPORT void visit_mapping (struct mapping *m, int action)
+{
+  switch (action) {
+#ifdef PIKE_DEBUG
+    default:
+      Pike_fatal ("Unknown visit action %d.\n", action);
+    case VISIT_NORMAL:
+    case VISIT_COMPLEX_ONLY:
+      break;
+#endif
+    case VISIT_COUNT_BYTES:
+      mc_counted_bytes += sizeof (struct mapping);
+      break;
+  }
+
+  visit_ref (m->data, REF_TYPE_INTERNAL,
+	     (visit_thing_fn *) &visit_mapping_data, m);
+}
+
 #ifdef MAPPING_SIZE_DEBUG
 #define DO_IF_MAPPING_SIZE_DEBUG(x) x
 #else
@@ -2317,6 +2459,54 @@ void check_all_mappings(void)
   }									\
 } while (0)
 
+#ifdef PIKE_MAPPING_KEYPAIR_LOOP
+#error Broken code below!
+#define GC_RECURSE(M, MD, REC_KEYPAIR, TYPE, IND_TYPES, VAL_TYPES) do {	\
+    int remove;								\
+    struct keypair *k,**prev_;						\
+    /* no locking required (no is_eq) */				\
+    for(k = MD_KEYPAIRS(md, md->hashsize);k < MD->free_list;)		\
+    {									\
+      REC_KEYPAIR(remove,						\
+		  PIKE_CONCAT(TYPE, _svalues),				\
+		  PIKE_CONCAT(TYPE, _weak_svalues),			\
+		  PIKE_CONCAT(TYPE, _without_recurse),			\
+		  PIKE_CONCAT(TYPE, _weak_without_recurse));		\
+      if (remove) {							\
+	/* Find and unlink k. */					\
+	unsigned INT32 h_;						\
+	h_ = k->hval % md->hashsize;					\
+	prev_ = md->hash + h_;						\
+	DO_IF_DEBUG(							\
+		    if (!*prev_) {					\
+		      Pike_fatal("Node to unlink not found!\n");	\
+		    }							\
+		    );							\
+	while (*prev_ != k) {						\
+	  prev_ = &((*prev_)->next);					\
+	  DO_IF_DEBUG(							\
+		      if (!*prev_) {					\
+			Pike_fatal("Node to unlink not found!\n");	\
+		      }							\
+		      );						\
+	}								\
+	(*prev_)->next = k->next;					\
+        FREE_KEYPAIR(MD, k);						\
+	mark_free_svalue (&MD->free_list->ind);				\
+	mark_free_svalue (&MD->free_list->val);				\
+	MD->size--;							\
+	DO_IF_MAPPING_SIZE_DEBUG(					\
+	  if(M->data ==MD)						\
+	    M->debug_size--;						\
+	);								\
+      } else {								\
+	VAL_TYPES |= 1 << k->val.type;					\
+	IND_TYPES |= 1 << k->ind.type;					\
+	k++;								\
+      }									\
+    }									\
+  } while (0)
+#else /* !PIKE_MAPPING_KEYPAIR_LOOP */
 #define GC_RECURSE(M, MD, REC_KEYPAIR, TYPE, IND_TYPES, VAL_TYPES) do {	\
   INT32 e;								\
   int remove;								\
@@ -2335,8 +2525,8 @@ void check_all_mappings(void)
       {									\
 	*prev=k->next;							\
         FREE_KEYPAIR(MD, k);						\
-        MD->free_list->ind.type = T_INT;				\
-        MD->free_list->val.type = T_INT;				\
+	mark_free_svalue (&MD->free_list->ind);				\
+	mark_free_svalue (&MD->free_list->val);				\
 	MD->size--;							\
 	DO_IF_MAPPING_SIZE_DEBUG(					\
 	  if(M->data ==MD)						\
@@ -2350,6 +2540,7 @@ void check_all_mappings(void)
     }									\
   }									\
 } while (0)
+#endif /* PIKE_MAPPING_KEYPAIR_LOOP */
 
 #define GC_REC_KP(REMOVE, N_REC, W_REC, N_TST, W_TST) do {		\
   if ((REMOVE = N_REC(&k->ind, 1)))					\
@@ -2571,7 +2762,7 @@ void gc_check_all_mappings(void)
   for(m=first_mapping;m;m=m->next)
   {
 #ifdef DEBUG_MALLOC
-    if (((int)m->data) == 0x55555555) {
+    if (((int) PTR_TO_INT (m->data)) == 0x55555555) {
       fprintf(stderr, "** Zapped mapping in list of active mappings!\n");
       describe_something(m, T_MAPPING, 0,2,0, NULL);
       Pike_fatal("Zapped mapping in list of active mappings!\n");
