@@ -2,12 +2,12 @@
 || This file is part of Pike. For copyright information see COPYRIGHT.
 || Pike is distributed under GPL, LGPL and MPL. See the file COPYING
 || for more information.
-|| $Id: mpz_glue.c,v 1.163 2006/10/20 07:20:17 nilsson Exp $
+|| $Id: mpz_glue.c,v 1.184 2009/06/28 10:08:26 grubba Exp $
 */
 
 #include "global.h"
-RCSID("$Id: mpz_glue.c,v 1.163 2006/10/20 07:20:17 nilsson Exp $");
 #include "gmp_machine.h"
+#include "pike_float.h"
 #include "module.h"
 
 /* Disable this for now to check that the fallbacks work correctly. */
@@ -37,6 +37,7 @@ RCSID("$Id: mpz_glue.c,v 1.163 2006/10/20 07:20:17 nilsson Exp $");
 #include "module_support.h"
 #include "bignum.h"
 #include "operators.h"
+#include "gc.h"
 
 #include "my_gmp.h"
 
@@ -58,7 +59,7 @@ long random(void)
 
 #undef THIS
 #define THIS ((MP_INT *)(fp->current_storage))
-#define THIS_PROGRAM (fp->context.prog)
+#define THIS_PROGRAM (fp->context->prog)
 
 struct program *mpzmod_program = NULL;
 #ifdef AUTO_BIGNUM
@@ -187,26 +188,30 @@ static int gmp_int64_from_bignum (INT64 *i, struct object *bignum)
 {
   MP_INT *mpz = OBTOMPZ (bignum);
   int neg = mpz_sgn (mpz) < 0;
-  INT64 res = 0;
 
-  /* Note: Similar code in mpzmod_reduce. */
+  /* Note: Similar code in mpzmod_reduce and gmp_ulongest_from_bignum. */
 
   /* Get the index of the highest limb that have bits within the range
    * of the INT64. */
   size_t pos = (INT64_BITS + GMP_NUMB_BITS - 1) / GMP_NUMB_BITS - 1;
 
 #ifdef PIKE_DEBUG
-  if (bignum->prog != bignum_program) Pike_fatal ("Not a Gmp.bignum.\n");
+  if ((bignum->prog != bignum_program) &&
+      (bignum->prog != mpzmod_program)) {
+    Pike_fatal("cast_to_int(): Not a Gmp.bignum or Gmp.mpz.\n");
+  }
 #endif
 
   if (mpz_size (mpz) <= pos + 1) {
+    INT64 res;
 #if INT64_BITS == GMP_NUMB_BITS
     res = MPZ_GETLIMBN (mpz, 0) & GMP_NUMB_MASK;
 #elif INT64_BITS < GMP_NUMB_BITS
     mp_limb_t val = MPZ_GETLIMBN (mpz, 0) & GMP_NUMB_MASK;
     if (val >= (mp_limb_t) 1 << INT64_BITS) goto overflow;
-    res = val;
+    res = DO_NOT_WARN ((INT64) val);
 #else
+    res = 0;
     for (;; pos--) {
       res |= MPZ_GETLIMBN (mpz, pos) & GMP_NUMB_MASK;
       if (pos == 0) break;
@@ -233,6 +238,87 @@ overflow:
 
 #endif /* INT64 */
 
+static void gmp_push_ulongest (unsigned LONGEST i)
+{
+  if (i <= MAX_INT_TYPE) {
+    push_int(DO_NOT_WARN((INT_TYPE)i));
+  }
+  else {
+    MP_INT *mpz;
+
+    push_object (fast_clone_object (bignum_program));
+    mpz = OBTOMPZ (sp[-1].u.object);
+
+#if SIZEOF_LONG >= SIZEOF_LONGEST
+    mpz_set_ui (mpz, i);
+#else
+    {
+#ifdef HAVE_MPZ_IMPORT
+      mpz_import (mpz, 1, 1, SIZEOF_LONGEST, 0, 0, &i);
+#else
+      {
+	size_t n =
+	  ((SIZEOF_LONGEST + SIZEOF_LONG - 1) / SIZEOF_LONG - 1)
+	  /* The above is the position of the top unsigned long in the INT64. */
+	  * ULONG_BITS;
+	mpz_set_ui (mpz, (unsigned long) (i >> n));
+	while (n) {
+	  n -= ULONG_BITS;
+	  mpz_mul_2exp (mpz, mpz, ULONG_BITS);
+	  mpz_add_ui (mpz, mpz, (unsigned long) (i >> n));
+	}
+      }
+#endif	/* !HAVE_MPZ_IMPORT */
+    }
+#endif	/* SIZEOF_LONG < SIZEOF_LONGEST */
+  }
+}
+
+static int gmp_ulongest_from_bignum (unsigned LONGEST *i, struct object *bignum)
+{
+  MP_INT *mpz = OBTOMPZ (bignum);
+
+  /* Note: Similar code in gmp_int64_from_bignum. */
+
+  /* Get the index of the highest limb that have bits within the range
+   * of LONGEST. */
+  size_t pos = (ULONGEST_BITS + GMP_NUMB_BITS - 1) / GMP_NUMB_BITS - 1;
+
+#ifdef PIKE_DEBUG
+  if ((bignum->prog != bignum_program) &&
+      (bignum->prog != mpzmod_program)) {
+    Pike_fatal("cast_to_int(): Not a Gmp.bignum or Gmp.mpz.\n");
+  }
+#endif
+
+  if (mpz_sgn (mpz) < 0) return 0;
+
+  if (mpz_size (mpz) <= pos + 1) {
+    unsigned LONGEST res;
+#if ULONGEST_BITS == GMP_NUMB_BITS
+    res = MPZ_GETLIMBN (mpz, 0) & GMP_NUMB_MASK;
+#elif ULONGEST_BITS < GMP_NUMB_BITS
+    mp_limb_t val = MPZ_GETLIMBN (mpz, 0) & GMP_NUMB_MASK;
+    res = DO_NOT_WARN ((unsigned LONGEST) val);
+    if (val != res) return 0;
+#else
+    res = 0;
+    for (;; pos--) {
+      res |= MPZ_GETLIMBN (mpz, pos) & GMP_NUMB_MASK;
+      if (pos == 0) break;
+      if (res >= (unsigned LONGEST) 1 << (ULONGEST_BITS - GMP_NUMB_BITS))
+	return 0;
+      res <<= GMP_NUMB_BITS;
+    }
+#endif
+
+    *i = res;
+    return 1;
+  }
+
+  return 0;
+}
+
 #else
 #define PUSH_REDUCED(o) push_object(o)
 #endif /* AUTO_BIGNUM */
@@ -250,6 +336,13 @@ overflow:
  *! can be used to explicitly type integers that are too big to be
  *! INT_TYPE. Best is however to not use this program unless you
  *! really know what you are doing.
+ *!
+ *! Due to the auto-bignum conversion, all integers can be treated as
+ *! @[Gmp.mpz] objects insofar as that they can be indexed with the
+ *! functions in the @[Gmp.mpz] class. For instance, to calculate the
+ *! greatest common divisor between @expr{51@} and @expr{85@}, you can
+ *! do @expr{51->gcd(85)@}. In other words, all the functions in
+ *! @[Gmp.mpz] are also available here.
  *! @endclass
  */
 
@@ -257,6 +350,11 @@ overflow:
  *! Gmp.mpz implements very large integers. In fact,
  *! the only limitation on these integers is the available memory.
  *! The mpz object implements all the normal integer operations.
+ *!
+ *! Note that the auto-bignum feature also makes these operations
+ *! available "in" normal integers. For instance, to calculate the
+ *! greatest common divisor between @expr{51@} and @expr{85@}, you can
+ *! do @expr{51->gcd(85)@}.
  */
 
 void get_mpz_from_digits(MP_INT *tmp,
@@ -372,7 +470,11 @@ int get_new_mpz(MP_INT *tmp, struct svalue *s,
     break;
     
   case T_FLOAT:
-    mpz_set_d(tmp, (double) s->u.float_number);
+    {
+      double val = (double)s->u.float_number;
+      if (PIKE_ISNAN(val) || PIKE_ISINF(val)) return 0;
+      mpz_set_d(tmp, val);
+    }
     break;
 
   case T_OBJECT:
@@ -509,7 +611,6 @@ static void mpzmod_get_int(INT32 args)
   if( Pike_sp[-1].type == T_OBJECT &&
       Pike_sp[-1].u.object->prog != bignum_program )
   {
-    extern struct svalue auto_bignum_program;
     apply_svalue(&auto_bignum_program, 1);
   }
 #else
@@ -610,9 +711,11 @@ static void mpzmod_get_string(INT32 args)
 }
 
 /*! @decl string digits(void|int(2..36)|int(256..256) base)
- *! This function converts an mpz to a string. If a @[base] is given the
+ *!
+ *! Convert this mpz object to a string. If a @[base] is given the
  *! number will be represented in that base. Valid bases are 2-36 and
  *! 256. The default base is 10.
+ *!
  *! @seealso
  *!   @[cast_to_string]
  */
@@ -848,8 +951,9 @@ static void mpzmod__is_type(INT32 args)
 }
 
 /*! @decl int size(void|int base)
- *! This function returns how long the mpz would be represented in the
- *! specified @[base]. The default base is 2.
+ *!
+ *! Return how long this mpz would be represented in the specified
+ *! @[base]. The default base is 2.
  */
 static void mpzmod_size(INT32 args)
 {
@@ -876,7 +980,10 @@ static void mpzmod_size(INT32 args)
 }
 
 /*! @decl mixed cast(string type)
- *! It is possible to cast an mpz to a string, int or float.
+ *!
+ *! Cast this mpz object to another type. Allowed types are string,
+ *! int and float.
+ *!
  *! @seealso
  *!   @[cast_to_int], @[cast_to_float], @[cast_to_string]
  */
@@ -991,7 +1098,7 @@ double double_from_sval(struct svalue *s)
 #define DO_IF_AUTO_BIGNUM(X)
 #endif
 
-#define BINFUN2(name, errmsg_op, fun, OP)				\
+#define BINFUN2(name, errmsg_op, fun, OP, f_op, LFUN)			\
 static void name(INT32 args)						\
 {									\
   INT32 e;								\
@@ -1004,6 +1111,33 @@ static void name(INT32 args)						\
     {									\
       switch(sp[e-args].type)						\
       {									\
+      case T_OBJECT:							\
+	{								\
+	  struct object *o = sp[e-args].u.object;			\
+	  struct program *p = NULL;					\
+	  int fun = -1;							\
+	  if (o->prog &&						\
+	      ((p = o->prog->inherits[sp[e-args].subtype].prog) !=	\
+	       bignum_program) &&					\
+	      ((fun = FIND_LFUN(p, PIKE_CONCAT(LFUN_R, LFUN))) !=	\
+	       -1)) {							\
+	    /* Found non-bignum program with double back operator. */	\
+	    memmove(Pike_sp+1-args, Pike_sp-args,			\
+		    args * sizeof(struct svalue));			\
+	    Pike_sp++;							\
+	    args++;							\
+	    e++;							\
+	    Pike_sp[-args].type = T_OBJECT;				\
+	    Pike_sp[-args].subtype = 0; /* FIXME? */			\
+	    add_ref(Pike_sp[-args].u.object = Pike_fp->current_object);	\
+	    args = low_rop(o, fun, e, args);				\
+	    if (args > 1) {						\
+	      f_op(args);						\
+	    }								\
+	    return;							\
+	  }								\
+	}								\
+	break;								\
        case T_FLOAT:                                                    \
 	ret=mpz_get_d(THIS);						\
 	for(e=0; e<args; e++)						\
@@ -1016,7 +1150,7 @@ static void name(INT32 args)						\
        case T_STRING:                                                   \
         MEMMOVE(sp-args+1, sp-args, sizeof(struct svalue)*args);        \
         sp++; args++;                                                   \
-        sp[-args].type=T_INT;                                           \
+	sp[-args].type=PIKE_T_FREE;					\
         sp[-args].u.string=low_get_mpz_digits(THIS, 10);                \
         sp[-args].type=T_STRING;                                        \
         f_add(args);                                                    \
@@ -1026,7 +1160,7 @@ static void name(INT32 args)						\
   } )									\
   for(e=0; e<args; e++)							\
     if(sp[e-args].type != T_INT || !FITS_ULONG (sp[e-args].u.integer))	\
-     get_mpz(sp+e-args, 1, "Gmp.mpz->`" errmsg_op, e + 1, args);	\
+      get_mpz(sp+e-args, 1, "Gmp.mpz->`" errmsg_op, e + 1, args);	\
   res = fast_clone_object(THIS_PROGRAM);				\
   mpz_set(OBTOMPZ(res), THIS);						\
   for(e=0;e<args;e++)							\
@@ -1070,7 +1204,7 @@ static void PIKE_CONCAT(name,_rhs)(INT32 args)				\
   } )									\
   for(e=0; e<args; e++)							\
     if(sp[e-args].type != T_INT || !FITS_ULONG (sp[e-args].u.integer))	\
-     get_mpz(sp+e-args, 1, "Gmp.mpz->``" errmsg_op, e + 1, args);	\
+      get_mpz(sp+e-args, 1, "Gmp.mpz->``" errmsg_op, e + 1, args);	\
   res = fast_clone_object(THIS_PROGRAM);				\
   mpz_set(OBTOMPZ(res), THIS);						\
   for(e=0;e<args;e++)							\
@@ -1107,7 +1241,7 @@ static void PIKE_CONCAT(name,_eq)(INT32 args)				\
        case T_STRING:                                                   \
         MEMMOVE(sp-args+1, sp-args, sizeof(struct svalue)*args);        \
         sp++; args++;                                                   \
-        sp[-args].type=T_INT;                                           \
+	sp[-args].type=PIKE_T_FREE;					\
         sp[-args].u.string=low_get_mpz_digits(THIS, 10);                \
         sp[-args].type=T_STRING;                                        \
         f_add(args);                                                    \
@@ -1135,7 +1269,7 @@ static void PIKE_CONCAT(name,_eq)(INT32 args)				\
  */
 /*! @decl Gmp.mpz `+=(int|float|Gmp.mpz ... x)
  */
-BINFUN2(mpzmod_add, "+", mpz_add, +);
+BINFUN2(mpzmod_add, "+", mpz_add, +, f_add, ADD)
 
 #undef STRINGCONV
 #define STRINGCONV(X)
@@ -1146,10 +1280,12 @@ BINFUN2(mpzmod_add, "+", mpz_add, +);
  */
 /*! @decl Gmp.mpz `*=(int|float|Gmp.mpz ... x)
  */
-BINFUN2(mpzmod_mul, "*", mpz_mul, *);
+BINFUN2(mpzmod_mul, "*", mpz_mul, *, f_multiply, MULTIPLY)
 
-/*! @decl Gmp.mpz gcd(object|int|float|string arg)
- *! This function returns the greatest common divisor for @[arg] and mpz.
+/*! @decl Gmp.mpz gcd(object|int|float|string... arg)
+ *! 
+ *! Return the greatest common divisor between this mpz object and
+ *! all the arguments.
  */
 static void mpzmod_gcd(INT32 args)
 {
@@ -1322,6 +1458,18 @@ static void mpzmod_rmod(INT32 args)
 }
 
 /*! @decl array(Gmp.mpz) gcdext(int|float|Gmp.mpz x)
+ *!
+ *! Compute the greatest common divisor between this mpz object and
+ *! @[x]. An array @expr{({g,s,t})@} is returned where @expr{g@} is
+ *! the greatest common divisor, and @expr{s@} and @expr{t@} are the
+ *! coefficients that satisfies
+ *!
+ *! @code
+ *!                this * s + @[x] * t = g
+ *! @endcode
+ *!
+ *! @seealso
+ *!   @[gcdext2], @[gcd]
  */
 static void mpzmod_gcdext(INT32 args)
 {
@@ -1344,6 +1492,20 @@ static void mpzmod_gcdext(INT32 args)
 }
 
 /*! @decl array(Gmp.mpz) gcdext2(int|float|Gmp.mpz x)
+ *!
+ *! Compute the greatest common divisor between this mpz object and
+ *! @[x]. An array @expr{({g,s})@} is returned where @expr{g@} is the
+ *! greatest common divisor, and @expr{s@} is a coefficient that
+ *! satisfies
+ *!
+ *! @code
+ *!                this * s + @[x] * t = g
+ *! @endcode
+ *!
+ *! where @expr{t@} is some integer value.
+ *!
+ *! @seealso
+ *!   @[gcdext], @[gcd]
  */
 static void mpzmod_gcdext2(INT32 args)
 {
@@ -1365,6 +1527,12 @@ static void mpzmod_gcdext2(INT32 args)
 }
 
 /*! @decl Gmp.mpz invert(int|float|Gmp.mpz x)
+ *!
+ *! Return the inverse of this mpz value modulo @[x]. The returned
+ *! value satisfies @expr{0 <= result < x@}.
+ *!
+ *! @throws
+ *!   An error is thrown if no inverse exists.
  */
 static void mpzmod_invert(INT32 args)
 {
@@ -1373,7 +1541,7 @@ static void mpzmod_invert(INT32 args)
 
   if (args != 1)
     SIMPLE_WRONG_NUM_ARGS_ERROR ("Gmp.mpz->invert", 1);
-  modulo = get_mpz(sp-args, 1, "Gmp.mpz->invert", 1, 1);
+  modulo = get_mpz(sp-1, 1, "Gmp.mpz->invert", 1, 1);
 
   if (!mpz_sgn(modulo))
     SIMPLE_DIVISION_BY_ZERO_ERROR ("Gmp.mpz->invert");
@@ -1387,6 +1555,73 @@ static void mpzmod_invert(INT32 args)
   PUSH_REDUCED(res);
 }
 
+/*! @decl Gmp.mpz fac()
+ *!
+ *! Return the factorial of this mpz object.
+ *!
+ *! @throws
+ *!   Since factorials grow very quickly, only small integers are
+ *!   supported. An error is thrown if the value in this mpz object is
+ *!   too large.
+ */
+static void mpzmod_fac(INT32 args)
+{
+  struct object *res;
+  if (mpz_sgn (THIS) < 0)
+    Pike_error ("Cannot calculate factorial for negative integer.\n");
+  if (!mpz_fits_ulong_p (THIS))
+    Pike_error ("Integer too large for factorial calculation.\n");
+  res = fast_clone_object (THIS_PROGRAM);
+  mpz_fac_ui (OBTOMPZ(res), mpz_get_ui (THIS));
+  pop_n_elems (args);
+  PUSH_REDUCED (res);
+}
+
+#ifdef HAVE_MPZ_BIN_UI
+/*! @decl Gmp.mpz bin(int k)
+ *! 
+ *! Return the binomial coefficient @expr{n@} over @[k], where
+ *! @expr{n@} is the value of this mpz object. Negative values of
+ *! @expr{n@} are supported using the identity
+ *!
+ *! @code
+ *!     (-n)->bin(k) == (-1)->pow(k) * (n+k-1)->bin(k)
+ *! @endcode
+ *!
+ *! (See Knuth volume 1, section 1.2.6 part G.)
+ *!
+ *! @throws
+ *!   The @[k] value can't be arbitrarily large. An error is thrown if
+ *!   it's too large.
+ *!
+ *! @note
+ *!   This function is currently (Pike 7.8) not available with old
+ *!   releases of the gmp libraries.
+ */
+static void mpzmod_bin(INT32 args)
+{
+  MP_INT *k;
+  struct object *res;
+
+  if (args != 1)
+    SIMPLE_WRONG_NUM_ARGS_ERROR ("Gmp.mpz->bin", 1);
+  k = get_mpz (sp-1, 1, "Gmp.mpz->bin", 1, 1);
+  if (mpz_sgn (k) < 0)
+    Pike_error ("Cannot calculate binomial with negative k value.\n");
+  if (!mpz_fits_ulong_p (k))
+    SIMPLE_ARG_ERROR ("Gmp.mpz->bin", 1, "Argument too large.\n");
+
+  res = fast_clone_object(THIS_PROGRAM);
+  mpz_bin_ui (OBTOMPZ (res), THIS, mpz_get_ui (k));
+
+  pop_n_elems(args);
+  PUSH_REDUCED(res);
+}
+#else
+/* FIXME: Implement mpz_bin_ui() using primitives. */
+/* Note: When fixing; note IF_MPZ_BIN_UI further below. */
+#endif /* HAVE_MPZ_BIN_UI */
+
 #define BINFUN(name, errmsg_name, fun)			\
 static void name(INT32 args)				\
 {							\
@@ -1405,15 +1640,15 @@ static void name(INT32 args)				\
 
 /*! @decl Gmp.mpz `&(int|float|Gmp.mpz ... x)
  */
-BINFUN(mpzmod_and, "Gmp.mpz->`&", mpz_and);
+BINFUN(mpzmod_and, "Gmp.mpz->`&", mpz_and)
 
 /*! @decl Gmp.mpz `|(int|float|Gmp.mpz ... x)
  */
-BINFUN(mpzmod_or, "Gmp.mpz->`|", mpz_ior);
+BINFUN(mpzmod_or, "Gmp.mpz->`|", mpz_ior)
 
 /*! @decl Gmp.mpz `^(int|float|Gmp.mpz ... x)
  */
-BINFUN(mpzmod_xor, "Gmp.mpz->`^", mpz_xor);
+BINFUN(mpzmod_xor, "Gmp.mpz->`^", mpz_xor)
 
 /*! @decl Gmp.mpz `~()
  */
@@ -1444,31 +1679,32 @@ static void name(INT32 args)				\
 
 /*! @decl int(0..1) `>(mixed with)
  */
-CMPEQU(mpzmod_gt, "Gmp.mpz->`>", >, RET_UNDEFINED);
+CMPEQU(mpzmod_gt, "Gmp.mpz->`>", >, RET_UNDEFINED)
 
 /*! @decl int(0..1) `<(mixed with)
  */
-CMPEQU(mpzmod_lt, "Gmp.mpz->`<", <, RET_UNDEFINED);
+CMPEQU(mpzmod_lt, "Gmp.mpz->`<", <, RET_UNDEFINED)
 
 /*! @decl int(0..1) `>=(mixed with)
  */
-CMPEQU(mpzmod_ge, "Gmp.mpz->`>=", >=, RET_UNDEFINED);
+CMPEQU(mpzmod_ge, "Gmp.mpz->`>=", >=, RET_UNDEFINED)
 
 /*! @decl int(0..1) `<=(mixed with)
  */
-CMPEQU(mpzmod_le, "Gmp.mpz->`<=", <=, RET_UNDEFINED);
+CMPEQU(mpzmod_le, "Gmp.mpz->`<=", <=, RET_UNDEFINED)
 
 /*! @decl int(0..1) `==(mixed with)
  */
-CMPEQU(mpzmod_eq, "Gmp.mpz->`==", ==, RET_UNDEFINED);
+CMPEQU(mpzmod_eq, "Gmp.mpz->`==", ==, RET_UNDEFINED)
 
 /*! @decl int(0..1) `!=(mixed with)
  */
-CMPEQU(mpzmod_nq, "Gmp.mpz->`!=", !=, i=1);
+CMPEQU(mpzmod_nq, "Gmp.mpz->`!=", !=, i=1)
 
 /*! @decl int(0..1) probably_prime_p()
- *! This function returns 1 if mpz is a prime, and 0 most of the time
- *! if it is not.
+ *!
+ *! Return 1 if this mpz object is a prime, and 0 most of the time if
+ *! it is not.
  */
 static void mpzmod_probably_prime_p(INT32 args)
 {
@@ -1544,8 +1780,9 @@ static void mpzmod_sgn(INT32 args)
 }
 
 /*! @decl Gmp.mpz sqrt()
- *! This function return the the truncated integer part of the square
- *! root of the value of mpz.
+ *!
+ *! Return the the truncated integer part of the square root of this
+ *! mpz object.
  */
 static void mpzmod_sqrt(INT32 args)
 {
@@ -1722,9 +1959,13 @@ static void mpzmod_rrsh(INT32 args)
   PUSH_REDUCED(res);
 }
 
-/*! @decl Gmp.mpz powm(int|string|float|Gmp.mpz a,@
- *!                    int|string|float|Gmp.mpz b)
- *! This function returns @tt{( mpz ** a ) % b@}.
+/*! @decl Gmp.mpz powm(int|string|float|Gmp.mpz exp,@
+ *!                    int|string|float|Gmp.mpz mod)
+ *!
+ *! Return @expr{( this->pow(@[exp]) ) % @[mod]@}.
+ *!
+ *! @seealso
+ *!   @[pow]
  */
 static void mpzmod_powm(INT32 args)
 {
@@ -1746,6 +1987,12 @@ static void mpzmod_powm(INT32 args)
 }
 
 /*! @decl Gmp.mpz pow(int|float|Gmp.mpz x)
+ *!
+ *! Return this mpz object raised to @[x]. The case when zero is
+ *! raised to zero yields one.
+ *!
+ *! @seealso
+ *!   @[powm]
  */
 static void mpzmod_pow(INT32 args)
 {
@@ -1864,7 +2111,7 @@ static void mpzmod_random(INT32 args)
   push_int(mpz_size(THIS)*sizeof(mp_limb_t) + 4);
   f_random_string(1);
   if (sp[-1].type != T_STRING) {
-    Pike_error("random_string(%"PRINTSIZET"d) returned non string.\n",
+    Pike_error("random_string(%ld) returned non string.\n",
 	       mpz_size(THIS)*sizeof(mp_limb_t) + 4);
   }
   get_mpz_from_digits(OBTOMPZ(res), sp[-1].u.string, 256);
@@ -1913,6 +2160,18 @@ static void exit_mpz_glue(struct object *o)
 #endif
   mpz_clear(THIS);
 }
+
+static void gc_recurse_mpz (struct object *o)
+{
+  if (mc_count_bytes (o))
+    mc_counted_bytes +=
+#ifdef MPZ_T_HAS__MP_ALLOC
+      THIS[0]._mp_alloc * sizeof (mp_limb_t) +
+#else
+      mpz_size (THIS) * sizeof (mp_limb_t) +
+#endif
+      sizeof (mpz_t);
+}
 #endif
 
 PIKE_MODULE_EXIT
@@ -1929,7 +2188,7 @@ PIKE_MODULE_EXIT
   {
     extern struct svalue auto_bignum_program;
     free_svalue(&auto_bignum_program);
-    auto_bignum_program.type=T_INT;
+    auto_bignum_program.type=PIKE_T_FREE;
     if(bignum_program)
     {
       free_program(bignum_program);
@@ -1938,20 +2197,57 @@ PIKE_MODULE_EXIT
     mpz_clear (mpz_int_type_min);
 #ifdef INT64
     mpz_clear (mpz_int64_min);
-    hook_in_int64_funcs (NULL, NULL, NULL);
 #endif
+    hook_in_gmp_funcs (
+#ifdef INT64
+      NULL, NULL, NULL,
+#endif
+      NULL, NULL);
   }
 #endif
 #endif
 }
 
 
+static void *pike_mp_alloc (size_t alloc_size)
+{
+  void *ret = malloc (alloc_size);
+  if (!ret)
+    /* According to gmp docs, we're neither allowed to return zero nor
+     * longjmp here. */
+    Pike_fatal ("Failed to allocate %"PRINTSIZET"db in gmp library.\n",
+		alloc_size);
+  return ret;
+}
+
+static void *pike_mp_realloc (void *ptr, size_t old_size, size_t new_size)
+{
+  void *ret = realloc (ptr, new_size);
+  if (!ret)
+    /* According to gmp docs, we're neither allowed to return zero nor
+     * longjmp here. */
+    Pike_fatal ("Failed to reallocate %"PRINTSIZET"db block "
+		"to %"PRINTSIZET"db in gmp library.\n", old_size, new_size);
+  return ret;
+}
+
+static void pike_mp_free (void *ptr, size_t size)
+{
+  free (ptr);
+}
+
+
 #define tMpz_arg tOr3(tInt,tFloat,tObj)
 #define tMpz_ret tObjIs_GMP_MPZ
 #define tMpz_shift_type tFunc(tMpz_arg,tMpz_ret)
 #define tMpz_binop_type tFuncV(tNone, tMpz_arg, tMpz_ret)
 #define tMpz_cmpop_type tFunc(tMixed, tInt01)
 
+#ifdef HAVE_MPZ_BIN_UI
+#define IF_MPZ_BIN_UI(X) X
+#else
+#define IF_MPZ_BIN_UI(X)
+#endif
 #define MPZ_DEFS()							\
   ADD_STORAGE(MP_INT);							\
   									\
@@ -1960,52 +2256,52 @@ PIKE_MODULE_EXIT
   ADD_FUNCTION("create", mpzmod_create,					\
 	       tOr(tFunc(tOr5(tVoid,tStr,tInt,tFlt,			\
 			      tObj),tVoid),				\
-		   tFunc(tStr tInt,tVoid)), ID_STATIC);			\
+		   tFunc(tStr tInt,tVoid)), ID_PROTECTED);		\
 									\
-  ADD_FUNCTION("`+",mpzmod_add,tMpz_binop_type, ID_STATIC);		\
-  ADD_FUNCTION("`+=",mpzmod_add_eq,tMpz_binop_type, ID_STATIC);		\
-  ADD_FUNCTION("``+",mpzmod_add_rhs,tMpz_binop_type, ID_STATIC);	\
-  ADD_FUNCTION("`-",mpzmod_sub,tMpz_binop_type, ID_STATIC);		\
-  ADD_FUNCTION("``-",mpzmod_rsub,tMpz_binop_type, ID_STATIC);		\
-  ADD_FUNCTION("`*",mpzmod_mul,tMpz_binop_type, ID_STATIC);		\
-  ADD_FUNCTION("``*",mpzmod_mul,tMpz_binop_type, ID_STATIC);		\
-  ADD_FUNCTION("`*=",mpzmod_mul_eq,tMpz_binop_type, ID_STATIC);		\
-  ADD_FUNCTION("`/",mpzmod_div,tMpz_binop_type, ID_STATIC);		\
-  ADD_FUNCTION("``/",mpzmod_rdiv,tMpz_binop_type, ID_STATIC);		\
-  ADD_FUNCTION("`%",mpzmod_mod,tMpz_binop_type, ID_STATIC);		\
-  ADD_FUNCTION("``%",mpzmod_rmod,tMpz_binop_type, ID_STATIC);		\
-  ADD_FUNCTION("`&",mpzmod_and,tMpz_binop_type, ID_STATIC);		\
-  ADD_FUNCTION("``&",mpzmod_and,tMpz_binop_type, ID_STATIC);		\
-  ADD_FUNCTION("`|",mpzmod_or,tMpz_binop_type, ID_STATIC);		\
-  ADD_FUNCTION("``|",mpzmod_or,tMpz_binop_type, ID_STATIC);		\
-  ADD_FUNCTION("`^",mpzmod_xor,tMpz_binop_type, ID_STATIC);		\
-  ADD_FUNCTION("``^",mpzmod_xor,tMpz_binop_type, ID_STATIC);		\
-  ADD_FUNCTION("`~",mpzmod_compl,tFunc(tNone,tObj), ID_STATIC);		\
+  ADD_FUNCTION("`+",mpzmod_add,tMpz_binop_type, ID_PROTECTED);		\
+  ADD_FUNCTION("`+=",mpzmod_add_eq,tMpz_binop_type, ID_PROTECTED);	\
+  ADD_FUNCTION("``+",mpzmod_add_rhs,tMpz_binop_type, ID_PROTECTED);	\
+  ADD_FUNCTION("`-",mpzmod_sub,tMpz_binop_type, ID_PROTECTED);		\
+  ADD_FUNCTION("``-",mpzmod_rsub,tMpz_binop_type, ID_PROTECTED);	\
+  ADD_FUNCTION("`*",mpzmod_mul,tMpz_binop_type, ID_PROTECTED);		\
+  ADD_FUNCTION("``*",mpzmod_mul,tMpz_binop_type, ID_PROTECTED);		\
+  ADD_FUNCTION("`*=",mpzmod_mul_eq,tMpz_binop_type, ID_PROTECTED);	\
+  ADD_FUNCTION("`/",mpzmod_div,tMpz_binop_type, ID_PROTECTED);		\
+  ADD_FUNCTION("``/",mpzmod_rdiv,tMpz_binop_type, ID_PROTECTED);	\
+  ADD_FUNCTION("`%",mpzmod_mod,tMpz_binop_type, ID_PROTECTED);		\
+  ADD_FUNCTION("``%",mpzmod_rmod,tMpz_binop_type, ID_PROTECTED);	\
+  ADD_FUNCTION("`&",mpzmod_and,tMpz_binop_type, ID_PROTECTED);		\
+  ADD_FUNCTION("``&",mpzmod_and,tMpz_binop_type, ID_PROTECTED);		\
+  ADD_FUNCTION("`|",mpzmod_or,tMpz_binop_type, ID_PROTECTED);		\
+  ADD_FUNCTION("``|",mpzmod_or,tMpz_binop_type, ID_PROTECTED);		\
+  ADD_FUNCTION("`^",mpzmod_xor,tMpz_binop_type, ID_PROTECTED);		\
+  ADD_FUNCTION("``^",mpzmod_xor,tMpz_binop_type, ID_PROTECTED);		\
+  ADD_FUNCTION("`~",mpzmod_compl,tFunc(tNone,tObj), ID_PROTECTED);	\
 									\
-  ADD_FUNCTION("`<<",mpzmod_lsh,tMpz_shift_type, ID_STATIC);		\
-  ADD_FUNCTION("`>>",mpzmod_rsh,tMpz_shift_type, ID_STATIC);		\
-  ADD_FUNCTION("``<<",mpzmod_rlsh,tMpz_shift_type, ID_STATIC);		\
-  ADD_FUNCTION("``>>",mpzmod_rrsh,tMpz_shift_type, ID_STATIC);		\
+  ADD_FUNCTION("`<<",mpzmod_lsh,tMpz_shift_type, ID_PROTECTED);		\
+  ADD_FUNCTION("`>>",mpzmod_rsh,tMpz_shift_type, ID_PROTECTED);		\
+  ADD_FUNCTION("``<<",mpzmod_rlsh,tMpz_shift_type, ID_PROTECTED);	\
+  ADD_FUNCTION("``>>",mpzmod_rrsh,tMpz_shift_type, ID_PROTECTED);	\
 									\
-  ADD_FUNCTION("`>", mpzmod_gt,tMpz_cmpop_type, ID_STATIC);		\
-  ADD_FUNCTION("`<", mpzmod_lt,tMpz_cmpop_type, ID_STATIC);		\
-  ADD_FUNCTION("`>=",mpzmod_ge,tMpz_cmpop_type, ID_STATIC);		\
-  ADD_FUNCTION("`<=",mpzmod_le,tMpz_cmpop_type, ID_STATIC);		\
+  ADD_FUNCTION("`>", mpzmod_gt,tMpz_cmpop_type, ID_PROTECTED);		\
+  ADD_FUNCTION("`<", mpzmod_lt,tMpz_cmpop_type, ID_PROTECTED);		\
+  ADD_FUNCTION("`>=",mpzmod_ge,tMpz_cmpop_type, ID_PROTECTED);		\
+  ADD_FUNCTION("`<=",mpzmod_le,tMpz_cmpop_type, ID_PROTECTED);		\
 									\
-  ADD_FUNCTION("`==",mpzmod_eq,tMpz_cmpop_type, ID_STATIC);		\
-  ADD_FUNCTION("`!=",mpzmod_nq,tMpz_cmpop_type, ID_STATIC);		\
+  ADD_FUNCTION("`==",mpzmod_eq,tMpz_cmpop_type, ID_PROTECTED);		\
+  ADD_FUNCTION("`!=",mpzmod_nq,tMpz_cmpop_type, ID_PROTECTED);		\
 									\
-  ADD_FUNCTION("`!",mpzmod_not,tFunc(tNone,tInt01), ID_STATIC);		\
+  ADD_FUNCTION("`!",mpzmod_not,tFunc(tNone,tInt01), ID_PROTECTED);	\
 									\
-  ADD_FUNCTION("__hash",mpzmod___hash,tFunc(tNone,tInt), ID_STATIC);	\
-  ADD_FUNCTION("cast",mpzmod_cast,tFunc(tStr,tMix), ID_STATIC);		\
+  ADD_FUNCTION("__hash",mpzmod___hash,tFunc(tNone,tInt), ID_PROTECTED);	\
+  ADD_FUNCTION("cast",mpzmod_cast,tFunc(tStr,tMix), ID_PROTECTED);	\
 									\
   ADD_FUNCTION("_is_type", mpzmod__is_type, tFunc(tStr,tInt01),         \
-               ID_STATIC);                                              \
+               ID_PROTECTED);                                           \
   									\
   ADD_FUNCTION("digits", mpzmod_digits,tFunc(tOr(tVoid,tInt),tStr), 0);	\
   ADD_FUNCTION("_sprintf", mpzmod__sprintf, tFunc(tInt tMapping,tStr),  \
-               ID_STATIC);                                              \
+               ID_PROTECTED);                                           \
   ADD_FUNCTION("size", mpzmod_size,tFunc(tOr(tVoid,tInt),tInt), 0);	\
 									\
   ADD_FUNCTION("cast_to_int",mpzmod_get_int,tFunc(tNone,tInt),0);	\
@@ -2024,6 +2320,10 @@ PIKE_MODULE_EXIT
   ADD_FUNCTION("gcdext2",mpzmod_gcdext2,tFunc(tMpz_arg,tArr(tMpz_ret)),0);\
   ADD_FUNCTION("invert", mpzmod_invert,tFunc(tMpz_arg,tMpz_ret),0);	\
 									\
+  ADD_FUNCTION("fac", mpzmod_fac, tFunc(tNone,tMpz_ret), 0);		\
+  IF_MPZ_BIN_UI(ADD_FUNCTION("bin", mpzmod_bin,				\
+			     tFunc(tMpz_arg,tMpz_ret), 0));		\
+									\
   ADD_FUNCTION("sgn", mpzmod_sgn, tFunc(tNone,tInt), 0);		\
   ADD_FUNCTION("sqrt", mpzmod_sqrt,tFunc(tNone,tMpz_ret),0);		\
   ADD_FUNCTION("_sqrt", mpzmod_sqrt,tFunc(tNone,tMpz_ret),0);		\
@@ -2036,7 +2336,8 @@ PIKE_MODULE_EXIT
   ADD_FUNCTION("_random",mpzmod_random,tFunc(tNone,tMpz_ret),0);	\
   									\
   set_init_callback(init_mpz_glue);					\
-  set_exit_callback(exit_mpz_glue);
+  set_exit_callback(exit_mpz_glue);					\
+  set_gc_recurse_callback (gc_recurse_mpz);
 
 
 PIKE_MODULE_INIT
@@ -2047,6 +2348,14 @@ PIKE_MODULE_INIT
     fatal("Gmp.mpz initialized twice!\n");
   }
 #endif /* PIKE_DEBUG */
+
+  /* Make sure that gmp uses the same malloc functions as we do since
+   * we got code that frees blocks allocated inside gmp (e.g.
+   * mpf.get_string). This also ensures that gmp uses dlmalloc if we
+   * do on Windows. In case gmp already uses the same malloc, this is
+   * essentially just a NOP. */
+  mp_set_memory_functions (pike_mp_alloc, pike_mp_realloc, pike_mp_free);
+
   start_new_program();
 
   MPZ_DEFS();
@@ -2113,9 +2422,13 @@ PIKE_MODULE_INIT
     mpz_init (mpz_int64_min);
     mpz_setbit (mpz_int64_min, INT64_BITS);
     mpz_neg (mpz_int64_min, mpz_int64_min);
-    hook_in_int64_funcs (gmp_push_int64, gmp_int64_from_bignum,
-			 gmp_reduce_stack_top_bignum);
 #endif
+    hook_in_gmp_funcs (
+#ifdef INT64
+      gmp_push_int64, gmp_int64_from_bignum,
+      gmp_reduce_stack_top_bignum,
+#endif
+      gmp_push_ulongest, gmp_ulongest_from_bignum);
 
 #if 0
     /* magic /Hubbe
