@@ -2,7 +2,7 @@
 || This file is part of Pike. For copyright information see COPYRIGHT.
 || Pike is distributed under GPL, LGPL and MPL. See the file COPYING
 || for more information.
-|| $Id: blit.c,v 1.57 2006/08/31 14:04:02 nilsson Exp $
+|| $Id: blit.c,v 1.60 2006/08/31 14:02:47 nilsson Exp $
 */
 
 #include "global.h"
@@ -172,8 +172,9 @@ void img_box_nocheck(INT32 x1,INT32 y1,INT32 x2,INT32 y2)
    {
      THREADS_ALLOW();
      do {
-       for (; foo<end; foo+=mod) for (x=x1; x<=x2; x++,foo++)
-	 set_rgb_group_alpha(*foo,rgb,this->alpha);
+       for (; foo<end; foo+=mod)
+         for (x=x1; x<=x2; x++,foo++)
+           set_rgb_group_alpha(*foo,rgb,this->alpha);
      } while(0);
      THREADS_DISALLOW();
    }
@@ -215,13 +216,12 @@ void img_crop(struct image *dest,
    if (x1>x2) tmp=x1, x1=x2, x2=tmp;
    if (y1>y2) tmp=y1, y1=y2, y2=tmp;
 
+   new=xalloc( (x2-x1+1)*(y2-y1+1)*sizeof(rgb_group)+1 );
+
    if (x1==0 && y1==0 &&
        img->xsize-1==x2 && img->ysize-1==y2)
    {
       *dest=*img;
-      new=malloc( (x2-x1+1)*(y2-y1+1)*sizeof(rgb_group) + 1);
-      if (!new) 
-	resource_error(NULL,0,0,"memory",0,"Out of memory.\n");
       THREADS_ALLOW();
       MEMCPY(new,img->img,(x2-x1+1)*(y2-y1+1)*sizeof(rgb_group));
       THREADS_DISALLOW();
@@ -229,10 +229,6 @@ void img_crop(struct image *dest,
       return;
    }
 
-   new=malloc( (x2-x1+1)*(y2-y1+1)*sizeof(rgb_group) +1);
-   if (!new)
-     resource_error(NULL,0,0,"memory",0,"Out of memory.\n");
-
    img_clear(new,THIS->rgb,(x2-x1+1)*(y2-y1+1));
 
    dest->xsize=x2-x1+1;
@@ -264,8 +260,7 @@ void img_crop(struct image *dest,
 void img_clone(struct image *newimg,struct image *img)
 {
    if (newimg->img) free(newimg->img);
-   newimg->img=malloc(sizeof(rgb_group)*img->xsize*img->ysize +1);
-   if (!newimg->img) resource_error(NULL,0,0,"memory",0,"Out of memory.\n");
+   newimg->img=xalloc(sizeof(rgb_group)*img->xsize*img->ysize+1);
    THREADS_ALLOW();
    MEMCPY(newimg->img,img->img,sizeof(rgb_group)*img->xsize*img->ysize);
    THREADS_DISALLOW();
