@@ -2,7 +2,7 @@
 || This file is part of Pike. For copyright information see COPYRIGHT.
 || Pike is distributed under GPL, LGPL and MPL. See the file COPYING
 || for more information.
-|| $Id: any.c,v 1.32 2006/09/12 09:47:24 grubba Exp $
+|| $Id: any.c,v 1.34 2006/09/12 09:47:33 grubba Exp $
 */
 
 /*
@@ -25,7 +25,6 @@
 #include <math.h>
 #include <ctype.h>
 
-RCSID("$Id: any.c,v 1.32 2006/09/12 09:47:24 grubba Exp $");
 #include "operators.h"
 #include "builtin_functions.h"
 #include "interpret.h"
@@ -179,6 +178,11 @@ void image_any__decode(INT32 args)
 	 image_tim_f__decode(1);
 	 return;
 
+      case CHAR2(0xc5, 0xd0):
+	/* FIXME: DOS EPS Binary File Header. */
+	goto unknown_format;
+	break;
+
       case CHAR2(0,0):
 	 switch (CHAR2(sp[-args].u.string->str[2],sp[-args].u.string->str[3]))
 	 {
@@ -325,15 +329,11 @@ void image_any_decode_alpha(INT32 args)
 
 void init_image_any(void)
 {
-   add_function("_decode",image_any__decode,
-		"function(string:mapping)",0);
-   add_function("decode_header",image_any_decode_header,
-		"function(string:mapping)",0);
+  ADD_FUNCTION("_decode",image_any__decode,tFunc(tStr,tMapping), 0);
+  ADD_FUNCTION("decode_header",image_any_decode_header,tFunc(tStr,tMapping),0);
 
-   add_function("decode",image_any_decode,
-		"function(string:object)",0);
-   add_function("decode_alpha",image_any_decode_alpha,
-		"function(string:object)",0);
+  ADD_FUNCTION("decode",image_any_decode,tFunc(tStr,tObj),0);
+  ADD_FUNCTION("decode_alpha",image_any_decode_alpha,tFunc(tStr,tObj),0);
 }
 
 void exit_image_any(void)
