@@ -2,7 +2,7 @@
 || This file is part of Pike. For copyright information see COPYRIGHT.
 || Pike is distributed under GPL, LGPL and MPL. See the file COPYING
 || for more information.
-|| $Id: avs.c,v 1.18 2006/09/22 12:30:37 grubba Exp $
+|| $Id: avs.c,v 1.22 2008/05/01 21:51:12 mast Exp $
 */
 
 #include "global.h"
@@ -18,7 +18,6 @@
 #endif
 
 #include "stralloc.h"
-RCSID("$Id: avs.c,v 1.18 2006/09/22 12:30:37 grubba Exp $");
 #include "object.h"
 #include "interpret.h"
 #include "pike_error.h"
@@ -64,7 +63,8 @@ void image_avs_f__decode(INT32 args)
   w = q[0]<<24 | q[1]<<16 | q[2]<<8 | q[3];
   h = q[4]<<24 | q[5]<<16 | q[6]<<8 | q[7];
 
-  if( w <= 0 || h <= 0 || (w>>16)*(h>>16))
+  /* Check for under and overflow. */
+  if ((w <= 0) || (h <= 0) || (w>>16)*(h>>16))
     Pike_error("This is not an AVS file (w=%d; h=%d)\n", w, h);
 
   if((size_t)w*h*4+8 != (size_t)s->len)
@@ -79,7 +79,7 @@ void image_avs_f__decode(INT32 args)
   push_int( h );
   ao = clone_object( image_program, 2);
 
-  for(c=0; c<w*h; c++)
+  for(c=0; c< (unsigned) w * h; c++)
   {
     rgb_group pix, apix;
     apix.r = apix.g = apix.b = q[c*4+8];
@@ -143,9 +143,9 @@ void image_avs_f_encode(INT32 args )
 
 void init_image_avs()
 {
-  add_function( "decode",  image_avs_f_decode,  "function(string:object)", 0);
-  add_function( "_decode", image_avs_f__decode, "function(string:mapping)", 0);
-  add_function( "encode",  image_avs_f_encode,  "function(object:string)", 0);
+  ADD_FUNCTION( "decode",  image_avs_f_decode,  tFunc(tStr,tObj), 0);
+  ADD_FUNCTION( "_decode", image_avs_f__decode, tFunc(tStr,tMapping), 0);
+  ADD_FUNCTION( "encode",  image_avs_f_encode,  tFunc(tObj,tStr), 0);
 }
 
 void exit_image_avs()
