@@ -2,7 +2,7 @@
 || This file is part of Pike. For copyright information see COPYRIGHT.
 || Pike is distributed under GPL, LGPL and MPL. See the file COPYING
 || for more information.
-|| $Id: ilbm.c,v 1.28 2004/03/06 00:06:59 nilsson Exp $
+|| $Id: ilbm.c,v 1.31 2005/01/23 13:30:04 nilsson Exp $
 */
 
 /*
@@ -17,7 +17,6 @@
 #include "global.h"
 
 #include "stralloc.h"
-RCSID("$Id: ilbm.c,v 1.28 2004/03/06 00:06:59 nilsson Exp $");
 #include "object.h"
 #include "interpret.h"
 #include "svalue.h"
@@ -826,7 +825,7 @@ struct program *image_encoding_ilbm_program=NULL;
 
 void init_image_ilbm(void)
 {
-   static char *str[] = { "BMHD", "CMAP", "CAMG", "BODY" };
+   static const char *str[] = { "BMHD", "CMAP", "CAMG", "BODY" };
    int n;
 
    for(n=0; n<4; n++) {
@@ -835,15 +834,12 @@ void init_image_ilbm(void)
      pop_stack();
    }
 
-   
-   add_function("__decode",image_ilbm___decode,
-		"function(string:array)",0);
-   add_function("_decode",image_ilbm__decode,
-		"function(string|array:mapping)",0);
-   add_function("decode",img_ilbm_decode,
-		"function(string|array:object)",0);
-   add_function("encode",image_ilbm_encode,
-		"function(object,void|mapping(string:mixed):string)",0);
+   ADD_FUNCTION("__decode",image_ilbm___decode,tFunc(tStr,tArray),0);
+   ADD_FUNCTION("_decode",image_ilbm__decode,
+		tFunc(tOr(tStr,tArray),tMapping),0);
+   ADD_FUNCTION("decode",img_ilbm_decode,tFunc(tOr(tStr,tArray),tObj),0);
+   ADD_FUNCTION("encode",image_ilbm_encode,
+		tFunc(tObj tOr(tVoid,tMap(tStr,tMix)),tStr),0);
 }
 
 void exit_image_ilbm(void)
