@@ -2,7 +2,7 @@
 || This file is part of Pike. For copyright information see COPYRIGHT.
 || Pike is distributed under GPL, LGPL and MPL. See the file COPYING
 || for more information.
-|| $Id: tga.c,v 1.35 2004/03/06 00:07:00 nilsson Exp $
+|| $Id: tga.c,v 1.42 2008/01/27 20:28:41 grubba Exp $
 */
 
 /*
@@ -61,10 +61,6 @@
 #include "image.h"
 #include "colortable.h"
 
-
-
-RCSID("$Id: tga.c,v 1.35 2004/03/06 00:07:00 nilsson Exp $");
-
 #define ROUNDUP_DIVIDE(n,d) (((n) + (d - 1)) / (d))
 
 extern struct program *image_colortable_program;
@@ -276,6 +272,8 @@ static ptrdiff_t rle_fread (guchar *buf, size_t datasize, size_t nelems,
     count = std_fgetc (fp);
     if (count == EOF)
     {
+      if (statebuf)
+	free(statebuf);
       return j / datasize;
     }
 
@@ -300,6 +298,8 @@ static ptrdiff_t rle_fread (guchar *buf, size_t datasize, size_t nelems,
       /* Fill the buffer with the next value. */
       if (std_fread (p, datasize, 1, fp) != 1)
       {
+	if (statebuf)
+	  free(statebuf);
         return j / datasize;
       }
 
@@ -313,8 +313,10 @@ static ptrdiff_t rle_fread (guchar *buf, size_t datasize, size_t nelems,
     else
     {
       /* Read in the buffer. */
-      if (std_fread (p, bytes, 1, fp) != 1)
+      if (std_fread (p, bytes, 1, fp) != 1) {
+	free(statebuf);
         return j / datasize;
+      }
     }
 
     /* We may need to copy bytes from the state buffer. */
@@ -323,6 +325,8 @@ static ptrdiff_t rle_fread (guchar *buf, size_t datasize, size_t nelems,
     else
       j += bytes;
   }
+  if (statebuf)
+    free(statebuf);
   return nelems;
 }
 
@@ -501,6 +505,11 @@ static struct image_alpha ReadImage(struct buffer *fp, struct tga_header *hdr)
     really_no_alpha = 1;
   }
 
+
+  if((double)width * (double)height * bpp  > (double)INT_MAX )
+    Pike_error("Too large image (width * height * bpp overflows)\n");
+
+
   switch (hdr->imageType)
   {
    case TGA_TYPE_MAPPED_RLE:
@@ -722,6 +731,7 @@ static struct buffer save_tga(struct image *img, struct image *alpha,
   struct buffer obuf;
   struct buffer *fp = &buf;
   struct tga_header hdr;
+  ONERROR err;
   ptrdiff_t (*myfwrite)(unsigned char *, size_t, size_t,
 			struct buffer *);
 
@@ -739,7 +749,7 @@ static struct buffer save_tga(struct image *img, struct image *alpha,
   /* We like our images top-to-bottom, thank you! */
   hdr.descriptor |= TGA_DESC_VERTICAL;
 
-  /* Choose the imageType based on alpha precense and compression options. */
+  /* Choose the imageType based on alpha presence and compression options. */
 
   hdr.bpp = 24;
   hdr.imageType = TGA_TYPE_COLOR;
@@ -771,20 +781,15 @@ static struct buffer save_tga(struct image *img, struct image *alpha,
 
   buf.len = width*height*(alpha?4:3)+strlen(SAVE_ID_STRING)+sizeof(hdr)+65535;
   buf.str = xalloc(buf.len);
-  obuf.len = buf.len;
-  obuf.str = buf.str;
+  obuf = buf;
+  SET_ONERROR(err, free, obuf.str);
 
   /* Just write the header. */
   if (std_fwrite((void *)&hdr, sizeof (hdr), 1, fp) != 1)
-  {
-    free(obuf.str);
-    Pike_error("Internal error: Out of space in buffer.\n");
-  }
+    Pike_error(msg_out_of_mem);
+
   if (std_fwrite ((void *)SAVE_ID_STRING, hdr.idLength, 1, fp) != 1)
-  {
-    free(obuf.str);
-    Pike_error("Internal error: Out of space in buffer.\n");
-  }
+    Pike_error(msg_out_of_mem);
 
   /* Allocate a new set of pixels. */
 
@@ -799,13 +804,8 @@ static struct buffer save_tga(struct image *img, struct image *alpha,
     {
       rgb_group *as = alpha->img;
       pixsize++;
-      p = data = malloc( width*height*4 );
+      p = data = xalloc( width*height*4 );
       datalen = width*height*4;
-      if(!data)
-      {
-        free(obuf.str);
-        Pike_error("Out of memory while encoding image\n");
-      }
       for(y=0; y<height; y++)
         for(x=0; x<width; x++)
         {
@@ -816,13 +816,8 @@ static struct buffer save_tga(struct image *img, struct image *alpha,
 	  as++;
         }
     } else {
-      p = data = malloc( width*height*3 );
+      p = data = xalloc( width*height*3 );
       datalen = width*height*3;
-      if(!data)
-      {
-        free(obuf.str);
-        Pike_error("Out of memory while encoding image\n");
-      }
       for(y=0; y<height; y++)
         for(x=0; x<width; x++)
         {
@@ -835,12 +830,13 @@ static struct buffer save_tga(struct image *img, struct image *alpha,
 	datalen/pixsize)
     {
       free(data);
-      free(obuf.str);
-      Pike_error("Internal error: Out of space in buffer.\n");
+      Pike_error(msg_out_of_mem);
     }
     free(data);
   }
-  obuf.len -= buf.len;
+  UNSET_ONERROR(err);
+
+  obuf.len -= fp->len;
   return obuf;
 }
 
@@ -984,12 +980,10 @@ void image_tga_encode( INT32 args )
 static struct program *image_encoding_tga_program=NULL;
 void init_image_tga( )
 {
-   pike_add_function( "_decode", image_tga__decode,
-                 "function(string:mapping(string:object))", 0);
-   pike_add_function( "decode", image_tga_decode,
-                 "function(string:object)", 0);
-   pike_add_function( "encode", image_tga_encode,
-                 "function(object,mapping|void:string)", 0);
+   ADD_FUNCTION( "_decode", image_tga__decode, tFunc(tStr,tMap(tStr,tObj)), 0);
+   ADD_FUNCTION( "decode", image_tga_decode, tFunc(tStr,tObj), 0);
+   ADD_FUNCTION( "encode", image_tga_encode,
+		 tFunc(tObj tOr(tVoid,tMapping),tStr), 0);
 
    param_alpha=make_shared_string("alpha");
    param_raw=make_shared_string("raw");
