@@ -2,12 +2,10 @@
 || This file is part of Pike. For copyright information see COPYRIGHT.
 || Pike is distributed under GPL, LGPL and MPL. See the file COPYING
 || for more information.
-|| $Id: xcf.c,v 1.48 2004/03/06 00:07:00 nilsson Exp $
+|| $Id: xcf.c,v 1.56 2008/01/22 21:23:51 grubba Exp $
 */
 
 #include "global.h"
-RCSID("$Id: xcf.c,v 1.48 2004/03/06 00:07:00 nilsson Exp $");
-
 #include "image_machine.h"
 
 #include "pike_macros.h"
@@ -52,10 +50,8 @@ extern struct program *image_program;
 struct buffer
 {
   struct pike_string *s;
-  ptrdiff_t base_offset;
-  ptrdiff_t base_len;
-  size_t len;
   unsigned char *str;
+  size_t len;
 };
 
 struct substring
@@ -85,7 +81,7 @@ static void f_substring_index( INT32 args )
 
   if( i < 0 ) i = s->len + i;
   if( i >= s->len ) {
-    Pike_error("Index out of bounds, %d > %ld\n", i,
+    Pike_error("Index out of bounds, %ld > %ld\n", i,
 	  DO_NOT_WARN((long)s->len-1) );
   }
   push_int( ((unsigned char *)s->s->str)[s->offset+i] );
@@ -331,10 +327,8 @@ static struct buffer read_string( struct buffer *data )
 {
   struct buffer res = *data;
   res.len = xcf_read_int( data );
-  res.base_offset = (data->base_offset+(data->base_len-data->len));
   res.str = (unsigned char *)read_data( data, res.len );
   if(res.len > 0)  res.len--;  /* len includes ending \0 */
-  res.base_len = res.len;
   if(!res.str)
     Pike_error("String read failed\n");
   return res;
@@ -350,14 +344,10 @@ static struct property read_property( struct buffer * data )
     read_uint(data); /* bogus 'len'... */
     foo = read_uint( data );
     res.data.len = foo*3;
-    res.data.base_offset = data->base_offset+(data->base_len-data->len);
-    res.data.base_len = res.data.len;
     res.data.str = (unsigned char *)read_data( data,foo*3 );
     res.data.s   = data->s;
   } else {
     res.data.len = read_uint( data );
-    res.data.base_offset = data->base_offset+(data->base_len-data->len);
-    res.data.base_len = res.data.len;
     res.data.str = (unsigned char *)read_data( data,res.data.len );
     res.data.s   = data->s;
   }
@@ -765,7 +755,7 @@ static struct gimp_image read_image( struct buffer * data )
 
 static void push_buffer( struct buffer *b )
 {
-  push_substring( b->s, b->base_offset+(b->base_len-b->len), b->len );
+  push_substring( b->s, b->str - (unsigned char *)b->s->str, b->len );
 /* push_string( make_shared_binary_string( (char *)b->str, b->len ) );*/
 }
 
@@ -902,8 +892,6 @@ static void image_xcf____decode( INT32 args )
     Pike_error("Too many arguments to Image.XCF.___decode()\n");
 
   b.s = s;
-  b.base_offset = 0;
-  b.base_len = s->len;
   b.len = s->len;
   b.str = (unsigned char *)s->str;
 
@@ -1192,6 +1180,7 @@ void image_xcf_f__decode_tiles( INT32 args )
 
   INT_TYPE rle, bpp, span, shrink;
   unsigned int l, x=0, y=0, cx, cy;
+  ONERROR err;
   get_all_args( "_decode_tiles", args, "%o%O%a%i%i%O%i%d%d",
                 &io, &ao, &tiles, &rle, &bpp, &cmapo, &shrink, &rxs, &rys);
 
@@ -1216,7 +1205,8 @@ void image_xcf_f__decode_tiles( INT32 args )
 
   if(cmap)
   {
-    colortable = malloc(sizeof(rgb_group)*image_colortable_size( cmap ));
+    colortable = xalloc(sizeof(rgb_group)*image_colortable_size( cmap ));
+    SET_ONERROR(err, free, colortable);
     image_colortable_write_rgb( cmap, (unsigned char *)colortable );
   }
 
@@ -1235,16 +1225,20 @@ void image_xcf_f__decode_tiles( INT32 args )
     if(!tile_ss)
       continue;
 
+    tile.s = NULL;
     tile.str = (unsigned char *)(tile_ss->s->str + tile_ss->offset);
     tile.len = tile_ss->len;
 
     ewidth = MINIMUM(TILE_WIDTH, (rxs-x));
     eheight = MINIMUM(TILE_HEIGHT, (rys-y));
 
+    if( (double)ewidth * eheight * bpp > INT_MAX )
+      Pike_error("Insanely large tiles not supported\n");
     if(rle)
     {
       struct buffer s = tile, od, d;
       int i;
+      od.s = NULL;
       od.len = eheight*ewidth*bpp;  /* Max and only size, really */
       df = (char *)(od.str = (unsigned char *)xalloc( eheight*ewidth*bpp+1 ));
       d = od;
@@ -1403,8 +1397,10 @@ void image_xcf_f__decode_tiles( INT32 args )
     }
   }
   THREADS_DISALLOW();
-  if(colortable) 
+  if(colortable) {
+    UNSET_ONERROR(err);
     free( colortable );
+  }
 
   pop_n_elems(args);
   push_int(0);
@@ -1414,10 +1410,9 @@ void image_xcf_f__decode_tiles( INT32 args )
 static struct program *image_encoding_xcf_program=NULL;
 void init_image_xcf()
 {
-  add_function( "___decode", image_xcf____decode,
-                "function(string:mapping)", 0);
+  ADD_FUNCTION( "___decode", image_xcf____decode, tFunc(tStr,tMapping), 0);
 
-  add_function( "_decode_tiles", image_xcf_f__decode_tiles, "mixed", 0);
+  ADD_FUNCTION( "_decode_tiles", image_xcf_f__decode_tiles, tFunction, 0);
 
   add_integer_constant( "PROP_END", PROP_END,0 );
   add_integer_constant( "PROP_COLORMAP", PROP_COLORMAP, 0 );
@@ -1496,14 +1491,14 @@ void init_image_xcf()
 
   start_new_program();
   ADD_STORAGE( struct substring );
-  add_function("cast", f_substring_cast, "function(string:mixed)",0);
-  add_function("`[]", f_substring_index, "function(int:int)",0);
-  add_function("get_short", f_substring_get_short, "function(int:int)", 0 );
-  add_function("get_ushort", f_substring_get_ushort, "function(int:int)", 0 );
-  add_function("get_int", f_substring_get_int, "function(int:int)", 0 );
-  add_function("get_uint", f_substring_get_uint, "function(int:int)", 0 );
-  add_function("_sprintf",f_substring__sprintf,
-               "function(int,mapping:mixed)", 0);
+  ADD_FUNCTION("cast", f_substring_cast, tFunc(tStr,tMix), 0);
+  ADD_FUNCTION("`[]", f_substring_index, tFunc(tInt,tInt), 0);
+  ADD_FUNCTION("get_short", f_substring_get_short, tFunc(tInt,tInt), 0 );
+  ADD_FUNCTION("get_ushort", f_substring_get_ushort, tFunc(tInt,tInt), 0 );
+  ADD_FUNCTION("get_int", f_substring_get_int, tFunc(tInt,tInt), 0 );
+  ADD_FUNCTION("get_uint", f_substring_get_uint, tFunc(tInt,tInt), 0 );
+  ADD_FUNCTION("_sprintf",f_substring__sprintf, tFunc(tInt tMapping,tMix), 0);
+
 /*   set_init_callback(init_substring); */
   set_exit_callback(free_substring);
   substring_program = end_program();  
