@@ -2,7 +2,7 @@
 || This file is part of Pike. For copyright information see COPYRIGHT.
 || Pike is distributed under GPL, LGPL and MPL. See the file COPYING
 || for more information.
-|| $Id: xwd.c,v 1.26 2004/03/06 00:07:01 nilsson Exp $
+|| $Id: xwd.c,v 1.30 2009/08/06 14:41:35 grubba Exp $
 */
 
 /*
@@ -28,7 +28,6 @@
 #include <ctype.h>
 
 #include "stralloc.h"
-RCSID("$Id: xwd.c,v 1.26 2004/03/06 00:07:01 nilsson Exp $");
 #include "object.h"
 #include "interpret.h"
 #include "svalue.h"
@@ -250,7 +249,8 @@ void img_xwd__decode(INT32 args,int header_only,int skipcmap)
 
    /* header.ncolors XWDColor structs */
 
-   if (!skipcmap || 
+   /* NB: The cmap needs to be decoded if the image is to be decoded. */
+   if (!skipcmap || !header_only ||
        header.visual_class==3 || 
        header.visual_class==5)
    {
@@ -305,7 +305,7 @@ void img_xwd__decode(INT32 args,int header_only,int skipcmap)
       push_text("image");
 
       if (s->len-(int)(header.header_size+header.ncolors*12)<0)
-	 push_text("");
+	 push_empty_string();
       else
 	 push_string(make_shared_binary_string(
 	    s->str+(header.header_size+header.ncolors*12),
@@ -355,8 +355,13 @@ void img_xwd__decode(INT32 args,int header_only,int skipcmap)
 	    push_int(header.bits_per_pixel);
 	    push_int(header.bitmap_pad);
 	    push_int(header.byte_order==1);
-	    ref_push_object(co);
-	    image_x_decode_pseudocolor(7);
+	    if (!co)
+	       image_x_decode_pseudocolor(6);
+	    else
+	    {
+	       ref_push_object(co);
+	       image_x_decode_pseudocolor(7);
+	    }
 	    break;
 	 default:
 	    pop_stack();
@@ -403,14 +408,13 @@ struct program *image_xwd_module_program=NULL;
 
 void init_image_xwd(void)
 {
-   add_function("_decode",image_xwd__decode,
-		"function(string,void|int:mapping(string:int|array|object))",0);
-   add_function("decode",image_xwd_decode,
-		"function(string:object)",0);
+   ADD_FUNCTION("_decode",image_xwd__decode,
+		tFunc(tStr tOr(tInt,tVoid),
+		      tMap(tStr,tOr3(tInt,tArray,tObj))),0);
 
-   add_function("decode_header",image_xwd_decode_header,
-		"function(string:object)",0);
+   ADD_FUNCTION("decode",image_xwd_decode,tFunc(tStr,tObj),0);
 
+   ADD_FUNCTION("decode_header",image_xwd_decode_header,tFunc(tStr,tObj),0);
 }
 
 void exit_image_xwd(void)
