@@ -2,7 +2,7 @@
 || This file is part of Pike. For copyright information see COPYRIGHT.
 || Pike is distributed under GPL, LGPL and MPL. See the file COPYING
 || for more information.
-|| $Id: layers.c,v 1.97 2004/04/18 03:30:02 nilsson Exp $
+|| $Id: layers.c,v 1.100 2008/01/16 19:42:57 grubba Exp $
 */
 
 /*
@@ -15,8 +15,6 @@
 
 #include <math.h> /* floor */
 
-RCSID("$Id: layers.c,v 1.97 2004/04/18 03:30:02 nilsson Exp $");
-
 #include "image_machine.h"
 
 #include "stralloc.h"
@@ -1103,7 +1101,7 @@ static void image_layer_mode(INT32 args)
 	 return;
       }
 
-   Pike_fatal("illegal mode: %p\n", (void *)layer_mode[i].func);
+   Pike_fatal("illegal mode: %p\n", (void *)layer_mode[i-1].func);
 }
 
 static void image_layer_available_modes(INT32 args)
@@ -2940,7 +2938,7 @@ void image_lay(INT32 args)
       return;
    }
 
-   l=(struct layer**)xalloc(sizeof(struct layer)*layers);
+   l=(struct layer**)xalloc(sizeof(struct layer *)*layers);
 
    SET_ONERROR(err, free, l);
 
