@@ -2,7 +2,7 @@
 || This file is part of Pike. For copyright information see COPYRIGHT.
 || Pike is distributed under GPL, LGPL and MPL. See the file COPYING
 || for more information.
-|| $Id: matrix_code.h,v 1.8 2003/07/09 01:09:12 agehall Exp $
+|| $Id: matrix_code.h,v 1.21 2008/09/17 12:05:13 mast Exp $
 */
 
 /*
@@ -16,7 +16,8 @@
  *
  *   FTYPE:     The type of the element
  *   PUSH_ELEM: Push a element on the stack
- *   PTYPE:     the pike type as a string
+ *   PTYPE:     the pike type
+ *   PNAME:     The class name on Pike level
  */
 
 
@@ -57,10 +58,10 @@ static void matrixX(_create)(INT32 args)
    FTYPE *m=NULL;
 
    if (!args)
-      SIMPLE_TOO_FEW_ARGS_ERROR("Matrix",1);
+      SIMPLE_TOO_FEW_ARGS_ERROR(PNAME,1);
 
    if (THIS->m)
-      bad_arg_error("Matrix", Pike_sp-args, args, 1, "", Pike_sp-args,
+      bad_arg_error(PNAME, Pike_sp-args, args, 1, "", Pike_sp-args,
 		    "Has already been called.\n");
    
    if (Pike_sp[-args].type==T_ARRAY)
@@ -80,18 +81,18 @@ static void matrixX(_create)(INT32 args)
       {
 	struct array *a;
 	 if (Pike_sp[-args].u.array->item[i].type!=T_ARRAY)
-	    SIMPLE_BAD_ARG_ERROR("matrix",1,"array(array)");
+	    SIMPLE_BAD_ARG_ERROR(PNAME,1,"array(array)");
 	 if (i==0) 
 	 {
 	    xs=Pike_sp[-args].u.array->item[i].u.array->size;
 	    THIS->m=m=malloc(sizeof(FTYPE)*xs*ys);
 	    if (!m)
-	       SIMPLE_OUT_OF_MEMORY_ERROR("matrix",
+	       SIMPLE_OUT_OF_MEMORY_ERROR(PNAME,
 					  sizeof(FTYPE)*xs*ys);
 	 }
 	 else
 	    if (xs!=Pike_sp[-args].u.array->item[i].u.array->size)
-	       SIMPLE_BAD_ARG_ERROR("matrix",1,
+	       SIMPLE_BAD_ARG_ERROR(PNAME,1,
 				    "array of equal sized arrays");
 
 	 a = Pike_sp[-args].u.array->item[i].u.array;
@@ -100,24 +101,45 @@ static void matrixX(_create)(INT32 args)
 	    switch (a->item[j].type)
 	    {
 	       case T_INT:
-		  *(m++)=(FTYPE)
-		     a->item[j].u.integer;
+		  *(m++) = (FTYPE)a->item[j].u.integer;
 		  break;
 	       case T_FLOAT:
-		  *(m++)=(FTYPE)
-  	            a->item[j].u.float_number;
+		  *(m++) = (FTYPE)a->item[j].u.float_number;
 		  break;
 	      case T_OBJECT:
 		{
 		  INT64 x;
-		  if( int64_from_bignum( &x, a->item[j].u.object ) )
-		  {
-		    *(m++)=(FTYPE)x;
+#ifdef AUTO_BIGNUM
+		  if (a->item[j].u.object->prog != get_auto_bignum_program()) {
+#endif
+		    /* Use push_svalue() so that we support subtypes... */
+		    push_svalue(a->item+j);
+		    o_cast_to_int();
+		    if (Pike_sp[-1].type == T_INT) {
+		      *(m++) = (FTYPE)Pike_sp[-1].u.integer;
+		      pop_stack();
+		      break;
+#ifdef AUTO_BIGNUM
+		    } else if ((Pike_sp[-1].type == T_OBJECT) &&
+			       (Pike_sp[-1].u.object->prog ==
+				get_auto_bignum_program()) &&
+			       int64_from_bignum(&x, Pike_sp[-1].u.object)) {
+		      *(m++) = (FTYPE)x;
+		      pop_stack();
+		      break;
+#endif
+		    }
+		    pop_stack();
+#ifdef AUTO_BIGNUM
+		  } else if (int64_from_bignum(&x, a->item[j].u.object)) {
+		    *(m++) = (FTYPE)x;
 		    break;
 		  }
+#endif
 		}
+		/* FALL_THROUGH */
 	      default:
-		SIMPLE_BAD_ARG_ERROR("matrix",1,
+		SIMPLE_BAD_ARG_ERROR(PNAME,1,
 				     "array(array(int|float))");
 	    }
       }
@@ -128,18 +150,18 @@ static void matrixX(_create)(INT32 args)
       FTYPE z = DO_NOT_WARN((FTYPE)0.0);
 
       if (args<2)
-	 SIMPLE_TOO_FEW_ARGS_ERROR("matrix",2);
+	 SIMPLE_TOO_FEW_ARGS_ERROR(PNAME,2);
       if (Pike_sp[1-args].type!=T_INT)
-	 SIMPLE_BAD_ARG_ERROR("matrix",2,"int");
+	 SIMPLE_BAD_ARG_ERROR(PNAME,2,"int");
 
       if ((THIS->xsize=xs=Pike_sp[-args].u.integer)<=0)
-	 SIMPLE_BAD_ARG_ERROR("matrix",1,"int > 0");
+	 SIMPLE_BAD_ARG_ERROR(PNAME,1,"int > 0");
       if ((THIS->ysize=ys=Pike_sp[1-args].u.integer)<=0)
-	 SIMPLE_BAD_ARG_ERROR("matrix",2,"int > 0");
+	 SIMPLE_BAD_ARG_ERROR(PNAME,2,"int > 0");
 
       THIS->m=m=malloc(sizeof(FTYPE)*xs*ys);
       if (!m)
-	 SIMPLE_OUT_OF_MEMORY_ERROR("matrix",
+	 SIMPLE_OUT_OF_MEMORY_ERROR(PNAME,
 				    sizeof(FTYPE)*xs*ys);
       
       if (args>2) {
@@ -161,12 +183,12 @@ static void matrixX(_create)(INT32 args)
 	       args=2;
 	    }
 	    else
-	       SIMPLE_BAD_ARG_ERROR("Matrix",3,
-				    "valid matrix mode (identity)");
+	       SIMPLE_BAD_ARG_ERROR(PNAME,3,
+				    "valid matrix mode");
 	    /* insert other base matrices here */
 	 }
 	 else
-	    SIMPLE_BAD_ARG_ERROR("Matrix",3,"int|float|string");
+	    SIMPLE_BAD_ARG_ERROR(PNAME,3,"int|float|string");
       }
       
       xs*=ys;
@@ -189,11 +211,11 @@ done_made:
 
       if (Pike_sp[-args].u.string==s_identity)
       {
-	 get_all_args("Matrix",args,"%s%i",&dummy,&side);
+	 get_all_args(PNAME,args,"%s%i",&dummy,&side);
 
 	 THIS->xsize=THIS->ysize=side;
 	 THIS->m=m=malloc(sizeof(FTYPE)*side*side);
-	 if (!m) SIMPLE_OUT_OF_MEMORY_ERROR("Matrix",sizeof(FTYPE)*side*side);
+	 if (!m) SIMPLE_OUT_OF_MEMORY_ERROR(PNAME,sizeof(FTYPE)*side*side);
 
   	 n=side*side;
 	 while (n--) *(m++) = DO_NOT_WARN((FTYPE)0.0); 
@@ -203,8 +225,8 @@ done_made:
       }
       else if (Pike_sp[-args].u.string==s_rotate)
       {
-	 float r;
-	 float x,y,z;
+	 FLOAT_TYPE r;
+	 FLOAT_TYPE x,y,z;
 	 double c,s;
 	 struct matrixX(_storage) *mx=NULL;
 
@@ -215,24 +237,24 @@ done_made:
 	       get_storage(Pike_sp[-1].u.object,XmatrixY(math_,_program)))))
 	 {
 	    if (mx->xsize*mx->ysize!=3)
-	       SIMPLE_BAD_ARG_ERROR("Matrix",4,"Matrix of size 1x3 or 3x1");
+	       SIMPLE_BAD_ARG_ERROR(PNAME,4,"Matrix of size 1x3 or 3x1");
 	    
 	    x = mx->m[0];
 	    y = mx->m[1];
 	    z = mx->m[2];
 
-	    get_all_args("Matrix",args,"%s%i%F",&dummy,&side,&r);
+	    get_all_args(PNAME,args,"%s%i%F",&dummy,&side,&r);
 	 }
 	 else
-	    get_all_args("Matrix",args,"%s%i%F%F%F%F",
+	    get_all_args(PNAME,args,"%s%i%F%F%F%F",
 			 &dummy,&side,&r,&x,&y,&z);
 	 
 	 if (side<2)
-	    SIMPLE_BAD_ARG_ERROR("Matrix",2,"int(2..)");
+	    SIMPLE_BAD_ARG_ERROR(PNAME,2,"int(2..)");
 
 	 THIS->xsize=THIS->ysize=side;
 	 THIS->m=m=malloc(sizeof(FTYPE)*side*side);
-	 if (!m) SIMPLE_OUT_OF_MEMORY_ERROR("Matrix",sizeof(FTYPE)*side*side);
+	 if (!m) SIMPLE_OUT_OF_MEMORY_ERROR(PNAME,sizeof(FTYPE)*side*side);
 
 	 n=side*side;
 	 while (n--) *(m++) = DO_NOT_WARN((FTYPE)0.0);
@@ -255,11 +277,11 @@ done_made:
 	 }
       }
       else
-	 SIMPLE_BAD_ARG_ERROR("Matrix",1,
+	 SIMPLE_BAD_ARG_ERROR(PNAME,1,
 			      "valid matrix mode (identity or rotate)");
    }
    else
-      SIMPLE_BAD_ARG_ERROR("Matrix",1,"array|int");
+      SIMPLE_BAD_ARG_ERROR(PNAME,1,"array|int");
 
    pop_n_elems(args);
    push_int(0);
@@ -275,7 +297,7 @@ void matrixX(_cast)(INT32 args)
    }
 
    if (args)
-      if (Pike_sp[-1].type==T_STRING)
+      if (Pike_sp[-1].type==T_STRING) {
 	 if (Pike_sp[-1].u.string==s_array)
 	 {
 	    int i,j;
@@ -292,8 +314,11 @@ void matrixX(_cast)(INT32 args)
 	    f_aggregate(ys);
 	    return;
 	 }
+         else
+           Pike_error("Can only cast to array.\n");
+      }
 
-   SIMPLE_BAD_ARG_ERROR("matrix->cast",1,"string");
+   SIMPLE_BAD_ARG_ERROR("cast",1,"string");
 }
 
 
@@ -334,14 +359,14 @@ void matrixX(__sprintf)(INT32 args)
 	 if (THIS->ysize>80 || THIS->xsize>80 ||
 	     THIS->xsize*THIS->ysize > 500)
 	 {
-	    sprintf(buf,"Math.Matrix( %d x %d elements )",
+	    sprintf(buf,"Math." PNAME "( %d x %d elements )",
 		    THIS->xsize,THIS->ysize);
 	    push_text(buf);
 	    stack_pop_n_elems_keep_top(args);
 	    return;
 	 }
 
-	 push_constant_text("Math.Matrix( ({ ({ ");
+	 push_constant_text("Math." PNAME "( ({ ({ ");
 	 n=1;
 	 for (y=0; y<THIS->ysize; y++)
 	 {
@@ -415,8 +440,8 @@ static void matrixX(_norm)(INT32 args)
    pop_n_elems(args);
 
    if (!(THIS->xsize==1 || THIS->ysize==1))
-      math_error("Matrix->norm",Pike_sp-args,args,0,
-		 "Cannot compute norm of non 1xn or nx1 matrices");
+      math_error("norm",Pike_sp-args,args,0,
+		 "Cannot compute norm of non 1xn or nx1 matrices.\n");
    
    z=0.0;
    s=THIS->m;
@@ -435,8 +460,8 @@ static void matrixX(_norm2)(INT32 args)
   pop_n_elems(args);
 
   if (!(THIS->xsize==1 || THIS->ysize==1))
-      math_error("Matrix->norm",Pike_sp-args,args,0,
-		 "Cannot compute norm of non 1xn or nx1 matrices");
+      math_error("norm2",Pike_sp-args,args,0,
+		 "Cannot compute norm of non 1xn or nx1 matrices.\n");
    
    z=0.0;
    s=THIS->m;
@@ -472,16 +497,16 @@ static void matrixX(_add)(INT32 args)
    FTYPE *s1,*s2,*d;
 
    if (args<1)
-      SIMPLE_TOO_FEW_ARGS_ERROR("matrix->`+",1);
+      SIMPLE_TOO_FEW_ARGS_ERROR("`+",1);
 
    if (Pike_sp[-1].type!=T_OBJECT ||
        !((mx=(struct matrixX(_storage)*)
 	  get_storage(Pike_sp[-1].u.object,XmatrixY(math_,_program)))))
-      SIMPLE_BAD_ARG_ERROR("matrix->`+",1,"object(Math.Matrix)");
+      SIMPLE_BAD_ARG_ERROR("`+",1,"object(Math.Matrix)");
 
    if (mx->xsize != THIS->xsize || mx->ysize != THIS->ysize)
-      math_error("Matrix->`+",Pike_sp-args,args,0,
-		 "Can't add matrices of different size");
+      math_error("`+",Pike_sp-args,args,0,
+		 "Cannot add matrices of different size.\n");
 
    pop_n_elems(args-1); /* shouldn't be needed */
    
@@ -511,12 +536,12 @@ static void matrixX(_sub)(INT32 args)
       if (Pike_sp[-1].type!=T_OBJECT ||
 	  !((mx=(struct matrixX(_storage)*)
 	     get_storage(Pike_sp[-1].u.object,XmatrixY(math_,_program)))))
-	 SIMPLE_BAD_ARG_ERROR("matrix->`-",1,"object(Math.Matrix)");
+	 SIMPLE_BAD_ARG_ERROR("`-",1,"object(Math.Matrix)");
 
       if (mx->xsize != THIS->xsize ||
 	  mx->ysize != THIS->ysize)
-	 math_error("Matrix->`-",Pike_sp-args,args,0,
-		    "Can't add matrices of different size");
+	 math_error("`-",Pike_sp-args,args,0,
+		    "Cannot add matrices of different size.\n");
 
       pop_n_elems(args-1); /* shouldn't be needed */
 
@@ -569,8 +594,8 @@ static void matrixX(_max)(INT32 args)
 
    n=THIS->xsize*THIS->ysize;
    s=THIS->m;
-   if (!n) math_error("Matrix->max", Pike_sp-args, args, 0,
-		      "Cannot do max() from a zero-sized matrix");
+   if (!n) math_error("max", Pike_sp-args, args, 0,
+		      "Cannot do max() from a zero-sized matrix.\n");
    max=*(s++);
    while (--n) { if (*s>max) max=*s; s++; }
    
@@ -588,8 +613,8 @@ static void matrixX(_min)(INT32 args)
 
    n=THIS->xsize*THIS->ysize;
    s=THIS->m;
-   if (!n) math_error("Matrix->min", Pike_sp-args, args, 0,
-		      "Cannot do min() from a zero-sized matrix");
+   if (!n) math_error("min", Pike_sp-args, args, 0,
+		      "Cannot do min() from a zero-sized matrix.\n");
    min=*(s++);
    while (--n) { if (*s<min) min=*s; s++; }
    
@@ -607,7 +632,7 @@ static void matrixX(_mult)(INT32 args)
    FTYPE z;
 
    if (args<1)
-      SIMPLE_TOO_FEW_ARGS_ERROR("matrix->`*",1);
+      SIMPLE_TOO_FEW_ARGS_ERROR("`*",1);
 
    pop_n_elems(args-1); /* shouldn't be needed */
 
@@ -637,20 +662,20 @@ scalar_mult:
    if (Pike_sp[-1].type!=T_OBJECT ||
        !((mx=(struct matrixX(_storage)*)
 	  get_storage(Pike_sp[-1].u.object,XmatrixY(math_,_program)))))
-      SIMPLE_BAD_ARG_ERROR("matrix->`*",1,"object(Math.Matrix)");
+      SIMPLE_BAD_ARG_ERROR("`*",1,"object(Math.Matrix)");
 
    if (mx->xsize != THIS->ysize)
-      math_error("Matrix->`*",Pike_sp-args,args,0,
-		 "Incompatible matrices");
+      math_error("`*",Pike_sp-args,args,0,
+		 "Incompatible matrices.\n");
 
    m=THIS->xsize;
    n=THIS->ysize; /* == mx->xsize */
    p=mx->ysize;
 
-   dmx=matrixX(_push_new_)(m,p);
+   dmx=matrixX(_push_new_)(p,m);
 
-   s1=THIS->m;
-   s2=mx->m;
+   s1=mx->m;
+   s2=THIS->m;
    d=dmx->m;
    for (k=0; k<p; k++)
       for (i=0; i<m; i++)
@@ -674,19 +699,19 @@ static void matrixX(_cross)(INT32 args)
    FTYPE *a,*b,*d;
 
    if (args<1)
-      SIMPLE_TOO_FEW_ARGS_ERROR("matrix->`",1);
+      SIMPLE_TOO_FEW_ARGS_ERROR("cross",1);
 
    pop_n_elems(args-1); /* shouldn't be needed */
 
    if (Pike_sp[-1].type!=T_OBJECT ||
        !((mx=(struct matrixX(_storage)*)
 	  get_storage(Pike_sp[-1].u.object,XmatrixY(math_,_program)))))
-      SIMPLE_BAD_ARG_ERROR("matrix->`",1,"object(Math.Matrix)");
+      SIMPLE_BAD_ARG_ERROR("cross",1,"object(Math.Matrix)");
 
    if (mx->xsize*mx->ysize != 3 ||
        THIS->ysize*THIS->xsize != 3)
-      math_error("Matrix->`",Pike_sp-args,args,0,
-		 "Matrices must both be of size 1x3 or 3x1");
+      math_error("cross",Pike_sp-args,args,0,
+		 "Matrices must both be of size 1x3 or 3x1.\n");
 
    dmx=matrixX(_push_new_)(THIS->xsize,THIS->ysize);
    a=THIS->m;
@@ -709,20 +734,20 @@ static void matrixX(_dot)(INT32 args)
   FTYPE *a,*b;
   
   if (args<1)
-     SIMPLE_TOO_FEW_ARGS_ERROR("matrix->`",1);
+     SIMPLE_TOO_FEW_ARGS_ERROR("dot_product",1);
   
   pop_n_elems(args-1); 
   
   if (Pike_sp[-1].type!=T_OBJECT ||
       !((mx=(struct matrixX(_storage)*)
 	 get_storage(Pike_sp[-1].u.object,XmatrixY(math_,_program)))))
-    SIMPLE_BAD_ARG_ERROR("matrix->`",1,"object(Math.Matrix)");
+    SIMPLE_BAD_ARG_ERROR("dot_product",1,"object(Math.Matrix)");
   
   if(!(mx->xsize==THIS->xsize &&
        mx->ysize==THIS->ysize &&
        (mx->xsize==1 || mx->ysize==1)))
-    math_error("Matrix->`",Pike_sp-args,args,0,
-	       "Matrices must be the same sizes, and one-dimensional\n");
+    math_error("dot_product",Pike_sp-args,args,0,
+	       "Matrices must be the same sizes, and one-dimensional.\n");
   
   res=(FTYPE)0;
   num=THIS->xsize+THIS->ysize;
@@ -748,17 +773,17 @@ static void matrixX(_convolve)(INT32 args)
    FTYPE *bs,*as,*d;
 
    if (args<1)
-      SIMPLE_TOO_FEW_ARGS_ERROR("matrix->`*",1);
+      SIMPLE_TOO_FEW_ARGS_ERROR("convolve",1);
 
    if (Pike_sp[-args].type!=T_OBJECT ||
        !((bmx=(struct matrixX(_storage)*)
 	  get_storage(Pike_sp[-args].u.object,XmatrixY(math_,_program)))))
-      SIMPLE_BAD_ARG_ERROR("matrix->something",1,"object(Math.Matrix)");
+      SIMPLE_BAD_ARG_ERROR("convolve",1,"object(Math.Matrix)");
 
    if (bmx->xsize==0 || bmx->ysize==0 ||
        THIS->xsize==0 || THIS->ysize==0)
-      math_error("matrix->something",Pike_sp-args,args,0,
-		 "source or argument matrix too small (zero size)");
+      math_error("convolve",Pike_sp-args,args,0,
+		 "Source or argument matrix too small (zero size).\n");
 
    bxz=bmx->xsize;
    byz=bmx->ysize;
@@ -901,6 +926,18 @@ static void matrixX(_convolve)(INT32 args)
 }
 
 
+static void matrixX(_xsize)(INT32 args)
+{
+  pop_n_elems(args);
+  push_int( THIS->xsize );
+}
+
+static void matrixX(_ysize)(INT32 args)
+{
+  pop_n_elems(args);
+  push_int( THIS->ysize );
+}
+
 
 /* ---------------------------------------------------------------- */
 
@@ -921,73 +958,55 @@ void Xmatrix(init_math_)(void)
    set_init_callback(Xmatrix(init_));
    set_exit_callback(Xmatrix(exit_));
 
-   add_function("create",matrixX(_create),
-		"function(array(array(int|float)):object)|"
-		"function(array(int|float):object)|"
-		"function(string,mixed...:object)|"
-		"function(int(1..),int(1..),int|float|string|void:object)",
-		0);
+   ADD_FUNCTION("create",matrixX(_create),
+		tOr4( tFunc(tArr(tArr(tOr(tInt,tFloat))), tVoid),
+		      tFunc(tArr(tOr(tInt,tFloat)), tVoid),
+		      tFuncV(tStr, tMix, tVoid),
+		      tFunc(tInt1Plus tInt1Plus tOr4(tInt,tFloat,tString,tVoid), tVoid)), ID_PROTECTED);
+
+   ADD_FUNCTION("cast",matrixX(_cast),
+		tFunc(tStr, tArr(tArr(tFloat))), 0);
+   ADD_FUNCTION("vect",matrixX(_vect), tFunc(tNone,tArr(PTYPE)), 0);
+   ADD_FUNCTION("_sprintf",matrixX(__sprintf), tFunc(tInt tMapping, tStr), 0);
+
+   ADD_FUNCTION("transpose",matrixX(_transpose), tFunc(tNone, tObj), 0);
+   ADD_FUNCTION("t",matrixX(_transpose), tFunc(tNone, tObj), 0);
+
+   ADD_FUNCTION("norm",matrixX(_norm), tFunc(tNone, tFloat), 0);
+   ADD_FUNCTION("norm2",matrixX(_norm2), tFunc(tNone, tFloat), 0);
+   ADD_FUNCTION("normv",matrixX(_normv), tFunc(tNone, tObj), 0);
+
+   ADD_FUNCTION("sum",matrixX(_sum), tFunc(tNone,PTYPE), 0);
+   ADD_FUNCTION("max",matrixX(_max), tFunc(tNone,PTYPE), 0);
+   ADD_FUNCTION("min",matrixX(_min), tFunc(tNone,PTYPE), 0);
+
+   ADD_FUNCTION("add",matrixX(_add), tFunc(tObj, tObj), 0);
+   ADD_FUNCTION("`+",matrixX(_add), tFunc(tObj, tObj), 0);
+   ADD_FUNCTION("sub",matrixX(_sub), tFunc(tObj, tObj), 0);
+   ADD_FUNCTION("`-",matrixX(_sub), tFunc(tObj, tObj), 0);
+
+   ADD_FUNCTION("mult",matrixX(_mult),
+		tFunc(tOr3(tObj,tFloat,tInt), tObj), 0);
+   ADD_FUNCTION("`*",matrixX(_mult),
+		tFunc(tOr3(tObj,tFloat,tInt), tObj), 0);
+   ADD_FUNCTION("``*",matrixX(_mult),
+		tFunc(tOr3(tObj,tFloat,tInt), tObj), 0);
+
+   ADD_FUNCTION("`",matrixX(_dot),
+		tFunc(tOr3(tObj,tFloat,tInt), tObj), 0);
+   ADD_FUNCTION("``",matrixX(_dot),
+		tFunc(tOr3(tObj,tFloat,tInt), tObj), 0);
+
+   ADD_FUNCTION("dot_product",matrixX(_dot), tFunc(tObj, tObj), 0);
+
+   ADD_FUNCTION("convolve",matrixX(_convolve), tFunc(tObj, tObj), 0);
    
-   add_function("cast",matrixX(_cast),
-		"function(string:array(array(float)))",0);
-   add_function("vect",matrixX(_vect),
-		"function(:array(" PTYPE "))",0);
-   add_function("_sprintf",matrixX(__sprintf),
-		"function(int,mapping:string)",0);
-
-   add_function("transpose",matrixX(_transpose),
-		"function(:object)",0);
-   add_function("t",matrixX(_transpose),
-		"function(:object)",0);
-
-   add_function("norm",matrixX(_norm),
-		"function(:float)",0);
-   add_function("norm2",matrixX(_norm2),
-		"function(:float)",0);
-   add_function("normv",matrixX(_normv),
-		"function(:object)",0);
-
-   add_function("sum",matrixX(_sum),
-		"function(:" PTYPE ")",0);
-   add_function("max",matrixX(_max),
-		"function(:" PTYPE ")",0);
-   add_function("min",matrixX(_min),
-		"function(:" PTYPE ")",0);
-
-   add_function("add",matrixX(_add),
-		"function(object:object)",0);
-   add_function("`+",matrixX(_add),
-		"function(object:object)",0);
-   add_function("sub",matrixX(_sub),
-		"function(object:object)",0);
-   add_function("`-",matrixX(_sub),
-		"function(object:object)",0);
-
-   add_function("mult",matrixX(_mult),
-		"function(object|float|int:object)",0);
-   add_function("`*",matrixX(_mult),
-		"function(object|float|int:object)",0);
-   add_function("``*",matrixX(_mult),
-		"function(object|float|int:object)",0);
-
-   add_function("`",matrixX(_dot),
-		"function(object|float|int:object)",0);
-   add_function("``",matrixX(_dot),
-		"function(object|float|int:object)",0);
-
-   add_function("dot_product",matrixX(_dot),
-		"function(object:object)",0);
-
-   add_function("convolve",matrixX(_convolve),
-		"function(object:object)",0);
-   
-   add_function("cross",matrixX(_cross),
-		"function(object:object)",0);
-   add_function("`",matrixX(_cross),
-		"function(object:object)",0);
-   add_function("``",matrixX(_cross),
-		"function(object:object)",0);
+   ADD_FUNCTION("cross",matrixX(_cross), tFunc(tObj, tObj), 0);
+   ADD_FUNCTION("`",matrixX(_cross), tFunc(tObj, tObj), 0);
+   ADD_FUNCTION("``",matrixX(_cross), tFunc(tObj, tObj), 0);
 
+   ADD_FUNCTION("xsize", matrixX(_xsize), tFunc(tNone, tInt), 0);
+   ADD_FUNCTION("ysize", matrixX(_ysize), tFunc(tNone, tInt), 0);
 
    Pike_compiler->new_program->flags |= 
      PROGRAM_CONSTANT |
