@@ -2,7 +2,7 @@
 || This file is part of Pike. For copyright information see COPYRIGHT.
 || Pike is distributed under GPL, LGPL and MPL. See the file COPYING
 || for more information.
-|| $Id: dummy.c,v 1.16 2007/03/10 16:07:59 marcus Exp $
+|| $Id: dummy.c,v 1.18 2007/03/20 16:39:21 grubba Exp $
 */
 
 /*
@@ -25,6 +25,10 @@
 
 #include <stdio.h>
 
+#ifdef HAVE_IO_H
+#include <io.h>
+#endif
+
 typedef INT64 _ll_t;
 typedef unsigned INT64 _ull_t;
 
@@ -49,7 +53,7 @@ _ll_t mysql_dummy_dum_dum(_ull_t a, _ull_t b, _ll_t c, _ll_t d) {
   fileno(stderr);
 #endif
 #ifdef HAVE_PUTS
-  puts(0);
+  puts(" ");
 #endif
 #ifdef HAVE_FGETS
   fgets(0, 0, 0);
@@ -63,7 +67,6 @@ _ll_t mysql_dummy_dum_dum(_ull_t a, _ull_t b, _ll_t c, _ll_t d) {
 #ifdef HAVE__FINDCLOSE
   _findclose(0);
 #endif
-
   return(a%b+(c%d)+(c/d)+(a/b));
 }
 
