@@ -2,7 +2,7 @@
 || This file is part of Pike. For copyright information see COPYRIGHT.
 || Pike is distributed under GPL, LGPL and MPL. See the file COPYING
 || for more information.
-|| $Id: mysql.c,v 1.92 2006/11/17 18:43:17 mast Exp $
+|| $Id: mysql.c,v 1.117 2009/07/24 12:42:39 mast Exp $
 */
 
 /*
@@ -98,8 +98,6 @@
  * Globals
  */
 
-RCSID("$Id: mysql.c,v 1.92 2006/11/17 18:43:17 mast Exp $");
-
 /*! @module Mysql
  *!
  *! This module enables access to the Mysql database from within Pike.
@@ -159,8 +157,6 @@ static MUTEX_T stupid_port_lock;
 #define MYSQL_DISALLOW()
 #endif /* _REENTRANT */
 
-#define string_has_null(STR)	((ptrdiff_t)strlen((STR)->str) != (STR)->len)
-
 #define CHECK_8BIT_NONBINARY_STRING(FUNC, ARG) do {			\
     if (sp[ARG-1-args].type != T_STRING ||				\
 	sp[ARG-1-args].u.string->size_shift ||				\
@@ -187,18 +183,15 @@ static void init_mysql_struct(struct object *o)
 {
   MEMSET(PIKE_MYSQL, 0, sizeof(struct precompiled_mysql));
   INIT_MYSQL_LOCK();
-  PIKE_MYSQL->mysql = (MYSQL *)xalloc(sizeof(MYSQL));
-#if defined(HAVE_MYSQL_REAL_CONNECT)
-  mysql_init(PIKE_MYSQL->mysql);
-#endif /* HAVE_MYSQL_REAL_CONNECT */
+  PIKE_MYSQL->mysql = mysql_init(NULL);
+  if (!PIKE_MYSQL->mysql)
+    Pike_error ("Out of memory when initializing mysql connection.\n");
 }
 
 static void exit_mysql_struct(struct object *o)
 {
-  MYSQL *socket = PIKE_MYSQL->socket;
   MYSQL *mysql = PIKE_MYSQL->mysql;
 
-  PIKE_MYSQL->socket = NULL;
   PIKE_MYSQL->mysql = NULL;
 
   if (PIKE_MYSQL->password) {
@@ -230,19 +223,16 @@ static void exit_mysql_struct(struct object *o)
 
   MYSQL_ALLOW();
 
-  if (socket) {
-    mysql_close(socket);
-  }
   if (mysql) {
-    free(mysql);
+    mysql_close(mysql);
   }
 
   MYSQL_DISALLOW();
 
   DESTROY_MYSQL_LOCK();
 }
-void pike_mysql_set_ssl(struct mapping *options) {
-
+void pike_mysql_set_ssl(struct mapping *options)
+{
 #ifdef HAVE_MYSQL_SSL
     char *ssl_key = NULL;
     char *ssl_cert = NULL;
@@ -250,7 +240,12 @@ void pike_mysql_set_ssl(struct mapping *options) {
     char *ssl_capath = NULL;
     char *ssl_cipher = NULL;
     struct svalue *val = NULL;
+#endif /* HAVE_MYSQL_SSL */
 
+    /* For some reason, we may get here without an options mapping. */
+    if(!options) return;
+
+#ifdef HAVE_MYSQL_SSL
     if ((val = simple_mapping_string_lookup(options, "ssl_key")) &&
 	(val->type == T_STRING) &&
 	(!val->u.string->size_shift))
@@ -402,10 +397,8 @@ static void pike_mysql_reconnect (int reconnect)
     password = PIKE_MYSQL->password->str;
   }
 
-  socket = PIKE_MYSQL->socket;
-  PIKE_MYSQL->socket = NULL;
-
-  if ((val = simple_mapping_string_lookup(PIKE_MYSQL->options,
+  if (PIKE_MYSQL->options &&
+      (val = simple_mapping_string_lookup(PIKE_MYSQL->options,
 					  "connect_options")) &&
       (val->type == T_INT) && (val->u.integer)) {
     options = (unsigned int)val->u.integer;
@@ -423,11 +416,6 @@ static void pike_mysql_reconnect (int reconnect)
   STUPID_PORT_LOCK();
 #endif /* HAVE_MYSQL_PORT || HAVE_MYSQL_UNIX_PORT */
 
-  if (socket) {
-    /* Disconnect the old connection */
-    mysql_close(socket);
-  }
-
 #ifdef HAVE_MYSQL_PORT
   if (port) {
     saved_port = mysql_port;
@@ -469,7 +457,7 @@ static void pike_mysql_reconnect (int reconnect)
     free(hostptr);
   }
   
-  if (!(PIKE_MYSQL->socket = socket)) {
+  if (!socket) {
     const char *err;
     MYSQL_ALLOW();
     err = mysql_error (mysql);
@@ -478,9 +466,9 @@ static void pike_mysql_reconnect (int reconnect)
 	       reconnect ? "reconnect" : "connect", err);
   }
 
-  if (socket->net.fd >= 0) {
+  if (mysql->net.fd >= 0) {
     /* Make sure the fd gets closed on exec. */
-    set_close_on_exec(socket->net.fd, 1);
+    set_close_on_exec(mysql->net.fd, 1);
   }
 
   if (database) {
@@ -488,17 +476,11 @@ static void pike_mysql_reconnect (int reconnect)
 
     MYSQL_ALLOW();
 
-    tmp = mysql_select_db(socket, database);
+    tmp = mysql_select_db(mysql, database);
 
     MYSQL_DISALLOW();
 
     if (tmp) {
-      PIKE_MYSQL->socket = NULL;
-      MYSQL_ALLOW();
-
-      mysql_close(socket);
-
-      MYSQL_DISALLOW();
       if (strlen(database) < 1024) {
 	Pike_error("Mysql.mysql(): Couldn't select database \"%s\"\n", database);
       } else {
@@ -663,7 +645,7 @@ static void f_create(INT32 args)
 #ifndef HAVE_MYSQL_SET_CHARACTER_SET
 #ifdef HAVE_MYSQL_CHARACTER_SET_NAME
   {
-    const char *charset = mysql_character_set_name (PIKE_MYSQL->socket);
+    const char *charset = mysql_character_set_name (PIKE_MYSQL->mysql);
     if (PIKE_MYSQL->conn_charset)
       free_string (PIKE_MYSQL->conn_charset);
     if (charset)
@@ -688,7 +670,6 @@ static void f_create(INT32 args)
 static void mysql__sprintf(INT32 args)
 {
   INT_TYPE type;
-  struct mapping * flags;
 
   if(args < 1)
     SIMPLE_TOO_FEW_ARGS_ERROR("_sprintf",2);
@@ -701,23 +682,23 @@ static void mysql__sprintf(INT32 args)
   {
     case 'O':
     {
-      MYSQL *socket = PIKE_MYSQL->socket;
+      MYSQL *mysql = PIKE_MYSQL->mysql;
 
-      if (socket) {
+      if (mysql) {
 	const char *info;
 	MYSQL_ALLOW();
-	info = mysql_get_host_info(socket);
+	info = mysql_get_host_info(mysql);
 	MYSQL_DISALLOW();
 	push_text("mysql(/*%s%s*/)");
-	push_text(info);
+	push_text(info ? info : "");
 #ifdef HAVE_MYSQL_SSL
 	if (PIKE_MYSQL->mysql->options.use_ssl) {
 	  push_text(", SSL");
 	}
 	else
-	  push_text("");
+	  push_empty_string();
 #else
-	push_text("");
+	push_empty_string();
 #endif /* HAVE_MYSQL_SSL */
 	f_sprintf(3);
       }
@@ -744,17 +725,17 @@ static void mysql__sprintf(INT32 args)
  */
 static void f_affected_rows(INT32 args)
 {
-  MYSQL *socket;
+  MYSQL *mysql;
   INT64 count;
 
-  if (!PIKE_MYSQL->socket) {
+  if (!PIKE_MYSQL->mysql) {
     pike_mysql_reconnect (1);
   }
   pop_n_elems(args);
-  socket = PIKE_MYSQL->socket;
+  mysql = PIKE_MYSQL->mysql;
 
   MYSQL_ALLOW();
-  count = mysql_affected_rows(socket);
+  count = mysql_affected_rows(mysql);
   MYSQL_DISALLOW();
 
   push_int64(count);
@@ -767,18 +748,18 @@ static void f_affected_rows(INT32 args)
  */
 static void f_insert_id(INT32 args)
 {
-  MYSQL *socket;
+  MYSQL *mysql;
   INT64 id;
 
-  if (!PIKE_MYSQL->socket) {
+  if (!PIKE_MYSQL->mysql) {
     pike_mysql_reconnect (1);
   }
   pop_n_elems(args);
 
-  socket = PIKE_MYSQL->socket;
+  mysql = PIKE_MYSQL->mysql;
 
   MYSQL_ALLOW();
-  id = mysql_insert_id(socket);
+  id = mysql_insert_id(mysql);
   MYSQL_DISALLOW();
 
   push_int64(id);
@@ -792,18 +773,18 @@ static void f_insert_id(INT32 args)
  */
 static void f_error(INT32 args)
 {
-  MYSQL *socket;
+  MYSQL *mysql;
   const char *error_msg;
 
-  if (!PIKE_MYSQL->socket) {
+  if (!PIKE_MYSQL->mysql) {
     pike_mysql_reconnect (1);
   }
 
-  socket = PIKE_MYSQL->socket;
+  mysql = PIKE_MYSQL->mysql;
 
   MYSQL_ALLOW();
 
-  error_msg = mysql_error(socket);
+  error_msg = mysql_error(mysql);
 
   MYSQL_DISALLOW();
 
@@ -828,7 +809,7 @@ static void f_error(INT32 args)
  */
 static void f_select_db(INT32 args)
 {
-  MYSQL *socket = PIKE_MYSQL->socket;
+  MYSQL *mysql = PIKE_MYSQL->mysql;
   char *database;
   int tmp = -1;
 
@@ -839,22 +820,22 @@ static void f_select_db(INT32 args)
 
   database = sp[-args].u.string->str;
 
-  if (socket) {
+  if (mysql) {
     MYSQL_ALLOW();
 
-    tmp = mysql_select_db(socket, database);
+    tmp = mysql_select_db(mysql, database);
 
     MYSQL_DISALLOW();
   }
-  if (!socket || tmp) {
+  if (!mysql || tmp) {
     /* The connection might have been closed. */
     pike_mysql_reconnect (1);
 
-    socket = PIKE_MYSQL->socket;
+    mysql = PIKE_MYSQL->mysql;
 
     MYSQL_ALLOW();
 
-    tmp = mysql_select_db(socket, database);
+    tmp = mysql_select_db(mysql, database);
 
     MYSQL_DISALLOW();
   }
@@ -863,7 +844,7 @@ static void f_select_db(INT32 args)
     const char *err;
 
     MYSQL_ALLOW();
-    err = mysql_error(socket);
+    err = mysql_error(mysql);
     MYSQL_DISALLOW();
 
     Pike_error("Mysql.mysql->select_db(): Couldn't select database \"%s\" (%s)\n",
@@ -879,69 +860,77 @@ static void f_select_db(INT32 args)
 
 #define PIKE_MYSQL_FLAG_STORE_RESULT	1
 
-static void f_big_query(INT32 args)
+static void low_query(INT32 args, char *name, int flags)
 {
-  MYSQL *socket = PIKE_MYSQL->socket;
+  MYSQL *mysql = PIKE_MYSQL->mysql;
   MYSQL_RES *result = NULL;
   char *query;
   int qlen;
   int tmp = -1;
 
   if (!args) {
-    SIMPLE_TOO_FEW_ARGS_ERROR ("big_query", 1);
+    SIMPLE_TOO_FEW_ARGS_ERROR (name, 1);
   }
 #ifdef HAVE_MYSQL_REAL_QUERY
-  CHECK_8BIT_STRING ("big_query", 1);
+  CHECK_8BIT_STRING (name, 1);
 #else
-  CHECK_8BIT_NONBINARY_STRING ("big_query", 1);
+  CHECK_8BIT_NONBINARY_STRING (name, 1);
 #endif
 
   query = sp[-args].u.string->str;
   qlen = sp[-args].u.string->len;
 
-  if (socket) {
+  if (mysql) {
     MYSQL_ALLOW();
 
 #ifdef HAVE_MYSQL_REAL_QUERY
-    tmp = mysql_real_query(socket, query, qlen);
+    tmp = mysql_real_query(mysql, query, qlen);
 #else
-    tmp = mysql_query(socket, query);
+    tmp = mysql_query(mysql, query);
 #endif /* HAVE_MYSQL_REAL_QUERY */
 
     if (!tmp) {
-      result = mysql_store_result(socket);
+      if (flags & PIKE_MYSQL_FLAG_STORE_RESULT) {
+	result = mysql_store_result(mysql);
+      } else {
+	result = mysql_use_result(mysql);
+      }
     }
 
     MYSQL_DISALLOW();
   }
-  if (socket && tmp) {
+  if (mysql && tmp) {
     /* Check if we need to reconnect. */
 #if defined(CR_SERVER_GONE_ERROR) && defined(CR_UNKNOWN_ERROR)
-    int eno = mysql_errno(socket);
+    int eno = mysql_errno(mysql);
     if ((eno == CR_SERVER_GONE_ERROR) ||
 	(eno == CR_UNKNOWN_ERROR)) {
-      socket = NULL;
+      mysql = NULL;
     }
 #else /* !CR_SERVER_GONE_ERROR || !CR_UNKNOWN_ERROR */
-    socket = NULL;
+    mysql = NULL;
 #endif /* CR_SERVER_GONE_ERROR && CR_UNKNOWN_ERROR */
   }
-  if (!socket) {
+  if (!mysql) {
     /* The connection might have been closed. */
     pike_mysql_reconnect (1);
 
-    socket = PIKE_MYSQL->socket;
+    mysql = PIKE_MYSQL->mysql;
 
     MYSQL_ALLOW();
 
 #ifdef HAVE_MYSQL_REAL_QUERY
-    tmp = mysql_real_query(socket, query, qlen);
+    tmp = mysql_real_query(mysql, query, qlen);
 #else
-    tmp = mysql_query(socket, query);
+    tmp = mysql_query(mysql, query);
 #endif /* HAVE_MYSQL_REAL_QUERY */
 
     if (!tmp) {
-      result = mysql_store_result(socket);
+      if (flags & PIKE_MYSQL_FLAG_STORE_RESULT) {
+	result = mysql_store_result(mysql);
+      } else {
+	result = mysql_use_result(mysql);
+      }
     }
 
     MYSQL_DISALLOW();
@@ -951,14 +940,14 @@ static void f_big_query(INT32 args)
     const char *err;
 
     MYSQL_ALLOW();
-    err = mysql_error(socket);
+    err = mysql_error(mysql);
     MYSQL_DISALLOW();
 
     if (sp[-args].u.string->len <= 512) {
-      Pike_error("big_query(): Query \"%s\" failed (%s)\n",
-		 query, err);
+      Pike_error("%s(): Query \"%s\" failed (%s)\n",
+		 name, query, err);
     } else {
-      Pike_error("big_query(): Query failed (%s)\n", err);
+      Pike_error("%s(): Query failed (%s)\n", name, err);
     }
   }
 
@@ -970,24 +959,24 @@ static void f_big_query(INT32 args)
     MYSQL_ALLOW();
     err = (
 #ifdef mysql_field_count
-	   mysql_field_count(socket)
+	   mysql_field_count(mysql)
 #else /* !mysql_field_count */
 #ifdef mysql_num_fields
-	   mysql_num_fields(socket)
+	   mysql_num_fields(mysql)
 #else /* !mysql_num_fields */
-	   socket->field_count
+	   mysql->field_count
 #endif /* mysql_num_fields */
 #endif /* mysql_field_count */
-	   && mysql_error(socket)[0]);
+	   && mysql_error(mysql)[0]);
     MYSQL_DISALLOW();
 
     if (err) {
       const char *msg;
       MYSQL_ALLOW();
-      msg = mysql_error (socket);
+      msg = mysql_error(mysql);
       MYSQL_DISALLOW();
-      Pike_error("big_query(): Couldn't create result for query (%s)\n",
-		 msg);
+      Pike_error("%s(): Couldn't create result for query (%s)\n",
+		 name, msg);
     }
     /* query was INSERT or similar - return 0 */
 
@@ -1005,12 +994,22 @@ static void f_big_query(INT32 args)
     if ((!(res = (struct precompiled_mysql_result *)
 	   get_storage(o, mysql_result_program))) || res->result) {
       mysql_free_result(result);
-      Pike_error("big_query(): Bad mysql result object!\n");
+      Pike_error("%s(): Bad mysql result object!\n", name);
     }
     res->result = result;
   }
 }
 
+static void f_big_query(INT32 args)
+{
+  low_query(args, "big_query", PIKE_MYSQL_FLAG_STORE_RESULT);
+}
+
+static void f_streaming_query(INT32 args)
+{
+  low_query(args, "streaming_query", 0);
+}
+
 
 /*! @decl void create_db(string database)
  *!
@@ -1025,7 +1024,7 @@ static void f_big_query(INT32 args)
 #ifdef USE_OLD_FUNCTIONS
 static void f_create_db(INT32 args)
 {
-  MYSQL *socket = PIKE_MYSQL->socket;
+  MYSQL *mysql = PIKE_MYSQL->mysql;
   char *database;
   int tmp = -1;
 
@@ -1044,21 +1043,21 @@ static void f_create_db(INT32 args)
   }
   database = sp[-args].u.string->str;
 
-  if (socket) {
+  if (mysql) {
     MYSQL_ALLOW();
 
-    tmp = mysql_create_db(socket, database);
+    tmp = mysql_create_db(mysql, database);
     MYSQL_DISALLOW();
   }
-  if (!socket || tmp) {
+  if (!mysql || tmp) {
     /* The connection might have been closed */
     pike_mysql_reconnect (1);
 
-    socket = PIKE_MYSQL->socket;
+    mysql = PIKE_MYSQL->mysql;
 
     MYSQL_ALLOW();
 
-    tmp = mysql_create_db(socket, database);
+    tmp = mysql_create_db(mysql, database);
 
     MYSQL_DISALLOW();
   }
@@ -1082,7 +1081,7 @@ static void f_create_db(INT32 args)
  */
 static void f_drop_db(INT32 args)
 {
-  MYSQL *socket = PIKE_MYSQL->socket;
+  MYSQL *mysql = PIKE_MYSQL->mysql;
   char *database;
   int tmp = -1;
 
@@ -1101,22 +1100,22 @@ static void f_drop_db(INT32 args)
   }
   database = sp[-args].u.string->str;
 
-  if (socket) {
+  if (mysql) {
     MYSQL_ALLOW();
 
-    tmp = mysql_drop_db(socket, database);
+    tmp = mysql_drop_db(mysql, database);
 
     MYSQL_DISALLOW();
   }
-  if (!socket || tmp) {
+  if (!mysql || tmp) {
     /* The connection might have been closed */
     pike_mysql_reconnect (1);
 
-    socket = PIKE_MYSQL->socket;
+    mysql = PIKE_MYSQL->mysql;
 
     MYSQL_ALLOW();
 
-    tmp = mysql_drop_db(socket, database);
+    tmp = mysql_drop_db(mysql, database);
 
     MYSQL_DISALLOW();
   }    
@@ -1141,34 +1140,34 @@ static void f_drop_db(INT32 args)
  */
 static void f_shutdown(INT32 args)
 {
-  MYSQL *socket = PIKE_MYSQL->socket;
+  MYSQL *mysql = PIKE_MYSQL->mysql;
   int tmp = -1;
 
-  if (socket) {
+  if (mysql) {
     MYSQL_ALLOW();
   
 #ifdef HAVE_SHUTDOWN_DEFAULT
     /* Mysql 4.1.3 added an extra shutdown_level argument. */
-    tmp = mysql_shutdown(socket, SHUTDOWN_DEFAULT);
+    tmp = mysql_shutdown(mysql, SHUTDOWN_DEFAULT);
 #else /* !HAVE_SHUTDOWN_DEFAULT */
-    tmp = mysql_shutdown(socket);
+    tmp = mysql_shutdown(mysql);
 #endif /* HAVE_SHUTDOWN_DEFAULT */
 
     MYSQL_DISALLOW();
   }
-  if (!socket || tmp) {
+  if (!mysql || tmp) {
     /* The connection might have been closed */
     pike_mysql_reconnect (1);
 
-    socket = PIKE_MYSQL->socket;
+    mysql = PIKE_MYSQL->mysql;
 
     MYSQL_ALLOW();
   
 #ifdef HAVE_SHUTDOWN_DEFAULT
     /* Mysql 4.1.3 added an extra shutdown_level argument. */
-    tmp = mysql_shutdown(socket, SHUTDOWN_DEFAULT);
+    tmp = mysql_shutdown(mysql, SHUTDOWN_DEFAULT);
 #else /* !HAVE_SHUTDOWN_DEFAULT */
-    tmp = mysql_shutdown(socket);
+    tmp = mysql_shutdown(mysql);
 #endif /* HAVE_SHUTDOWN_DEFAULT */
 
     MYSQL_DISALLOW();
@@ -1192,25 +1191,25 @@ static void f_shutdown(INT32 args)
  */
 static void f_reload(INT32 args)
 {
-  MYSQL *socket = PIKE_MYSQL->socket;
+  MYSQL *mysql = PIKE_MYSQL->mysql;
   int tmp = -1;
 
-  if (socket) {
+  if (mysql) {
     MYSQL_ALLOW();
 
-    tmp = mysql_reload(socket);
+    tmp = mysql_reload(mysql);
 
     MYSQL_DISALLOW();
   }
-  if (!socket || tmp) {
+  if (!mysql || tmp) {
     /* The connection might have been closed */
     pike_mysql_reconnect (1);
 
-    socket = PIKE_MYSQL->socket;
+    mysql = PIKE_MYSQL->mysql;
 
     MYSQL_ALLOW();
 
-    tmp = mysql_reload(socket);
+    tmp = mysql_reload(mysql);
 
     MYSQL_DISALLOW();
   }
@@ -1233,19 +1232,19 @@ static void f_reload(INT32 args)
  */
 static void f_statistics(INT32 args)
 {
-  MYSQL *socket = PIKE_MYSQL->socket;
+  MYSQL *mysql = PIKE_MYSQL->mysql;
   const char *stats;
 
-  if (!socket) {
+  if (!mysql) {
     pike_mysql_reconnect (1);
-    socket = PIKE_MYSQL->socket;
+    mysql = PIKE_MYSQL->mysql;
   }
 
   pop_n_elems(args);
 
   MYSQL_ALLOW();
 
-  stats = mysql_stat(socket);
+  stats = mysql_stat(mysql);
 
   MYSQL_DISALLOW();
 
@@ -1261,12 +1260,12 @@ static void f_statistics(INT32 args)
  */
 static void f_server_info(INT32 args)
 {
-  MYSQL *socket = PIKE_MYSQL->socket;
+  MYSQL *mysql = PIKE_MYSQL->mysql;
   const char *info;
 
-  if (!socket) {
+  if (!mysql) {
     pike_mysql_reconnect (1);
-    socket = PIKE_MYSQL->socket;
+    mysql = PIKE_MYSQL->mysql;
   }
 
   pop_n_elems(args);
@@ -1275,7 +1274,7 @@ static void f_server_info(INT32 args)
 
   MYSQL_ALLOW();
 
-  info = mysql_get_server_info(socket);
+  info = mysql_get_server_info(mysql);
 
   MYSQL_DISALLOW();
 
@@ -1292,20 +1291,20 @@ static void f_server_info(INT32 args)
  */
 static void f_host_info(INT32 args)
 {
-  MYSQL *socket;
+  MYSQL *mysql;
   const char *info;
 
-  if (!PIKE_MYSQL->socket) {
+  if (!PIKE_MYSQL->mysql) {
     pike_mysql_reconnect (1);
   }
 
-  socket = PIKE_MYSQL->socket;
+  mysql = PIKE_MYSQL->mysql;
 
   pop_n_elems(args);
 
   MYSQL_ALLOW();
 
-  info = mysql_get_host_info(socket);
+  info = mysql_get_host_info(mysql);
 
   MYSQL_DISALLOW();
 
@@ -1324,19 +1323,19 @@ static void f_host_info(INT32 args)
  */
 static void f_protocol_info(INT32 args)
 {
-  MYSQL *socket;
+  MYSQL *mysql;
   int prot;
 
-  if (!PIKE_MYSQL->socket) {
+  if (!PIKE_MYSQL->mysql) {
     pike_mysql_reconnect (1);
   }
 
   pop_n_elems(args);
 
-  socket = PIKE_MYSQL->socket;
+  mysql = PIKE_MYSQL->mysql;
 
   MYSQL_ALLOW();
-  prot = mysql_get_proto_info(socket);
+  prot = mysql_get_proto_info(mysql);
   MYSQL_DISALLOW();
 
   push_int(prot);
@@ -1357,7 +1356,7 @@ static void f_protocol_info(INT32 args)
  */
 static void f_list_dbs(INT32 args)
 {
-  MYSQL *socket = PIKE_MYSQL->socket;
+  MYSQL *mysql = PIKE_MYSQL->mysql;
   MYSQL_RES *result = NULL;
   char *wild = NULL;
 
@@ -1375,22 +1374,22 @@ static void f_list_dbs(INT32 args)
     wild = sp[-args].u.string->str;
   }
 
-  if (socket) {
+  if (mysql) {
     MYSQL_ALLOW();
 
-    result = mysql_list_dbs(socket, wild);
+    result = mysql_list_dbs(mysql, wild);
 
     MYSQL_DISALLOW();
   }
-  if (!socket || !result) {
+  if (!mysql || !result) {
     /* The connection might have been closed */
     pike_mysql_reconnect (1);
 
-    socket = PIKE_MYSQL->socket;
+    mysql = PIKE_MYSQL->mysql;
 
     MYSQL_ALLOW();
 
-    result = mysql_list_dbs(socket, wild);
+    result = mysql_list_dbs(mysql, wild);
 
     MYSQL_DISALLOW();
   }
@@ -1400,7 +1399,7 @@ static void f_list_dbs(INT32 args)
 
     MYSQL_ALLOW();
 
-    err = mysql_error(socket);
+    err = mysql_error(mysql);
 
     MYSQL_DISALLOW();
 
@@ -1443,7 +1442,7 @@ static void f_list_dbs(INT32 args)
  */
 static void f_list_tables(INT32 args)
 {
-  MYSQL *socket = PIKE_MYSQL->socket;
+  MYSQL *mysql = PIKE_MYSQL->mysql;
   MYSQL_RES *result = NULL;
   char *wild = NULL;
 
@@ -1461,22 +1460,22 @@ static void f_list_tables(INT32 args)
     wild = sp[-args].u.string->str;
   }
 
-  if (socket) {
+  if (mysql) {
     MYSQL_ALLOW();
 
-    result = mysql_list_tables(socket, wild);
+    result = mysql_list_tables(mysql, wild);
 
     MYSQL_DISALLOW();
   }
-  if (!socket || !result) {
+  if (!mysql || !result) {
     /* The connection might have been closed */
     pike_mysql_reconnect (1);
 
-    socket = PIKE_MYSQL->socket;
+    mysql = PIKE_MYSQL->mysql;
 
     MYSQL_ALLOW();
 
-    result = mysql_list_tables(socket, wild);
+    result = mysql_list_tables(mysql, wild);
 
     MYSQL_DISALLOW();
   }
@@ -1486,7 +1485,7 @@ static void f_list_tables(INT32 args)
 
     MYSQL_ALLOW();
 
-    err =  mysql_error(socket);
+    err =  mysql_error(mysql);
 
     MYSQL_DISALLOW();
 
@@ -1535,9 +1534,9 @@ static void f_list_tables(INT32 args)
  *!   @member string "type"
  *!	The SQL type of the field.
  *!   @member int "length"
- *!	The length of the field.
- *!   @member int "max_length"
- *!	The length of the longest element in this field.
+ *!	The length of the longest possible value that can be stored in
+ *!	the field. Note that this measures the display length in
+ *!	string form.
  *!   @member multiset(string) "flags"
  *!	Some flags.
  *!   @member int decimals
@@ -1572,7 +1571,7 @@ static void f_list_tables(INT32 args)
  */
 static void f_list_fields(INT32 args)
 {
-  MYSQL *socket = PIKE_MYSQL->socket;
+  MYSQL *mysql = PIKE_MYSQL->mysql;
   MYSQL_RES *result = NULL;
   MYSQL_FIELD *field;
   int i = 0;
@@ -1612,22 +1611,22 @@ static void f_list_fields(INT32 args)
     wild = sp[-args+1].u.string->str;
   }
 
-  if (socket) {
+  if (mysql) {
     MYSQL_ALLOW();
 
-    result = mysql_list_fields(socket, table, wild);
+    result = mysql_list_fields(mysql, table, wild);
 
     MYSQL_DISALLOW();
   }
-  if (!socket || !result) {
+  if (!mysql || !result) {
     /* The connection might have been closed */
     pike_mysql_reconnect (1);
 
-    socket = PIKE_MYSQL->socket;
+    mysql = PIKE_MYSQL->mysql;
 
     MYSQL_ALLOW();
 
-    result = mysql_list_fields(socket, table, wild);
+    result = mysql_list_fields(mysql, table, wild);
 
     MYSQL_DISALLOW();
   }
@@ -1637,7 +1636,7 @@ static void f_list_fields(INT32 args)
 
     MYSQL_ALLOW();
 
-    err = mysql_error(socket);
+    err = mysql_error(mysql);
 
     MYSQL_DISALLOW();
 
@@ -1667,27 +1666,27 @@ static void f_list_fields(INT32 args)
  */
 static void f_list_processes(INT32 args)
 {
-  MYSQL *socket = PIKE_MYSQL->socket;
+  MYSQL *mysql = PIKE_MYSQL->mysql;
   MYSQL_RES *result = NULL;
 
   pop_n_elems(args);
 
-  if (socket) {
+  if (mysql) {
     MYSQL_ALLOW();
 
-    result = mysql_list_processes(socket);
+    result = mysql_list_processes(mysql);
 
     MYSQL_DISALLOW();
   }
-  if (!socket || !result) {
+  if (!mysql || !result) {
     /* The connection might have been closed */
     pike_mysql_reconnect (1);
 
-    socket = PIKE_MYSQL->socket;
+    mysql = PIKE_MYSQL->mysql;
 
     MYSQL_ALLOW();
 
-    result = mysql_list_processes(socket);
+    result = mysql_list_processes(mysql);
 
     MYSQL_DISALLOW();
   }
@@ -1697,7 +1696,7 @@ static void f_list_processes(INT32 args)
 
     MYSQL_ALLOW();
 
-    err = mysql_error(socket);
+    err = mysql_error(mysql);
 
     MYSQL_DISALLOW();
 
@@ -1755,14 +1754,14 @@ static void f_set_charset (INT32 args)
 #ifdef HAVE_MYSQL_SET_CHARACTER_SET
   {
     int res;
-    MYSQL *socket = PIKE_MYSQL->socket;
+    MYSQL *mysql = PIKE_MYSQL->mysql;
     MYSQL_ALLOW();
-    res = mysql_set_character_set (socket, charset->str);
+    res = mysql_set_character_set (mysql, charset->str);
     MYSQL_DISALLOW();
     if (res) {
       const char *err;
       MYSQL_ALLOW();
-      err = mysql_error(socket);
+      err = mysql_error(mysql);
       MYSQL_DISALLOW();
       Pike_error("Setting the charset failed: %s\n", err);
     }
@@ -1773,7 +1772,7 @@ static void f_set_charset (INT32 args)
   ref_push_string (charset);
   push_constant_text ("'");
   f_add (3);
-  f_big_query(1);
+  low_query (1, "set_charset", PIKE_MYSQL_FLAG_STORE_RESULT);
   args++;
   if (PIKE_MYSQL->conn_charset)
     free_string (PIKE_MYSQL->conn_charset);
@@ -1790,7 +1789,7 @@ static void f_get_charset (INT32 args)
   {
     /* mysql_character_set_name should always exist if
      * mysql_set_character_set exists. */
-    const char *charset = mysql_character_set_name (PIKE_MYSQL->socket);
+    const char *charset = mysql_character_set_name (PIKE_MYSQL->mysql);
     if (charset)
       push_text (charset);
     else
@@ -1861,23 +1860,6 @@ static void f__can_send_as_latin1 (INT32 args)
 PIKE_MODULE_INIT
 {
 #ifdef HAVE_MYSQL
-  /*
-   * start_new_program();
-   *
-   * add_storage();
-   *
-   * add_function();
-   * add_function();
-   * ...
-   *
-   * set_init_callback();
-   * set_exit_callback();
-   *
-   * program = end_c_program();
-   * program->refs++;
-   *
-   */
-
   start_new_program();
   ADD_STORAGE(struct precompiled_mysql);
 
@@ -1895,7 +1877,10 @@ PIKE_MODULE_INIT
   /* function(string:void) */
   ADD_FUNCTION("select_db", f_select_db,tFunc(tStr,tVoid), ID_PUBLIC);
   /* function(string:int|object) */
-  ADD_FUNCTION("big_query", f_big_query,tFunc(tStr,tOr(tInt,tObj)), ID_PUBLIC);
+  ADD_FUNCTION("big_query", f_big_query,tFunc(tStr,tObj), ID_PUBLIC);
+  /* function(string:int|object) */
+  ADD_FUNCTION("streaming_query",
+	       f_streaming_query,tFunc(tStr,tObj), ID_PUBLIC);
 #ifdef USE_OLD_FUNCTIONS
   /* function(string:void) */
   ADD_FUNCTION("create_db", f_create_db,tFunc(tStr,tVoid), ID_PUBLIC);
@@ -1931,7 +1916,7 @@ PIKE_MODULE_INIT
   ADD_FUNCTION ("set_charset", f_set_charset, tFunc(tStr,tVoid), ID_PUBLIC);
   ADD_FUNCTION ("get_charset", f_get_charset, tFunc(tVoid,tStr), ID_PUBLIC);
   ADD_FUNCTION ("_can_send_as_latin1", f__can_send_as_latin1,
-		tFunc(tStr,tInt01), ID_STATIC);
+		tFunc(tStr,tInt01), ID_PROTECTED);
 
   add_integer_constant( "CLIENT_COMPRESS", CLIENT_COMPRESS, 0);
   add_integer_constant( "CLIENT_FOUND_ROWS", CLIENT_FOUND_ROWS, 0);
@@ -1961,6 +1946,9 @@ PIKE_MODULE_INIT
 #endif /* HAVE_MYSQL_PORT */
 
   init_mysql_res_programs();
+#else
+  if(!TEST_COMPAT(7,6))
+    HIDE_MODULE();
 #endif /* HAVE_MYSQL */
 }
 
