@@ -2,7 +2,7 @@
 || This file is part of Pike. For copyright information see COPYRIGHT.
 || Pike is distributed under GPL, LGPL and MPL. See the file COPYING
 || for more information.
-|| $Id: precompiled_mysql.h,v 1.17 2006/08/14 13:37:11 grubba Exp $
+|| $Id: precompiled_mysql.h,v 1.19 2008/08/20 13:34:00 grubba Exp $
 */
 
 /*
@@ -56,7 +56,7 @@ struct precompiled_mysql {
   DEFINE_MUTEX(lock);
 #endif /* PIKE_THREADS */
 
-  MYSQL		*mysql, *socket;
+  MYSQL		*mysql;
   struct pike_string	*host, *database, *user, *password;	/* Reconnect */
   struct mapping   *options;
 #ifndef HAVE_MYSQL_SET_CHARACTER_SET
