@@ -2,15 +2,20 @@
 || This file is part of Pike. For copyright information see COPYRIGHT.
 || Pike is distributed under GPL, LGPL and MPL. See the file COPYING
 || for more information.
-|| $Id: odbc_result.c,v 1.42 2006/07/04 11:09:13 grubba Exp $
+|| $Id: odbc_result.c,v 1.64 2009/05/11 11:36:13 grubba Exp $
 */
 
 /*
- * Pike  interface to ODBC compliant databases
+ * Pike interface to ODBC compliant databases
  *
  * Henrik Grubbstrm
  */
 
+/* FIXME: ODBC allows for multiple result sets from the same query.
+ * Support for SQLMoreResults should be added to support that (and it
+ * needs to work also in the case when the first result set has no
+ * columns). */
+
 /*
  * Includes
  */
@@ -21,8 +26,6 @@
 #include "config.h"
 #endif /* HAVE_CONFIG_H */
 
-RCSID("$Id: odbc_result.c,v 1.42 2006/07/04 11:09:13 grubba Exp $");
-
 #include "interpret.h"
 #include "object.h"
 #include "threads.h"
@@ -113,10 +116,13 @@ static void exit_res_struct(struct object *o)
 {
   if (PIKE_ODBC_RES->hstmt != SQL_NULL_HSTMT) {
     SQLHSTMT hstmt = PIKE_ODBC_RES->hstmt;
+    RETCODE code;
     PIKE_ODBC_RES->hstmt = SQL_NULL_HSTMT;
+    ODBC_ALLOW();
+    code = SQLFreeStmt(hstmt, SQL_DROP);
+    ODBC_DISALLOW();
     odbc_check_error("exit_res_struct", "Freeing of HSTMT failed",
-		     SQLFreeStmt(hstmt, SQL_DROP),
-		     (void (*)(void *))clean_sql_res, NULL);
+		     code, (void (*)(void *))clean_sql_res, NULL);
   }
   clean_sql_res();
 }
@@ -127,8 +133,11 @@ static void exit_res_struct(struct object *o)
 
 static void odbc_fix_fields(void)
 {
+  SQLHSTMT hstmt = PIKE_ODBC_RES->hstmt;
   int i;
   SWORD *odbc_field_types = alloca(sizeof(SWORD) * PIKE_ODBC_RES->num_fields);
+  SQLULEN *odbc_field_sizes =
+    alloca(sizeof(SQLULEN) * PIKE_ODBC_RES->num_fields);
   size_t buf_size = 1024;
 #ifdef SQL_WCHAR
   SQLWCHAR *buf = alloca(buf_size * sizeof(SQLWCHAR));
@@ -152,21 +161,25 @@ static void odbc_fix_fields(void)
     SWORD nullable = 0;
 
     while (1) {
-      odbc_check_error("odbc_fix_fields", "Failed to fetch field info",
+      RETCODE code;
+      ODBC_ALLOW();
+      code =
 #ifdef SQL_WCHAR
-		       SQLDescribeColW
+	SQLDescribeColW
 #else
-		       SQLDescribeCol
+	SQLDescribeCol
 #endif
-		       (PIKE_ODBC_RES->hstmt, i+1,
-			buf,
-			DO_NOT_WARN((SQLSMALLINT)buf_size),
-			&name_len,
-			&sql_type, &precision, &scale, &nullable),
-		       NULL, NULL);
+	(hstmt, i+1,
+	 buf,
+	 DO_NOT_WARN((SQLSMALLINT)buf_size),
+	 &name_len,
+	 &sql_type, &precision, &scale, &nullable);
+      ODBC_DISALLOW();
+      odbc_check_error("odbc_fix_fields", "Failed to fetch field info",
+		       code, NULL, NULL);
       if (name_len
 #ifdef SQL_WCHAR
-	  * sizeof(SQLWCHAR)
+	  * (ptrdiff_t)sizeof(SQLWCHAR)
 #endif
 	  < (ptrdiff_t)buf_size) {
 	break;
@@ -175,7 +188,7 @@ static void odbc_fix_fields(void)
 	buf_size *= 2;
       } while (name_len
 #ifdef SQL_WCHAR
-	       * sizeof(SQLWCHAR)
+	       * (ptrdiff_t)sizeof(SQLWCHAR)
 #endif
 	       >= (ptrdiff_t)buf_size);
       if (!(buf = alloca(
@@ -194,16 +207,19 @@ static void odbc_fix_fields(void)
 #else
 	    "name:%s\n"
 #endif
+	    "", buf);
+    fprintf(stderr,
+	    "name_len:%d\n"
 	    "sql_type:%d\n"
 	    "precision:%ld\n"
 	    "scale:%d\n"
 	    "nullable:%d\n",
-	    buf, sql_type, precision, scale, nullable);
+	    name_len, sql_type, precision, scale, nullable);
 #endif /* ODBC_DEBUG */
     /* Create the mapping */
     push_text("name");
 #ifdef SQL_WCHAR
-    push_sqlwchar(buf, name_len * sizeof(SQLWCHAR));
+    push_sqlwchar(buf, name_len);
 #else
     push_string(make_shared_binary_string((char *)buf, name_len));
 #endif
@@ -219,33 +235,42 @@ static void odbc_fix_fields(void)
 #endif /* ODBC_DEBUG */
     odbc_field_types[i] = SQL_C_CHAR;
 #endif
+    odbc_field_sizes[i] = precision;
     switch(sql_type) {
     case SQL_CHAR:
       push_text("char");
       break;
     case SQL_NUMERIC:
       push_text("numeric");
+      odbc_field_types[i] = SQL_C_CHAR;
       break;
     case SQL_DECIMAL:
       push_text("decimal");
+      odbc_field_types[i] = SQL_C_CHAR;
       break;
     case SQL_INTEGER:
       push_text("integer");
+      odbc_field_types[i] = SQL_C_CHAR;
       break;
     case SQL_SMALLINT:
       push_text("short");
+      odbc_field_types[i] = SQL_C_CHAR;
       break;
     case SQL_FLOAT:
       push_text("float");
+      odbc_field_types[i] = SQL_C_CHAR;
       break;
     case SQL_REAL:
       push_text("real");
+      odbc_field_types[i] = SQL_C_CHAR;
       break;
     case SQL_DOUBLE:
       push_text("double");
+      odbc_field_types[i] = SQL_C_CHAR;
       break;
     case SQL_VARCHAR:
       push_text("var string");
+      odbc_field_sizes[i] = 0;	/* Variable length */
       break;
     case SQL_DATE:
       push_text("date");
@@ -255,6 +280,7 @@ static void odbc_fix_fields(void)
       break;
     case SQL_LONGVARCHAR:
       push_text("var string");
+      odbc_field_sizes[i] = 0;	/* Variable length */
       break;
     case SQL_BINARY:
       push_text("binary");
@@ -269,6 +295,7 @@ static void odbc_fix_fields(void)
       fprintf(stderr, "SQL_C_BINARY\n");
 #endif /* ODBC_DEBUG */
       odbc_field_types[i] = SQL_C_BINARY;
+      odbc_field_sizes[i] = 0;	/* Variable length */
       break;
     case SQL_LONGVARBINARY:
       push_text("long blob");
@@ -276,6 +303,7 @@ static void odbc_fix_fields(void)
       fprintf(stderr, "SQL_C_BINARY\n");
 #endif /* ODBC_DEBUG */
       odbc_field_types[i] = SQL_C_BINARY;
+      odbc_field_sizes[i] = 0;	/* Variable length */
       break;
     case SQL_BIGINT:
       push_text("long integer");
@@ -306,7 +334,6 @@ static void odbc_fix_fields(void)
     f_aggregate_multiset(nbits);
 
     f_aggregate_mapping(5*2);
-
   }
   f_aggregate(PIKE_ODBC_RES->num_fields);
 
@@ -319,8 +346,19 @@ static void odbc_fix_fields(void)
   /*
    * Now it's time to bind the columns
    */
-  for (i=0; i < PIKE_ODBC_RES->num_fields; i++)
+  for (i=0; i < PIKE_ODBC_RES->num_fields; i++) {
     PIKE_ODBC_RES->field_info[i].type = odbc_field_types[i];
+#ifdef SQL_WCHAR
+    if (odbc_field_types[i] == SQL_C_WCHAR) {
+      /* NOTE: Field size is in characters, while SQLGetData()
+       *       wants bytes.
+       */
+      PIKE_ODBC_RES->field_info[i].size = odbc_field_sizes[i]
+	* (ptrdiff_t)sizeof(SQLWCHAR);
+    } else
+#endif 
+      PIKE_ODBC_RES->field_info[i].size = odbc_field_sizes[i];
+  }
 }
 
 /*
@@ -343,9 +381,16 @@ static void f_create(INT32 args)
   }
   add_ref(PIKE_ODBC_RES->obj = Pike_sp[-args].u.object);
 
-  odbc_check_error("odbc_result", "Statement allocation failed",
-		   SQLAllocStmt(PIKE_ODBC_RES->odbc->hdbc, &hstmt),
-		   NULL, NULL);
+  /* It's doubtful if this call really can block, but we play safe. /mast */
+  {
+    HDBC hdbc = PIKE_ODBC_RES->odbc->hdbc;
+    RETCODE code;
+    ODBC_ALLOW();
+    code = SQLAllocStmt(hdbc, &hstmt);
+    ODBC_DISALLOW();
+    odbc_check_error("odbc_result", "Statement allocation failed",
+		     code, NULL, NULL);
+  }
   PIKE_ODBC_RES->hstmt = hstmt;
 }
 
@@ -353,6 +398,10 @@ static void f_execute(INT32 args)
 {
   struct pike_string *q = NULL;
   SQLHSTMT hstmt = PIKE_ODBC_RES->hstmt;
+  RETCODE code;
+  const char *err_msg = NULL;
+  SWORD num_fields;
+  SQLLEN num_rows;
 
 #ifdef SQL_WCHAR
   get_all_args("odbc_result->execute", args, "%W", &q);
@@ -368,48 +417,76 @@ static void f_execute(INT32 args)
     char *to_free = NULL;
     SQLWCHAR *p;
     if ((sizeof(SQLWCHAR) == 4) && (q->size_shift == 1)) {
-      p = require_wstring2(q, &to_free);
+      p = (SQLWCHAR *)require_wstring2(q, &to_free);
     } else {
       p = (SQLWCHAR *)q->str;
     }
-    odbc_check_error("odbc_result->execute", "Query failed",
-		     SQLExecDirectW(hstmt, p,
-				    DO_NOT_WARN((SQLINTEGER)(q->len))),
-		     to_free?(void (*)(void *))free:NULL, to_free);
-  } else {
+    ODBC_ALLOW();
+    code = SQLExecDirectW(hstmt, p, DO_NOT_WARN((SQLINTEGER)(q->len)));
+    if (code != SQL_SUCCESS && code != SQL_SUCCESS_WITH_INFO)
+      err_msg = "Query failed";
+    else {
+      code = SQLNumResultCols(hstmt, &num_fields);
+      if (code != SQL_SUCCESS && code != SQL_SUCCESS_WITH_INFO)
+	err_msg = "Couldn't get the number of fields";
+      else {
+	code = SQLRowCount(hstmt, &num_rows);
+	if (code != SQL_SUCCESS && code != SQL_SUCCESS_WITH_INFO)
+	  err_msg = "Couldn't get the number of rows";
+      }
+    }
+    ODBC_DISALLOW();
+    if (to_free) free (to_free);
+  } else
 #endif
-    odbc_check_error("odbc_result->execute", "Query failed",
-		     SQLExecDirect(hstmt, STR0(q),
-				   DO_NOT_WARN((SQLINTEGER)(q->len))),
-		     NULL, NULL);
-#ifdef SQL_WCHAR
-  }
-#endif
-
-  odbc_check_error("odbc_result->execute", "Couldn't get the number of fields",
-		   SQLNumResultCols(hstmt, &(PIKE_ODBC_RES->num_fields)),
-		   NULL, NULL);
-
-  odbc_check_error("odbc_result->execute", "Couldn't get the number of rows",
-		   SQLRowCount(hstmt, &(PIKE_ODBC_RES->num_rows)),
-		   NULL, NULL);
-
-  PIKE_ODBC_RES->odbc->affected_rows = PIKE_ODBC_RES->num_rows;
-
-  if (PIKE_ODBC_RES->num_fields) {
-    odbc_fix_fields();
+  {
+    ODBC_ALLOW();
+    code = SQLExecDirect(hstmt, STR0(q), DO_NOT_WARN((SQLINTEGER)(q->len)));
+    if (code != SQL_SUCCESS && code != SQL_SUCCESS_WITH_INFO)
+      err_msg = "Query failed";
+    else {
+      code = SQLNumResultCols(hstmt, &num_fields);
+      if (code != SQL_SUCCESS && code != SQL_SUCCESS_WITH_INFO)
+	err_msg = "Couldn't get the number of fields";
+      else {
+	code = SQLRowCount(hstmt, &num_rows);
+	if (code != SQL_SUCCESS && code != SQL_SUCCESS_WITH_INFO)
+	  err_msg = "Couldn't get the number of rows";
+      }
+    }
+    ODBC_DISALLOW();
   }
 
-  pop_n_elems(args);
+#ifdef ODBC_DEBUG
+  fprintf (stderr, "ODBC:execute: SQLExecDirect returned %d, "
+	   "cols %d, rows %d\n", code, num_fields, num_rows);
+#endif
 
-  /* Result */
-  push_int(PIKE_ODBC_RES->num_fields);
+  if (err_msg)
+    odbc_error ("odbc_result->execute", err_msg, PIKE_ODBC_RES->odbc, hstmt,
+		code, NULL, NULL);
+  else {
+    PIKE_ODBC_RES->odbc->affected_rows = PIKE_ODBC_RES->num_rows = num_rows;
+    if (num_fields) {
+      PIKE_ODBC_RES->num_fields = num_fields;
+      odbc_fix_fields();
+    }
+
+    pop_n_elems(args);
+
+    /* Result */
+    push_int(num_fields);
+  }
 }
  
 static void f_list_tables(INT32 args)
 {
   struct pike_string *table_name_pattern = NULL;
   SQLHSTMT hstmt = PIKE_ODBC_RES->hstmt;
+  RETCODE code;
+  const char *err_msg = NULL;
+  SWORD num_fields;
+  SQLLEN num_rows;
 
   if (!args) {
     push_constant_text("%");
@@ -422,33 +499,40 @@ static void f_list_tables(INT32 args)
 
   table_name_pattern = Pike_sp[-args].u.string;
 
-  odbc_check_error("odbc_result->list_tables", "Query failed",
-		   SQLTables(hstmt, "%", 1, "%", 1,
-			     table_name_pattern->str,
-			     DO_NOT_WARN((SQLSMALLINT)table_name_pattern->len),
-			     "%", 1),
-		   NULL, NULL);
-
-  odbc_check_error("odbc_result->list_tables",
-		   "Couldn't get the number of fields",
-		   SQLNumResultCols(hstmt, &(PIKE_ODBC_RES->num_fields)),
-		   NULL, NULL);
-
-  odbc_check_error("odbc_result->list_tables",
-		   "Couldn't get the number of rows",
-		   SQLRowCount(hstmt, &(PIKE_ODBC_RES->num_rows)),
-		   NULL, NULL);
-
-  PIKE_ODBC_RES->odbc->affected_rows = PIKE_ODBC_RES->num_rows;
-
-  if (PIKE_ODBC_RES->num_fields) {
-    odbc_fix_fields();
+  ODBC_ALLOW();
+  code = SQLTables(hstmt, (SQLCHAR *) "%", 1, (SQLCHAR *) "%", 1,
+		   (SQLCHAR *) table_name_pattern->str,
+		   DO_NOT_WARN((SQLSMALLINT)table_name_pattern->len),
+		   (SQLCHAR *) "%", 1);
+  if (code != SQL_SUCCESS && code != SQL_SUCCESS_WITH_INFO)
+    err_msg = "Query failed";
+  else {
+    code = SQLNumResultCols(hstmt, &num_fields);
+    if (code != SQL_SUCCESS && code != SQL_SUCCESS_WITH_INFO)
+      err_msg = "Couldn't get the number of fields";
+    else {
+      code = SQLRowCount(hstmt, &num_rows);
+      if (code != SQL_SUCCESS && code != SQL_SUCCESS_WITH_INFO)
+	err_msg = "Couldn't get the number of rows";
+    }
   }
+  ODBC_DISALLOW();
 
-  pop_n_elems(args);
+  if (err_msg)
+    odbc_error ("odbc_result->list_tables", err_msg, PIKE_ODBC_RES->odbc, hstmt,
+		code, NULL, NULL);
+  else {
+    PIKE_ODBC_RES->odbc->affected_rows = PIKE_ODBC_RES->num_rows = num_rows;
+    if (num_fields) {
+      PIKE_ODBC_RES->num_fields = num_fields;
+      odbc_fix_fields();
+    }
 
-  /* Result */
-  push_int(PIKE_ODBC_RES->num_fields);
+    pop_n_elems(args);
+
+    /* Result */
+    push_int(PIKE_ODBC_RES->num_fields);
+  }
 }
  
 /* int num_rows() */
@@ -476,152 +560,234 @@ static void f_fetch_fields(INT32 args)
 /* int|array(string|float|int) fetch_row() */
 static void f_fetch_row(INT32 args)
 {
+  SQLHSTMT hstmt = PIKE_ODBC_RES->hstmt;
   int i;
   RETCODE code;
  
   pop_n_elems(args);
 
-  code = SQLFetch(PIKE_ODBC_RES->hstmt);
+  ODBC_ALLOW();
+  code = SQLFetch(hstmt);
+  ODBC_DISALLOW();
   
   if (code == SQL_NO_DATA_FOUND) {
     /* No rows left in result */
-    push_int(0);
+    push_undefined();
   } else {
     odbc_check_error("odbc->fetch_row", "Couldn't fetch row",
 		     code, NULL, NULL);
  
     for (i=0; i < PIKE_ODBC_RES->num_fields; i++) {
-	/* BLOB */
-      char blob_buf[(BLOB_BUFSIZ+1)
-#ifdef SQL_WCHAR
-		    * sizeof(SQLWCHAR)
-#endif
-		    ];
-	int num_strings = 0;
-	SQLLEN len = 0;
+      SQLLEN len = PIKE_ODBC_RES->field_info[i].size;
+      SWORD field_type = PIKE_ODBC_RES->field_info[i].type;
+      static char dummy_buf[4];
 
-	while(1) {
- #ifdef ODBC_DEBUG
-	  fprintf(stderr, "ODBC:fetch_row(): field_info[%d].type: ", i);
-	  if (PIKE_ODBC_RES->field_info[i].type == SQL_C_CHAR) {
-	    fprintf(stderr, "SQL_C_CHAR\n");
+      /* First get the size of the data.
+       *
+       * Note that this method of getting the size apparently isn't
+       * always supported for all data types (eg for INTEGER/MSSQL),
+       * where we instead use the width returned by SQLDescribeCol()
+       * earlier.
+       *
+       * Note also that FreeTDS/UnixODBC apparently has both a broken
+       * SQLDescribeCol() (which returns the SQL_C_BINARY size for
+       * eg INTEGER even if SQL_C_CHAR has been requested, and a
+       * SQLGetData() which doesn't properly support streamed fetching.
+       *
+       * We tried solving this by performing SQLGetData with zero size,
+       * and using the result if it succeeded, using the SQLDescribeCol()
+       * size if not zero, and otherwise throw an error. Unfortunately
+       * this didn't work, since MSSQL believes that it returned all the
+       * data in the first call (which failed with code "22003")...
+       *
+       * On unbroken ODBC implementations the UnixODBC method is
+       * preferred, since less buffer space is wasted.
+       */
+
+#ifdef __NT__
+      if (!len) {
+#endif /* __NT__ */
+
+	ODBC_ALLOW();
+
+#ifdef ODBC_DEBUG
+	fprintf(stderr, "ODBC:fetch_row(): Field %d: "
+		"SQLGetData(X, %d, %d, X, 0, X)...\n",
+		i + 1, i+1, field_type);
+#endif /* ODBC_DEBUG */
+	code = SQLGetData(hstmt, (SQLUSMALLINT)(i+1),
+			  field_type, dummy_buf, 0, &len);
 #ifdef SQL_WCHAR
-	  } else if (PIKE_ODBC_RES->field_info[i].type == SQL_C_WCHAR) {
-	    fprintf(stderr, "SQL_C_WCHAR\n");
+	if (code == SQL_NULL_DATA && (field_type == SQL_C_WCHAR)) {
+	  /* Kludge for FreeTDS which doesn't support WCHAR.
+	   * Refetch as a normal char.
+	   */
+#ifdef ODBC_DEBUG
+	  fprintf(stderr, "ODBC:fetch_row(): Field %d: WCHAR not supported.\n",
+		  i + 1);
+#endif /* ODBC_DEBUG */
+	  field_type = SQL_C_CHAR;
+#ifdef ODBC_DEBUG
+	  fprintf(stderr, "ODBC:fetch_row(): Field %d: "
+		  "SQLGetData(X, %d, %d, X, 0, X)...\n",
+		  i + 1, i+1, field_type);
+#endif /* ODBC_DEBUG */
+	  code = SQLGetData(hstmt, (SQLUSMALLINT)(i+1),
+			    field_type, dummy_buf, 0, &len);
+	}
+#endif
+
+#ifdef ODBC_DEBUG
+	fprintf(stderr, "ODBC:fetch_row(): Field %d: "
+		"SQLGetData(X, %d, %d, X, 0, X, X) ==> code: %d, len: %ld.\n",
+		i + 1,
+		i+1, field_type, code, len);
+#endif /* ODBC_DEBUG */
+
+	ODBC_DISALLOW();
+
+#ifdef SQL_WCHAR
+	/* In case the type got changed in the kludge above. */
+	PIKE_ODBC_RES->field_info[i].type = field_type;
+#endif
+
+#ifdef __NT__
+      }
+#endif /* __NT__ */
+
+#ifdef ODBC_DEBUG
+      fprintf(stderr, "ODBC:fetch_row(): Field %d: "
+	      "field_type: %d, len: %ld, code: %d.\n",
+	      i + 1, field_type, len, code);
+#endif /* ODBC_DEBUG */
+
+      if (code == SQL_NO_DATA_FOUND) {
+	/* All data already returned. */
+#ifdef ODBC_DEBUG
+	fprintf(stderr, "ODBC:fetch_row(): NO DATA FOUND\n");
+#endif /* ODBC_DEBUG */
+	push_empty_string();
+	continue;
+      }
+      if (!len) {
+	odbc_check_error("odbc->fetch_row", "SQLGetData() failed",
+			 code, NULL, NULL);
+      }
+      if (len == SQL_NULL_DATA) {
+#ifdef ODBC_DEBUG
+	fprintf(stderr, "ODBC:fetch_row(): NULL\n");
+#endif /* ODBC_DEBUG */
+	/* NULL */
+	push_undefined();
+      } else if (len == 0) {
+	push_empty_string();
+      } else {
+	struct pike_string *s;
+	SQLLEN bytes;
+	SQLLEN pad = 0;
+	int num_strings = 0;
+
+	switch(field_type) {
+	case SQL_C_CHAR:
+	  /* Adjust for NUL. */
+	  pad = 1;
+	  break;
+	default:
+	case SQL_C_BINARY:
+	  break;
+#ifdef SQL_WCHAR
+	case SQL_C_WCHAR:
+	  /* Adjust for NUL. */
+	  pad = sizeof(SQLWCHAR);
+	  break;
 #endif /* SQL_WCHAR */
-	  } else {
-	    fprintf(stderr, "%d\n", PIKE_ODBC_RES->field_info[i].type);
-	  }
-#endif /* ODBC_DEBUG */
-	  code = SQLGetData(PIKE_ODBC_RES->hstmt, (SQLUSMALLINT)(i+1),
-			    PIKE_ODBC_RES->field_info[i].type,
-			    blob_buf, BLOB_BUFSIZ
-#ifdef SQL_WCHAR
-			    * sizeof(SQLWCHAR)
-#endif
-			    , &len);
-	  if (code == SQL_NO_DATA_FOUND) {
-#ifdef ODBC_DEBUG
-	    fprintf(stderr, "ODBC:fetch_row(): NO DATA\n");
-#endif /* ODBC_DEBUG */
-	    if (!num_strings) {
-	      num_strings++;
-	      push_constant_text("");
-	    }
-	    break;
-	  }
-	  odbc_check_error("odbc->fetch_row", "SQLGetData() failed",
-			   code, NULL, NULL);
-	  if (len == SQL_NULL_DATA) {
-#ifdef ODBC_DEBUG
-	    fprintf(stderr, "ODBC:fetch_row(): NULL\n");
-#endif /* ODBC_DEBUG */
-	    if (!num_strings) {
-	      /* NULL */
-	      push_int(0);
-	    }
-	    break;
-	  } else if (code == SQL_SUCCESS_WITH_INFO) {
-	    /* Data truncated. */
-	    num_strings++;
-#ifdef ODBC_DEBUG
-	    fprintf(stderr, "[%d] ", num_strings);
-#endif /* ODBC_DEBUG */
-	    if (PIKE_ODBC_RES->field_info[i].type == SQL_C_BINARY) {
-	      push_string(make_shared_binary_string(blob_buf, BLOB_BUFSIZ));
-	    } else {
-	      /* SQL_C_CHAR and SQL_C_WCHAR's are NUL-terminated... */
-#ifdef SQL_WCHAR
-	      push_sqlwchar(blob_buf, BLOB_BUFSIZ - 1);
-#else
-	      push_string(make_shared_binary_string(blob_buf, BLOB_BUFSIZ - 1));
-#endif
-	    }
-	  } else {
-	    num_strings++;
-#ifdef ODBC_DEBUG
-	    fprintf(stderr, "[%d] ", num_strings);
-#endif /* ODBC_DEBUG */
-	    if (len < BLOB_BUFSIZ
+	}
+
+	while((bytes = len)) {
 #ifdef SQL_NO_TOTAL
-                && (len != SQL_NO_TOTAL)
+	  if (bytes == SQL_NO_TOTAL) {
+	    bytes = BLOB_BUFSIZ;
+	  }
 #endif /* SQL_NO_TOTAL */
-                ) {
-#ifdef SQL_WCHAR
-	      if (PIKE_ODBC_RES->field_info[i].type == SQL_C_WCHAR) {
-		push_sqlwchar(blob_buf, len / sizeof(SQLWCHAR));
-	      } else {
-#endif
-		push_string(make_shared_binary_string(blob_buf, len));
-#ifdef SQL_WCHAR
-	      }
-#endif
-	    } else {
-	      /* Truncated, but no support for chained SQLGetData calls. */
-	      char *buf = xalloc((len+2)
-#ifdef SQL_WCHAR
-				 * sizeof(SQLWCHAR)
-#endif
-				 );
-	      SQLLEN newlen = 0;
-	      code = SQLGetData(PIKE_ODBC_RES->hstmt, (SQLUSMALLINT)(i+1),
-				PIKE_ODBC_RES->field_info[i].type,
-				buf, (len+1)
-#ifdef SQL_WCHAR
-				* sizeof(SQLWCHAR)
-#endif
-				, &newlen);
-	      if (code != SQL_SUCCESS) {
-		Pike_error("odbc->fetch_row(): "
-			   "Unexpected code from SQLGetData(): %d\n",
-			   code);
-	      }
-	      if (len != newlen) {
-		Pike_error("odbc->fetch_row(): "
-			   "Unexpected length from SQLGetData(): "
-			   "%d (expected %d)\n", newlen, len);
-	      }
-#ifdef SQL_WCHAR
-	      if (PIKE_ODBC_RES->field_info[i].type == SQL_C_WCHAR) {
-		push_sqlwchar(buf, len / sizeof(SQLWCHAR));
-	      } else {
-#endif
-		push_string(make_shared_binary_string(buf, len));
-#ifdef SQL_WCHAR
-	      }
-#endif
-	      free(buf);
-	    }
+	  switch(field_type) {
+	  case SQL_C_CHAR:
+	  default:
+	  case SQL_C_BINARY:
+	    s = begin_shared_string(bytes);
 	    break;
+#ifdef SQL_WCHAR
+	  case SQL_C_WCHAR:
+	    s = begin_wide_shared_string(bytes/sizeof(SQLWCHAR),
+					 sizeof(SQLWCHAR)>2?2:1);
+	    break;
+#endif /* SQL_WCHAR */
+	  }
+
+#ifdef ODBC_DEBUG
+	  fprintf(stderr, "ODBC:fetch_row(): Field %d: "
+		  "SQLGetData(X, %d, %d, X, %d, X)...\n",
+		  i + 1, i+1, field_type, bytes+pad);
+#endif /* ODBC_DEBUG */
+
+	  ODBC_ALLOW();
+
+	  code = SQLGetData(hstmt, (SQLUSMALLINT)(i+1),
+			    field_type, s->str, bytes + pad, &len);
+
+	  ODBC_DISALLOW();
+
+	  num_strings++;
+#ifdef ODBC_DEBUG
+	  fprintf(stderr,
+		  "ODBC:fetch_row(): %d:%d: Got %ld/%ld bytes (pad: %ld).\n"
+		  "     Code: %d\n",
+		  i+1, num_strings, (long)bytes, (long)len, (long)pad,
+		  code);
+#endif /* ODBC_DEBUG */
+	  if (code == SQL_NO_DATA_FOUND) {
+	    /* No data or end marker. */
+	    free_string(s);
+	    push_empty_string();
+	    break;
+	  } else {
+	    odbc_check_error("odbc->fetch_row", "SQLGetData() failed",
+			     code, NULL, NULL);
+	    if (!len) {
+	      free_string(s);
+	      push_empty_string();
+	      break;
+#ifdef SQL_NO_TOTAL
+	    } else if (len == SQL_NO_TOTAL) {
+	      /* More data remaining... */
+	      push_string(end_shared_string(s));
+#ifdef ODBC_DEBUG
+	      fprintf(stderr, "ODBC:fetch_row(): More data remaining.\n");
+#endif /* ODBC_DEBUG */
+#endif /* SQL_NO_TOTAL */
+	    } else {
+	      SQLLEN str_len = len;
+	      if (len > bytes) {
+		/* Possibly truncated result. */
+		str_len = bytes;
+		len -= bytes;
+#ifdef ODBC_DEBUG
+		fprintf(stderr, "ODBC:fetch_row(): %ld bytes remaining.\n",
+			(long)len);
+#endif /* ODBC_DEBUG */
+	      } else len = 0;
+#ifdef SQL_WCHAR
+	      str_len = str_len>>s->size_shift;
+#endif
+	      push_string(end_and_resize_shared_string(s, str_len));
+	    }
 	  }
 	}
-	if (num_strings > 1) {
-#ifdef ODBC_DEBUG
-	  fprintf(stderr, "ODBC:fetch_row(): Joining %d strings\n",
-		  num_strings);
-#endif /* ODBC_DEBUG */
+	if (!num_strings) {
+	  push_empty_string();
+	} else if (num_strings > 1) {
 	  f_add(num_strings);
 	}
+      }
     }
     f_aggregate(PIKE_ODBC_RES->num_fields);
   }
@@ -645,23 +811,6 @@ static void f_seek(INT32 args)
  
 void init_odbc_res_programs(void)
 {
-  /*
-   * start_new_program();
-   *
-   * add_storage();
-   *
-   * add_function();
-   * add_function();
-   * ...
-   *
-   * set_init_callback();
-   * set_exit_callback();
-   *
-   * program = end_c_program();
-   * program->refs++;
-   *
-   */
- 
   start_new_program();
   ADD_STORAGE(struct precompiled_odbc_result);
 
