@@ -2,7 +2,7 @@
 || This file is part of Pike. For copyright information see COPYRIGHT.
 || Pike is distributed under GPL, LGPL and MPL. See the file COPYING
 || for more information.
-|| $Id: postgres.c,v 1.43 2007/04/18 13:25:32 grubba Exp $
+|| $Id: postgres.c,v 1.65 2009/02/15 16:15:41 srb Exp $
 */
 
 /*
@@ -29,6 +29,7 @@
 #include <string.h>
 
 /* Pike includes */
+#include "svalue.h"
 #include "las.h"
 #include "machine.h"
 #include "pike_memory.h"
@@ -44,7 +45,7 @@
 #define THIS ((struct pgres_object_data *) (Pike_fp->current_storage))
 
 /* Actual code */
-#ifdef _REENTRANT
+#ifdef PIKE_THREADS
 # ifdef PQ_THREADSAFE
 #  define PQ_FETCH() PIKE_MUTEX_T *pg_mutex = &THIS->mutex;
 #  define PQ_LOCK() mt_lock(pg_mutex)
@@ -72,8 +73,6 @@ static void pgdebug (char * a, ...) {}
 
 struct program * postgres_program;
 
-RCSID("$Id: postgres.c,v 1.43 2007/04/18 13:25:32 grubba Exp $");
-
 static void set_error (char * newerror)
 {
 	pgdebug("set_error(%s).\n",newerror);
@@ -87,12 +86,12 @@ static void pgres_create (struct object * o) {
 	pgdebug ("pgres_create().\n");
 	THIS->dblink=NULL;
 	THIS->last_error=NULL;
-	THIS->notify_callback=(struct svalue*)xalloc(sizeof(struct svalue));
-	THIS->notify_callback->type=PIKE_T_INT;
+	mark_free_svalue (&THIS->notify_callback);
 	THIS->docommit=0;
 	THIS->dofetch=0;
 	THIS->lastcommit=0;
-#ifdef PQ_THREADSAFE
+        THIS->last_rows=-1;
+#if defined(PIKE_THREADS) && defined(PQ_THREADSAFE)
 	mt_init(&THIS->mutex);
 #endif
 
@@ -116,11 +115,11 @@ static void pgres_destroy (struct object * o)
 			free_string(THIS->last_error);
 		THIS->last_error=NULL;
 	}
-	if (THIS->notify_callback->type!=PIKE_T_INT) {
-		free_svalue(THIS->notify_callback);
+	if (THIS->notify_callback.type!=PIKE_T_FREE) {
+		free_svalue(&THIS->notify_callback);
+		mark_free_svalue(&THIS->notify_callback);
 	}
-	free(THIS->notify_callback);
-#ifdef PQ_THREADSAFE
+#if defined(PIKE_THREADS) && defined(PQ_THREADSAFE)
 	mt_destroy(&THIS->mutex);
 #endif
 }
@@ -129,8 +128,8 @@ static void pgres_destroy (struct object * o)
  *!
  *! @class postgres
  *!
- *! This is an interface to the Postgres (Postgres95, pgsql) database server.
- *! This module may or may not be availible on your Pike, depending
+ *! This is an interface to the Postgres database server using libpq.
+ *! This module may or may not be available on your Pike, depending
  *! whether the appropriate include and library files 
  *! could be found at compile-time. Note that you @b{do not@}
  *! need to have a Postgres server running on your host to use this module:
@@ -144,22 +143,22 @@ static void pgres_destroy (struct object * o)
  *! a single line of code, at least for most common (and simple) operations.
  *!
  *! The program @[Postgres.postgres] provides the @b{raw@} interface
- *! to the database. Many functions are @b{not@} availible
+ *! to the database. Many functions are @b{not@} available
  *! for this program. Therefore, its use is DEPRECATED.
  *! It is included in this documentation only for completeness' sake.
  *! Use @[Sql.postgres] instead, or even better @[Sql.Sql]
  *!
  *! @note
+ *! This driver is based on libpq and is DEPRECATED.  There is a newer driver
+ *! called @ref{Sql.pgsql@} which is faster and more robust than this driver
+ *! and does not depend on any libraries.
+ *!
+ *! @note
  *! There is no testsuite for this module, since to test anything would
  *! require a working Postgres server. You can try to use the included scripts
  *! in the "pike/src/modules/Postgres/extras" directory but you'll probably
  *! have to patch them to reflect your site's settings.
  *!
- *! Also note that @b{this module uses blocking I/O@} I/O to connect to the server.
- *! Postgres is quite slow, and so you might want to consider this
- *! particular aspect. It is (at least should be) thread-safe, and so it can be used
- *! in a multithread environment.
- *!
  *! The behavior of the Postgres C API also depends on certain environment variables
  *! defined in the environment of the pike interpreter.
  *!
@@ -184,7 +183,7 @@ static void pgres_destroy (struct object * o)
  *!   	this, so I can't help you.
  *! @endstring
  *!
- *! Refer to the Postgres documentation for further details.
+ *! Refer to the libpq documentation for further details.
  *!
  *! @seealso
  *!   @[Sql.Sql], @[Sql.postgres], @[Sql.postgres_result]
@@ -233,82 +232,37 @@ static void f_create (INT32 args)
 {
         /*port will be used as a string with a sprintf()*/
 	char * host=NULL, *db=NULL, *user=NULL, *pass=NULL, *port=NULL;
+	int port_no = -1;
+	char port_buffer[10]; /*it's enough, we need only 6*/
 	PGconn * conn;
 	PQ_FETCH();
 
-	check_all_args("postgres->create",args,
-			BIT_STRING|BIT_VOID,
-		        BIT_STRING|BIT_VOID,
-			BIT_STRING|BIT_VOID,
-		        BIT_STRING|BIT_VOID,
-			BIT_INT|BIT_VOID,
-		        0);
+	get_all_args("postgres->create",args,
+		     ".%s%s%s%s%d",
+		     &host,
+		     &db,
+		     &user,
+		     &pass,
+		     &port_no);
 
-	if (THIS->dblink) {
-		conn=THIS->dblink;
-		THREADS_ALLOW();
-		PQ_LOCK();
-		PQfinish(conn);
-		PQ_UNLOCK();
-		THREADS_DISALLOW();
+	if (port_no > 0) {
+	  if (port_no < 65536) {
+	    sprintf(port = port_buffer, "%d", port_no);
+	  } else {
+	    SIMPLE_ARG_TYPE_ERROR("create", 5, "int(0..65535)");
+	  }
 	}
 
-	/*no break;'s here, it's intentional*/
-	switch(args) {
-		default:
-		case 5:
-			if (Pike_sp[4-args].type==PIKE_T_INT &&  /*this check is maybe redundant*/
-					Pike_sp[4-args].u.integer <=65535 &&
-			    		Pike_sp[4-args].u.integer >= 0) {
-				if (Pike_sp[4-args].u.integer>0) {
-					port=xalloc(10*sizeof(char)); /*we only need 6, we just checked.*/
-					sprintf(port,"%"PRINTPIKEINT"d",Pike_sp[4-args].u.integer);
-				}
-			}
-		case 4:
-			if (Pike_sp[3-args].type==PIKE_T_STRING && Pike_sp[3-args].u.string->len)
-				pass=Pike_sp[3-args].u.string->str;
-		case 3:
-			if (Pike_sp[2-args].type==PIKE_T_STRING && Pike_sp[2-args].u.string->len)
-				user=Pike_sp[2-args].u.string->str;
-		case 2:
-			if (Pike_sp[1-args].type==PIKE_T_STRING && Pike_sp[1-args].u.string->len)
-				db=Pike_sp[1-args].u.string->str;
-		case 1:
-			if(Pike_sp[-args].type==PIKE_T_STRING && Pike_sp[-args].u.string->len)
-				host=Pike_sp[-args].u.string->str;
-		case 0:
-			;
-	}
-#if 0
-	/* Old arguments-checking code */
-	if (args>=1)
-		if(Pike_sp[-args].type==PIKE_T_STRING && Pike_sp[-args].u.string->len)
-			host=Pike_sp[-args].u.string->str;
-		/* postgres docs say they use hardwired defaults if no variable is found*/
-	if (args>=2)
-		if (Pike_sp[1-args].type==PIKE_T_STRING && Pike_sp[1-args].u.string->len)
-			db=Pike_sp[1-args].u.string->str;
-		/* This is not beautiful code, but it works:
-		 * it specifies the port to connect to if there is a third
-		 * argument greater than 0. It accepts integer arguments >= 0
-		 * to allow simpler code in pike wrappers part.
-		 */
-	if (args==3)
-		if (Pike_sp[2-args].type==PIKE_T_INT && Pike_sp[2-args].u.integer <=65535 && Pike_sp[2-args].u.integer >= 0) {
-			if (Pike_sp[2-args].u.integer>0) {
-				port=xalloc(10*sizeof(char)); /*it's enough, we need only 6*/
-				sprintf(port,"%d",Pike_sp[2-args].u.integer);
-			}
-		}
-		else
-			Pike_error ("You must specify a TCP/IP port number as argument 5 "
-					"to Sql.postgres->create().\n");
-#endif
-	pgdebug("f_create(host=%s, port=%s, db=%s, user=%s, pass=%s).\n",
-			host,port,db,user,pass);
+	conn=THIS->dblink;
+	THIS->dblink = NULL;
+
+	/* FIXME: Race-condition! */
+
 	THREADS_ALLOW();
 	PQ_LOCK();
+	if (conn) {
+		PQfinish(conn);
+	}
 #ifdef HAVE_PQSETDBLOGIN
 	conn=PQsetdbLogin(host,port,NULL,NULL,db,user,pass);
 #else
@@ -316,6 +270,7 @@ static void f_create (INT32 args)
 #endif
 	PQ_UNLOCK();
 	THREADS_DISALLOW();
+
 	if (!conn)
 		Pike_error ("Could not conneect to server\n");
 	if (PQstatus(conn)!=CONNECTION_OK) {
@@ -325,11 +280,11 @@ static void f_create (INT32 args)
 		PQfinish(conn);
 		PQ_UNLOCK();
 		THREADS_DISALLOW();
-		Pike_error("Could not connect to database. Reason: \"%s\".\n",THIS->last_error->str);
+		Pike_error("Could not connect to database. Reason: \"%S\".\n",
+			   THIS->last_error);
+		conn=NULL;
 	}
 	THIS->dblink=conn;
-	if (!THIS->dblink)
-		Pike_error ("Huh? Weirdness here! Internal error!\n");
 	pop_n_elems(args);
 }
 
@@ -355,35 +310,29 @@ static void f_select_db (INT32 args)
 	PGconn * conn, *newconn;
 	PQ_FETCH();
 
-	check_all_args("Postgres->select_db",args,BIT_STRING,0);
+	get_all_args("Postgres->select_db",args,"%s", &db);
 	
 	if (!THIS->dblink)
 	  Pike_error ("Driver error. How can you possibly not be linked to a "
 		      "database already?\n");
 	conn=THIS->dblink;
+#if 0
+	/* This is an optimization, but people may want to reset a connection
+	 * by re-selecting its database.
+	 */
+	if (!strcmp(db, PQdb(conn))) {
+		pop_n_elems(args);
+		return;
+	}
+#endif
 	THREADS_ALLOW();
 	PQ_LOCK();
 	host=PQhost(conn);
 	port=PQport(conn);
 	options=PQoptions(conn);
 	tty=PQtty(conn);
-	db=PQdb(conn);
-	PQ_UNLOCK();
-	THREADS_DISALLOW();
-#if 0
-	/* This is an optimization, but people may want to reset a connection
-	 * re-selecting its database.
-	 */
-	if (!strcmp(Pike_sp[-args].u.string->str,db)) {
-		pop_n_elems(args);
-		return;
-	}
-#endif
-	db=Pike_sp[-args].u.string->str;
 	/* This could be really done calling f_create, but it's more efficient this
 	 * way */
-	THREADS_ALLOW();
-	PQ_LOCK();
 	/* using newconn is necessary or otherwise the datastructures I use
 	 * as arguments get freed by PQfinish. Could be a problem under extreme
 	 * situations (i.e. if the temporary use of _one_ more filedescriptor
@@ -396,8 +345,11 @@ static void f_select_db (INT32 args)
 	THREADS_DISALLOW();
 	if (PQstatus(conn)==CONNECTION_BAD) {
 		set_error(PQerrorMessage(conn));
+		PQ_LOCK();
 		PQfinish(conn);
+		PQ_UNLOCK();
 		Pike_error("Could not connect to database.\n");
+		conn=NULL;
 	}
 	THIS->dblink=conn;
 	pop_n_elems(args);
@@ -418,8 +370,9 @@ static void f_select_db (INT32 args)
  *! @note
  *! This function @b{can@} raise exceptions.
  *!
- *! The program @[Sql.postgres_result] is exactly
- *! the same as @[Postgres.postgres_result].
+ *! The program @[Sql.postgres_result] is a superset of
+ *! the raw @[Postgres.postgres_result] which does not do chunking to
+ *! avoid excessive memory usage in case of large resultsets.
  *!
  *! @seealso
  *!  @[Sql.Sql], @[Sql.sql_result]
@@ -430,25 +383,102 @@ static void f_big_query(INT32 args)
 	PGconn *conn = THIS->dblink;
 	PGresult * res;
 	PGnotify * notification;
-	char *query;
+        struct array *bnds = 0;
+	char *query = 0;
 	int lastcommit,docommit,dofetch;
+        int cnt;
+        int nParams = 0;
+        const char** paramValues = 0;
+        int* paramLengths = 0;
+        int* paramFormats = 0;
+        int resultFormat = 0;
+        
 	PQ_FETCH();
 	docommit=dofetch=0;
 	lastcommit=THIS->lastcommit;
 
-	check_all_args("Postgres->big_query",args,BIT_STRING,0);
+	check_all_args("Postgres->big_query",args,
+	  BIT_STRING,
+	  BIT_VOID | BIT_ARRAY,
+	  0);
 
 	if (!conn)
 		Pike_error ("Not connected.\n");
 
-	if (Pike_sp[-args].u.string->len)
+        switch(args)
+        {
+          default:
+            if(Pike_sp[1-args].type == PIKE_T_ARRAY)
+              bnds=Pike_sp[1-args].u.array;
+      
+          case 1:
+	    query=" ";
+	    if(Pike_sp[-args].u.string->len)
 		query=Pike_sp[-args].u.string->str;
-	else
-		query=" ";
+        }
+
+        if(bnds && (cnt=bnds->size)) {
+          int i;
+	  struct svalue *item;
+
+	  nParams=cnt;
+	  
+	  paramValues = xalloc(nParams*sizeof*paramValues);
+	  paramLengths = xalloc(nParams*sizeof*paramLengths);
+	  paramFormats = xalloc(nParams*sizeof*paramFormats);
+
+	  for (i=0,item=bnds->item; cnt--; item++,i++) {
+	    int sendbin=0;
+	    switch(item->type)
+	    { case PIKE_T_STRING: {
+	          long len;
+		  const char*p;
+		  paramLengths[i] = len = item->u.string->len;
+	          paramValues[i] = p = item->u.string->str;
+	          if( len >= BINARYCUTOFF)
+	            sendbin = 1;
+	          else {
+	            while(len--) {
+		      switch(*p++) {
+			default:
+			  continue;
+			case '\0':case '\\':
+			  sendbin = 1;
+		      }
+		      break;
+		    }
+	          }
+	        break;
+	      }
+	      case PIKE_T_INT:
+	      case T_VOID:
+	          paramValues[i] = 0;	     /* NULL */
+		  paramLengths[i] = 0;
+	        break;
+	      default:
+                  Pike_error("Expected string or UNDEFINED element, Got %d.\n",
+		   item->type);
+	        break;
+	    }
+	    paramFormats[i] = sendbin;
+	  }
+        }
 
 	THREADS_ALLOW();
 	PQ_LOCK();
 	pgdebug("f_big_query(\"%s\")\n",query);
+        /*
+         * libpq implements partial reads from a SELECT through
+         * a mechanism of a CURSOR incombination with FETCH.
+         *
+         * Cursors are only used if the query starts with "SELECT "
+         * and does not end in "LIMIT 1" or "LIMIT 1;"
+         * The checks are case-sensitive by design.
+         *
+         * Cursors can only be used within a transaction, so you'll end
+         * up trying to start a transaction; if a transaction already is
+         * in progress, that will be noticed and it will not be disrupted.
+         */
 #define SELECTSTR	"SELECT "
 #define LIMIT1STR	"LIMIT 1"
 #define LIMIT1STRSC	LIMIT1STR";"
@@ -468,11 +498,13 @@ static void f_big_query(INT32 args)
 	    strcpy(nquery+CPREFLEN-1,query);
 	    if(lastcommit)
 	      goto yupbegin;
-	    res=PQexec(conn,nquery);
+	    res=PQexecParams(conn,nquery,
+             nParams,0,paramValues,paramLengths,paramFormats,resultFormat);
 	    if(PQstatus(conn) != CONNECTION_OK) {
 	      PQclear(res);
 	      PQreset(conn);
-	      res=PQexec(conn,nquery);
+	      res=PQexecParams(conn,nquery,
+               nParams,0,paramValues,paramLengths,paramFormats,resultFormat);
 	    }
 	    if(res)
 	      switch(PQresultStatus(res)) {
@@ -481,7 +513,8 @@ static void f_big_query(INT32 args)
 yupbegin:       res=PQexec(conn,"BEGIN");
 		if(res && PQresultStatus(res)==PGRES_COMMAND_OK) {
 		  PQclear(res);
-		  res=PQexec(conn,nquery);
+	          res=PQexecParams(conn,nquery,nParams,0,
+		   paramValues,paramLengths,paramFormats,resultFormat);
 		  if(res && PQresultStatus(res)==PGRES_COMMAND_OK)
 		    docommit=1;
 		  else {
@@ -507,7 +540,8 @@ yupbegin:       res=PQexec(conn,"BEGIN");
 	}
 	lastcommit=0;
 	if(!res)
-	  res=PQexec(conn,query);
+	  res=PQexecParams(conn,query,
+           nParams,0,paramValues,paramLengths,paramFormats,resultFormat);
 	/* A dirty hack to fix the reconnect bug.
 	 * we don't need to store the host/user/pass/db... etc..
 	 * PQreset() does all the job.
@@ -518,7 +552,8 @@ yupbegin:       res=PQexec(conn,"BEGIN");
 	   (PQresultStatus(res) == PGRES_BAD_RESPONSE)) {
 	  PQclear(res);
 	  PQreset(conn);
-	  res=PQexec(conn,query);
+	  res=PQexecParams(conn,query,
+           nParams,0,paramValues,paramLengths,paramFormats,resultFormat);
 	}
 
 	notification=PQnotifies(conn);
@@ -528,11 +563,15 @@ yupbegin:       res=PQexec(conn,"BEGIN");
 	THIS->dofetch=dofetch;
 	THIS->lastcommit=lastcommit;
 
+        if (bnds) {
+	  xfree(paramValues); xfree(paramLengths); xfree(paramFormats);
+        }
+
 	pop_n_elems(args);
 	if (notification!=NULL) {
 		pgdebug("Incoming notification: \"%s\"\n",notification->relname);
 		push_text(notification->relname);
-		apply_svalue(THIS->notify_callback,1); 
+		apply_svalue(&THIS->notify_callback,1); 
 		/* apply_svalue simply returns if the first argument is a PIKE_T_INT */
 		free (notification);
 	}
@@ -552,6 +591,7 @@ yupbegin:       res=PQexec(conn,"BEGIN");
 		case PGRES_COMMAND_OK:
 			pgdebug("\tOk.\n");
 			THIS->last_result=NULL;
+                        THIS->last_rows=atoi(PQcmdTuples(res));
 			PQclear(res);
 			push_int(0);
 			return;
@@ -569,13 +609,14 @@ yupbegin:       res=PQexec(conn,"BEGIN");
 		case PGRES_TUPLES_OK:
 			pgdebug("\tResult.\n");
 			THIS->last_result=res;
+                        THIS->last_rows=PQntuples(res);
 			push_object(this_object());
 			push_object(clone_object(pgresult_program,1));
 			return;
 		default:
 			Pike_error ("Unimplemented server feature.\n");
 	}
-	Pike_error ("Internal error in postgresmodule.\n");
+	Pike_error ("Internal error in postgres module.\n");
 }
 
 
@@ -592,7 +633,7 @@ yupbegin:       res=PQexec(conn,"BEGIN");
 
 static void f_error (INT32 args)
 {
-	check_all_args("Postgres->error",args,0);
+        pop_n_elems(args);
 
 	if (THIS->last_error)
 		ref_push_string(THIS->last_error);
@@ -608,7 +649,7 @@ static void f_error (INT32 args)
  *! a variety of reasons, for example to detect the status of a connection.
  *!
  *! @note
- *! This function is Postgres-specific, and thus it is not availible
+ *! This function is Postgres-specific, and thus it is not available
  *! through the generic SQL-interface.
  */
 
@@ -617,7 +658,7 @@ static void f_reset (INT32 args)
 	PGconn * conn;
 	PQ_FETCH();
 
-	check_all_args("Postgres->reset",args,0);
+	pop_n_elems(args);
 
 	if (!THIS->dblink)
 		Pike_error ("Not connected.\n");
@@ -676,7 +717,7 @@ static void f_trace (INT32 args)
  *! The @[Sql.postgres] program adds support for automatic delivery of
  *! messages (see it for explanation on the inner workings of this feature).
  *!
- *! This function is Postgres-specific, and thus it is not availible
+ *! This function is Postgres-specific, and thus it is not available
  *! through the generic SQL-interface
  *!
  *! @seealso
@@ -689,15 +730,15 @@ static void f_callback(INT32 args)
 		       args, BIT_INT|BIT_FUNCTION, 0);
 
 	if (Pike_sp[-args].type==PIKE_T_INT) {
-		if (THIS->notify_callback->type!=PIKE_T_INT) {
-			free_svalue(THIS->notify_callback);
-			THIS->notify_callback->type=PIKE_T_INT;
+		if (THIS->notify_callback.type!=PIKE_T_FREE) {
+			free_svalue(&THIS->notify_callback);
+			mark_free_svalue (&THIS->notify_callback);
 		}
 		pop_n_elems(args);
 		return;
 	}
 	/*let's assume it's a function otherwise*/
-	assign_svalue(THIS->notify_callback,Pike_sp-args);
+	assign_svalue(&THIS->notify_callback,Pike_sp-args);
 	pop_n_elems(args);
 }
 
@@ -710,7 +751,7 @@ static void f_callback(INT32 args)
 
 static void f_host_info (INT32 args)
 {
-	check_all_args("Postgres->host_info",args,0);
+        pop_n_elems(args);
 
 	if (PQstatus(THIS->dblink)!=CONNECTION_BAD) {
 		char buf[64];
@@ -729,6 +770,17 @@ static void f_host_info (INT32 args)
 	Pike_error ("Bad connection.\n");
 }
 
+/*! @decl int affected_rows()
+ *!
+ *! This function returns the number of rows affected by the last query.
+ */
+
+static void f_affected_rows (INT32 args)
+{
+  pop_n_elems(args);
+  push_int(THIS->last_rows);
+}
+
 /*! @decl string _quote(string s)
  *!
  *! Escape a string to prevent SQL injection, using the current connection's
@@ -761,7 +813,6 @@ static void f_quote(INT32 args)
 }
 #endif /* HAVE_PQESCAPESTRINGCONN || HAVE_PQESCAPESTRING */
 
-
 /*! @endclass
  *!
  *! @endmodule
@@ -769,61 +820,67 @@ static void f_quote(INT32 args)
 
 PIKE_MODULE_INIT
 {
-	start_new_program();
-	ADD_STORAGE(struct pgres_object_data);
-	set_init_callback(pgres_create);
-	set_exit_callback(pgres_destroy);
+  start_new_program();
+  ADD_STORAGE(struct pgres_object_data);
+  set_init_callback(pgres_create);
+  set_exit_callback(pgres_destroy);
 
-	/* sql-interface compliant functions */
-	/* function(void|string,void|string,void|string,void|string,int|void:void) */
-  ADD_FUNCTION("create",f_create,tFunc(tOr(tVoid,tStr) tOr(tVoid,tStr) tOr(tVoid,tStr) tOr(tVoid,tStr) tOr(tInt,tVoid),tVoid),
-		OPT_EXTERNAL_DEPEND);
-	/* That is: create(hostname,database,port)
-	 * It depends on the environment variables:
-	 * PGHOST, PGOPTIONS, PGPORT, PGTTY(don't use!), PGDATABASE
-	 * Notice: Postgres _requires_ a database to be selected upon connection
-	 */
-	/* function(string:void) */
-  ADD_FUNCTION("select_db",f_select_db,tFunc(tStr,tVoid),
-			OPT_EXTERNAL_DEPEND);
-	/* function(string:int|object) */
-  ADD_FUNCTION("big_query",f_big_query,tFunc(tStr,tOr(tInt,tObj)),
-			OPT_EXTERNAL_DEPEND|OPT_RETURN);
-	/* function(void:string) */
-  ADD_FUNCTION("error",f_error,tFunc(tVoid,tStr),
-			OPT_EXTERNAL_DEPEND|OPT_RETURN);
-	/* function(void:string) */
-  ADD_FUNCTION("host_info",f_host_info,tFunc(tVoid,tStr),
-			OPT_EXTERNAL_DEPEND|OPT_RETURN);
+  /* sql-interface compliant functions */
+  /* function(void|string,void|string,void|string,void|string,int|void:void) */
+  ADD_FUNCTION("create", f_create,
+	       tFunc(tOr(tVoid,tStr) tOr(tVoid,tStr) tOr(tVoid,tStr)
+		     tOr(tVoid,tStr) tOr(tInt,tVoid), tVoid), 0);
+
+  /* That is: create(hostname,database,port)
+   * It depends on the environment variables:
+   * PGHOST, PGOPTIONS, PGPORT, PGTTY(don't use!), PGDATABASE
+   * Notice: Postgres _requires_ a database to be selected upon connection
+   */
+  /* function(string:void) */
+  ADD_FUNCTION("select_db", f_select_db, tFunc(tStr,tVoid), 0);
+
+  /* function(string:int|object) */
+  ADD_FUNCTION("big_query", f_big_query,
+       tFunc(tStr tOr(tVoid,tArr(tOr3(tInt,tStr,tVoid))), tOr(tInt,tObj)), 0);
+
+  /* function(void:string) */
+  ADD_FUNCTION("error", f_error, tFunc(tVoid,tStr), 0);
+
+  /* function(void:string) */
+  ADD_FUNCTION("host_info", f_host_info, tFunc(tVoid,tStr), 0);
+
+  /* function(void:int) */
+  ADD_FUNCTION("affected_rows", f_affected_rows, tFunc(tVoid,tInt), 0);
 
   /* function(string:string) */
 #if defined(HAVE_PQESCAPESTRINGCONN) || defined(HAVE_PQESCAPESTRING)
   ADD_FUNCTION("_quote", f_quote, tFunc(tStr,tStr), 0);
 #endif
 
-	/* postgres-specific functions */
-	/* function(void:void) */
-  ADD_FUNCTION("reset",f_reset,tFunc(tVoid,tVoid),OPT_EXTERNAL_DEPEND|OPT_SIDE_EFFECT);
+
+  /* postgres-specific functions */
+  /* function(void:void) */
+  ADD_FUNCTION("reset", f_reset, tFunc(tVoid,tVoid), 0);
 
 #if 0
-	 /* function(object|int:void) */
-  ADD_FUNCTION("trace",f_trace,tFunc(tOr(tObj,tInt),tVoid),
-		 OPT_EXTERNAL_DEPEND|OPT_SIDE_EFFECT);
-	 /* If given a clone of /precompiled/file, traces to that file.
-		* If given 0, stops tracing.
-		* See note on the implementation.
-		*/
+  /* function(object|int:void) */
+  ADD_FUNCTION("trace", f_trace, tFunc(tOr(tObj,tInt),tVoid), 0);
+
+  /* If given a clone of /precompiled/file, traces to that file.
+   * If given 0, stops tracing.
+   * See note on the implementation.
+   */
 #endif
-	/* function(int|function(string:void):void) */
-  ADD_FUNCTION("_set_notify_callback",f_callback,tFunc(tOr(tInt,tFunc(tStr,tVoid)),tVoid),
-			OPT_SIDE_EFFECT);
+  /* function(int|function(string:void):void) */
+  ADD_FUNCTION("_set_notify_callback", f_callback,
+	       tFunc(tOr(tInt,tFunc(tStr,tVoid)),tVoid), 0);
 
-	postgres_program = end_program();
-	add_program_constant("postgres",postgres_program,0);
+  postgres_program = end_program();
+  add_program_constant("postgres",postgres_program,0);
 
-	add_string_constant("version",PGSQL_VERSION,0);
+  add_string_constant("version",PGSQL_VERSION,0);
 
-	pgresult_init(); 
+  pgresult_init(); 
 }
 
 PIKE_MODULE_EXIT
@@ -844,7 +901,12 @@ PIKE_MODULE_EXIT
 }
 
 #else /* HAVE_POSTGRES */
+#include "program.h"
 #include "module.h"
-PIKE_MODULE_INIT {}
+#include "module_support.h"
+PIKE_MODULE_INIT {
+  if(!TEST_COMPAT(7,6))
+    HIDE_MODULE();
+}
 PIKE_MODULE_EXIT {}
 #endif /* HAVE_POSTGRES */
