@@ -2,7 +2,7 @@
 || This file is part of Pike. For copyright information see COPYRIGHT.
 || Pike is distributed under GPL, LGPL and MPL. See the file COPYING
 || for more information.
-|| $Id: roxen.c,v 1.45 2006/11/10 13:44:48 nilsson Exp $
+|| $Id: roxen.c,v 1.55 2009/07/29 15:48:47 nilsson Exp $
 */
 
 #define NO_PIKE_SHORTHAND
@@ -42,6 +42,8 @@
  */
 
 /*! @class HeaderParser
+ *!
+ *! Class for parsing HTTP-requests.
  */
 
 #define THP ((struct header_buf *)Pike_fp->current_storage)
@@ -51,6 +53,7 @@ struct  header_buf
   unsigned char *pnt;
   ptrdiff_t hsize, left;
   int slash_n, tslash_n, spc;
+  int mode;
 };
 
 static void f_hp_init( struct object *o )
@@ -60,6 +63,7 @@ static void f_hp_init( struct object *o )
   THP->hsize = 0;
   THP->left = 0;
   THP->spc = THP->slash_n = THP->tslash_n = 0;
+  THP->mode = 0;
 }
 
 static void f_hp_exit( struct object *o )
@@ -73,6 +77,16 @@ static void f_hp_exit( struct object *o )
 
 static void f_hp_feed( INT32 args )
 /*! @decl array(string|mapping) feed(string data)
+ *!
+ *! @returns
+ *!   @array
+ *!     @elem string 0
+ *!       Trailing data.
+ *!     @elem string 1
+ *!       First line of request.
+ *!     @elem mapping(string:string|array(string)) 2
+ *!       Headers.
+ *!   @endarray
  */
 {
   struct pike_string *str = Pike_sp[-1].u.string;
@@ -85,6 +99,8 @@ static void f_hp_feed( INT32 args )
   ptrdiff_t os=0, i, j, l;
   unsigned char *in;
 
+  if (args != 1)
+    Pike_error("Bad number of arguments to feed().\n");
   if( Pike_sp[-1].type != PIKE_T_STRING )
     Pike_error("Wrong type of argument to feed()\n");
   if( str->size_shift )
@@ -146,7 +162,7 @@ static void f_hp_feed( INT32 args )
      */
     if( (spc < 2) && tot_slash_n )
     {
-      ref_push_string(empty_pike_string);
+      push_empty_string();
       /* This includes (all eventual) \r\n etc. */
       push_text((char *)hp->headers); 
       f_aggregate_mapping( 0 );
@@ -179,7 +195,8 @@ static void f_hp_feed( INT32 args )
   /* Parse headers. */
   for(i = 0; i < l; i++)
   {
-    if(in[i] > 64 && in[i] < 91) in[i]+=32;	/* lower_case */
+    if(in[i] > 64 && in[i] < 91)
+      in[i]+=32;	/* lower_case */
     else if( in[i] == ':' )
     {
       /* FIXME: Does not support white space before the colon. */
@@ -224,7 +241,7 @@ static void f_hp_feed( INT32 args )
 	  f_add(2);
 	} else {
 	  ref_push_string(tmp->u.string);
-          stack_swap();
+	  stack_swap(); 
 	  map_delete(headers, Pike_sp-3);
 	  f_aggregate(2);
 	}
@@ -233,19 +250,32 @@ static void f_hp_feed( INT32 args )
 
       pop_n_elems(2);
     }
+    else if( in[i]=='\r' || in[i]=='\n' )
+    {
+      if( THP->mode == 1 )
+      {
+        /* FIXME: Reset stack so that backtrace shows faulty header. */
+        Pike_error("Malformed HTTP header.\n");
+      }
+      else
+        os = i+1;
+    }
   }
   push_mapping( headers );
   f_aggregate( 3 );             /* data, firstline, headers */
 }
 
 static void f_hp_create( INT32 args )
-/*! @decl void create(void)
+/*! @decl void create(int throw_errors)
  */
 {
   if (THP->headers) {
     free(THP->headers);
     THP->headers = NULL;
   }
+
+  get_all_args("create",args,".%i",&THP->mode);
+
   THP->headers = xalloc( 8192 );
   THP->pnt = THP->headers;
   THP->hsize = 8192;
@@ -582,7 +612,7 @@ PIKE_MODULE_INIT
   set_init_callback( f_hp_init );
   set_exit_callback( f_hp_exit );
   ADD_FUNCTION( "feed", f_hp_feed, tFunc(tStr,tArr(tOr(tStr,tMapping))), 0 );
-  ADD_FUNCTION( "create", f_hp_create, tFunc(tNone,tVoid), ID_STATIC );
+  ADD_FUNCTION( "create", f_hp_create, tFunc(tOr(tInt,tVoid),tVoid), ID_PROTECTED );
   end_class( "HeaderParser", 0 );
 }
 
