@@ -2,7 +2,7 @@
 || This file is part of Pike. For copyright information see COPYRIGHT.
 || Pike is distributed under GPL, LGPL and MPL. See the file COPYING
 || for more information.
-|| $Id: math.c,v 1.78 2004/03/20 19:28:31 grubba Exp $
+|| $Id: math.c,v 1.87 2008/01/26 22:34:27 mast Exp $
 */
 
 #include "global.h"
@@ -37,8 +37,6 @@
   if(sp[-args].type!=T_FLOAT) SIMPLE_BAD_ARG_ERROR(X, 1, "float"); \
   TRIM_STACK(1)
 
-RCSID("$Id: math.c,v 1.78 2004/03/20 19:28:31 grubba Exp $");
-
 #ifndef M_PI
 #define M_PI 3.1415926535897932384626433832795080
 #endif
@@ -141,7 +139,13 @@ void f_sin(INT32 args)
 void f_asin(INT32 args)
 {
   ARG_CHECK("asin");
-  sp[-1].u.float_number = FL1(asin,sp[-1].u.float_number);
+  if ((sp[-1].u.float_number >= -1.0) &&
+      (sp[-1].u.float_number <= 1.0)) {
+    sp[-1].u.float_number = FL1(asin,sp[-1].u.float_number);
+  } else {
+    DECLARE_NAN;
+    sp[-1].u.float_number = (FLOAT_TYPE) MAKE_NAN();
+  }
 }
 
 /*! @decl float cos(float f)
@@ -169,7 +173,13 @@ void f_cos(INT32 args)
 void f_acos(INT32 args)
 {
   ARG_CHECK("acos");
-  sp[-1].u.float_number = FL1(acos,sp[-1].u.float_number);
+  if ((sp[-1].u.float_number >= -1.0) &&
+      (sp[-1].u.float_number <= 1.0)) {
+    sp[-1].u.float_number = FL1(acos,sp[-1].u.float_number);
+  } else {
+    DECLARE_NAN;
+    sp[-1].u.float_number = (FLOAT_TYPE) MAKE_NAN();
+  }
 }
 
 /*! @decl float tan(float f)
@@ -576,7 +586,45 @@ void f_ceil(INT32 args)
 void f_round(INT32 args)
 {
   ARG_CHECK("round");
-  sp[-1].u.float_number = FL1(RINT,sp[-1].u.float_number);
+  sp[-1].u.float_number = FL1(rint,sp[-1].u.float_number);
+}
+
+
+/*! @decl int|float|object limit(int|float|object minval, int|float|object x, int|float|object maxval)
+ *!
+ *! Limits the value @[x] so that it's between @[minval] and @[maxval].
+ *! If @[x] is an object, it must implement the @[lfun::`<] method.
+ *!
+ *! @seealso
+ *!   @[max()] and @[min()]
+ */
+void f_limit(INT32 args)
+{
+  INT32 minpos = 0;
+
+  if(args != 3) 
+  {
+      Pike_error("limit needs 3 arguments\n");
+  }
+
+  /* -3  -2  -1 */
+  /*  a < X < b */
+  if( is_lt( Pike_sp-3,  Pike_sp-2 ) )      /* a < X */
+  { 
+      if( is_lt( Pike_sp-2,  Pike_sp-1 ) )      /* X < b */
+      {
+	  /* return X (-2) */
+	  pop_stack();
+	  stack_pop_keep_top();
+      }
+      else
+      {
+         /* X > b, return b (-1) */
+	  stack_unlink( 2 );
+      }
+  }
+  else
+      pop_n_elems(2); /* a > X, return a (-3)*/
 }
 
 /*! @decl int|float|object min(int|float|object ... args)
@@ -587,7 +635,7 @@ void f_round(INT32 args)
  *! must implement the @[lfun::`<] method.
  *!
  *! @seealso
- *!   @[max()]
+ *!   @[max()] and @[limit()]
  */
 void f_min(INT32 args)
 {
@@ -619,7 +667,7 @@ void f_min(INT32 args)
  *! must implement the @[lfun::`<] method.
  *!
  *! @seealso
- *!   @[min()]
+ *!   @[min()] and @[limit()]
  */
 void f_max(INT32 args)
 {
@@ -655,6 +703,7 @@ void f_abs(INT32 args)
 {
   struct svalue zero;
   zero.type=T_INT;
+  zero.subtype = NUMBER_NUMBER;
   zero.u.integer=0;
 
   check_all_args("abs",args,BIT_INT|BIT_FLOAT|BIT_OBJECT,0);
@@ -774,31 +823,27 @@ PIKE_MODULE_INIT
 /* function(float:float) */
   ADD_EFUN("round",f_round,tFunc(tFlt,tFlt),0);
 
-#define CMP_TYPE \
-  tOr4(tFuncV(tNone,tString,tString), \
-       tFunc(tVoid,tInt0), \
-       tFuncV(tNone,tSetvar(0,tOr(tInt,tFloat)),tVar(0)), \
-       tIfnot(tFuncV(tNone,tString,tMix), \
-              tIfnot(tFuncV(tNone,tOr(tInt,tFloat),tMix), \
-	             tIfnot(tFuncV(tNone,tNot(tOr(tObj,tMix)),tMix), \
-		            tFuncV(tNone,tMix,tMix)))))
-    /*
-     * "!function(!object...:mixed)&function(mixed...:mixed)|" \
-     * "function(int|zero...:int)|" \
-     * "function(float...:float)|" \
-     * "function(string...:string)|"  \
-     * "function(:int(0..0)|" \
-     * "!function(int...:mixed)&!function(float...:mixed)&function(int|float...:int|float)"
-     */
+#define CMP_TYPE							\
+  tOr4(tFuncV(tNone,tString,tString),					\
+       tFunc(tVoid,tInt0),						\
+       tFuncV(tNone,tSetvar(0,tOr(tInt,tFloat)),tVar(0)),		\
+       tIfnot(tFuncV(tNone,tNot(tOr(tObj,tMix)),tMix),			\
+	      tFuncV(tNone,tMix,tMix)))
 
   ADD_EFUN("max",f_max,CMP_TYPE,0);
   ADD_EFUN("min",f_min,CMP_TYPE,0);
   
+  ADD_EFUN("limit",f_limit,
+	   tFunc(tSetvar(0,tOr3(tFlt,tInt,tObj))
+		 tSetvar(1,tOr3(tFlt,tInt,tObj))
+		 tSetvar(2,tOr3(tFlt,tInt,tObj)),
+		 tOr3(tVar(0),tVar(1),tVar(2))),0);
+
 /* function(float|int|object:float|int|object) */
-  ADD_EFUN("abs",f_abs,tFunc(tOr3(tFlt,tInt,tObj),tOr3(tFlt,tInt,tObj)),0);
+  ADD_EFUN("abs",f_abs,tFunc(tSetvar(0,tOr3(tFlt,tInt,tObj)),tVar(0)),0);
   
 /* function(mixed,mixed|void:int) */
-  ADD_EFUN("sgn",f_sgn,tFunc(tMix tOr(tMix,tVoid),tInt),0);
+  ADD_EFUN("sgn",f_sgn,tFunc(tMix tOr(tMix,tVoid),tInt_11),0);
 }
 
 PIKE_MODULE_EXIT {}
