@@ -2,7 +2,7 @@
 || This file is part of Pike. For copyright information see COPYRIGHT.
 || Pike is distributed under GPL, LGPL and MPL. See the file COPYING
 || for more information.
-|| $Id: efuns.c,v 1.150 2006/07/05 18:07:23 mast Exp $
+|| $Id: efuns.c,v 1.189 2009/08/07 13:06:37 grubba Exp $
 */
 
 #include "global.h"
@@ -26,8 +26,6 @@
 #include "file_machine.h"
 #include "file.h"
 
-RCSID("$Id: efuns.c,v 1.150 2006/07/05 18:07:23 mast Exp $");
-
 #ifdef HAVE_SYS_TYPES_H
 #include <sys/types.h>
 #endif
@@ -76,6 +74,19 @@ RCSID("$Id: efuns.c,v 1.150 2006/07/05 18:07:23 mast Exp $");
 #include <process.h>
 #endif
 
+#ifdef HAVE_SYS_XATTR_H
+#include <sys/xattr.h>
+#endif /* HAVE_SYS_XATTR_H */
+
+#if 0
+#ifdef HAVE_LIBZFS_INIT
+#ifdef HAVE_LIBZFS_H
+#include <libzfs.h>
+#endif /* HAVE_LIBZFS_H */
+static libzfs_handle_t *libzfs_handle;
+#endif /* HAVE_LIBZFS_INIT */
+#endif /* 0 */
+
 #define sp Pike_sp
 
 /* #define DEBUG_FILE */
@@ -84,6 +95,7 @@ RCSID("$Id: efuns.c,v 1.150 2006/07/05 18:07:23 mast Exp $");
 #ifdef __NT__
 
 #include <winbase.h>
+#include <io.h>
 
 /* Old versions of the headerfiles don't have this constant... */
 #ifndef INVALID_SET_FILE_POINTER
@@ -104,7 +116,7 @@ LINKFUNC(BOOL, movefileex, (
   DWORD dwFlags                /* move options  */
 ));
 
-#endif
+#endif /* __NT__ */
 
 struct array *encode_stat(PIKE_STAT_T *s)
 {
@@ -131,14 +143,267 @@ struct array *encode_stat(PIKE_STAT_T *s)
     ITEM(a)[1].u.integer=-4;
     break;
   }
-  ITEM(a)[2].u.integer=s->st_atime;
-  ITEM(a)[3].u.integer=s->st_mtime;
-  ITEM(a)[4].u.integer=s->st_ctime;
+  ITEM(a)[2].u.integer = DO_NOT_WARN ((INT_TYPE) s->st_atime);
+  ITEM(a)[3].u.integer = DO_NOT_WARN ((INT_TYPE) s->st_mtime);
+  ITEM(a)[4].u.integer = DO_NOT_WARN ((INT_TYPE) s->st_ctime);
   ITEM(a)[5].u.integer=s->st_uid;
   ITEM(a)[6].u.integer=s->st_gid;
   return a;
 }
 
+#if defined(HAVE_FSETXATTR) && defined(HAVE_FGETXATTR) && defined(HAVE_FLISTXATTR)
+/*! @decl array(string) listxattr( string file, void|int(0..1) symlink )
+ *! 
+ *! Return an array of all extended attributes set on the file
+ */
+
+#ifdef HAVE_DARWIN_XATTR
+#define LISTXATTR(PATH, BUF, SZ)	listxattr(PATH, BUF, SZ, 0)
+#define LLISTXATTR(PATH, BUF, SZ)	listxattr(PATH, BUF, SZ, XATTR_NOFOLLOW)
+#else
+#define LISTXATTR(PATH, BUF, SZ)	listxattr(PATH, BUF, SZ)
+#define LLISTXATTR(PATH, BUF, SZ)	llistxattr(PATH, BUF, SZ)
+#endif /* !HAVE_DARWIN_XATTR */
+
+static void f_listxattr(INT32 args)
+{
+  char buffer[1024];
+  char *ptr = buffer;
+  char *name;
+  int do_free = 0;
+  int nofollow = 0;
+  ssize_t res;
+  get_all_args( "listxattr", args, "%s.%d", &name, &nofollow );
+
+  THREADS_ALLOW();
+  do {
+    /* First try, for speed.*/
+    if (nofollow)
+      res = LLISTXATTR(name, buffer, sizeof(buffer));
+    else
+      res = LISTXATTR( name, buffer, sizeof(buffer));
+  } while( res < 0 && errno == EINTR );
+  THREADS_DISALLOW();
+
+  if( res<0 && errno==ERANGE )
+  {
+    /* Too little space in stackbuffer.*/
+    size_t blen = 65536;
+    ptr = NULL;
+    do_free = 1;
+    do {
+      if (!ptr) {
+	ptr = xalloc(blen);
+      } else {
+	char *tmp = realloc( ptr, blen );
+	if( !tmp )
+	  break;
+	ptr = tmp;
+      }
+      THREADS_ALLOW();
+      do {
+	if (nofollow)
+	  res = LLISTXATTR(name, ptr, sizeof(blen));
+	else
+	  res = LISTXATTR(name, ptr, blen);
+      } while( res < 0 && errno == EINTR );
+      THREADS_DISALLOW();
+      blen *= 2;
+    } while( (res < 0) && (errno == ERANGE) );
+  }
+
+  pop_n_elems( args );
+  if (res < 0)
+  {
+    if (do_free)
+      free(ptr);
+    push_int(0);
+    return;
+  }
+
+  push_string( make_shared_binary_string( ptr, res ) );
+  ptr[0]=0;
+  push_string( make_shared_binary_string( ptr, 1 ) );
+  o_divide();
+  push_empty_string();
+  f_aggregate(1);
+  o_subtract();
+
+  if (do_free) 
+    free(ptr);
+}
+
+#ifdef HAVE_DARWIN_XATTR
+#define GETXATTR(PATH, NAME, BUF, SZ)	getxattr(PATH, NAME, BUF, SZ, 0, 0)
+#define LGETXATTR(PATH, NAME, BUF, SZ)	getxattr(PATH, NAME, BUF, SZ, 0, XATTR_NOFOLLOW)
+#else
+#define GETXATTR(PATH, NAME, BUF, SZ)	getxattr(PATH, NAME, BUF, SZ)
+#define LGETXATTR(PATH, NAME, BUF, SZ)	lgetxattr(PATH, NAME, BUF, SZ)
+#endif /* !HAVE_DARWIN_XATTR */
+
+/*! @decl string getxattr(string file, string attr, void|int(0..1) symlink)
+ *! 
+ *! Return the value of a specified attribute, or 0 if it does not exist.
+ */
+static void f_getxattr(INT32 args)
+{
+  char buffer[1024];
+  char *ptr = buffer;
+  int do_free = 0;
+  ssize_t res;
+  char *name, *file;
+  int nofollow=0;
+  get_all_args( "getxattr", args, "%s%s.%d", &file, &name, &nofollow );
+
+  THREADS_ALLOW();
+  do {
+    /* First try, for speed.*/
+    if (nofollow)
+      res = LGETXATTR(file, name, buffer, sizeof(buffer));
+    else
+      res = GETXATTR(file, name, buffer, sizeof(buffer));
+  } while( res < 0 && errno == EINTR );
+  THREADS_DISALLOW();
+
+  if( res<0 && errno==ERANGE )
+  {
+    /* Too little space in buffer.*/
+    size_t blen = 65536;
+    do_free = 1;
+    ptr = NULL;
+    do {
+      if (!ptr) {
+	ptr = xalloc(blen);
+      } else {
+	char *tmp = realloc( ptr, blen );
+	if( !tmp )
+	  break;
+	ptr = tmp;
+      }
+      THREADS_ALLOW();
+      do {
+	if (nofollow)
+	  res = LGETXATTR(file, name, ptr, blen);
+	else
+	  res = GETXATTR(file, name, ptr, blen);
+      } while( res < 0 && errno == EINTR );
+      THREADS_DISALLOW();
+      blen *= 2;
+    } while( (res < 0) && (errno == ERANGE) );
+  }
+
+  if( res < 0 )
+  {
+    if( do_free && ptr )
+      free(ptr);
+    push_int(0);
+    return;
+  }
+
+  push_string( make_shared_binary_string( ptr, res ) );
+  if( do_free && ptr ) 
+    free( ptr );
+}
+
+
+#ifdef HAVE_DARWIN_XATTR
+#define REMOVEXATTR(PATH, NAME)		removexattr(PATH, NAME, 0)
+#define LREMOVEXATTR(PATH, NAME)	removexattr(PATH, NAME, XATTR_NOFOLLOW)
+#else
+#define REMOVEXATTR(PATH, NAME)		removexattr(PATH, NAME)
+#define LREMOVEXATTR(PATH, NAME)	lremovexattr(PATH, NAME)
+#endif /* !HAVE_DARWIN_XATTR */
+
+/*! @decl void removexattr( string file, string attr , void|int(0..1) symlink)
+ *! Remove the specified extended attribute.
+ */
+static void f_removexattr( INT32 args )
+{
+  char *name, *file;
+  int nofollow=0, rv;
+  
+  get_all_args( "removexattr", args, "%s%s.%d", &file, &name, &nofollow );
+
+  THREADS_ALLOW();
+  if (nofollow) {
+    while(((rv=LREMOVEXATTR(file, name)) < 0) && (errno == EINTR))
+      ;
+  } else {
+    while(((rv=REMOVEXATTR(file, name)) < 0) && (errno == EINTR))
+      ;
+  }
+  THREADS_DISALLOW();
+
+  pop_n_elems(args);
+  if( rv < 0 )
+  {
+    push_int(0);
+  }
+  else
+  {
+    push_int(1);
+  }
+}
+
+#ifdef HAVE_DARWIN_XATTR
+#define SETXATTR(PATH, NAME, BUF, SZ, FL)	setxattr(PATH, NAME, BUF, SZ, 0, FL)
+#define LSETXATTR(PATH, NAME, BUF, SZ, FL)	setxattr(PATH, NAME, BUF, SZ, 0, (FL)|XATTR_NOFOLLOW)
+#else
+#define SETXATTR(PATH, NAME, BUF, SZ, FL)	setxattr(PATH, NAME, BUF, SZ, FL)
+#define LSETXATTR(PATH, NAME, BUF, SZ, FL)	lsetxattr(PATH, NAME, BUF, SZ, FL)
+#endif /* !HAVE_DARWIN_XATTR */
+
+/*! @decl void setxattr( string file, string attr, string value, int flags,void|int(0..1) symlink)
+ *!
+ *! Set the attribute @[attr] to the value @[value].
+ *!
+ *! The flags parameter can be used to refine the semantics of the operation.  
+ *!
+ *! @[Stdio.XATTR_CREATE] specifies a pure create, which
+ *! fails if the named attribute exists already.  
+ *!
+ *! @[Stdio.XATTR_REPLACE] specifies a pure replace operation, which
+ *! fails if the named attribute does not already exist.
+ *!
+ *! By default (no flags), the extended attribute will be created if need be, 
+ *! or will simply replace the value if the attribute exists.
+ *!
+ *! @returns
+ *! 1 if successful, 0 otherwise, setting errno.
+ */
+static void f_setxattr( INT32 args )
+{
+  char *ind, *file;
+  struct pike_string *val;  
+  int flags;
+  int rv;
+  int nofollow=0;
+  get_all_args( "setxattr", args, "%s%s%S%d.%d", &file, &ind, &val, &flags, &nofollow );
+
+  THREADS_ALLOW();
+  if (nofollow) {
+    while(((rv=LSETXATTR(file, ind, val->str,
+			  (val->len<<val->size_shift), flags )) < 0) &&
+	  (errno == EINTR))
+      ;
+  } else {
+    while(((rv=SETXATTR(file, ind, val->str,
+			(val->len<<val->size_shift), flags )) < 0) &&
+	  (errno == EINTR))
+      ;
+  }
+  THREADS_DISALLOW();
+  pop_n_elems(args);
+  if( rv < 0 )
+  {
+    push_int(0);
+  }
+  else
+    push_int(1);
+}
+#endif
+
+
 /*! @decl Stdio.Stat file_stat(string path, void|int(0..1) symlink)
  *!
  *! Stat a file.
@@ -183,7 +448,7 @@ void f_file_stat(INT32 args)
   str = sp[-args].u.string;
   l = (args>1 && !UNSAFE_IS_ZERO(sp+1-args))?1:0;
 
-  if (strlen(str->str) != (size_t)str->len) {
+  if (string_has_null(str)) {
     /* Filenames with NUL are not supported. */
     errno = ENOENT;
     pop_n_elems(args);
@@ -220,10 +485,10 @@ void f_file_stat(INT32 args)
  */
 void f_file_truncate(INT32 args)
 {
-#if defined (INT64) || defined (HAVE_TRUNCATE64)
-  INT64 len;
+#if defined(INT64)
+  INT64 len = 0;
 #else
-  off_t len;
+  off_t len = 0;
 #endif
   struct pike_string *str;
   int res;
@@ -251,7 +516,7 @@ void f_file_truncate(INT32 args)
 
   str = sp[-args].u.string;
 
-  if (strlen(str->str) != (size_t)str->len) {
+  if (string_has_null(str)) {
     /* Filenames with NUL are not supported. */
     errno = ENOENT;
     pop_n_elems(args);
@@ -271,7 +536,7 @@ void f_file_truncate(INT32 args)
       LONG high;
       DWORD err;
 #ifdef INT64
-      high = len >> 32;
+      high = DO_NOT_WARN ((LONG) (len >> 32));
       len &= ((INT64) 1 << 32) - 1;
 #else
       high = 0;
@@ -471,7 +736,7 @@ void f_filesystem_stat(INT32 args)
 
   str = sp[-args].u.string;
 
-  if (strlen(str->str) != (size_t)str->len) {
+  if (string_has_null(str)) {
     /* Filenames with NUL are not supported. */
     errno = ENOENT;
     pop_n_elems(args);
@@ -506,12 +771,12 @@ void f_filesystem_stat(INT32 args)
   {
     push_int(0);
   }else{
-    int num_fields = 8;
+    int num_fields = 0;
+    char *fsname = NULL;
 #ifdef HAVE_STATVFS
 #if 0
     push_text("id");         push_int(st.f_fsid);
-#else
-    num_fields--;
+    num_fields++;
 #endif
     push_text("blocksize");  push_int(st.f_frsize);
     push_text("blocks");     push_int(st.f_blocks);
@@ -520,22 +785,21 @@ void f_filesystem_stat(INT32 args)
     push_text("files");      push_int(st.f_files);
     push_text("ffree");      push_int(st.f_ffree);
     push_text("favail");     push_int(st.f_favail);
+    num_fields += 7;
 #ifdef HAVE_STATVFS_F_FSTR
-    push_text("fsname");     push_text(st.f_fstr);
+    push_text("fsname");     push_text(fsname = st.f_fstr);
     num_fields++;
 #endif /* HAVE_STATVFS_F_FSTR */
 #ifdef HAVE_STATVFS_F_BASETYPE
     push_text("fstype");     push_text(st.f_basetype);
     num_fields++;
 #endif /* HAVE_STATVFS_F_BASETYPE */
-    f_aggregate_mapping(num_fields*2);
 #else /* !HAVE_STATVFS */
 #ifdef HAVE_STATFS
 #ifdef HAVE_STRUCT_STATFS
 #if 0 && HAVE_STATFS_F_FSID
     push_text("id");           push_int(st.f_fsid);
-#else
-    num_fields--;
+    num_fields++;
 #endif
     push_text("blocksize");    push_int(st.f_bsize);
     push_text("blocks");       push_int(st.f_blocks);
@@ -543,11 +807,10 @@ void f_filesystem_stat(INT32 args)
     push_text("files");        push_int(st.f_files);
     push_text("ffree");        push_int(st.f_ffree);
     push_text("favail");       push_int(st.f_ffree);
+    num_fields += 6;
 #ifdef HAVE_STATFS_F_BAVAIL
     push_text("bavail");       push_int(st.f_bavail);
-    f_aggregate_mapping((num_fields-1)*2);
-#else
-    f_aggregate_mapping((num_fields-2)*2);
+    num_fields++;
 #endif /* HAVE_STATFS_F_BAVAIL */
 #else /* !HAVE_STRUCT_STATFS */
 #ifdef HAVE_STRUCT_FS_DATA
@@ -556,7 +819,7 @@ void f_filesystem_stat(INT32 args)
     push_text("blocks");       push_int(st.fd_btot);
     push_text("bfree");        push_int(st.fd_bfree);
     push_text("bavail");       push_int(st.fd_bfreen);
-    f_aggregate_mapping(4*2);
+    num_fields += 4;
 #else /* !HAVE_STRUCT_FS_DATA */
     /* Should not be reached */
 #error No struct to hold statfs() data.
@@ -566,46 +829,53 @@ void f_filesystem_stat(INT32 args)
 #ifdef HAVE_USTAT
     push_text("bfree");      push_int(st.f_tfree);
     push_text("ffree");      push_int(st.f_tinode);
-    push_text("fsname");     push_text(st.f_fname);
-    f_aggregate_mapping(3*2);
+    push_text("fsname");     push_text(fsname = st.f_fname);
+    num_fields += 3;
 #else
     /* Should not be reached */
 #error No stat function for filesystems.
 #endif /* HAVE_USTAT */
 #endif /* HAVE_STATFS */
 #endif /* HAVE_STATVFS */
+#if 0
+#if defined(HAVE_LIBZFS_INIT) && defined(HAVE_ZFS_PATH_TO_ZHANDLE)
+    /* zfs_path_to_zhandle() has an unfortunate tendency to output stuff
+     * to stderr when it fails...
+     */
+    if (fsname && !strcmp(fsname, "zfs")) {
+      zfs_handle_t *zfs_handle;
+      if ((str->len == 1) && (str->str[0] == '.')) {
+	/* Workaround for bug in zfs_path_to_zhandle(). */
+	zfs_handle = zfs_path_to_zhandle(libzfs_handle, "./",
+					 ZFS_TYPE_FILESYSTEM);
+      } else {
+	zfs_handle = zfs_path_to_zhandle(libzfs_handle, str->str,
+					 ZFS_TYPE_FILESYSTEM);
+      }
+      if (zfs_handle) {
+#ifndef HAVE_ZFS_PROP_UTF8ONLY
+	/* NOTE: ZFS_PROP_UTF8ONLY is not present in Solaris 10 127112-05.
+	 *       Attempt to be forward compatible.
+	 */
+#define ZFS_PROP_UTF8ONLY	(ZFS_PROP_XATTR + 4)
+#endif /* !HAVE_ZFS_PROP_UTF8ONLY */
+	if (zfs_prop_get_int(zfs_handle, ZFS_PROP_UTF8ONLY)) {
+	  push_text("filename_encoding");
+	  push_text("utf8");
+	  num_fields++;
+	}
+      }
+    }
+#endif /* HAVE_LIBZFS_INIT && HAVE_ZFS_PATH_TO_ZHANDLE */
+#endif /* 0 */
+
+    f_aggregate_mapping(num_fields*2);
   }
 }
   
 #endif /* HAVE_STATVFS || HAVE_STATFS || HAVE_USTAT */
 #endif /* __NT__ */
 
-/*! @decl void werror(string msg, mixed ... args)
- *!
- *! Write to standard error.
- */
-void f_werror(INT32 args)
-{
-  VALID_FILE_IO("werror","werror");
-
-  if(!args)
-    SIMPLE_TOO_FEW_ARGS_ERROR("werror", 1);
-  if(sp[-args].type != T_STRING)
-    SIMPLE_BAD_ARG_ERROR("werror", 1, "string");
-
-  if(args> 1)
-  {
-    f_sprintf(args);
-    args=1;
-  }
-
-  /* FIXME: Wide string handling. */
-  if(sp[-args].u.string->size_shift!=0)
-    Pike_error("Wide strings can not be written to stderr.\n");
-  write_to_stderr(sp[-args].u.string->str, sp[-args].u.string->len);
-  pop_n_elems(args);
-}
-
 /*! @decl int rm(string f)
  *!
  *! Remove a file or directory.
@@ -634,7 +904,7 @@ void f_rm(INT32 args)
 
   str = sp[-args].u.string;
   
-  if (strlen(str->str) != (size_t)str->len) {
+  if (string_has_null(str)) {
     /* Filenames with NUL are not supported. */
     errno = ENOENT;
     pop_n_elems(args);
@@ -701,10 +971,10 @@ void f_rm(INT32 args)
 void f_mkdir(INT32 args)
 {
   struct pike_string *str;
-  char *s, *s_dup;
   int mode;
   int i;
-
+  char *s, *s_dup;
+  
   VALID_FILE_IO("mkdir","write");
 
   if(!args)
@@ -725,7 +995,7 @@ void f_mkdir(INT32 args)
 
   str = sp[-args].u.string;
 
-  if (strlen(str->str) != (size_t)str->len) {
+  if (string_has_null(str)) {
     /* Filenames with NUL are not supported. */
     errno = EINVAL;
     pop_n_elems(args);
@@ -733,24 +1003,21 @@ void f_mkdir(INT32 args)
     return;
   }
 
-#if MKDIR_ARGS == 2
-  /* Remove trailing / which is not accepted by all mkdir()
-     implementations (e.g. Mac OS X) */
+  /* Remove trailing / or \ which is not accepted by all mkdir()
+     implementations (e.g. Mac OS X and Windows) */
   s = str->str;
-  s_dup = 0;
-  if (str->len && s[str->len - 1] == '/') {
+  s_dup = NULL;
+  if (str->len && (s[str->len - 1] == '/' || s[str->len - 1] == '\\')) {
     if ((s_dup = strdup(s))) {
       s = s_dup;
       s[str->len - 1] = '\0';
     }
   }
   
+#if MKDIR_ARGS == 2
   THREADS_ALLOW_UID();
   i = mkdir(s, mode) != -1;
   THREADS_DISALLOW_UID();
-  
-  if (s_dup)
-    free(s_dup);
 #else
 
 #ifdef HAVE_LSTAT
@@ -763,8 +1030,9 @@ void f_mkdir(INT32 args)
     /* Most OS's should have MKDIR_ARGS == 2 nowadays fortunately. */
     int mask = umask(0);
     THREADS_ALLOW_UID();
-    i = mkdir(str->str) != -1;
+    i = mkdir(s) != -1;
     umask(mask);
+    REVEAL_GLOBAL_VARIABLES();
     if (i) {
       /* Attempt to set the mode.
        *
@@ -772,27 +1040,32 @@ void f_mkdir(INT32 args)
        */
       struct stat statbuf1;
       struct stat statbuf2;
-      i = LSTAT(str->str, &statbuf1) != -1;
+      i = LSTAT(s, &statbuf1) != -1;
       if (i) {
 	i = ((statbuf1.st_mode & S_IFMT) == S_IFDIR);
       }
       if (i) {
 	mode = ((mode & 0777) | (statbuf1.st_mode & ~0777)) & ~mask;
 	do {
-	  i = chmod(str->str, mode) != -1;
+	  i = chmod(s, mode) != -1;
 	  if (i || errno != EINTR) break;
 	  /* Must have do { ... } while(0) around these since
 	   * THREADS_DISALLOW_UID contains "} while (0)" and
-	   * THREADS_ALLOW_UID "do {". */
+	   * THREADS_ALLOW_UID "do {".
+	   *
+	   * The same thing applies to {HIDE,REVEAL}_HIDDEN_VARIABLES().
+	   */
 	  do {
+	    HIDE_GLOBAL_VARIABLES();
 	    THREADS_DISALLOW_UID();
 	    check_threads_etc();
 	    THREADS_ALLOW_UID();
+	    REVEAL_GLOBAL_VARIABLES();
 	  } while (0);
 	} while (1);
       }
       if (i) {
-	i = LSTAT(str->str, &statbuf2) != -1;
+	i = LSTAT(s, &statbuf2) != -1;
       }
       if (i) {
 	i = (statbuf2.st_mode == mode) && (statbuf1.st_ino == statbuf2.st_ino);
@@ -801,12 +1074,17 @@ void f_mkdir(INT32 args)
 	}
       }
       if (!i) {
-	rmdir(str->str);
+	rmdir(s);
       }
     }
+    HIDE_GLOBAL_VARIABLES();
     THREADS_DISALLOW_UID();
   }
 #endif
+  
+  if (s_dup)
+    free(s_dup);
+  
   pop_n_elems(args);
   push_int(i);
 }
@@ -838,16 +1116,22 @@ void f_mkdir(INT32 args)
 
 #endif /* _REENTRANT && HAVE_READDIR_R */
 
-/*! @decl array(string) get_dir(string dirname)
+/*! @decl array(string) get_dir(void|string dirname)
  *!
  *! Returns an array of all filenames in the directory @[dirname], or
- *! @expr{0@} (zero) if the directory does not exist.
+ *! @expr{0@} (zero) if the directory does not exist. When no
+ *! @[dirname] is given, current work directory is used.
  *!
  *! @seealso
  *!   @[mkdir()], @[cd()]
  */
 void f_get_dir(INT32 args)
 {
+#ifdef __NT__
+  HANDLE dir;
+  WIN32_FIND_DATAW d;
+  struct string_builder sb;
+#else /* !__NT__ */
 #ifdef USE_FDOPENDIR
   int dir_fd;
 #endif
@@ -855,13 +1139,105 @@ void f_get_dir(INT32 args)
 #ifdef HAVE_READDIR_R
   ptrdiff_t name_max = -1;
 #endif
-  struct pike_string *str;
+#endif /* __NT__ */
+  struct pike_string *str=0;
 
   VALID_FILE_IO("get_dir","read");
 
-  get_all_args("get_dir",args,"%S",&str);
+#ifdef __NT__
+  get_all_args("get_dir",args,".%T",&str);
+#else /* !__NT__ */
+  get_all_args("get_dir",args,".%N",&str);
+#endif /* __NT__ */
 
-  if (strlen(str->str) != (size_t)str->len) {
+  if(!str) {
+#if defined(__amigaos4__)
+    push_empty_string();
+#else
+    push_constant_text(".");
+#endif
+    str = Pike_sp[-1].u.string;
+    args++;
+  }
+
+#ifdef __NT__
+
+  if (str->size_shift == 2) {
+    /* Filenames that are too wide are not supported. */
+    errno = ENOENT;
+    pop_n_elems(args);
+    push_int(0);
+    return;
+  }
+
+  init_string_builder_alloc(&sb, str->len+2, 1);
+
+  string_builder_shared_strcat(&sb, str);
+
+  /* Append "/" "*". */
+  if (sb.s->len && (STR1(sb.s)[sb.s->len-1] != '/') &&
+      (STR1(sb.s)[sb.s->len-1] != '\\')) {
+    STR1(sb.s)[sb.s->len++] = '/';
+  }
+  STR1(sb.s)[sb.s->len++] = '*';
+  STR1(sb.s)[sb.s->len] = '\0';
+
+  if (wcslen(STR1(sb.s)) != (size_t)sb.s->len) {
+    /* Filenames with NUL are not supported. */
+    free_string_builder(&sb);
+    errno = ENOENT;
+    pop_n_elems(args);
+    push_int(0);
+    return;
+  }
+
+#ifdef READDIR_DEBUG
+  fprintf(stderr, "FindFirstFile(\"%S\")...\n", STR1(str));
+#endif /* READDIR_DEBUG */
+
+  dir = FindFirstFileW(STR1(sb.s), &d);
+
+  free_string_builder(&sb);
+
+  if (dir == DO_NOT_WARN(INVALID_HANDLE_VALUE)) {
+#ifdef READDIR_DEBUG
+    fprintf(stderr, "  INVALID_HANDLE_VALUE\n");
+#endif /* READDIR_DEBUG */
+
+    pop_n_elems(args);
+    push_int(0);
+    return;
+  }
+
+  BEGIN_AGGREGATE_ARRAY(10);
+
+  do {
+#ifdef READDIR_DEBUG
+    fprintf(stderr, "  \"%S\"\n", d.cFileName);
+#endif /* READDIR_DEBUG */
+    /* Filter "." and ".." from the list. */
+    if(d.cFileName[0]=='.')
+    {
+      if(!d.cFileName[1]) continue;
+      if(d.cFileName[1]=='.' && !d.cFileName[2]) continue;
+    }
+    push_string(make_shared_binary_string1(d.cFileName, wcslen(d.cFileName)));
+    DO_AGGREGATE_ARRAY(120);
+  } while(FindNextFileW(dir, &d));
+
+#ifdef READDIR_DEBUG
+  fprintf(stderr, "  DONE\n");
+#endif /* READDIR_DEBUG */
+
+  FindClose(dir);
+
+  END_AGGREGATE_ARRAY;
+
+  stack_pop_n_elems_keep_top(args);
+
+#else /* !__NT__ */
+
+  if (string_has_null(str)) {
     /* Filenames with NUL are not supported. */
     errno = ENOENT;
     pop_n_elems(args);
@@ -1013,8 +1389,12 @@ void f_get_dir(INT32 args)
       }
       THREADS_DISALLOW();
       if ((!d) && err) {
-	Pike_error("get_dir(): readdir_r(\"%s\") failed: %d\n", str->str, err);
+	free(tmp);
+	Pike_error("get_dir(): readdir_r(\"%S\") failed: %d\n", str, err);
       }
+#ifdef READDIR_DEBUG
+      fprintf(stderr, "Pushing %d filenames...\n", num_files);
+#endif /* READDIR_DEBUG */
       for(e=0;e<num_files;e++)
       {
 	push_string(make_shared_binary_string(ptrs[e],lens[e]));
@@ -1049,6 +1429,7 @@ void f_get_dir(INT32 args)
     pop_n_elems(args);
     push_int(0);
   }
+#endif /* __NT__ */
 }
 
 /*! @decl int cd(string s)
@@ -1076,7 +1457,7 @@ void f_cd(INT32 args)
 
   str = sp[-args].u.string;
 
-  if (strlen(str->str) != (size_t)str->len) {
+  if (string_has_null(str)) {
     /* Filenames with NUL are not supported. */
     errno = ENOENT;
     pop_n_elems(args);
@@ -1106,7 +1487,7 @@ void f_getcwd(INT32 args)
   size=1000;
   do {
     tmp=(char *)xalloc(size);
-    e = getcwd(tmp,1000); 
+    e = getcwd(tmp,size);
     if (e || errno!=ERANGE) break;
     free(tmp);
     tmp=0;
@@ -1129,7 +1510,7 @@ void f_getcwd(INT32 args)
 
   pop_n_elems(args);
   push_text(e);
-  free(e);
+  free(tmp);
 }
 
 #ifdef HAVE_EXECVE
@@ -1196,9 +1577,9 @@ void f_exece(INT32 args)
   case 2:
     if(sp[1-args].type != T_ARRAY)
       SIMPLE_BAD_ARG_ERROR("exece", 2, "array(string)");
-    array_fix_type_field(sp[1-args].u.array);
 
-    if(sp[1-args].u.array->type_field & ~BIT_STRING)
+
+    if(array_fix_type_field(sp[1-args].u.array) & ~BIT_STRING)
       SIMPLE_BAD_ARG_ERROR("exece", 2, "array(string)");
 
   case 1:
@@ -1221,26 +1602,24 @@ void f_exece(INT32 args)
   if(en)
   {
     INT32 e;
-    struct array *i,*v;
+    struct keypair *k;
 
-    env=(char **)xalloc((1+m_sizeof(en)) * sizeof(char *));
+    env=(char **)malloc((1+m_sizeof(en)) * sizeof(char *));
+    if(!env) {
+      free(argv);
+      SIMPLE_OUT_OF_MEMORY_ERROR("exece", (1+m_sizeof(en)*sizeof(char *)));
+    }
 
-    i=mapping_indices(en);
-    v=mapping_values(en);
-    
-    for(e=0;e<i->size;e++)
-    {
-      push_string(ITEM(i)[e].u.string);
+    NEW_MAPPING_LOOP(en->data) {
+      push_string(k->ind.u.string);
       push_constant_text("=");
-      push_string(ITEM(v)[e].u.string);
+      push_string(k->val.u.string);
       f_add(3);
       env[e]=sp[-1].u.string->str;
       sp--;
       dmalloc_touch_svalue(sp);
     }
-      
-    free_array(i);
-    free_array(v);
+
     env[e]=0;
   }else{
     env=environ;
@@ -1316,10 +1695,9 @@ void f_mv(INT32 args)
   str1 = sp[-args].u.string;
   str2 = sp[1-args].u.string;
 
-  if ((strlen(str1->str) != (size_t)str1->len) ||
-      (strlen(str2->str) != (size_t)str2->len)) {
+  if (string_has_null(str1) || string_has_null(str2)) {
     /* Filenames with NUL are not supported. */
-    if (strlen(str1->str) != (size_t)str1->len) {
+    if (string_has_null(str1)) {
       errno = ENOENT;
     } else {
       errno = EINVAL;
@@ -1397,8 +1775,8 @@ void f_mv(INT32 args)
       if (h != INVALID_HANDLE_VALUE) {
 	do {
 	  if (dir->cFileName[0] != '.' ||
-	      dir->cFileName[1] &&
-	      (dir->cFileName[1] != '.' || dir->cFileName[2])) {
+	      (dir->cFileName[1] &&
+	       (dir->cFileName[1] != '.' || dir->cFileName[2]))) {
 	    /* Target dir not empty. */
 	    FindClose (h);
 	    free (dir);
@@ -1415,15 +1793,17 @@ void f_mv(INT32 args)
     for (p[0] = 'A'; p[0] != 'Z'; p[0]++)
       for (p[1] = 'A'; p[1] != 'Z'; p[1]++) {
 	if (movefileex) {
-	  if (!movefileex (str2->str, s, 0))
+	  if (!movefileex (str2->str, s, 0)) {
 	    if (GetLastError() == ERROR_ALREADY_EXISTS) continue;
 	    else goto nt_rename_kludge_end;
+	  }
 	}
-	else
-	  if (rename (str2->str, s))
+	else {
+	  if (rename (str2->str, s)) {
 	    if (errno == EEXIST) continue;
 	    else goto nt_rename_kludge_end;
-
+	  }
+	}
 	if (movefileex ?
 	    movefileex (str1->str, str2->str, MOVEFILE_REPLACE_EXISTING) :
 	    !rename (str1->str, str2->str)) {
@@ -1553,6 +1933,12 @@ void init_files_efuns(void)
   set_close_on_exec(1,1);
   set_close_on_exec(2,1);
 
+#if 0
+#ifdef HAVE_LIBZFS_INIT
+  libzfs_handle = libzfs_init();
+#endif
+#endif /* 0 */
+
 #ifdef __NT__
   {
     /* MoveFileEx doesn't exist in W98 and earlier. */
@@ -1576,11 +1962,17 @@ void init_files_efuns(void)
 #endif
 
 /* function(string,int|void:object) */
-  ADD_EFUN("file_stat",f_file_stat,tFunc(tStr tOr(tInt,tVoid),tObj), OPT_EXTERNAL_DEPEND|OPT_SIDE_EFFECT);
+  ADD_EFUN("file_stat",f_file_stat,tFunc(tStr tOr(tInt,tVoid),tObjIs_STDIO_STAT), OPT_EXTERNAL_DEPEND|OPT_SIDE_EFFECT);
 
   /* function(string,int:int(0..1)) */
   ADD_EFUN("file_truncate",f_file_truncate,tFunc(tStr tInt,tInt),OPT_EXTERNAL_DEPEND|OPT_SIDE_EFFECT);
 
+#if defined(HAVE_FSETXATTR) && defined(HAVE_FGETXATTR) && defined(HAVE_FLISTXATTR)
+  ADD_EFUN( "listxattr", f_listxattr, tFunc(tStr tOr(tVoid,tInt),tArr(tStr)), OPT_EXTERNAL_DEPEND|OPT_SIDE_EFFECT);
+  ADD_EFUN( "setxattr", f_setxattr, tFunc(tStr tStr tStr tInt tOr(tVoid,tInt),tInt), OPT_EXTERNAL_DEPEND|OPT_SIDE_EFFECT);
+  ADD_EFUN( "getxattr", f_getxattr, tFunc(tStr tStr tOr(tVoid,tInt),tStr), OPT_EXTERNAL_DEPEND|OPT_SIDE_EFFECT);
+  ADD_EFUN( "removexattr", f_removexattr, tFunc(tStr tStr  tOr(tVoid,tInt),tInt), OPT_EXTERNAL_DEPEND|OPT_SIDE_EFFECT);
+#endif
 
 #if defined(HAVE_STATVFS) || defined(HAVE_STATFS) || defined(HAVE_USTAT) || defined(__NT__)
   
@@ -1591,9 +1983,6 @@ void init_files_efuns(void)
 /* function(:int) */
   ADD_EFUN("errno",f_errno,tFunc(tNone,tInt),OPT_EXTERNAL_DEPEND);
   
-/* function(string,void|mixed...:void) */
-  ADD_EFUN("werror",f_werror,tFuncV(tStr,tOr(tVoid,tMix),tVoid),OPT_SIDE_EFFECT);
-  
 /* function(string:int) */
   ADD_EFUN("rm",f_rm,tFunc(tStr,tInt),OPT_SIDE_EFFECT);
   
@@ -1604,7 +1993,7 @@ void init_files_efuns(void)
   ADD_EFUN("mv", f_mv,tFunc(tStr tStr,tInt), OPT_SIDE_EFFECT);
   
 /* function(string:string *) */
-  ADD_EFUN("get_dir",f_get_dir,tFunc(tStr,tArr(tStr)),OPT_EXTERNAL_DEPEND|OPT_SIDE_EFFECT);
+  ADD_EFUN("get_dir",f_get_dir,tFunc(tOr(tVoid,tStr),tArr(tStr)),OPT_EXTERNAL_DEPEND|OPT_SIDE_EFFECT);
   
 /* function(string:int) */
   ADD_EFUN("cd",f_cd,tFunc(tStr,tInt),OPT_SIDE_EFFECT);
@@ -1630,4 +2019,9 @@ void exit_files_efuns(void)
     kernel32lib = 0;
   }
 #endif
+#if 0
+#ifdef HAVE_LIBZFS_INIT
+  libzfs_fini(libzfs_handle);
+#endif
+#endif /* 0 */
 }
