@@ -2,7 +2,7 @@
 || This file is part of Pike. For copyright information see COPYRIGHT.
 || Pike is distributed under GPL, LGPL and MPL. See the file COPYING
 || for more information.
-|| $Id: socket.c,v 1.87 2004/05/13 17:53:44 bill Exp $
+|| $Id: socket.c,v 1.107 2009/08/05 13:59:03 grubba Exp $
 */
 
 #define NO_PIKE_SHORTHAND
@@ -20,20 +20,15 @@
 #include "fd_control.h"
 #include "threads.h"
 #include "program_id.h"
+#include "module_support.h"
 
 #include "file_machine.h"
 #include "file.h"
 
-RCSID("$Id: socket.c,v 1.87 2004/05/13 17:53:44 bill Exp $");
-
 #ifdef HAVE_SYS_TYPE_H
 #include <sys/types.h>
 #endif
 
-#ifdef HAVE_SYS_UN_H
-#include <sys/un.h>
-#endif
-
 #ifdef HAVE_SYS_PARAM_H
 #include <sys/param.h>
 #endif
@@ -65,6 +60,10 @@ RCSID("$Id: socket.c,v 1.87 2004/05/13 17:53:44 bill Exp $");
 #include <sys/socketvar.h>
 #endif
 
+#ifdef HAVE_SYS_UN_H
+#include <sys/un.h>
+#endif
+
 #include "dmalloc.h"
 
 /*! @module Stdio
@@ -109,6 +108,7 @@ static void assign_accept_cb (struct port *p, struct svalue *cb)
   if (UNSAFE_IS_ZERO (cb)) {
     if (p->box.backend)
       set_fd_callback_events (&p->box, 0);
+    set_nonblocking(p->box.fd,0);
   }
   else {
     if (!p->box.backend)
@@ -144,9 +144,7 @@ static void do_close(struct port *p)
  */
 static void port_set_id(INT32 args)
 {
-  if(args < 1)
-    SIMPLE_TOO_FEW_ARGS_ERROR("Port->set_id", 1);
-
+  check_all_args("Port->set_id", args, BIT_MIXED, 0);
   assign_svalue(& THIS->id, Pike_sp-args);
   pop_n_elems(args-1);
 }
@@ -167,9 +165,9 @@ static void port_query_id(INT32 args)
 
 /*! @decl int errno()
  *!
- *! If the last call done on this port failed, errno will return an
- *! integer describing what went wrong. Refer to your unix manual for
- *! further information.
+ *! If the last call done on this port failed, this function will
+ *! return an integer describing what went wrong. Refer to your unix
+ *! manual for further information.
  */
 static void port_errno(INT32 args)
 {
@@ -179,9 +177,9 @@ static void port_errno(INT32 args)
 
 /*! @decl int listen_fd(int fd, void|function accept_callback)
  *!
- *! This function does the same as port->bind, except that instead
- *! of creating a new socket and bind it to a port, it expects that
- *! the filedescriptor 'fd' is an already open port.
+ *! This function does the same as @[bind], except that instead of
+ *! creating a new socket and bind it to a port, it expects the file
+ *! descriptor @[fd] to be an already open port.
  *!
  *! @note
  *!  This function is only for the advanced user, and is generally used
@@ -193,16 +191,11 @@ static void port_errno(INT32 args)
 static void port_listen_fd(INT32 args)
 {
   struct port *p = THIS;
+  struct svalue *cb;
   int fd;
   do_close(p);
 
-  if(args < 1)
-    SIMPLE_TOO_FEW_ARGS_ERROR("Port->bind_fd", 1);
-
-  if(Pike_sp[-args].type != PIKE_T_INT)
-    SIMPLE_BAD_ARG_ERROR("Port->bind_fd", 1, "int");
-
-  fd=Pike_sp[-args].u.integer;
+  get_all_args("Port->listen_fd", args, "%d.%*", &fd, &cb);
 
   if(fd<0)
   {
@@ -221,25 +214,31 @@ static void port_listen_fd(INT32 args)
   }
 
   change_fd_for_box (&p->box, fd);
-  if(args > 1) assign_accept_cb (p, Pike_sp+1-args);
+  if(args > 1) assign_accept_cb (p, cb);
   p->my_errno=0;
   pop_n_elems(args);
   push_int(1);
 }
 
-/*! @decl int bind(int|string port, void|function accept_callback, void|string ip)
+/*! @decl int bind(int|string port, void|function accept_callback, @
+ *!                void|string ip)
  *!
- *! Bind opens a socket and binds it to port number on the local machine.
- *! If the second argument is present, the socket is set to nonblocking
- *! and the callback funcition is called whenever something connects to
- *! it. The callback will receive the id for this port as argument.
- *! Bind returns 1 on success, and zero on failiure.
- *! 
- *! If the optional argument 'ip' is given, bind will try to bind to
- *! this ip name or number.
+ *! Opens a socket and binds it to port number on the local machine.
+ *! If the second argument is present, the socket is set to
+ *! nonblocking and the callback funcition is called whenever
+ *! something connects to it. The callback will receive the id for
+ *! this port as argument and should typically call @[accept] to
+ *! establish a connection.
+ *!
+ *! If the optional argument @[ip] is given, @[bind] will try to bind
+ *! to an interface with that host name or IP number.
+ *!
+ *! @returns
+ *!   1 is returned on success, zero on failure. @[errno] provides
+ *!   further details about the error in the latter case.
  *!
  *! @seealso
- *!   @[accept]
+ *!   @[accept], @[set_id]
  */
 static void port_bind(INT32 args)
 {
@@ -295,6 +294,13 @@ static void port_bind(INT32 args)
   tmp=fd_bind(fd, (struct sockaddr *)&addr, addr_len) < 0 || fd_listen(fd, 16384) < 0;
   THREADS_DISALLOW_UID();
 
+  if(!Pike_fp->current_object->prog)
+  {
+    if (fd >= 0)
+      while (fd_close(fd) && errno == EINTR) {}
+    Pike_error("Object destructed in Stdio.Port->bind()\n");
+  }
+
   if(tmp)
   {
     p->my_errno=errno;
@@ -317,50 +323,55 @@ static void port_bind(INT32 args)
 
 /*! @decl int bind_unix(string path, void|function accept_callback)
  *!
- *! Bind opens a Unix domain socket at the filesystem location path.
- *! If the second argument is present, the socket is set to nonblocking
- *! and the callback funcition is called whenever something connects to
- *! it. The callback will receive the id for this port as argument.
- *! 
- *! @returns 
- *!    1 on success, and zero on failiure.
+ *! Opens a Unix domain socket at the given path in the file system.
+ *! If the second argument is present, the socket is set to
+ *! nonblocking and the callback funcition is called whenever
+ *! something connects to it. The callback will receive the id for
+ *! this port as argument and should typically call @[accept] to
+ *! establish a connection.
+ *!
+ *! @returns
+ *!   1 is returned on success, zero on failure. @[errno] provides
+ *!   further details about the error in the latter case.
  *!
  *! @note
- *!   this function is only available on systems that support Unix domain 
+ *!   This function is only available on systems that support Unix domain
  *!   sockets.
- *! 
+ *!
+ *! @note
+ *!   @[path] had a quite restrictive length limit (~100 characters)
+ *!   prior to Pike 7.8.334.
+ *!
  *! @seealso
- *!   @[accept]
+ *!   @[accept], @[set_id]
  */
 static void unix_bind(INT32 args)
 {
   struct port *p = THIS;
-  struct sockaddr_un addr;
+  struct sockaddr_un *addr;
+  struct pike_string *path;
+  struct svalue *cb;
   int addr_len,fd,tmp;
 
   do_close(p);
 
-  if(args < 1)
-    SIMPLE_TOO_FEW_ARGS_ERROR("Port->bind", 1);
+  get_all_args("Port->bind_unix", args, "%n%*", &path, &cb);
 
-  if((Pike_sp[-args].type != PIKE_T_STRING ||
-      Pike_sp[-args].u.string->size_shift))
-    SIMPLE_BAD_ARG_ERROR("Port->bind", 1, "string (8bit)");
+  addr_len = sizeof(struct sockaddr_un) + path->len;
+  addr = xalloc(addr_len);
 
-  tmp = sizeof(addr.sun_path);
-
-  if(Pike_sp[-args].u.string->len > tmp)
-    SIMPLE_BAD_ARG_ERROR("Port->Bind", 1, "path too long");
-
-  strcpy(addr.sun_path, Pike_sp[-args].u.string->str);
-  addr.sun_family = AF_UNIX;
-
-  addr_len = sizeof(struct sockaddr_un);
+  strcpy(addr->sun_path, path->str);
+  addr->sun_family = AF_UNIX;
+#ifdef HAVE_STRUCT_SOCKADDR_UN_SUN_LEN
+  /* Length including NUL. */
+  addr->sun_len = path->len + 1;
+#endif
 
   fd=fd_socket(AF_UNIX, SOCK_STREAM, 0);
 
   if(fd < 0)
   {
+    free(addr);
     p->my_errno=errno;
     pop_n_elems(args);
     push_int(0);
@@ -370,9 +381,11 @@ static void unix_bind(INT32 args)
 #ifndef __NT__
   {
     int o=1;
-    if(fd_setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, (char *)&o, sizeof(int)) < 0)
+    if(fd_setsockopt(fd, SOL_SOCKET,
+		     SO_REUSEADDR, (char *)&o, sizeof(int)) < 0)
     {
       p->my_errno=errno;
+      free(addr);
       while (fd_close(fd) && errno == EINTR) {}
       errno = p->my_errno;
       pop_n_elems(args);
@@ -385,10 +398,26 @@ static void unix_bind(INT32 args)
   my_set_close_on_exec(fd,1);
 
   THREADS_ALLOW_UID();
-  tmp=fd_bind(fd, (struct sockaddr *)&addr, addr_len) < 0 || fd_listen(fd, 16384) < 0;
+  do {
+    tmp = fd_bind(fd, (struct sockaddr *)&addr, addr_len);
+  } while ((tmp < 0) && (errno == EINTR));
+  if (tmp >= 0) {
+    do {
+      tmp = fd_listen(fd, 16384);
+    } while ((tmp < 0) && (errno == EINTR));
+  }
   THREADS_DISALLOW_UID();
 
-  if(tmp)
+  free(addr);
+
+  if(!Pike_fp->current_object->prog)
+  {
+    if (fd >= 0)
+      while (fd_close(fd) && errno == EINTR) {}
+    Pike_error("Object destructed in Stdio.Port->unix_bind()\n");
+  }
+
+  if(tmp < 0)
   {
     p->my_errno=errno;
     while (fd_close(fd) && errno == EINTR) {}
@@ -399,7 +428,7 @@ static void unix_bind(INT32 args)
   }
 
   change_fd_for_box (&p->box, fd);
-  if(args > 1) assign_accept_cb (p, Pike_sp+1-args);
+  if(args > 1) assign_accept_cb (p, cb);
   p->my_errno=0;
   pop_n_elems(args);
   push_int(1);
@@ -417,14 +446,20 @@ static void port_close (INT32 args)
   do_close (THIS);
 }
 
-/* @decl void create("stdin", void|function accept_callback)
- * @decl void create(int|string port, void|function accept_callback, void|string ip)
- *
- * When create is called with 'stdin' as argument, a socket is created
- * out of the file descriptor 0. This is only useful if that actually
- * IS a socket to begin with. When create is called with an int or any
- * other string as first argument, it does the same as bind() would do
- * with the same arguments.
+/*! @decl void create(int|string port, void|function accept_callback, @
+ *!                   void|string ip)
+ *! @decl void create("stdin", void|function accept_callback)
+ *!
+ *! When called with an int or any string except @expr{"stdin"@} as
+ *! first argument, this function does the same as @[bind()] would do
+ *! with the same arguments.
+ *!
+ *! When called with @expr{"stdin"@} as argument, a socket is created
+ *! out of the file descriptor 0. This is only useful if that actually
+ *! IS a socket to begin with.
+ *!
+ *! @seealso
+ *!   @[bind], @[listen_fd]
  */
 static void port_create(INT32 args)
 {
@@ -432,7 +467,8 @@ static void port_create(INT32 args)
   {
     if(Pike_sp[-args].type == PIKE_T_INT ||
        (Pike_sp[-args].type == PIKE_T_STRING &&
-	strcmp("stdin",Pike_sp[-args].u.string->str)))
+	(Pike_sp[-args].u.string->len != 5 ||
+	 strcmp("stdin",Pike_sp[-args].u.string->str))))
     {
       port_bind(args); /* pops stack */
       return;
@@ -442,8 +478,7 @@ static void port_create(INT32 args)
       if(Pike_sp[-args].type != PIKE_T_STRING)
 	SIMPLE_TOO_FEW_ARGS_ERROR("Port->create", 1);
 
-      if(strcmp("stdin",Pike_sp[-args].u.string->str))
-	Pike_error("port->create() called with string other than 'stdin'\n");
+      /* FIXME: Check that the argument is "stdin". */
 
       do_close(p);
       change_fd_for_box (&p->box, 0);
@@ -461,7 +496,27 @@ static void port_create(INT32 args)
 
 extern struct program *file_program;
 
+static int port_fd_factory_fun_num = -1;
+static void port_fd_factory(INT32 args)
+{
+  pop_n_elems(args);
+  push_object(clone_object(file_program, 0));
+}
+
 /*! @decl Stdio.File accept()
+ *!
+ *! Get the first connection request waiting for this port and return
+ *! it as a connected socket.
+ *!
+ *! If no connection request is waiting and the port is in nonblocking
+ *! mode (i.e. an accept callback is installed) then zero is returned.
+ *! Otherwise this function waits until a connection has arrived.
+ *!
+ *! In Pike 7.8 and later the returned object is created via @[fd_factory()].
+ *!
+ *! @note
+ *!   In Pike 7.7 and later the resulting file object will be assigned
+ *!   to the same backend as the port object.
  */
 
 static void port_accept(INT32 args)
@@ -469,12 +524,15 @@ static void port_accept(INT32 args)
   PIKE_SOCKADDR addr;
   struct port *this=THIS;
   int fd, err;
-  struct object *o;
   ACCEPT_SIZE_T len=0;
+  int one = 1;
 
   if(this->box.fd < 0)
     Pike_error("port->accept(): Port not open.\n");
 
+  /* FIXME: Race. */
+  THIS->box.revents = 0;
+
   THREADS_ALLOW();
   len=sizeof(addr);
   do {
@@ -491,30 +549,51 @@ static void port_accept(INT32 args)
     return;
   }
 
-  my_set_close_on_exec(fd,1);
-  o=file_make_object_from_fd(fd,FILE_READ | FILE_WRITE, SOCKET_CAPABILITIES);
+  /* We don't really care if setsockopt fails, since it's just a hint. */
+  while ((fd_setsockopt(fd, SOL_SOCKET, SO_REUSEADDR,
+			(char *)&one, sizeof(int)) < 0) &&
+	 (errno == EINTR))
+    one = 1;
 
-  pop_n_elems(args);
-  push_object(o);
+  my_set_close_on_exec(fd,1);
+  push_new_fd_object(port_fd_factory_fun_num,
+		     fd, FILE_READ | FILE_WRITE, SOCKET_CAPABILITIES);
+
+  if (this->box.backend) {
+    struct object *o = Pike_sp[-1].u.object;
+    struct my_file *f = (struct my_file *)
+      (o->storage + o->prog->inherits[Pike_sp[-1].subtype].storage_offset);
+    change_backend_for_box(&f->box, this->box.backend);
+  }
+
+  stack_pop_n_elems_keep_top(args);
 }
 
-/*! @decl string query_address(string arg1)
+/*! @decl string query_address()
  *!
- *! @fixme
- *!   Document this function.
+ *! Get the address and port of the local socket end-point.
+ *!
+ *! @returns
+ *!   This function returns the address and port of a socket end-point
+ *!   on the form @expr{"x.x.x.x port"@} (IPv4) or
+ *!   @expr{"x:x:x:x:x:x:x:x port"@} (IPv6).
+ *!
+ *!   If there is some error querying or formatting the address,
+ *!   @expr{0@} (zero) is returned and @[errno()] will return the
+ *!   error code.
+ *!
+ *! @throws
+ *!   An error is thrown if the socket isn't bound.
  */
 static void socket_query_address(INT32 args)
 {
   PIKE_SOCKADDR addr;
   int i;
   char buffer[496];
-#ifndef HAVE_INET_NTOP
-  char *q;
-#endif
   ACCEPT_SIZE_T len;
 
   if(THIS->box.fd <0)
-    Pike_error("socket->query_address(): Socket not bound yet.\n");
+    Pike_error("Stdio.Port->query_address(): Socket not bound yet.\n");
 
   len=sizeof(addr);
   i=fd_getsockname(THIS->box.fd,(struct sockaddr *)&addr,&len);
@@ -526,13 +605,29 @@ static void socket_query_address(INT32 args)
     return;
   }
 
-#ifdef HAVE_INET_NTOP
-  inet_ntop(SOCKADDR_FAMILY(addr), SOCKADDR_IN_ADDR(addr),
-	    buffer, sizeof(buffer)-20);
+#ifdef fd_inet_ntop
+  if(!fd_inet_ntop(SOCKADDR_FAMILY(addr), SOCKADDR_IN_ADDR(addr),
+		   buffer, sizeof(buffer)-20))
+  {
+    THIS->my_errno = errno;
+    push_int(0);
+    return;
+  }
 #else
-  q=inet_ntoa(*SOCKADDR_IN_ADDR(addr));
-  strncpy(buffer,q,sizeof(buffer)-20);
-  buffer[sizeof(buffer)-20]=0;
+  if(SOCKADDR_FAMILY(addr) == AF_INET)
+  {
+    char *q = inet_ntoa(*SOCKADDR_IN_ADDR(addr));
+    strncpy(buffer,q,sizeof(buffer)-20);
+    buffer[sizeof(buffer)-20]=0;
+  }else{
+#ifdef EAFNOSUPPORT
+    THIS->my_errno = EAFNOSUPPORT;
+#else
+    THIS->my_errno = EINVAL;
+#endif
+    push_int(0);
+    return;
+  }
 #endif
   sprintf(buffer+strlen(buffer)," %d",(int)(ntohs(addr.ipv4.sin_port)));
 
@@ -611,7 +706,7 @@ static void exit_port_struct(struct object *o)
 /*! @endmodule
  */
 
-struct program *port_program = NULL;
+PMOD_EXPORT struct program *port_program = NULL;
 
 void port_exit_program(void)
 {
@@ -623,9 +718,9 @@ void port_setup_program(void)
   ptrdiff_t offset;
   START_NEW_PROGRAM_ID (STDIO_PORT);
   offset = ADD_STORAGE(struct port);
-  map_variable("_accept_callback", "mixed", 0,
+  MAP_VARIABLE("_accept_callback", tMix, 0,
 	       offset + OFFSETOF(port, accept_callback), PIKE_T_MIXED);
-  map_variable("_id", "mixed", 0,
+  MAP_VARIABLE("_id", tMix, 0,
 	       offset + OFFSETOF(port, id), PIKE_T_MIXED);
   /* function(int|string,void|mixed,void|string:int) */
   ADD_FUNCTION("bind", port_bind,
@@ -633,7 +728,7 @@ void port_setup_program(void)
 #ifdef HAVE_SYS_UN_H
   /* function(int|string,void|mixed,void|string:int) */
   ADD_FUNCTION("bind_unix", unix_bind,
-	       tFunc(tStr tOr(tVoid,tMix),tInt), 0);
+               tFunc(tStr tOr(tVoid,tMix),tInt), 0);
 #endif /* HAVE_SYS_UN_H */
   ADD_FUNCTION("close",port_close,tFunc(tNone,tVoid),0);
   /* function(int,void|mixed:int) */
@@ -647,6 +742,9 @@ void port_setup_program(void)
   /* function(:int) */
   ADD_FUNCTION("errno",port_errno,tFunc(tNone,tInt),0);
   /* function(:object) */
+  port_fd_factory_fun_num =
+    ADD_FUNCTION("fd_factory", port_fd_factory, tFunc(tNone,tObjIs_STDIO_FD),
+		 ID_STATIC);
   ADD_FUNCTION("accept",port_accept,tFunc(tNone,tObjIs_STDIO_FD),0);
   /* function(void|string|int,void|mixed,void|string:void) */
   ADD_FUNCTION("create", port_create,
