@@ -2,15 +2,13 @@
 || This file is part of Pike. For copyright information see COPYRIGHT.
 || Pike is distributed under GPL, LGPL and MPL. See the file COPYING
 || for more information.
-|| $Id: spider.c,v 1.126 2004/04/06 13:00:50 nilsson Exp $
+|| $Id: spider.c,v 1.135 2008/06/28 18:35:59 mast Exp $
 */
 
 #include "global.h"
 #include "config.h"
 
 
-#include "machine.h"
-
 #include <sys/types.h>
 #include <sys/stat.h>
 #ifdef HAVE_SYS_MMAN_H
@@ -34,10 +32,19 @@
 #include <unistd.h>
 #endif
 
+/* Some <sys/mman.h>'s (eg AIX 5L/ia64) contain a #define of MAP_VARIABLE
+ * for use as the opposite of MAP_FIXED.
+ *
+ * "program.h" included indirectly below has a conflicting definition of
+ * MAP_VARIABLE.
+ */
+#ifdef MAP_VARIABLE
+#undef MAP_VARIABLE
+#endif /* MAP_VARIABLE */
+
 #include "fdlib.h"
 #include "stralloc.h"
 #include "pike_macros.h"
-#include "machine.h"
 #include "object.h"
 #include "constants.h"
 #include "interpret.h"
@@ -51,8 +58,6 @@
 #include "operators.h"
 #include "pike_security.h"
 
-RCSID("$Id: spider.c,v 1.126 2004/04/06 13:00:50 nilsson Exp $");
-
 #ifdef HAVE_PWD_H
 #include <pwd.h>
 #undef HAVE_PWD_H
@@ -60,10 +65,6 @@ RCSID("$Id: spider.c,v 1.126 2004/04/06 13:00:50 nilsson Exp $");
 
 #include "defs.h"
 
-#ifdef HAVE_SYS_CONF_H
-#include <sys/conf.h>
-#endif
-
 #ifdef HAVE_STROPTS_H
 #include <stropts.h>
 #endif
@@ -126,7 +127,7 @@ void f_parse_accessed_database(INT32 args)
     SIMPLE_TOO_FEW_ARGS_ERROR("parse_accessed_database",1);
 
   if ((sp[-args].type != T_STRING) || (sp[-args].u.string->size_shift)) {
-    Pike_error("Bad argument 1 to parse_accessed_database(string(8)).\n");
+    Pike_error("Bad argument 1 to parse_accessed_database(string(0..255)).\n");
   }
 
   /* Pop all but the first argument */
@@ -191,7 +192,7 @@ void f_parse_html(INT32 args)
   if(!ss->len)
   {
     pop_n_elems(args);
-    push_text("");
+    push_empty_string();
     return;
   }
 
@@ -232,7 +233,7 @@ void f_parse_html(INT32 args)
   if(strings > 1)
     f_add(strings);
   else if(!strings)
-    push_text("");
+    push_empty_string();
 }
 
 
@@ -258,11 +259,11 @@ void f_parse_html_lines(INT32 args)
   if(!ss->len)
   {
     pop_n_elems(args);
-    push_text("");
+    push_empty_string();
     return;
   }
 
-  sp[-args].type=T_INT;
+  mark_free_svalue (sp - args);
 
   add_ref(single=sp[1-args].u.mapping);
   add_ref(cont=sp[2-args].u.mapping);
@@ -298,12 +299,12 @@ void f_parse_html_lines(INT32 args)
   if(strings > 1)
     f_add(strings);
   else if(!strings)
-    push_text("");
+    push_empty_string();
 /*   fprintf(stderr, "sp=%p (strings=%d)\n", sp, strings); */
 }
 
-char start_quote_character = '\000';
-char end_quote_character = '\000';
+static char start_quote_character = '\000';
+static char end_quote_character = '\000';
 
 /*! @decl void set_end_quote(int quote)
  */
@@ -311,7 +312,7 @@ void f_set_end_quote(INT32 args)
 {
   if(args < 1 || sp[-1].type != T_INT)
     Pike_error("Wrong argument to set_end_quote(int CHAR)\n");
-  end_quote_character = sp[-1].u.integer;
+  end_quote_character = (char) sp[-1].u.integer;
 }
 
 /*! @decl void set_start_quote(int quote)
@@ -320,7 +321,7 @@ void f_set_start_quote(INT32 args)
 {
   if(args < 1 || sp[-1].type != T_INT)
     Pike_error("Wrong argument to set_start_quote(int CHAR)\n");
-  start_quote_character = sp[-1].u.integer;
+  start_quote_character = (char) sp[-1].u.integer;
 }
 
 
@@ -399,7 +400,7 @@ done:
   if(strs > 1)
     f_add(strs);
   else if(!strs)
-    push_text("");
+    push_empty_string();
 
   SKIP_SPACE();
   return i;
@@ -622,7 +623,7 @@ void do_html_parse(struct pike_string *ss,
 	  do_html_parse(ss2,cont,single,strings,recurse_left-1,extra_args);
 	  continue;
 	} else if (sp[-1].type==T_ARRAY) {
-	  push_text("");
+	  push_empty_string();
 	  f_multiply(2);
 	  copy_shared_string(ss2,sp[-1].u.string);
 	  pop_stack();
@@ -705,7 +706,7 @@ void do_html_parse(struct pike_string *ss,
 	  continue;
 
 	} else if (sp[-1].type==T_ARRAY) {
-	  push_text("");
+	  push_empty_string();
 	  f_multiply(2);
 	  copy_shared_string(ss2,sp[-1].u.string);
 	  pop_stack();
@@ -765,7 +766,7 @@ void do_html_parse(struct pike_string *ss,
 
 
 #define PARSE_RETURN(END) do{					\
-  push_text("");						\
+  push_empty_string();						\
   f_multiply(2);						\
   (*strings)++;							\
   if (last!=i-1)						\
@@ -1100,7 +1101,7 @@ void f__dump_obj_table(INT32 args)
 
 PIKE_MODULE_INIT
 {
-  push_constant_text("");
+  push_empty_string();
   empty_string_svalue = sp[-1];
   pop_stack();
 
@@ -1173,18 +1174,9 @@ PIKE_MODULE_INIT
 
   /* function(int:string) */
   ADD_FUNCTION("fd_info", f_fd_info,tFunc(tInt,tStr), OPT_EXTERNAL_DEPEND);
-
-  {
-    extern void init_xml(void);
-    init_xml();
-  }
 }
 
 
 PIKE_MODULE_EXIT
 {
-  {
-    extern void exit_xml(void);
-    exit_xml();
-  }
 }
