@@ -2,7 +2,7 @@
 || This file is part of Pike. For copyright information see COPYRIGHT.
 || Pike is distributed under GPL, LGPL and MPL. See the file COPYING
 || for more information.
-|| $Id: memory.c,v 1.28 2003/10/10 19:28:34 nilsson Exp $
+|| $Id: memory.c,v 1.35 2008/01/22 17:17:44 grubba Exp $
 */
 
 /*! @module System
@@ -19,8 +19,6 @@
  *!	Don't blame Pike if you shoot your foot off.
  */
 #include "global.h"
-RCSID("$Id: memory.c,v 1.28 2003/10/10 19:28:34 nilsson Exp $");
-
 #include "system_machine.h"
 
 #ifdef HAVE_UNISTD_H
@@ -241,12 +239,12 @@ static void memory_shm( INT32 args )
   THIS->flags = MEM_READ|MEM_WRITE|MEM_FREE_SHMDEL;
   pop_n_elems(args);
   push_int(1);
-#else /* HAVE_MMAP */
+#else /* HAVE_SYS_SHM_H */
 #ifdef WIN32SHM
   {
     HANDLE handle;
     char id[4711];
-    sprintf( id, "pike.%d", Pike_sp[-args].u.integer );
+    sprintf( id, "pike.%ld", Pike_sp[-args].u.integer );
     THIS->size = Pike_sp[1-args].u.integer;
     THIS->flags = MEM_READ|MEM_WRITE|MEM_FREE_SHMDEL;
     pop_n_elems(args);
@@ -271,10 +269,10 @@ static void memory_shm( INT32 args )
 	Pike_error("Failed to create segment\n");
     }
   }
-#else
+#else /* !WIN32SHM */
    Pike_error("Memory.shmat(): system has no shmat() (sorry)\n");
-#endif
-#endif
+#endif /* WIN32SHM */
+#endif /* HAVE_SYS_SHM_H */
 }
 
 static void memory__mmap(INT32 args,int complain,int private)
@@ -371,7 +369,7 @@ static void memory__mmap(INT32 args,int complain,int private)
 #ifdef PAGE_SIZE
    if (offset%PAGE_SIZE)
       Pike_error("Memory.mmap(): mapped offset not aligned to PAGE_SIZE "
-		 "(%d aka system.PAGE_SIZE)\n",(int)offset);
+		 "(%d aka System.PAGE_SIZE)\n",(int)offset);
 #endif
 
    if (private) flags|=MAP_PRIVATE;
@@ -595,7 +593,7 @@ static void memory_cast(INT32 args)
    	 Pike_error(NAME": reading (some) outside allocation\n");	\
    									\
       if (!rlen)							\
-   	 push_text("");							\
+   	 push_empty_string();						\
       else								\
    	 push_string(MAKER((void*)(THIS->p+pos),len));			\
    									\
@@ -668,13 +666,14 @@ static void copy_reverse_string1_to_2(unsigned char *d,
 }
 
 #define MAKE_REVERSE_ORDER_STRINGN(N)					\
-   static struct pike_string*						\
-      make_reverse_order_string##N(unsigned char *s,size_t len)		\
+  static struct pike_string *PIKE_CONCAT(make_reverse_order_string, N)	\
+       (unsigned char *s, size_t len)					\
    {									\
-      struct pike_string *ps;						\
-      ps=begin_wide_shared_string(len,N);				\
-      copy_reverse_string##N(ps->str, (unsigned char *)s, len);		\
-      return end_shared_string(ps);					\
+     struct pike_string *ps;						\
+     ps = begin_wide_shared_string(len, N);				\
+     PIKE_CONCAT(copy_reverse_string, N)				\
+       ((unsigned char *) PIKE_CONCAT(STR, N)(ps), s, len);		\
+     return end_shared_string(ps);					\
    }
 
 MAKE_REVERSE_ORDER_STRINGN(1)
@@ -739,34 +738,34 @@ static void pwrite_n(INT32 args,int shift,int reverse,char *func)
 #endif
 
    if (rlen) 
-      switch (ps->size_shift*10 + shift)
+      switch (ps->size_shift*010 + shift)
       {
-	 case 22: /* 2 -> 2 */
+	 case 022: /* 2 -> 2 */
 	    if (reverse)
 	      copy_reverse_string2(d, (unsigned char *)ps->str, ps->len);
 	    else MEMCPY(d,ps->str,ps->len*4);
 	    break;
-	 case 12: /* 1 -> 2 */
+	 case 012: /* 1 -> 2 */
 	    if (reverse)
 	      copy_reverse_string1_to_2(d, (unsigned char *)ps->str, ps->len);
-	    else convert_1_to_2((p_wchar2*)d,(p_wchar1*)ps->str,ps->len);
+	    else convert_1_to_2((p_wchar2*)d, STR1(ps), ps->len);
 	    break;
-	 case 02: /* 0 -> 2 */
+	 case 002: /* 0 -> 2 */
 	    if (reverse)
 	      copy_reverse_string0_to_2(d, (unsigned char *)ps->str, ps->len);
-	    else convert_0_to_2((p_wchar2*)d,ps->str,ps->len);
+	    else convert_0_to_2((p_wchar2*)d, STR0(ps), ps->len);
 	    break;
-	 case 11: /* 1 -> 1 */
+	 case 011: /* 1 -> 1 */
 	    if (reverse)
 	      copy_reverse_string1(d, (unsigned char *)ps->str,ps->len);
 	    else MEMCPY(d,ps->str,ps->len*2);
 	    break;
-	 case 01: /* 0 -> 1 */
+	 case 001: /* 0 -> 1 */
 	    if (reverse)
 	      copy_reverse_string0_to_1(d, (unsigned char *)ps->str,ps->len);
-	    else convert_0_to_1((p_wchar1*)d,ps->str,ps->len);
+	    else convert_0_to_1((p_wchar1*)d, STR0(ps), ps->len);
 	    break;
-	 case 00:
+	 case 000:
 	    MEMCPY(d,ps->str,ps->len);
 	    break;
 	 default:
@@ -829,7 +828,7 @@ static void memory_index(INT32 args)
    else
    {
       if (THIS->size==0)
-	 push_text("");
+	 push_empty_string();
       else
       {
 	 INT_TYPE pos1,pos2;
@@ -841,7 +840,7 @@ static void memory_index(INT32 args)
 	 else rpos2=(size_t)pos2;
 
 	 if (rpos2<rpos1)
-	    push_text("");
+	    push_empty_string();
 	 else
 	    push_string(make_shared_binary_string((char *)THIS->p+rpos1,
 						  rpos2-rpos1+1));
@@ -863,7 +862,7 @@ static void memory_index_write(INT32 args)
    if (args==2)
    {
       INT_TYPE pos,ch;
-      size_t rpos;
+      size_t rpos = 0;
       get_all_args("Memory.`[]=",args,"%i%i",&pos,&ch);
       if (pos<0) 
 	 if ((off_t)-pos>=DO_NOT_WARN((off_t)THIS->size))
@@ -884,29 +883,22 @@ static void memory_index_write(INT32 args)
    }
    else
    {
-      if (THIS->size==0)
-	 push_text("");
-      else
-      {
-	 INT_TYPE pos1,pos2;
-	 size_t rpos1,rpos2;
-	 struct pike_string *ps;
+     INT_TYPE pos1, pos2;
+     struct pike_string *ps;
 
-	 get_all_args("Memory.`[]=",args,"%i%i%S",&pos1,&pos2,&ps);
-	 if (pos1<0) rpos1=0; else rpos1=(size_t)pos1;
-	 if ((size_t)pos2>=THIS->size) rpos2=THIS->size-1; 
-	 else rpos2=(size_t)pos2;
+     get_all_args("Memory.`[]=", args, "%i%i%S", &pos1, &pos2, &ps);
 
-	 if (pos2<pos1 ||
-	     ps->len != pos2-pos1+1)
-	    Pike_error("Memory.`[]=: source and destination "
-		       "not equally long (%ld v/s %ld; can't resize memory)\n",
-		       DO_NOT_WARN((long)ps->len),(long)pos2-(long)pos1);
-	 else
-	    MEMCPY(THIS->p+pos1,ps->str,ps->len);
+     if (pos1 < 0) pos1 = 0;
+     if (pos2 < 0) pos2 = 0;
 
-	 ref_push_string(ps);
-      }
+     if ((pos2<pos1) || (ps->len != pos2-pos1+1))
+       Pike_error("Memory.`[]=: source and destination "
+		  "not equally long (%ld v/s %ld; can't resize memory)\n",
+		  DO_NOT_WARN((long)ps->len), (long)pos2-(long)pos1);
+     else
+       MEMCPY(THIS->p+pos1, ps->str, ps->len);
+
+     ref_push_string(ps);
    }
    stack_pop_n_elems_keep_top(args);
 }
