@@ -2,7 +2,7 @@
 || This file is part of Pike. For copyright information see COPYRIGHT.
 || Pike is distributed under GPL, LGPL and MPL. See the file COPYING
 || for more information.
-|| $Id: passwords.c,v 1.44 2004/09/26 15:31:15 marcus Exp $
+|| $Id: passwords.c,v 1.47 2005/05/06 00:42:35 nilsson Exp $
 */
 
 /*
@@ -26,9 +26,6 @@
 
 #include "system_machine.h"
 #include "system.h"
-
-RCSID("$Id: passwords.c,v 1.44 2004/09/26 15:31:15 marcus Exp $");
-
 #include "module_support.h"
 #include "interpret.h"
 #include "stralloc.h"
@@ -109,7 +106,11 @@ struct group *getgrnam(char *name)
 #endif
 
 
-#define SAFE_PUSH_TEXT(X) do { char *text_ = (X); if(text_) push_text(text_); else push_constant_text(""); }while(0);
+#define SAFE_PUSH_TEXT(X) do {						\
+    char *text_ = (X);							\
+    if(text_) push_text(text_);						\
+    else push_empty_string();						\
+  } while(0)
 
 /*
  * Functions
