@@ -2,7 +2,7 @@
 || This file is part of Pike. For copyright information see COPYRIGHT.
 || Pike is distributed under GPL, LGPL and MPL. See the file COPYING
 || for more information.
-|| $Id: multiset.h,v 1.37 2004/09/28 23:58:35 mast Exp $
+|| $Id: multiset.h,v 1.48 2008/07/22 20:03:20 mast Exp $
 */
 
 #ifndef MULTISET_H
@@ -20,6 +20,9 @@
 #include "rbtree.h"
 #include "block_alloc_h.h"
 
+/* Keep this defined so that code can test which multiset API is in use. */
+#define PIKE_NEW_MULTISETS
+
 /* Note: Don't access the ind svalue (or at least not its type field)
  * in the following directly, since the rbtree flags overlay that. Use
  * assign_multiset_index_no_free() or similar instead. */
@@ -83,8 +86,8 @@ struct multiset
  *    nodes. Note that deleted nodes might still be on the free list
  *    even when there are no node references (see below).
  *
- * o  multiset_data.cmp_less.type is T_INT when the internal set order
- *    is used.
+ * o  multiset_data.cmp_less is the integer zero when the internal set
+ *    order is used. It must not have any other integer value.
  *
  * o  multset_data.refs counts the number of "independent" references
  *    to the data block. When it's greater than one, the data block
@@ -147,8 +150,6 @@ struct multiset
 extern struct multiset *first_multiset;
 extern struct multiset *gc_internal_multiset;
 
-PMOD_EXPORT extern struct svalue svalue_int_one;
-
 PMOD_EXPORT void multiset_clear_node_refs (struct multiset *l);
 
 #ifdef PIKE_DEBUG
@@ -195,7 +196,9 @@ union msnode *low_multiset_find_eq (struct multiset *l, struct svalue *key);
    &(VAR))
 
 #define low_get_multiset_value(MSD, NODE)				\
-  ((MSD)->flags & MULTISET_INDVAL ? &(NODE)->iv.val : &svalue_int_one)
+  ((MSD)->flags & MULTISET_INDVAL ? &(NODE)->iv.val :			\
+   /* Caller better not try to change this. */				\
+   (struct svalue *) &svalue_int_one)
 #define low_set_multiset_value(MSD, NODE, VAL) do {			\
     if ((MSD)->flags & MULTISET_INDVAL)					\
       assign_svalue (&(NODE)->iv.val, VAL);				\
@@ -221,7 +224,13 @@ PMOD_EXPORT INT32 multiset_sizeof (struct multiset *l);
  * check whether or not the multiset has no elements at all. */
 #define multiset_is_empty(L) (!(L)->msd->root)
 
-PMOD_PROTO void really_free_multiset (struct multiset *l);
+#if defined(USE_DLL) && defined(DYNAMIC_MODULE)
+/* Use the function in modules so we don't have to export the block
+ * alloc stuff. */
+#define free_multiset(M) do_free_multiset (M)
+#else
+
+void really_free_multiset (struct multiset *l);
 
 #define free_multiset(L) do {						\
     struct multiset *_ms_ = (L);					\
@@ -233,9 +242,11 @@ PMOD_PROTO void really_free_multiset (struct multiset *l);
     if (!sub_ref (_ms_)) really_free_multiset (_ms_);			\
   } while (0)
 
+#endif	/* !(USE_DLL && DYNAMIC_MODULE) */
+
 #ifdef PIKE_DEBUG
 
-union msnode *debug_check_msnode (
+PMOD_EXPORT union msnode *debug_check_msnode (
   struct multiset *l, ptrdiff_t nodepos, int allow_deleted,
   char *file, int line);
 #define check_msnode(L, NODEPOS, ALLOW_DELETED)				\
@@ -250,7 +261,7 @@ union msnode *debug_check_msnode (
 
 #endif
 
-BLOCK_ALLOC_FILL_PAGES (multiset, 2)
+BLOCK_ALLOC_FILL_PAGES (multiset, 2);
 
 /* See rbtree.h for a description of the operations.
  *
@@ -343,7 +354,9 @@ PMOD_EXPORT int msnode_is_deleted (struct multiset *l, ptrdiff_t nodepos);
 
 #define get_multiset_value(L, NODEPOS)					\
   ((L)->msd->flags & MULTISET_INDVAL ?					\
-   &access_msnode ((L), (NODEPOS))->iv.val : &svalue_int_one)
+   &access_msnode ((L), (NODEPOS))->iv.val :				\
+   /* Caller better not try to change this. */				\
+   (struct svalue *) &svalue_int_one)
 #define set_multiset_value(L, NODEPOS, VAL) do {			\
     if ((L)->msd->flags & MULTISET_INDVAL)				\
       assign_svalue (&access_msnode ((L), (NODEPOS))->iv.val, VAL);	\
@@ -358,11 +371,15 @@ PMOD_EXPORT int msnode_is_deleted (struct multiset *l, ptrdiff_t nodepos);
 #define push_multiset_value(L, NODEPOS)					\
   push_svalue (get_multiset_value (L, NODEPOS))
 
-PMOD_EXPORT struct multiset *allocate_multiset (int allocsize,
-						int flags,
-						struct svalue *cmp_less);
+#define allocate_multiset(allocsize, flags, cmp_less)			\
+  dmalloc_touch (struct multiset *,					\
+		 real_allocate_multiset ((allocsize), (flags), (cmp_less)))
+
+PMOD_EXPORT struct multiset *real_allocate_multiset (int allocsize,
+						     int flags,
+						     struct svalue *cmp_less);
 PMOD_EXPORT void do_free_multiset (struct multiset *l);
-void multiset_fix_type_field (struct multiset *l);
+PMOD_EXPORT void multiset_fix_type_field (struct multiset *l);
 PMOD_EXPORT void multiset_set_flags (struct multiset *l, int flags);
 PMOD_EXPORT void multiset_set_cmp_less (struct multiset *l,
 					struct svalue *cmp_less);
@@ -414,9 +431,10 @@ int multiset_is_constant (struct multiset *l, struct processing *p);
 node *make_node_from_multiset (struct multiset *l);
 PMOD_EXPORT void f_aggregate_multiset (int args);
 struct multiset *copy_multiset_recursively (struct multiset *l,
-					    struct processing *p);
+					    struct mapping *p);
 PMOD_EXPORT ptrdiff_t multiset_get_nth (struct multiset *l, size_t n);
 
+PMOD_EXPORT void visit_multiset (struct multiset *l, int action);
 unsigned gc_touch_all_multisets (void);
 void gc_check_all_multisets (void);
 void gc_mark_multiset_as_referenced (struct multiset *l);
@@ -425,6 +443,10 @@ void gc_zap_ext_weak_refs_in_multisets (void);
 void real_gc_cycle_check_multiset (struct multiset *l, int weak);
 void gc_cycle_check_all_multisets (void);
 size_t gc_free_all_unreferenced_multisets (void);
+
+#define visit_multiset_ref(L, REF_TYPE)				\
+  visit_ref (pass_multiset (L), (REF_TYPE),			\
+	     (visit_thing_fn *) &visit_multiset, NULL)
 #define gc_cycle_check_multiset(X, WEAK) \
   gc_cycle_enqueue ((gc_cycle_check_cb *) real_gc_cycle_check_multiset, (X), (WEAK))
 
@@ -434,7 +456,7 @@ void check_all_multisets (int safe);
 void debug_dump_multiset (struct multiset *l);
 #endif
 
-void count_memory_in_multisets (INT32 *num, INT32 *size);
+void count_memory_in_multisets (size_t *num, size_t *size);
 void init_multiset (void);
 void exit_multiset (void);
 void test_multiset (void);
