@@ -2,17 +2,16 @@
 || This file is part of Pike. For copyright information see COPYRIGHT.
 || Pike is distributed under GPL, LGPL and MPL. See the file COPYING
 || for more information.
-|| $Id: peep.c,v 1.100 2006/09/05 12:10:15 grubba Exp $
+|| $Id: peep.c,v 1.112 2007/04/19 11:39:40 mast Exp $
 */
 
 #include "global.h"
-#include "language.h"
 #include "stralloc.h"
 #include "dynamic_buffer.h"
 #include "program.h"
 #include "las.h"
 #include "docode.h"
-#include "main.h"
+#include "pike_embed.h"
 #include "pike_error.h"
 #include "lex.h"
 #include "pike_memory.h"
@@ -26,12 +25,6 @@
 #include "interpret.h"
 #include "pikecode.h"
 
-RCSID("$Id: peep.c,v 1.100 2006/09/05 12:10:15 grubba Exp $");
-
-static void asm_opt(void);
-
-dynamic_buffer instrbuf;
-
 #ifdef PIKE_DEBUG
 static int hasarg(int opcode)
 {
@@ -58,10 +51,17 @@ static void dump_instr(p_instr *p)
 #endif
 
 
+static int asm_opt(void);
+
+/* Output buffer. The optimization eye is at the end of the buffer. */
+dynamic_buffer instrbuf;
+long num_instrs = 0;
+
 
 void init_bytecode(void)
 {
   initialize_buf(&instrbuf);
+  num_instrs = 0;
 }
 
 void exit_bytecode(void)
@@ -72,11 +72,29 @@ void exit_bytecode(void)
   c=(p_instr *)instrbuf.s.str;
   length=instrbuf.s.len / sizeof(p_instr);
 
-  for(e=0;e<length;e++) free_string(c[e].file);
+  for(e=0;e<length;e++) {
+    free_string(dmalloc_touch_named(struct pike_string *, c[e].file,
+				    "exit_bytecode"));
+#ifdef PIKE_DEBUG
+    c[e].file = (void *)(ptrdiff_t)~0;
+#endif
+  }
   
   toss_buffer(&instrbuf);
 }
 
+/* insert_opcode{,0,1,2}() append an opcode to the instrbuf. */
+
+ptrdiff_t insert_opcode(p_instr *opcode)
+{
+  /* Note: Steals references from opcode. */
+  p_instr *p = (p_instr *)low_make_buf_space(sizeof(p_instr), &instrbuf);
+  if (!p) Pike_fatal("Out of memory in peep.\n");
+  *p = *opcode;
+  num_instrs++;
+  return p - (p_instr *)instrbuf.s.str;
+}
+
 ptrdiff_t insert_opcode2(unsigned int f,
 			 INT32 b,
 			 INT32 c,
@@ -100,7 +118,9 @@ ptrdiff_t insert_opcode2(unsigned int f,
 
   p->opcode=f;
   p->line=current_line;
-  copy_shared_string(p->file, current_file);
+  copy_shared_string(p->file, dmalloc_touch_named(struct pike_string *,
+						  current_file,
+						  "insert_opcode"));
   p->arg=b;
   p->arg2=c;
 
@@ -151,10 +171,9 @@ INT32 assemble(int store_linenumbers)
 {
   INT32 entry_point;
   INT32 max_label=-1,tmp;
-  INT32 *labels, *jumps, *uses;
+  INT32 *labels, *jumps, *uses, *aliases;
   ptrdiff_t e, length;
   p_instr *c;
-  int reoptimize=!(debug_options & NO_PEEP_OPTIMIZING);
 #ifdef PIKE_PORTABLE_BYTECODE
   struct pike_string *tripples = NULL;
 #endif /* PIKE_PORTABLE_BYTECODE */
@@ -163,6 +182,8 @@ INT32 assemble(int store_linenumbers)
   int synch_depth = 0;
   size_t fun_start = Pike_compiler->new_program->num_program;
 #endif
+  int relabel;
+  int reoptimize = relabel = !(debug_options & NO_PEEP_OPTIMIZING);
 
   c=(p_instr *)instrbuf.s.str;
   length=instrbuf.s.len / sizeof(p_instr);
@@ -195,7 +216,8 @@ INT32 assemble(int store_linenumbers)
     /* Count the number of F_FILENAME/F_LINE pseudo-ops we need to add. */
     for (e=0; e < length; e++) {
       if (c[e].file != previous_file) {
-	previous_file = c[e].file;
+	previous_file = dmalloc_touch_named(struct pike_string *,
+					    c[e].file, "prev_file");
 	num_linedirectives++;
       }
       if (c[e].line != previous_line) {
@@ -219,10 +241,14 @@ INT32 assemble(int store_linenumbers)
     for (e = 0; e < length; e++) {
       if (c[e].file != previous_file) {
 	current_tripple[0] = F_FILENAME;
-	current_tripple[1] = store_prog_string(c[e].file);
+	current_tripple[1] =
+	  store_prog_string(dmalloc_touch_named(struct pike_string *,
+						c[e].file,
+						"store_prog_string"));
 	current_tripple[2] = 0;
 	current_tripple += 3;
-	previous_file = c[e].file;
+	previous_file = dmalloc_touch_named(struct pike_string *,
+					    c[e].file, "prev_file");
       }
       if (c[e].line != previous_line) {
 	current_tripple[0] = F_LINE;
@@ -288,16 +314,23 @@ INT32 assemble(int store_linenumbers)
   }
 #endif /* PIKE_DEBUG */
 
-  labels=(INT32 *)xalloc(sizeof(INT32) * 3 * (max_label+2));
+#ifndef INS_ENTRY
+  /* Replace F_ENTRY with F_NOP if we have no entry prologue. */
+  for (c = (p_instr *) instrbuf.s.str, e = 0; e < length; e++, c++)
+    if (c->opcode == F_ENTRY) c->opcode = F_NOP;
+#endif
+
+  labels=(INT32 *)xalloc(sizeof(INT32) * 4 * (max_label+2));
   jumps = labels + max_label + 2;
   uses = jumps + max_label + 2;
+  aliases = uses + max_label + 2;
 
-  while(reoptimize)
+  while(relabel)
   {
-    reoptimize=0;
+    /* First do the relabel pass. */
     for(e=0;e<=max_label;e++)
     {
-      labels[e]=jumps[e]=-1;
+      labels[e]=jumps[e]= aliases[e] = -1;
       uses[e]=0;
     }
     
@@ -305,13 +338,25 @@ INT32 assemble(int store_linenumbers)
     length=instrbuf.s.len / sizeof(p_instr);
     for(e=0;e<length;e++)
       if(c[e].opcode == F_LABEL && c[e].arg>=0) {
-	labels[c[e].arg]=DO_NOT_WARN((INT32)e);
+	INT32 l = c[e].arg;
+	labels[l]=DO_NOT_WARN((INT32)e);
+	while (e+1 < length &&
+	       c[e+1].opcode == F_LABEL && c[e+1].arg >= 0) {
+	  /* aliases is used to compact several labels at the same
+	   * position to one label. That's necessary for some peep
+	   * optimizations to work well. */
+	  e++;
+	  labels[c[e].arg] = DO_NOT_WARN((INT32)e);
+	  aliases[c[e].arg] = l;
+	}
       }
     
     for(e=0;e<length;e++)
     {
       if(instrs[c[e].opcode-F_OFFSET].flags & I_POINTER)
       {
+	if (aliases[c[e].arg] >= 0) c[e].arg = aliases[c[e].arg];
+
 	while(1)
 	{
 	  int tmp;
@@ -353,9 +398,15 @@ INT32 assemble(int store_linenumbers)
 	    case TWOO(F_BRANCH, F_RETURN_0):
 	    case TWOO(F_BRANCH, F_RETURN_1):
 	    case TWOO(F_BRANCH, F_RETURN_LOCAL):
-	      if(c[e].file) free_string(c[e].file);
+	      if(c[e].file) {
+		free_string(dmalloc_touch_named(struct pike_string *,
+						c[e].file, "branch_opt 1"));
+	      }
 	      c[e]=c[tmp];
-	      if(c[e].file) add_ref(c[e].file);
+	      if(c[e].file) {
+		add_ref(dmalloc_touch_named(struct pike_string *,
+					    c[e].file, "branch_opt 2"));
+	      }
 	      goto pointer_opcode_done;
 	  }
 	  break;
@@ -374,19 +425,37 @@ INT32 assemble(int store_linenumbers)
       if(!uses[e] && labels[e]>=0)
       {
 	c[labels[e]].opcode=F_NOP;
-	reoptimize++;
+	reoptimize = 1;
       }
     }
-    if(!reoptimize) break;
-    
-    asm_opt();
+
+    if(!reoptimize) {
+#ifdef PIKE_DEBUG
+      if (a_flag > 3) fprintf (stderr, "Optimizer done after relabel.\n");
+#endif
+      break;
+    }
+
+    /* Then do the optimize pass. */
+
+    relabel = asm_opt();
+
+    reoptimize = 0;
+
+    if (!relabel) {
+#ifdef PIKE_DEBUG
+      if (a_flag > 3) fprintf (stderr, "Optimizer done after asm_opt.\n");
+#endif
+      break;
+    }
+
 #if 1
 #ifdef PIKE_DEBUG
     if (a_flag > 3)
       fprintf(stderr, "Rerunning optimizer.\n");
 #endif
 #else /* !1 */
-    reoptimize=0;
+    relabel = 0;
 #endif /* 1 */
   }
 
@@ -407,7 +476,7 @@ INT32 assemble(int store_linenumbers)
     ins_data(store_prog_string(tripples));
     free_string(tripples);
   } else {
-    ins_data(NULL);
+    ins_data(0);
   }
 #endif /* PIKE_PORTABLE_BYTECODE */
 
@@ -437,7 +506,9 @@ INT32 assemble(int store_linenumbers)
 #endif
 
     if(store_linenumbers) {
-      store_linenumber(c->line, c->file);
+      store_linenumber(c->line, dmalloc_touch_named(struct pike_string *,
+						    c->file,
+						    "store_line"));
 #ifdef PIKE_DEBUG
       if (c->opcode < F_MAX_OPCODE)
 	ADD_COMPILED(c->opcode);
@@ -465,11 +536,19 @@ INT32 assemble(int store_linenumbers)
     case F_ENTRY:
 #ifdef INS_ENTRY
       INS_ENTRY();
-#endif /* INS_ENTRY */
+#else
+      Pike_fatal ("F_ENTRY is supposed to be gone here.\n");
+#endif
       break;
 
     case F_LABEL:
-      if(c->arg == -1) break;
+      if(c->arg == -1) {
+#ifdef PIKE_DEBUG
+	if (!(debug_options & NO_PEEP_OPTIMIZING))
+	  Pike_fatal ("Failed to optimize away an unused label.\n");
+#endif
+	break;
+      }
 #ifdef PIKE_DEBUG
       if(c->arg > max_label || c->arg < 0)
 	Pike_fatal("max_label calculation failed!\n");
@@ -703,13 +782,31 @@ INT32 assemble(int store_linenumbers)
 
 /**** Peephole optimizer ****/
 
+static void do_optimization(int topop, int topush, ...);
+static INLINE int opcode(int offset);
+static INLINE int argument(int offset);
+static INLINE int argument2(int offset);
+
+#include "peep_engine.c"
+
+#ifndef PEEP_STACK_SIZE
+#define PEEP_STACK_SIZE 256
+#endif
+
+/* Stack holding pending instructions.
+ * Note that instructions must be pushed in reverse order.
+ */
+static long stack_depth = 0;
+static p_instr instrstack[PEEP_STACK_SIZE];
+
 int remove_clear_locals=0x7fffffff;
-static int fifo_len;
 static ptrdiff_t eye, len;
 static p_instr *instructions;
 
-static INLINE ptrdiff_t insopt2(int f, INT32 a, INT32 b,
-				int cl, struct pike_string *cf)
+/* insopt{0,1,2} push an instruction on instrstack. */
+
+static INLINE p_instr *insopt2(int f, INT32 a, INT32 b,
+			       int cl, struct pike_string *cf)
 {
   p_instr *p;
 
@@ -718,29 +815,24 @@ static INLINE ptrdiff_t insopt2(int f, INT32 a, INT32 b,
     Pike_fatal("hasarg2(%d /*%s */) is wrong!\n",f,get_f_name(f));
 #endif
 
-  p=(p_instr *)low_make_buf_space(sizeof(p_instr), &instrbuf);
-
-  if(fifo_len)
-  {
-    MEMMOVE(p-fifo_len+1,p-fifo_len,fifo_len*sizeof(p_instr));
-    p-=fifo_len;
-  }
+  p = instrstack + stack_depth++;
 
 #ifdef PIKE_DEBUG
-  if(!instrbuf.s.len)
-    Pike_fatal("Low make buf space failed!!!!!!\n");
+  if(stack_depth > PEEP_STACK_SIZE)
+    Pike_fatal("Instructions stacked too high!!!!!!\n");
 #endif
 
   p->opcode=f;
   p->line=cl;
-  copy_shared_string(p->file, cf);
+  copy_shared_string(p->file, dmalloc_touch_named(struct pike_string *,
+						  cf, "insopt2"));
   p->arg=a;
   p->arg2=b;
 
-  return p - (p_instr *)instrbuf.s.str;
+  return p;
 }
 
-static INLINE ptrdiff_t insopt1(int f, INT32 a, int cl, struct pike_string *cf)
+static INLINE p_instr *insopt1(int f, INT32 a, int cl, struct pike_string *cf)
 {
 #ifdef PIKE_DEBUG
   if(!hasarg(f) && a)
@@ -750,7 +842,7 @@ static INLINE ptrdiff_t insopt1(int f, INT32 a, int cl, struct pike_string *cf)
   return insopt2(f,a,0,cl, cf);
 }
 
-static INLINE ptrdiff_t insopt0(int f, int cl, struct pike_string *cf)
+static INLINE p_instr *insopt0(int f, int cl, struct pike_string *cf)
 {
 #ifdef PIKE_DEBUG
   if(hasarg(f))
@@ -759,14 +851,13 @@ static INLINE ptrdiff_t insopt0(int f, int cl, struct pike_string *cf)
   return insopt2(f,0,0,cl, cf);
 }
 
+#ifdef PIKE_DEBUG
 static void debug(void)
 {
-  if(fifo_len > (long)instrbuf.s.len / (long)sizeof(p_instr))
-    fifo_len=(long)instrbuf.s.len / (long)sizeof(p_instr);
-#ifdef PIKE_DEBUG
-  if(eye < 0)
-    Pike_fatal("Popped beyond start of code.\n");
-
+  if (num_instrs != (long)instrbuf.s.len / (long)sizeof(p_instr)) {
+    Pike_fatal("PEEP: instrbuf lost count (%d != %d)\n",
+	       num_instrs, (long)instrbuf.s.len / (long)sizeof(p_instr));
+  }
   if(instrbuf.s.len)
   {
     p_instr *p;
@@ -774,29 +865,17 @@ static void debug(void)
     if(!p[-1].file)
       Pike_fatal("No file name on last instruction!\n");
   }
-#endif
 }
+#else
+#define debug()
+#endif
 
 
+/* Offset from the end of instrbuf backwards. */
 static INLINE p_instr *instr(int offset)
 {
-  p_instr *p;
-
-  debug();
-
-  if(offset < fifo_len)
-  {
-    p=(p_instr *)low_make_buf_space(0, &instrbuf);
-    p-=fifo_len;
-    p+=offset;
-    if(((char *)p)<instrbuf.s.str)  return 0;
-    return p;
-  }else{
-    offset-=fifo_len;
-    offset+=eye;
-    if(offset >= len) return 0;
-    return instructions+offset;
-  }
+  if (offset >= num_instrs) return NULL;
+  return ((p_instr *)low_make_buf_space(0, &instrbuf)) - (offset + 1);
 }
 
 static INLINE int opcode(int offset)
@@ -823,101 +902,82 @@ static INLINE int argument2(int offset)
   return -1;
 }
 
-static void advance(void)
+static int advance(void)
 {
-  if(fifo_len)
+  p_instr *p;
+  if(stack_depth)
   {
-    fifo_len--;
+    p = instrstack + --stack_depth;
   }else{
-    p_instr *p;
-    if((p=instr(0)))
-      insert_opcode2(p->opcode, p->arg, p->arg2, p->line, p->file);
+    if (eye >= len) return 0;
+    p = instructions + eye;
     eye++;
   }
+  insert_opcode(p);
+  debug_malloc_touch_named(p->file, "advance");
   debug();
+  return 1;
 }
 
 static void pop_n_opcodes(int n)
 {
-  int e,d;
-  if(fifo_len)
-  {
-    p_instr *p;
+  int e;
+  p_instr *p;
 
-    d=n;
-    if(d>fifo_len) d=fifo_len;
 #ifdef PIKE_DEBUG
-    if((long)d > (long)instrbuf.s.len / (long)sizeof(p_instr))
-      Pike_fatal("Popping out of instructions.\n");
+  if (n > num_instrs)
+    Pike_fatal("Popping out of instructions.\n");
 #endif
 
-    /* FIXME: It looks like the fifo could be optimized.
-     *	/grubba 2000-11-21 (in Versailles)
-     */
-
-    p=(p_instr *)low_make_buf_space(0, &instrbuf);
-    p-=fifo_len;
-    for(e=0;e<d;e++) free_string(p[e].file);
-    fifo_len-=d;
-    if(fifo_len) MEMMOVE(p,p+d,fifo_len*sizeof(p_instr));
-    n-=d;
-    low_make_buf_space(-((INT32)sizeof(p_instr))*d, &instrbuf);
+  p = ((p_instr *)low_make_buf_space(0, &instrbuf)) - n;
+  for (e = 0; e < n; e++) {
+    free_string(dmalloc_touch_named(struct pike_string *, p[e].file,
+				    "pop_n_opcodes"));
   }
-  eye+=n;
+  num_instrs -= n;
+  low_make_buf_space(-((INT32)sizeof(p_instr))*n, &instrbuf);
 }
 
-#define DO_OPTIMIZATION_PREQUEL(topop) do {	\
-    struct pike_string *cf; 			\
-    INT32 cl=instr(0)->line;			\
-  						\
-    DO_IF_DEBUG(				\
-      if(a_flag>5)				\
-      {						\
-  	int e;					\
-  	fprintf(stderr,"PEEP at %d:",cl);	\
-  	for(e=0;e<topop;e++)			\
-  	{					\
-  	  fprintf(stderr," ");			\
-  	  dump_instr(instr(e));			\
-  	}					\
-  	fprintf(stderr," => ");			\
-      }						\
-    )						\
-    						\
-    copy_shared_string(cf,instr(0)->file);	\
-    pop_n_opcodes(topop)
 
-#define DO_OPTIMIZATION_POSTQUEL(q)	\
-    fifo_len+=q;			\
-    free_string(cf);			\
-    debug();				\
-  					\
-    DO_IF_DEBUG(			\
-      if(a_flag>5)			\
-      {					\
-  	int e;				\
-  	for(e=0;e<q;e++)		\
-  	{				\
-  	  fprintf(stderr," ");		\
-  	  dump_instr(instr(e));		\
-  	}				\
-  	fprintf(stderr,"\n");		\
-      }					\
-    )					\
-  					\
-    fifo_len += q + 3;			\
-  }  while(0)
-
-
-static void do_optimization(int topop, ...)
+/* NOTE: Called with opcodes in reverse order! */
+static void do_optimization(int topop, int topush, ...)
 {
   va_list arglist;
   int q=0;
   int oplen;
+  struct pike_string *cf;
+  INT32 cl=instr(0)->line;
 
-  DO_OPTIMIZATION_PREQUEL(topop);
+#ifdef PIKE_DEBUG
+  if(a_flag>5)
+  {
+    int e;
+    fprintf(stderr,"PEEP at %d:",cl);
+    for(e = topop; e--;)
+    {
+      fprintf(stderr," ");
+      dump_instr(instr(e));
+    }
+    fprintf(stderr," => ");
+  }
+#endif
 
-  va_start(arglist, topop);
+  if (stack_depth + topush > PEEP_STACK_SIZE) {
+    /* No place left on stack. Ignore the optimization. */
+#ifdef PIKE_DEBUG
+    if (a_flag) {
+      fprintf(stderr, "PEEP stack full.\n");
+    }
+#endif
+    return;
+  }
+
+  copy_shared_string(cf,dmalloc_touch_named(struct pike_string *,
+					    instr(0)->file,
+					    "do_optimization"));
+  pop_n_opcodes(topop);
+
+  va_start(arglist, topush);
   
   while((oplen = va_arg(arglist, int)))
   {
@@ -958,13 +1018,39 @@ static void do_optimization(int topop, ...)
 
   va_end(arglist);
 
-  DO_OPTIMIZATION_POSTQUEL(q);
+  /*fifo_len+=q;*/
+  free_string(dmalloc_touch_named(struct pike_string *, cf,
+				  "do_optimization"));
+  debug();
+
+#ifdef PIKE_DEBUG
+  if(a_flag>5)
+  {
+    p_instr *p = instrstack + stack_depth;
+    int e;
+    for(e=0;e<q;e++)
+    {
+      fprintf(stderr," ");
+      dump_instr(p-(e+1));
+    }
+    fprintf(stderr,"\n");
+  }
+  if (q != topush) {
+    Pike_fatal("PEEP: Lost track of instructions to push (%d != %d)\n",
+	       q, topush);
+  }
+#endif
+
+  /* Note: The 5 below is the longest
+   *       match prefix in the ruleset
+   */
+  /*fifo_len += q + 5;*/
 }
 
-#include "peep_engine.c"
-
-static void asm_opt(void)
+static int asm_opt(void)
 {
+  int relabel = 0;
+
 #ifdef PIKE_DEBUG
   if(a_flag > 3)
   {
@@ -990,31 +1076,28 @@ static void asm_opt(void)
   len=instrbuf.s.len/sizeof(p_instr);
   instructions=(p_instr *)instrbuf.s.str;
   instrbuf.s.str=0;
-  fifo_len=0;
   init_bytecode();
 
-  for(eye=0;eye<len || fifo_len;)
+  for(eye = 0; advance();)
   {
-
 #ifdef PIKE_DEBUG
     if(a_flag>6) {
       int e;
-      fprintf(stderr, "#%ld,%d:",
+      fprintf(stderr, "#%ld,%ld:",
               DO_NOT_WARN((long)eye),
-              fifo_len);
-      for(e=0;e<4;e++) {
+              stack_depth);
+      for(e = 4;e--;) {
         fprintf(stderr," ");
         dump_instr(instr(e));
       }
+      /* FIXME: Show instrstack too? */
       fprintf(stderr,"\n");
     }
 #endif
 
-    low_asm_opt();
-    advance();
+    relabel |= low_asm_opt();
   }
 
-  for(eye=0;eye<len;eye++) free_string(instructions[eye].file);
   free((char *)instructions);
 
 #ifdef PIKE_DEBUG
@@ -1038,4 +1121,6 @@ static void asm_opt(void)
     }
   }
 #endif
+
+  return relabel;
 }
