@@ -2,14 +2,12 @@
 || This file is part of Pike. For copyright information see COPYRIGHT.
 || Pike is distributed under GPL, LGPL and MPL. See the file COPYING
 || for more information.
-|| $Id: pike_error.h,v 1.34 2006/05/10 19:43:54 mast Exp $
+|| $Id: pike_error.h,v 1.54 2009/08/07 14:05:59 grubba Exp $
 */
 
 #ifndef PIKE_ERROR_H
 #define PIKE_ERROR_H
 
-#include <stdarg.h>
-
 #ifdef CONFIGURE_TEST
 
 #include <stdio.h>
@@ -26,7 +24,7 @@ static inline void Pike_fatal (const char *fmt, ...)
 
 #else  /* !CONFIGURE_TEST */
 
-#include "machine.h"
+#include "global.h"
 
 #ifdef HAVE_SETJMP_H
 #include <setjmp.h>
@@ -65,7 +63,8 @@ PMOD_EXPORT extern const char msg_fatal_error[];
 #define fatal Pike_fatal
 #endif
 
-PMOD_EXPORT DECLSPEC(noreturn) void debug_fatal(const char *fmt, ...) ATTRIBUTE((noreturn,format (printf, 1, 2)));
+PMOD_EXPORT DECLSPEC(noreturn) void debug_va_fatal(const char *fmt, va_list args) ATTRIBUTE((noreturn));
+PMOD_EXPORT DECLSPEC(noreturn) void debug_fatal(const char *fmt, ...) ATTRIBUTE((noreturn));
 
 
 #include "svalue.h"
@@ -78,6 +77,7 @@ typedef void (*error_call)(void *);
 struct pike_frame;
 #endif
 
+#define THROW_N_A 0
 #define THROW_ERROR 10
 #define THROW_THREAD_EXIT 20
 #define THROW_THREAD_KILLED 30
@@ -95,6 +95,7 @@ struct pike_frame;
 typedef struct ONERROR
 {
   struct ONERROR *previous;
+  struct pike_frame *frame_pointer;
   error_call func;
   void *arg;
 #ifdef PIKE_DEBUG
@@ -115,6 +116,7 @@ typedef struct JMP_BUF
 #ifdef PIKE_DEBUG
   int line;
   char *file;
+  int on_stack;
 #endif
 } JMP_BUF;
 
@@ -124,18 +126,19 @@ extern int throw_severity;
 #ifdef PIKE_DEBUG
 PMOD_EXPORT extern const char msg_unsetjmp_nosync_1[];
 PMOD_EXPORT extern const char msg_unsetjmp_nosync_2[];
-#define UNSETJMP(X) do{ \
-   check_recovery_context(); \
-   OED_FPRINTF((stderr, "unsetjmp(%p) %s:%d\n", \
-                &(X),  __FILE__, __LINE__)); \
-  if(Pike_interpreter.recoveries != &X) { \
-    if(Pike_interpreter.recoveries) \
-      Pike_fatal(msg_unsetjmp_nosync_1,Pike_interpreter.recoveries->file); \
-    else \
-      Pike_fatal(msg_unsetjmp_nosync_2); \
-    } \
-    Pike_interpreter.recoveries=X.previous; \
-   check_recovery_context(); \
+#define UNSETJMP(X) do{						\
+    check_recovery_context();					\
+    OED_FPRINTF((stderr, "unsetjmp(%p) %s:%d\n",		\
+		 &(X),  __FILE__, __LINE__));			\
+    if(Pike_interpreter.recoveries != &X) {			\
+      if(Pike_interpreter.recoveries)				\
+	Pike_fatal(msg_unsetjmp_nosync_1,			\
+		   Pike_interpreter.recoveries->file);		\
+      else							\
+	Pike_fatal("%s", msg_unsetjmp_nosync_2);		\
+    }								\
+    Pike_interpreter.recoveries=X.previous;			\
+    check_recovery_context();					\
   }while (0)
 
 #ifdef DMALLOC_LOCATION
@@ -144,54 +147,59 @@ PMOD_EXPORT extern const char msg_unsetjmp_nosync_2[];
 #define PERR_LOCATION() ( __FILE__ ":" DEFINETOSTR(__LINE__) )
 #endif
 
-#define DEBUG_LINE_ARGS ,char *location
-#define SETJMP(X) LOW_SETJMP((init_recovery(&X, 0, PERR_LOCATION())->recovery))
+#define DEBUG_INIT_REC_ARGS , int on_stack, const char *location
+#define SETJMP(X) LOW_SETJMP((init_recovery(&X, 0, 1, PERR_LOCATION())->recovery))
 #define SETJMP_SP(jmp, stack_pop_levels)				\
-  LOW_SETJMP((init_recovery(&jmp, stack_pop_levels, PERR_LOCATION())->recovery))
-#else
-#define DEBUG_LINE_ARGS 
+  LOW_SETJMP((init_recovery(&jmp, stack_pop_levels, 1, PERR_LOCATION())->recovery))
+#else  /* !PIKE_DEBUG */
+#define DEBUG_INIT_REC_ARGS
 #define SETJMP(X) LOW_SETJMP((init_recovery(&X, 0)->recovery))
 #define SETJMP_SP(jmp, stack_pop_levels)				\
   LOW_SETJMP((init_recovery(&jmp, stack_pop_levels)->recovery))
 #define UNSETJMP(X) Pike_interpreter.recoveries=X.previous
-#endif
+#endif	/* !PIKE_DEBUG */
 
 
 #ifdef PIKE_DEBUG
-#define SET_ONERROR(X,Y,Z) \
-  do{ \
-     check_recovery_context(); \
-     OED_FPRINTF((stderr, "SET_ONERROR(%p, %p, %p) %s:%d\n", \
-                  &(X), (Y), (void *)(Z), __FILE__, __LINE__)); \
-     X.func=(error_call)(Y); \
-     DO_IF_DMALLOC( if( X.func == free ) X.func=dmalloc_free;) \
-     X.arg=(void *)(Z); \
-     if(!Pike_interpreter.recoveries) break; \
-     X.previous=Pike_interpreter.recoveries->onerror; \
-     X.file = __FILE__; \
-     X.line = __LINE__; \
-     Pike_interpreter.recoveries->onerror=&X; \
+#define SET_ONERROR(X,Y,Z)					\
+  do{								\
+    check_recovery_context();					\
+    OED_FPRINTF((stderr, "SET_ONERROR(%p, %p, %p) %s:%d\n",	\
+		 &(X), (Y), (void *)(Z), __FILE__, __LINE__));	\
+    X.frame_pointer = Pike_interpreter.frame_pointer;		\
+    X.func=(error_call)(Y);					\
+    DO_IF_DMALLOC( if( X.func == free ) X.func=dmalloc_free);	\
+    X.arg=(void *)(Z);						\
+    if(!Pike_interpreter.recoveries) {				\
+      X.previous = NULL;					\
+      break;							\
+    }								\
+    X.previous=Pike_interpreter.recoveries->onerror;		\
+    X.file = __FILE__;						\
+    X.line = __LINE__;						\
+    Pike_interpreter.recoveries->onerror=&X;			\
   }while(0)
 
 PMOD_EXPORT extern const char msg_last_setjmp[];
 PMOD_EXPORT extern const char msg_unset_onerr_nosync_1[];
 PMOD_EXPORT extern const char msg_unset_onerr_nosync_2[];
-#define UNSET_ONERROR(X) do {\
-    check_recovery_context(); \
-    OED_FPRINTF((stderr, "UNSET_ONERROR(%p) %s:%d\n", \
-                 &(X), __FILE__, __LINE__)); \
-    if(!Pike_interpreter.recoveries) break; \
-    if(Pike_interpreter.recoveries->onerror != &(X)) { \
-      fprintf(stderr,msg_last_setjmp,Pike_interpreter.recoveries->file); \
-      if (Pike_interpreter.recoveries->onerror) { \
-	Pike_fatal(msg_unset_onerr_nosync_1,\
-	           Pike_interpreter.recoveries->onerror, &(X), \
+#define UNSET_ONERROR(X) do {					\
+    check_recovery_context();					\
+    OED_FPRINTF((stderr, "UNSET_ONERROR(%p) %s:%d\n",		\
+                 &(X), __FILE__, __LINE__));			\
+    if(!Pike_interpreter.recoveries) break;			\
+    if(Pike_interpreter.recoveries->onerror != &(X)) {		\
+      fprintf(stderr,msg_last_setjmp,				\
+	      Pike_interpreter.recoveries->file);		\
+      if (Pike_interpreter.recoveries->onerror) {		\
+	Pike_fatal(msg_unset_onerr_nosync_1,			\
+	           Pike_interpreter.recoveries->onerror, &(X),	\
                    Pike_interpreter.recoveries->onerror->file); \
-      } else { \
-        Pike_fatal(msg_unset_onerr_nosync_2); \
-      } \
-    } \
-    Pike_interpreter.recoveries->onerror=(X).previous; \
+      } else {							\
+	Pike_fatal("%s", msg_unset_onerr_nosync_2);		\
+      }								\
+    }								\
+    Pike_interpreter.recoveries->onerror=(X).previous;		\
   } while(0)
 
 PMOD_EXPORT extern const char msg_assert_onerr[];
@@ -208,7 +216,11 @@ PMOD_EXPORT extern const char msg_assert_onerr[];
   do{ \
      X.func=(error_call)(Y); \
      X.arg=(void *)(Z); \
-     if(!Pike_interpreter.recoveries) break; \
+     X.frame_pointer = Pike_interpreter.frame_pointer; \
+     if(!Pike_interpreter.recoveries) {		       \
+       X.previous = NULL;			       \
+       break;					       \
+     }						       \
      X.previous=Pike_interpreter.recoveries->onerror; \
      Pike_interpreter.recoveries->onerror=&X; \
   }while(0)
@@ -235,61 +247,62 @@ PMOD_EXPORT extern const char msg_assert_onerr[];
 /* Prototypes begin here */
 PMOD_EXPORT void check_recovery_context(void);
 PMOD_EXPORT void pike_gdb_breakpoint(INT32 args);
-PMOD_EXPORT JMP_BUF *init_recovery(JMP_BUF *r, size_t stack_pop_levels DEBUG_LINE_ARGS);
+PMOD_EXPORT JMP_BUF *init_recovery(JMP_BUF *r, size_t stack_pop_levels DEBUG_INIT_REC_ARGS);
 PMOD_EXPORT DECLSPEC(noreturn) void pike_throw(void) ATTRIBUTE((noreturn));
 PMOD_EXPORT void push_error(const char *description);
 PMOD_EXPORT DECLSPEC(noreturn) void low_error(const char *buf) ATTRIBUTE((noreturn));
-void DECLSPEC(noreturn) va_error(const char *fmt, va_list args) ATTRIBUTE((noreturn));
+PMOD_EXPORT void Pike_vsnprintf(char *str, size_t size, const char *fmt, va_list ap);
+PMOD_EXPORT void va_make_error (const char *fmt, va_list args);
+PMOD_EXPORT void DECLSPEC(noreturn) va_error(const char *fmt, va_list args) ATTRIBUTE((noreturn));
+PMOD_EXPORT void make_error (const char *fmt, ...);
+PMOD_EXPORT DECLSPEC(noreturn) void Pike_error(const char *fmt,...) ATTRIBUTE((noreturn));
 PMOD_EXPORT DECLSPEC(noreturn) void new_error(const char *name, const char *text, struct svalue *oldsp,
 	       INT32 args, const char *file, int line) ATTRIBUTE((noreturn));
 PMOD_EXPORT void exit_on_error(const void *msg);
 PMOD_EXPORT void fatal_on_error(const void *msg);
-PMOD_EXPORT DECLSPEC(noreturn) void Pike_error(const char *fmt,...) ATTRIBUTE((noreturn,format (printf, 1, 2)));
-PMOD_EXPORT DECLSPEC(noreturn) void debug_fatal(const char *fmt, ...) ATTRIBUTE((noreturn,format (printf, 1, 2)));
-void DECLSPEC(noreturn) generic_error_va(struct object *o,
-		      const char *func,
-		      struct svalue *base_sp,  int args,
-		      const char *fmt,
-		      va_list foo)
+PMOD_EXPORT DECLSPEC(noreturn) void debug_fatal(const char *fmt, ...) ATTRIBUTE((noreturn));
+PMOD_EXPORT DECLSPEC(noreturn) void generic_error_va(
+  struct object *o, const char *func, const struct svalue *base_sp, int args,
+  const char *fmt, va_list *fmt_args)
   ATTRIBUTE((noreturn));
 PMOD_EXPORT DECLSPEC(noreturn) void throw_error_object(
   struct object *o,
   const char *func,
   struct svalue *base_sp,  int args,
-  const char *desc, ...) ATTRIBUTE((noreturn,format (printf, 5, 6)));
+  const char *desc, ...) ATTRIBUTE((noreturn));
 PMOD_EXPORT void DECLSPEC(noreturn) generic_error(
   const char *func,
   struct svalue *base_sp,  int args,
-  const char *desc, ...) ATTRIBUTE((noreturn,format (printf, 4, 5)));
+  const char *desc, ...) ATTRIBUTE((noreturn));
 PMOD_EXPORT DECLSPEC(noreturn) void index_error(
   const char *func,
   struct svalue *base_sp,  int args,
   struct svalue *val,
   struct svalue *ind,
-  const char *desc, ...) ATTRIBUTE((noreturn,format (printf, 6, 7)));
+  const char *desc, ...) ATTRIBUTE((noreturn));
 PMOD_EXPORT DECLSPEC(noreturn) void bad_arg_error(
   const char *func,
   struct svalue *base_sp,  int args,
   int which_arg,
   const char *expected_type,
   struct svalue *got,
-  const char *desc, ...)  ATTRIBUTE((noreturn,format (printf, 7, 8)));
+  const char *desc, ...)  ATTRIBUTE((noreturn));
 PMOD_EXPORT void DECLSPEC(noreturn) math_error(
   const char *func,
   struct svalue *base_sp,  int args,
   struct svalue *number,
-  const char *desc, ...) ATTRIBUTE((noreturn,format (printf, 5, 6)));
+  const char *desc, ...) ATTRIBUTE((noreturn));
 PMOD_EXPORT void DECLSPEC(noreturn) resource_error(
   const char *func,
   struct svalue *base_sp,  int args,
   const char *resource_type,
   size_t howmuch,
-  const char *desc, ...) ATTRIBUTE((noreturn,format (printf, 6, 7)));
+  const char *desc, ...) ATTRIBUTE((noreturn));
 PMOD_EXPORT void DECLSPEC(noreturn) permission_error(
   const char *func,
   struct svalue *base_sp, int args,
   const char *permission_type,
-  const char *desc, ...) ATTRIBUTE((noreturn, format(printf, 5, 6)));
+  const char *desc, ...) ATTRIBUTE((noreturn));
 PMOD_EXPORT void wrong_number_of_args_error(const char *name, int args, int expected)
   ATTRIBUTE((noreturn));
 void init_error(void);
@@ -309,8 +322,8 @@ PMOD_EXPORT extern const char msg_bad_arg[];
    bad_arg_error(FUNC, Pike_sp-args, args, ARG, EXPECT, Pike_sp+ARG-1-args,\
 		 msg_bad_arg, ARG, FUNC, EXPECT)
 
-/* This one should be used when there's some other problem with the
- * argument than a bogus type. PROBLEM is a full sentence without a
+/* This one should be used when there's some problem with the argument
+ * other than a bogus type. PROBLEM is a full sentence without a
  * trailing newline. */
 PMOD_EXPORT extern const char msg_bad_arg_2[];
 #define SIMPLE_ARG_ERROR(FUNC, ARG, PROBLEM) \
@@ -326,6 +339,7 @@ PMOD_EXPORT extern const char msg_bad_arg_2[];
 PMOD_EXPORT extern const char msg_out_of_mem[];
 PMOD_EXPORT extern const char msg_out_of_mem_2[];
 
+#if 1
 static INLINE void DECLSPEC(noreturn) out_of_memory_error (
   const char *func,
   struct svalue *base_sp,  int args,
@@ -336,8 +350,17 @@ static INLINE void DECLSPEC(noreturn) out_of_memory_error (
   size_t amount)
 {
   resource_error (func, base_sp, args, "memory", amount,
-		  msg_out_of_mem_2, amount);
+		  amount ? msg_out_of_mem_2 : msg_out_of_mem, amount);
 }
+#else
+#define out_of_memory_error(FUNC, BASE_SP, ARGS, AMOUNT)	\
+  do {								\
+    size_t amount_ = (AMOUNT);					\
+    int args_ = (ARGS);						\
+    resource_error((FUNC), (BASE_SP), args_, "memory", amount_,	\
+		   msg_out_of_mem_2, amount_);			\
+  } while(0)
+#endif
 
 #define SIMPLE_OUT_OF_MEMORY_ERROR(FUNC, AMOUNT) \
    out_of_memory_error(FUNC, Pike_sp-args, args, AMOUNT)
@@ -383,7 +406,7 @@ PMOD_EXPORT extern const char msg_div_by_zero[];
                 __exception_rethrow = 1; \
             if(__is_exception) { \
 		free_svalue(&throw_value); \
-		throw_value.type = T_INT; \
+		mark_free_svalue (&throw_value); \
 	    } \
 	    else \
                 UNSETJMP(exception); \
