@@ -1,6 +1,6 @@
 /* Misc stuff for dealing with floats.
  *
- * $Id: pike_float.h,v 1.5 2004/03/21 20:14:45 nilsson Exp $
+ * $Id: pike_float.h,v 1.8 2005/07/15 18:16:52 grubba Exp $
  */
 
 #ifndef PIKE_FLOAT_H
@@ -35,7 +35,7 @@
 #define PIKE_ISNAN(X)	_isnan(X)
 #else /* !HAVE__ISNAN */
 /* Fallback function */
-static inline int pike_isnan(double x)
+static INLINE int pike_isnan(double x)
 {
   return ((x == 0.0) == (x < 0.0)) &&
     ((x == 0.0) == (x > 0.0));
@@ -44,10 +44,31 @@ static inline int pike_isnan(double x)
 #endif /* HAVE__ISNAN */
 #endif /* HAVE_ISNAN */
 
+/* isinf()...
+ */
+#ifdef HAVE_ISINF
+#define PIKE_ISINF(X)	isinf(X)
+#else /* HAVE_ISINF */
+#define PIKE_ISINF(X)	((X) && ((X)+(X) == (X)))
+#endif /* HAVE_ISINF */
+
 #ifdef HAVE_ISUNORDERED
 #define PIKE_ISUNORDERED(X,Y) isunordered(X,Y)
 #else
 #define PIKE_ISUNORDERED(X,Y) (PIKE_ISNAN(X)||PIKE_ISNAN(Y))
 #endif /* HAVE_ISUNORDERED */
 
+#ifndef FLOAT_IS_IEEE_BIG
+#ifndef FLOAT_IS_IEEE_LITTLE
+#define NEED_CUSTOM_IEEE
+#endif
+#endif
+#ifndef NEED_CUSTOM_IEEE
+#ifndef DOUBLE_IS_IEEE_BIG
+#ifndef DOUBLE_IS_IEEE_LITTLE
+#define NEED_CUSTOM_IEEE
+#endif
+#endif
+#endif
+
 #endif	/* !PIKE_FLOAT_H */
