@@ -2,7 +2,7 @@
 || This file is part of Pike. For copyright information see COPYRIGHT.
 || Pike is distributed under GPL, LGPL and MPL. See the file COPYING
 || for more information.
-|| $Id: port.h,v 1.55 2003/11/14 06:55:58 mast Exp $
+|| $Id: port.h,v 1.66 2009/05/28 13:10:48 grubba Exp $
 */
 
 #ifndef PORT_H
@@ -11,6 +11,95 @@
 #include "global.h"
 #include <math.h>
 
+#ifdef __MINGW32__
+/******************************************************/
+/* First we must ensure that all defines are in mingw */
+/******************************************************/
+#ifndef PROCESSOR_PPC_601
+#define PROCESSOR_PPC_601       601
+#endif
+
+#ifndef PROCESSOR_PPC_604
+#define PROCESSOR_PPC_603       603
+#endif
+
+#ifndef PROCESSOR_PPC_604
+#define PROCESSOR_PPC_604       604
+#endif
+
+#ifndef PROCESSOR_PPC_620
+#define PROCESSOR_PPC_620       620
+#endif
+
+#ifndef PROCESSOR_OPTIL
+#define PROCESSOR_OPTIL         0x494f  /* MSIL */
+#endif
+
+#ifndef PROCESSOR_ARCHITECTURE_MSIL
+#define PROCESSOR_ARCHITECTURE_MSIL 8
+#endif
+
+#ifndef PROCESSOR_ARCHITECTURE_AMD64
+#define PROCESSOR_ARCHITECTURE_AMD64            9
+#endif
+
+#ifndef PROCESSOR_ARCHITECTURE_IA32_ON_WIN64
+#define PROCESSOR_ARCHITECTURE_IA32_ON_WIN64    10
+#endif
+
+#ifndef PROCESSOR_HITACHI_SH3
+#define PROCESSOR_HITACHI_SH3   10003   /* Windows CE */
+#endif
+
+#ifndef PROCESSOR_HITACHI_SH3E
+#define PROCESSOR_HITACHI_SH3E  10004   /* Windows CE */
+#endif
+
+#ifndef PROCESSOR_HITACHI_SH4
+#define PROCESSOR_HITACHI_SH4   10005   /* Windows CE */
+#endif
+
+#ifndef PROCESSOR_SHx_SH3
+#define PROCESSOR_SHx_SH3       103     /* Windows CE */
+#endif
+
+#ifndef PROCESSOR_SHx_SH4
+#define PROCESSOR_SHx_SH4       104     /* Windows CE */
+#endif
+
+#ifndef PROCESSOR_STRONGARM
+#define PROCESSOR_STRONGARM     2577    /* Windows CE - 0xA11 */
+#endif
+
+#ifndef PROCESSOR_ARM720
+#define PROCESSOR_ARM720        1824    /* Windows CE - 0x720 */
+#endif
+
+#ifndef PROCESSOR_ARM820
+#define PROCESSOR_ARM820        2080    /* Windows CE - 0x820 */
+#endif
+
+#ifndef PROCESSOR_ARM920
+#define PROCESSOR_ARM920        2336    /* Windows CE - 0x920 */
+#endif
+
+#ifndef PROCESSOR_ARM_7TDMI
+#define PROCESSOR_ARM_7TDMI     70001   /* Windows CE */
+#endif
+
+#ifndef LOGON32_LOGON_NETWORK
+#define LOGON32_LOGON_NETWORK 3
+#endif
+
+/* FP_CLASS compleation */
+
+/* Now for some functions */
+#define Emulate_GetLongPathName GetLongPathNameA
+
+
+
+#endif /* __MINGW32__ */
+
 #ifndef STRUCT_TIMEVAL_DECLARED
 #define STRUCT_TIMEVAL_DECLARED
 struct timeval;
@@ -19,17 +108,14 @@ struct timeval;
 #ifdef HAVE_ISSPACE
 #define ISSPACE(X) isspace(X)
 #else
-#define ISSPACE(X) ("0012345678SSSSS456789012345678901" \
-                     "S3456789012345678901234567890123" \
-                     "45678901234567890123456789012345" \
-                     "67890123456789012345678901234567" \
-                     "89012345678901234567890123456789" \
-                     "01234567890123456789012345678901" \
-                     "23456789012345678901234567890123" \
-                     "45678901234567890123456789000000"[(X)+1] == 'S')
+PMOD_EXPORT extern const char Pike_isspace_vector[];
+#define ISSPACE(X) (Pike_isspace_vector[(X)+1] == 'S')
 #endif
 
 /* Warning, these run 'C' more than once */
+/* FIXME: Is it that great that every wide char is considered an
+ * identifier char? Doesn't strike me as very unicode compliant.
+ * isalnum, isdigit and islower also look seriously borken. /mast */
 #define WIDE_ISSPACE(C)	(((C) < 256)?ISSPACE(C):0)
 #define WIDE_ISIDCHAR(C) (((C) < 256)?isidchar(C):1)
 #define WIDE_ISALNUM(C)	(((C) < 256)?isalnum(C):0)
@@ -53,15 +139,24 @@ time_t TIME(time_t *);
 #  define TIME time
 #endif
 
+#ifndef HAVE_SNPRINTF
+#ifdef HAVE__SNPRINTF
+/* In WIN32 snprintf is known as _snprintf... */
+#define snprintf _snprintf
+#define HAVE_SNPRINTF
+#endif
+#endif
+
 #ifndef HAVE_RINT
-#define RINTF(X) floorf ((X) + 0.5)
-#define RINT(X) floor( (X) + 0.5 )
-#define RINTL(X) floorl ((X) + 0.5)
-#else
+#define rintf(X) floorf ((X) + 0.5)
+#define rint(X) floor( (X) + 0.5 )
+#define rintl(X) floorl ((X) + 0.5)
+#endif
+
+/* Compat. */
 #define RINTF rintf
 #define RINT rint
 #define RINTL rintl
-#endif
 
 long STRTOL(const char *str,char **ptr,int base);
 PMOD_EXPORT double STRTOD(const char * nptr, char **endptr);
@@ -144,32 +239,6 @@ char *STRCHR(char *s,int c);
 #  endif
 #endif
 
-#ifndef HAVE_STRRCHR
-#  ifdef HAVE_RINDEX
-#    define STRRCHR(X,Y) ((char *)rindex(X,Y))
-#  else
-PMOD_EXPORT char *STRRCHR(char *s,int c);
-#  endif
-#else
-#  define STRRCHR strrchr
-#endif
-
-#ifndef HAVE_STRSTR
-PMOD_EXPORT char *STRSTR(char *s1,const char *s2);
-#else
-#  define STRSTR strstr
-#endif
-
-#ifndef HAVE_STRTOK
-PMOD_EXPORT char *STRTOK(char *s1,char *s2);
-#else
-#  define STRTOK strtok
-#endif
-
-#if !defined(HAVE_VFPRINTF) || !defined(HAVE_VSPRINTF) || !defined(HAVE_VSNPRINTF)
-#  include <stdarg.h>
-#endif
-
 #ifndef HAVE_VFPRINTF
 PMOD_EXPORT int VFPRINTF(FILE *f,const char *s,va_list args);
 #else
@@ -217,21 +286,21 @@ PMOD_EXPORT unsigned INT16 EXTRACT_UWORD_(unsigned char *p);
 PMOD_EXPORT INT16 EXTRACT_WORD_(unsigned char *p);
 PMOD_EXPORT INT32 EXTRACT_INT_(unsigned char *p);
 #else
-static INLINE unsigned EXTRACT_UWORD_(unsigned char *p)
+/*@unused@*/ static INLINE unsigned EXTRACT_UWORD_(unsigned char *p)
 {
   unsigned INT16 a;
   MEMCPY((char *)&a,p,sizeof(a));
   return a;
 }
 
-static INLINE int EXTRACT_WORD_(unsigned char *p)
+/*@unused@*/ static INLINE int EXTRACT_WORD_(unsigned char *p)
 {
   INT16 a;
   MEMCPY((char *)&a,p,sizeof(a));
   return a;
 }
 
-static INLINE INT32 EXTRACT_INT_(unsigned char *p)
+/*@unused@*/ static INLINE INT32 EXTRACT_INT_(unsigned char *p)
 {
   INT32 a;
   MEMCPY((char *)&a,p,sizeof(a));
@@ -248,7 +317,7 @@ static INLINE INT32 EXTRACT_INT_(unsigned char *p)
 PMOD_EXPORT unsigned INT32 my_rand(void);
 PMOD_EXPORT void my_srand(INT32 seed);
 
-PMOD_EXPORT void *pike_realloc(void *ptr, size_t sz);
+PMOD_EXPORT /*@null@*/ void *pike_realloc(void *ptr, size_t sz);
 
 #ifdef OWN_GETHRTIME
 void own_gethrtime_init(void);
@@ -259,33 +328,33 @@ long long gethrtime(void);
 #endif
 
 #ifdef DOUBLE_IS_IEEE_BIG
-#define DECLARE_INF static struct { unsigned char c[8]; double d[1]; } \
+#define DECLARE_INF static const struct { unsigned char c[8]; double d[1]; } \
 	inf_ = { { 0x7f, 0xf0, 0, 0, 0, 0, 0, 0 }, { 0.0 } };
-#define DECLARE_NAN static struct { unsigned char c[8]; double d[1]; } \
+#define DECLARE_NAN static const struct { unsigned char c[8]; double d[1]; } \
 	nan_ = { { 0x7f, 0xf8, 0, 0, 0, 0, 0, 0 }, { 0.0 } };
 #define MAKE_INF(s) ((s)*inf_.d[-1])
 #define MAKE_NAN() (nan_.d[-1])
 #else
 #ifdef DOUBLE_IS_IEEE_LITTLE
-#define DECLARE_INF static struct { unsigned char c[8]; double d[1]; } \
+#define DECLARE_INF static const struct { unsigned char c[8]; double d[1]; } \
 	inf_ = { { 0, 0, 0, 0, 0, 0, 0xf0, 0x7f }, { 0.0 } };
-#define DECLARE_NAN static struct { unsigned char c[8]; double d[1]; } \
+#define DECLARE_NAN static const struct { unsigned char c[8]; double d[1]; } \
 	nan_ = { { 0, 0, 0, 0, 0, 0, 0xf8, 0x7f }, { 0.0 } };
 #define MAKE_INF(s) ((s)*inf_.d[-1])
 #define MAKE_NAN() (nan_.d[-1])
 #else
 #ifdef FLOAT_IS_IEEE_BIG
-#define DECLARE_INF static struct { unsigned char c[4]; float f[1]; } \
+#define DECLARE_INF static const struct { unsigned char c[4]; float f[1]; } \
 	inf_ = { { 0x7f, 0x80, 0, 0 }, { 0.0 } };
-#define DECLARE_NAN static struct { unsigned char c[4]; float f[1]; } \
+#define DECLARE_NAN static const struct { unsigned char c[4]; float f[1]; } \
 	nan_ = { { 0x7f, 0xc0, 0, 0 }, { 0.0 } };
 #define MAKE_INF(s) ((s)*inf_.f[-1])
 #define MAKE_NAN() (nan_.f[-1])
 #else
 #ifdef FLOAT_IS_IEEE_LITTLE
-#define DECLARE_INF static struct { unsigned char c[4]; float f[1]; } \
+#define DECLARE_INF static const struct { unsigned char c[4]; float f[1]; } \
 	inf_ = { { 0, 0, 0x80, 0x7f }, { 0.0 } };
-#define DECLARE_NAN static struct { unsigned char c[4]; float f[1]; } \
+#define DECLARE_NAN static const struct { unsigned char c[4]; float f[1]; } \
 	nan_ = { { 0, 0, 0xc0, 0x7f }, { 0.0 } };
 #define MAKE_INF(s) ((s)*inf_.f[-1])
 #define MAKE_NAN() (nan_.f[-1])
@@ -344,4 +413,8 @@ double FREXP(double x, int *exp);
 double LDEXP(double x, int exp);
 #endif
 
+#ifdef __MINGW32__
+void _dosmaperr(int x);
+#endif
+
 #endif
