@@ -2,23 +2,24 @@
 || This file is part of Pike. For copyright information see COPYRIGHT.
 || Pike is distributed under GPL, LGPL and MPL. See the file COPYING
 || for more information.
-|| $Id: program.h,v 1.202 2004/04/18 02:16:06 mast Exp $
+|| $Id: program.h,v 1.267 2009/09/12 13:31:38 grubba Exp $
 */
 
 #ifndef PROGRAM_H
 #define PROGRAM_H
 
-#include <stdarg.h>
 #include "global.h"
 #include "pike_macros.h"
 #include "pike_error.h"
 #include "svalue.h"
+#include "dmalloc.h"
 #include "time_stuff.h"
 #include "program_id.h"
+#include "pike_rusage.h"
 #include "block_alloc_h.h"
 
 /* Needed to support dynamic loading on NT */
-PMOD_PROTO extern struct program_state * Pike_compiler;
+PMOD_EXPORT extern struct program_state * Pike_compiler;
 
 /* Compilation flags */
 #define COMPILATION_CHECK_FINAL         0x01
@@ -51,6 +52,10 @@ PMOD_PROTO extern struct program_state * Pike_compiler;
 
 extern struct pike_string *this_program_string;
 
+/* Common compiler subsystems */
+extern struct pike_string *parser_system_string;
+extern struct pike_string *type_check_system_string;
+
 #define LFUN___INIT 0
 #define LFUN_CREATE 1
 #define LFUN_DESTROY 2
@@ -95,15 +100,16 @@ extern struct pike_string *this_program_string;
 #define LFUN__EQUAL 41
 #define LFUN__M_DELETE 42
 #define LFUN__GET_ITERATOR 43
+#define LFUN_RANGE 44
 
-#define NUM_LFUNS 44
+#define NUM_LFUNS 45
 
 /* NOTE: After this point there are only fake lfuns.
  *       ie use low_find_lfun(), and NOT FIND_LFUN()!
  */
-#define LFUN__SEARCH 44
+#define LFUN__SEARCH 45
 
-extern char *lfun_names[];
+extern const char *const lfun_names[];
 
 extern struct pike_string *lfun_strings[];
 
@@ -118,6 +124,12 @@ typedef struct node_s node;
 struct object;
 #endif
 
+#undef EXTERN
+#undef STRUCT
+#undef PUSH
+#undef POP
+#undef DECLARE
+
 #define STRUCT
 #include "compilation.h"
 
@@ -130,14 +142,24 @@ struct object;
 #define PIKE_BYTECODE_GOTO	1
 #define PIKE_BYTECODE_SPARC	2
 #define PIKE_BYTECODE_IA32	3
-#define PIKE_BYTECODE_PPC32     4
+#define PIKE_BYTECODE_PPC32    4
+#define PIKE_BYTECODE_AMD64    5
+#define PIKE_BYTECODE_PPC64    6
+
+#ifndef PIKE_BYTECODE_METHOD
+#error PIKE_BYTECODE_METHOD not set.
+#endif
 
 #if PIKE_BYTECODE_METHOD == PIKE_BYTECODE_IA32
 #define PIKE_OPCODE_T unsigned INT8
+#elif PIKE_BYTECODE_METHOD == PIKE_BYTECODE_AMD64
+#define PIKE_OPCODE_T unsigned INT8
 #elif PIKE_BYTECODE_METHOD == PIKE_BYTECODE_SPARC
 #define PIKE_OPCODE_T unsigned INT32
 #elif PIKE_BYTECODE_METHOD == PIKE_BYTECODE_PPC32
 #define PIKE_OPCODE_T unsigned INT32
+#elif PIKE_BYTECODE_METHOD == PIKE_BYTECODE_PPC64
+#define PIKE_OPCODE_T unsigned INT32
 #elif PIKE_BYTECODE_METHOD == PIKE_BYTECODE_GOTO
 #define PIKE_OPCODE_T void *
 #define PIKE_INSTR_T void *
@@ -176,48 +198,69 @@ struct object;
 
 union idptr
 {
-  /* C function pointer. */
+  /* External symbol reference. (IDENTIFIER_EXTERN)
+   *
+   * Note that this bit MUST be checked to be zero
+   * before looking at the other four cases!
+   */
+  struct {
+    unsigned short depth;	/* Scope count. */
+    unsigned short id;		/* Reference number. */
+  } ext_ref;
+
+  /* Getter/setter reference pair. (IDENTIFIER_VARIABLE && PIKE_T_GET_SET)
+   */
+  struct {
+    INT16 getter;		/* Reference to getter. */
+    INT16 setter;		/* Reference to setter. */
+  } gs_info;
+
+  /* C function pointer. (IDENTIFIER_C_FUNCTION) */
   void (*c_fun)(INT32);
 
   /* For variables: Offset of the variable in the storage pointed to
    * by inherit.storage_offset in the struct inherit that corresponds
    * to the identifier. See LOW_GET_GLOBAL and GET_GLOBAL. The stored
    * variable may be either a normal or a short svalue, depending on
-   * identifier.run_time_type.
+   * identifier.run_time_type. (IDENTIFIER_VARIABLE)
    *
    * For constants: Offset of the struct program_constant in
    * program.constants in the program pointed to by prog in the struct
-   * inherit that corresponds to the identifier.
+   * inherit that corresponds to the identifier. (IDENTIFIER_CONSTANT)
    *
    * For pike functions: Offset to the start of the function in
    * program.program in the program pointed to by prog in the struct
    * inherit that corresponds to the identifier. Or -1 if a prototype.
+   * (IDENTIFIER_PIKE_FUNCTION)
    */
   ptrdiff_t offset;
 };
 
 #define IDENTIFIER_VARIABLE 0
-#define IDENTIFIER_PIKE_FUNCTION 1
+#define IDENTIFIER_CONSTANT 1
 #define IDENTIFIER_C_FUNCTION 2
-#define IDENTIFIER_FUNCTION 3
-#define IDENTIFIER_CONSTANT 4
-#define IDENTIFIER_TYPE_MASK 7
+#define IDENTIFIER_PIKE_FUNCTION 3
+#define IDENTIFIER_FUNCTION 2
+#define IDENTIFIER_TYPE_MASK 3
 
+#define IDENTIFIER_ALIAS 4  /* Identifier is an alias for another
+			     * (possibly extern) symbol.
+			     */
 #define IDENTIFIER_VARARGS 8	/* Used for functions only. */
 #define IDENTIFIER_NO_THIS_REF 8 /* Used for variables only: Don't count refs to self. */
 #define IDENTIFIER_HAS_BODY 16  /* Function has a body (set already in pass 1). */
 #define IDENTIFIER_SCOPED 32   /* This is used for local functions only */
 #define IDENTIFIER_SCOPE_USED 64 /* contains scoped local functions */
-#define IDENTIFIER_ALIAS 128   /* This identifier is an alias. */
 
 #define IDENTIFIER_IS_FUNCTION(X) ((X) & IDENTIFIER_FUNCTION)
-#define IDENTIFIER_IS_PIKE_FUNCTION(X) ((X) & IDENTIFIER_PIKE_FUNCTION)
-#define IDENTIFIER_IS_C_FUNCTION(X) ((X) & IDENTIFIER_C_FUNCTION)
-#define IDENTIFIER_IS_CONSTANT(X) ((X) & IDENTIFIER_CONSTANT)
+#define IDENTIFIER_IS_PIKE_FUNCTION(X) (((X) & IDENTIFIER_TYPE_MASK) == IDENTIFIER_PIKE_FUNCTION)
+#define IDENTIFIER_IS_C_FUNCTION(X) (((X) & IDENTIFIER_TYPE_MASK) == IDENTIFIER_C_FUNCTION)
+#define IDENTIFIER_IS_CONSTANT(X) (((X) & IDENTIFIER_TYPE_MASK) == IDENTIFIER_CONSTANT)
 #define IDENTIFIER_IS_VARIABLE(X) (!((X) & IDENTIFIER_TYPE_MASK))
 #define IDENTIFIER_IS_ALIAS(X)	((X) & IDENTIFIER_ALIAS)
+#define IDENTIFIER_IS_SCOPED(X) ((X) & IDENTIFIER_SCOPED)
 
-#define IDENTIFIER_MASK 255
+#define IDENTIFIER_MASK 127
 
 /*
  * Every constant, class, function and variable
@@ -227,13 +270,15 @@ struct identifier
 {
   struct pike_string *name;
   struct pike_type *type;
+  unsigned INT32 filename_strno;	/* Index in strings. */
+  unsigned INT32 linenumber;
   unsigned INT8 identifier_flags;	/* IDENTIFIER_??? */
   unsigned INT8 run_time_type;		/* PIKE_T_??? */
   unsigned INT16 opt_flags;		/* OPT_??? */
 #ifdef PROFILING
-  unsigned INT32 num_calls;
-  unsigned INT32 total_time;
-  unsigned INT32 self_time;
+  unsigned INT32 num_calls;		/* Total number of calls. */
+  cpu_time_t total_time;		/* Total time with children. */
+  cpu_time_t self_time;			/* Total time excluding children. */
 #endif /* PROFILING */
   union idptr func;
 };
@@ -245,8 +290,8 @@ struct identifier
  */
 struct program_constant
 {
-  struct svalue sval;
-  struct pike_string *name;
+  struct svalue sval;	/* Value. */
+  ptrdiff_t offset;	/* Offset in identifiers to initialization function. */
 };
 
 /*
@@ -256,24 +301,29 @@ struct program_constant
  * char code[]
  */
 
+#define ID_PROTECTED       0x01 /* Symbol is not visible by indexing */
 #define ID_STATIC          0x01	/* Symbol is not visible by indexing */
 #define ID_PRIVATE         0x02	/* Symbol is not visible by inherit */
-#define ID_NOMASK          0x04	/* Symbol may not be overloaded */
+#define ID_FINAL           0x04	/* Symbol may not be overloaded */
+#define ID_NOMASK          0x04	/* Symbol may not be overloaded (deprecated) */
 #define ID_PUBLIC          0x08 /* Anti private */
-#define ID_PROTECTED       0x10 /* Not currently used at all */
+#define ID_USED		   0x10 /* This identifier has been used. Check
+				 * that the type is compatible when
+				 * overloading. */
+#define ID_LOCAL           0x20 /* Locally referenced symbol (not virtual) */
 #define ID_INLINE          0x20 /* Same as local */
 #define ID_HIDDEN          0x40	/* Symbols that are private and inherited one step later */
 #define ID_INHERITED       0x80 /* Symbol is inherited */
 #define ID_OPTIONAL       0x100	/* Symbol is not required by the interface */
 #define ID_EXTERN         0x200	/* Symbol is defined later */
 #define ID_VARIANT	  0x400 /* Function is overloaded by argument. */
-#define ID_ALIAS	  0x800 /* Variable is an overloaded alias. */
 
-#define ID_MODIFIER_MASK 0x0fff
+#define ID_MODIFIER_MASK 0x07ff
 
-#define ID_STRICT_TYPES  0x8000	/* #pragma strict_types */
-#define ID_SAVE_PARENT  0x10000 /* #pragma save_parent */
-#define ID_DONT_SAVE_PARENT 0x20000 /* #pragma dont_save_parent */
+#define ID_STRICT_TYPES             0x8000 /* #pragma strict_types */
+#define ID_SAVE_PARENT             0x10000 /* #pragma save_parent */
+#define ID_DONT_SAVE_PARENT        0x20000 /* #pragma dont_save_parent */
+#define ID_NO_DEPRECATION_WARNINGS 0x40000 /* #pragma no_deprecation_warnings */
 
 
 /*
@@ -299,8 +349,8 @@ struct reference
    * above. See ID_FROM_PTR and ID_FROM_INT. */
   unsigned INT16 identifier_offset;
 
-  /* ID_* flags - static, private etc.. */
-  INT16 id_flags;
+  /* ID_* flags - protected, private etc.. */
+  unsigned INT16 id_flags;
 };
 
 /* Magic value used as identifier reference integer to refer to this. */
@@ -467,9 +517,35 @@ struct pike_trampoline
 #define PROGRAM_VIRGIN 0x800
 
 /* Don't allow the program to be inherited or cloned if there's no
- * parent object. Only set if PROGRAM_USES_PARENT is. */
+ * parent object. Might only be set if PROGRAM_USES_PARENT is. */
 #define PROGRAM_NEEDS_PARENT 0x1000
 
+/* Program has code to be executed when it's destructed. This causes
+ * the gc to consider objects of this program to be "live" and will
+ * take care to ensure a good destruct sequence so that other
+ * referenced things are intact as far as possible when the object is
+ * destructed. See the blurb near the top of gc.c for further
+ * discussion.
+ *
+ * This flag gets set automatically when an event handler that might
+ * act on PROG_EVENT_EXIT is installed (i.e. through set_exit_callback
+ * or pike_set_prog_event_callback). However, if the handler is very
+ * simple then this flag may be cleared again to allow the gc to
+ * handle objects of this program more efficiently (by considering
+ * them "dead" and destroy them in an arbitrary order).
+ *
+ * Such a "very simple" handler should only do things like freeing
+ * pointers and clearing variables in the object storage. It can not
+ * assume that any other context is intact when it's called. It might
+ * even get called after the module exit function, which means that
+ * this optimization trick can never be used in a dynamically loaded
+ * module. */
+#define PROGRAM_LIVE_OBJ 0x2000
+
+/* Indicates that the class is a facet or product_class. */
+#define PROGRAM_IS_FACET 0x4000
+#define PROGRAM_IS_PRODUCT 0x8000
+
 /* Using define instead of enum allows for ifdefs - Hubbe */
 #define PROG_EVENT_INIT 0
 #define PROG_EVENT_EXIT 1
@@ -497,6 +573,13 @@ struct parent_info
   INT16 parent_identifier;
 };
 
+/* Single entry cache for object indexing. */
+struct identifier_lookup_cache
+{
+  INT32 program_id;
+  INT32 identifier_id;
+};
+
 struct program
 {
   PIKE_MEMORY_OBJECT_MEMBERS; /* Must be first */
@@ -537,14 +620,23 @@ struct program
 #include "program_areas.h"
   
   INT16 lfuns[NUM_LFUNS];
+
+#ifdef WITH_FACETS
+  /* Facet related stuff */
+  INT32 facet_index;   /* Index to the facet this facet class belongs to */
+  struct object *facet_group;
+#endif
 };
 
-#if 0
-static inline int CHECK_IDREF_RANGE (int x, const struct program *p)
+void dump_program_tables (const struct program *p, int indent);
+#ifdef PIKE_DEBUG
+static INLINE int CHECK_IDREF_RANGE (int x, const struct program *p)
 {
-  if (x < 0 || x >= p->num_identifier_references)
+  if (x < 0 || x >= p->num_identifier_references) {
+    dump_program_tables(p, 4);
     debug_fatal ("Identifier reference index %d out of range 0..%d\n", x,
 		 p->num_identifier_references - 1);
+  }
   return x;
 }
 #else
@@ -564,63 +656,103 @@ static inline int CHECK_IDREF_RANGE (int x, const struct program *p)
 #define FIND_LFUN(P,N) ( dmalloc_touch(struct program *,(P))->flags & PROGRAM_FIXED?((P)->lfuns[(N)]):low_find_lfun((P), (N)) )
 #define QUICK_FIND_LFUN(P,N) (dmalloc_touch(struct program *,(P))->lfuns[N])
 
+#ifdef DO_PIKE_CLEANUP
+PMOD_EXPORT extern int gc_external_refs_zapped;
+PMOD_EXPORT void gc_check_zapped (void *a, TYPE_T type, const char *file, int line);
+#endif
+
+#if defined (USE_DLL) && defined (DYNAMIC_MODULE)
+/* Use the function in modules so we don't have to export the block
+ * alloc stuff. */
+#define free_program(p) do_free_program(p)
+#else
 #define free_program(p) do{						\
     struct program *_=(p);						\
     debug_malloc_touch(_);						\
-    DO_IF_PIKE_CLEANUP (						\
-      if (gc_external_refs_zapped)					\
-	gc_check_zapped (_, PIKE_T_PROGRAM, __FILE__, __LINE__);	\
-    );									\
+    DO_IF_DEBUG (							\
+      DO_IF_PIKE_CLEANUP (						\
+	if (gc_external_refs_zapped)					\
+	  gc_check_zapped (_, PIKE_T_PROGRAM, __FILE__, __LINE__)));	\
     if(!sub_ref(_))							\
       really_free_program(_);						\
   }while(0)
+#endif
 
-BLOCK_ALLOC_FILL_PAGES(program, n/a)
+BLOCK_ALLOC_FILL_PAGES(program, n/a);
 
 
-extern struct object *error_handler;
-extern struct object *compat_handler;
-
 extern struct program *first_program;
 extern struct program *null_program;
+extern struct program *compilation_env_program;
+extern struct program *compilation_program;
+extern struct object *compilation_environment;
 extern struct program *pike_trampoline_program;
 extern struct program *gc_internal_program;
 extern struct program *placeholder_program;
 extern struct object *placeholder_object;
 
-extern int compilation_depth;
-
 /* Flags for identifier finding... */
 #define SEE_STATIC 1
+#define SEE_PROTECTED 1
 #define SEE_PRIVATE 2
 
+/* Report levels */
+#define REPORT_NOTICE	0	/* FYI. */
+#define REPORT_WARNING	1	/* Compiler warning. */
+#define REPORT_ERROR	2	/* Compilation error. */
+#define REPORT_FATAL	3	/* Unrecoverable error. */
+
 
 #define COMPILER_IN_CATCH 1
 
 #define ADD_STORAGE(X) low_add_storage(sizeof(X), ALIGNOF(X),0)
+
 #define STORAGE_NEEDED(X) ((X)->storage_needed - (X)->inherits[0].storage_offset)
 
-#define FOO(NUMTYPE,TYPE,ARGTYPE,NAME) void PIKE_CONCAT(add_to_,NAME(ARGTYPE ARG));
+#define FOO(NUMTYPE,TYPE,ARGTYPE,NAME) \
+  void PIKE_CONCAT(add_to_,NAME)(ARGTYPE ARG);
+#define BAR(NUMTYPE,TYPE,ARGTYPE,NAME) \
+  void PIKE_CONCAT(low_add_many_to_,NAME)(struct program_state *state, ARGTYPE *ARG, NUMTYPE cnt); \
+  void PIKE_CONCAT(add_to_,NAME)(ARGTYPE ARG);
 #include "program_areas.h"
 
 typedef int supporter_callback (void *, int);
+
 struct Supporter
 {
 #ifdef PIKE_DEBUG
   int magic;
 #endif
+
   struct Supporter *previous;
+  /* Makes up a linked list of supporters with the first one in
+   * current_supporter. Supporters are linked onto this list during
+   * the (recursive) compilation of each compilation unit (i.e.
+   * compiled string). Thus nested programs and programs built from C
+   * don't have supporters. */
+
   struct Supporter *depends_on;
-  struct Supporter *dependants;
-  struct Supporter *next_dependant;
+  /* The supporter furthest in on the current_supporter linked list
+   * that this one depends on. When it gets unlinked from that list,
+   * this becomes a back pointer for the dependants linked list
+   * below. */
+
+  struct Supporter *dependants, *next_dependant;
+  /* dependants points to a linked list of supporters that depends on
+   * this one, and next_dependant makes up the links between those
+   * supporters. A supporter is linked onto this list when it is
+   * unlinked from the current_supporter list. */
+
+  struct object *self;
   supporter_callback *fun;
   void *data;
+
   struct program *prog;
+  /* The top level program in the compilation unit. */
 };
 
-
-
 /* Prototypes begin here */
+PMOD_EXPORT void do_free_program (struct program *p);
 void ins_int(INT32 i, void (*func)(char tmp));
 void ins_short(int i, void (*func)(char tmp));
 void add_relocated_int_to_program(INT32 i);
@@ -651,7 +783,6 @@ void low_start_new_program(struct program *p,
 PMOD_EXPORT void debug_start_new_program(int line, const char *file);
 void dump_program_desc(struct program *p);
 int sizeof_variable(int run_time_type);
-void dump_program_tables (struct program *p, int indent);
 void check_program(struct program *p);
 struct program *end_first_pass(int finish);
 PMOD_EXPORT struct program *debug_end_program(void);
@@ -661,8 +792,8 @@ PMOD_EXPORT void set_init_callback(void (*init_callback)(struct object *));
 PMOD_EXPORT void set_exit_callback(void (*exit_callback)(struct object *));
 PMOD_EXPORT void set_gc_recurse_callback(void (*m)(struct object *));
 PMOD_EXPORT void set_gc_check_callback(void (*m)(struct object *));
-void pike_set_prog_event_callback(void (*cb)(int));
-void pike_set_prog_optimize_callback(node *(*opt)(node *));
+PMOD_EXPORT void pike_set_prog_event_callback(void (*cb)(int));
+PMOD_EXPORT void pike_set_prog_optimize_callback(node *(*opt)(node *));
 int really_low_reference_inherited_identifier(struct program_state *q,
 					      int e,
 					      int i);
@@ -674,12 +805,12 @@ int find_inherit(struct program *p, struct pike_string *name);
 node *reference_inherited_identifier(struct pike_string *super_name,
 				     struct pike_string *function_name);
 void rename_last_inherit(struct pike_string *n);
-void low_inherit(struct program *p,
-		 struct object *parent,
-		 int parent_identifier,
-		 int parent_offset,
-		 INT32 flags,
-		 struct pike_string *name);
+PMOD_EXPORT void low_inherit(struct program *p,
+			     struct object *parent,
+			     int parent_identifier,
+			     int parent_offset,
+			     INT32 flags,
+			     struct pike_string *name);
 PMOD_EXPORT void do_inherit(struct svalue *s,
 		INT32 flags,
 		struct pike_string *name);
@@ -689,6 +820,10 @@ void simple_do_inherit(struct pike_string *s,
 		       INT32 flags,
 		       struct pike_string *name);
 int isidentifier(struct pike_string *s);
+int low_define_alias(struct pike_string *name, struct pike_type *type,
+		     int flags, int depth, int refno);
+PMOD_EXPORT int define_alias(struct pike_string *name, struct pike_type *type,
+			     int flags, int depth, int refno);
 int low_define_variable(struct pike_string *name,
 			struct pike_type *type,
 			INT32 flags,
@@ -752,7 +887,7 @@ INT32 define_function(struct pike_string *name,
 int really_low_find_shared_string_identifier(struct pike_string *name,
 					     struct program *prog,
 					     int flags);
-int low_find_lfun(struct program *p, ptrdiff_t lfun);
+PMOD_EXPORT int low_find_lfun(struct program *p, ptrdiff_t lfun);
 int lfun_lookup_id(struct pike_string *lfun_name);
 int low_find_shared_string_identifier(struct pike_string *name,
 				      struct program *prog);
@@ -766,8 +901,8 @@ int store_constant(struct svalue *foo,
 		   struct pike_string *constant_name);
 struct array *program_indices(struct program *p);
 struct array *program_values(struct program *p);
-void program_index_no_free(struct svalue *to, struct program *p,
-			   struct svalue *ind);
+int program_index_no_free(struct svalue *to, struct svalue *what,
+			  struct svalue *ind);
 int get_small_number(char **q);
 void ext_store_program_line (struct program *prog, INT32 line, struct pike_string *file);
 void start_line_numbering(void);
@@ -786,11 +921,31 @@ PMOD_EXPORT struct pike_string *get_line(PIKE_OPCODE_T *pc,
 					 struct program *prog, INT32 *linep);
 PMOD_EXPORT struct pike_string *low_get_function_line (struct object *o,
 						       int fun, INT32 *linep);
-void va_yyerror(const char *fmt, va_list args);
-void my_yyerror(const char *fmt,...)  ATTRIBUTE((format(printf,1,2)));
+PMOD_EXPORT struct pike_string *get_identifier_line(struct program *p,
+						    int fun, INT32 *linep);
+PMOD_EXPORT void va_yyreport(int severity_level,
+			     struct pike_string *file, INT32 line,
+			     struct pike_string *system, INT32 args,
+			     const char *fmt, va_list vargs);
+PMOD_EXPORT void low_yyreport(int severity_level,
+			      struct pike_string *file, INT32 line,
+			      struct pike_string *system,
+			      INT32 args, const char *fmt, ...);
+PMOD_EXPORT void yyreport(int severity_level, struct pike_string *system,
+			  INT32 args, const char *fmt, ...);
+PMOD_EXPORT void yywarning(char *fmt, ...);
+PMOD_EXPORT void my_yyerror(const char *fmt,...);
+PMOD_EXPORT void yyerror(const char *s);
+void yytype_report(int severity_level,
+		   struct pike_string *expect_file, INT32 expect_line, 
+		   struct pike_type *expected_t,
+		   struct pike_string *got_file, INT32 got_line,
+		   struct pike_type *got_t,
+		   INT32 args, const char *fmt, ...);
+void yytype_error(const char *msg, struct pike_type *expected_t,
+		  struct pike_type *got_t, unsigned int flags);
 struct pike_string *format_exception_for_error_msg (struct svalue *thrown);
-void handle_compile_exception (const char *yyerror_fmt, ...)
-  ATTRIBUTE((format(printf,1,2)));
+void handle_compile_exception (const char *yyerror_fmt, ...);
 struct supporter_marker;
 void verify_supporters(void);
 void init_supporter(struct Supporter *s,
@@ -801,6 +956,8 @@ int call_dependants(struct Supporter *s, int finish);
 int report_compiler_dependency(struct program *p);
 struct compilation;
 void run_pass2(struct compilation *c);
+PMOD_EXPORT void enter_compiler(struct pike_string *filename, int linenumber);
+PMOD_EXPORT void exit_compiler(void);
 struct program *compile(struct pike_string *aprog,
 			struct object *ahandler,
 			int amajor, int aminor,
@@ -820,6 +977,7 @@ void check_all_programs(void);
 void placeholder_index(INT32 args);
 void init_program(void);
 void cleanup_program(void);
+PMOD_EXPORT void visit_program (struct program *p, int action);
 void gc_mark_program_as_referenced(struct program *p);
 void real_gc_cycle_check_program(struct program *p, int weak);
 unsigned gc_touch_all_programs(void);
@@ -832,27 +990,27 @@ void push_compiler_frame(int lexical_scope);
 void low_pop_local_variables(int level);
 void pop_local_variables(int level);
 void pop_compiler_frame(void);
-ptrdiff_t low_get_storage(struct program *o, struct program *p);
+PMOD_EXPORT ptrdiff_t low_get_storage(struct program *o, struct program *p);
 PMOD_EXPORT char *get_storage(struct object *o, struct program *p);
-struct program *low_program_from_function(struct program *p,
-					  INT32 i);
+struct program *low_program_from_function(struct object *o, INT32 i);
 PMOD_EXPORT struct program *program_from_function(const struct svalue *f);
 PMOD_EXPORT struct program *program_from_svalue(const struct svalue *s);
 struct find_child_cache_s;
 int find_child(struct program *parent, struct program *child);
-void yywarning(char *fmt, ...) ATTRIBUTE((format(printf,1,2)));
 struct implements_cache_s;
 PMOD_EXPORT int implements(struct program *a, struct program *b);
 PMOD_EXPORT int is_compatible(struct program *a, struct program *b);
-void yyexplain_not_compatible(struct program *a, struct program *b, int flags);
-void yyexplain_not_implements(struct program *a, struct program *b, int flags);
+void yyexplain_not_compatible(int severity_level,
+			      struct program *a, struct program *b);
+void yyexplain_not_implements(int severity_level,
+			      struct program *a, struct program *b);
 PMOD_EXPORT void *parent_storage(int depth);
 PMOD_EXPORT void change_compiler_compatibility(int major, int minor);
+void make_area_executable (char *start, size_t len);
 void make_program_executable(struct program *p);
 /* Prototypes end here */
 
-void really_free_program(struct program *);
-void count_memory_in_programs(INT32*,INT32*);
+void count_memory_in_programs(size_t *, size_t *);
 
 #ifndef PIKE_USE_MACHINE_CODE
 #define make_program_executable(X)
@@ -883,6 +1041,9 @@ void count_memory_in_programs(INT32*,INT32*);
   quick_map_variable(NAME, CONSTANT_STRLEN(NAME), OFFSET, \
                      TYPE, CONSTANT_STRLEN(TYPE), RTTYPE, FLAGS)
 
+#define MAP_VARIABLE(NAME, TYPE, FLAGS, OFFSET, RTYPE) \
+  PIKE_MAP_VARIABLE(NAME, OFFSET, TYPE, RTYPE, FLAGS)
+
 #define ADD_FUNCTION_DTYPE(NAME,FUN,DTYPE,FLAGS) do {		\
   DTYPE_START;							\
   {DTYPE}							\
@@ -923,6 +1084,9 @@ void count_memory_in_programs(INT32*,INT32*);
 
 #define start_new_program() debug_start_new_program(__LINE__,__FILE__)
 
+#define visit_program_ref(P, REF_TYPE)				\
+  visit_ref (pass_program (P), (REF_TYPE),			\
+	     (visit_thing_fn *) &visit_program, NULL)
 #define gc_cycle_check_program(X, WEAK) \
   gc_cycle_enqueue((gc_cycle_check_cb *) real_gc_cycle_check_program, (X), (WEAK))
 
