@@ -2,7 +2,7 @@
 || This file is part of Pike. For copyright information see COPYRIGHT.
 || Pike is distributed under GPL, LGPL and MPL. See the file COPYING
 || for more information.
-|| $Id: sscanf.c,v 1.160 2004/03/23 18:06:11 nilsson Exp $
+|| $Id: sscanf.c,v 1.190 2008/12/13 08:22:22 nilsson Exp $
 */
 
 #include "global.h"
@@ -15,12 +15,13 @@
 #include "module_support.h"
 #include "operators.h"
 #include "bignum.h"
+#include "pike_float.h"
+#include "pike_types.h"
+#include "sscanf.h"
 
 #define sp Pike_sp
 
-RCSID("$Id: sscanf.c,v 1.160 2004/03/23 18:06:11 nilsson Exp $");
-
-/* 
+/*
  * helper functions for sscanf %O
  */
 
@@ -252,6 +253,7 @@ static void *pcharp_to_svalue_percent_o(struct svalue *res,
    *
   operators:
   %d
+  %H
   %s
   %f
   %c
@@ -282,18 +284,20 @@ static ptrdiff_t PIKE_CONCAT(read_set,SIZE) (			\
   PIKE_CONCAT(p_wchar,SIZE) *match,				\
   ptrdiff_t cnt,						\
   struct sscanf_set *set,					\
-  ptrdiff_t match_len)						\
+  ptrdiff_t match_len, INT32 flags)				\
 {								\
-  size_t e, last = 0;						\
-  CHAROPT( int set_size=0; )					\
+  p_wchar2 e, last = 0;						\
+  MATCH_IS_WIDE( int set_size=0; )				\
 								\
   if(cnt>=match_len)						\
-    Pike_error("Error in sscanf format string.\n");			\
+    Pike_error("Error in sscanf format string.\n");		\
 								\
   MEMSET(set->c, 0, sizeof(set->c));				\
   set->a=0;							\
 								\
-  if(match[cnt]=='^')						\
+  if(match[cnt]=='^' &&						\
+     (cnt+2>=match_len || match[cnt+1]!='-' ||			\
+      match[cnt+2]==']' || (flags & SSCANF_FLAG_76_COMPAT)))	\
   {								\
     set->neg=1;							\
     cnt++;							\
@@ -311,7 +315,7 @@ static ptrdiff_t PIKE_CONCAT(read_set,SIZE) (			\
       Pike_error("Error in sscanf format string.\n");		\
   }								\
 								\
-  for(;match[cnt]!=']';cnt++)					\
+  for(;match[cnt]!=']';)					\
   {								\
     if(match[cnt]=='-')						\
     {								\
@@ -325,19 +329,19 @@ static ptrdiff_t PIKE_CONCAT(read_set,SIZE) (			\
 	break;							\
       }								\
 								\
-      if(last >= match[cnt])					\
+      if(last > match[cnt])					\
 	Pike_error("Error in sscanf format string.\n");		\
 								\
-CHAROPT(							\
-      if(last < (size_t)sizeof(set->c))				\
+MATCH_IS_WIDE(							\
+      if(last < (p_wchar2)sizeof(set->c) && last >= 0)		\
       {								\
-	if(match[cnt] < (size_t)sizeof(set->c))			\
+	if(match[cnt] < (p_wchar2)sizeof(set->c))		\
 	{							\
 )								\
 	  for(e=last;e<=match[cnt];e++) set->c[e]=1;		\
-CHAROPT(							\
+MATCH_IS_WIDE(							\
 	}else{							\
-	  for(e=last;e<(size_t)sizeof(set->c);e++)		\
+	  for(e=last;e<(p_wchar2)sizeof(set->c);e++)		\
             set->c[e]=1;					\
 								\
 	  check_stack(2);					\
@@ -351,41 +355,46 @@ CHAROPT(							\
 	sp[-1].u.integer=match[cnt];				\
       }								\
 )								\
-      continue;							\
-    }								\
-    last=match[cnt];						\
-    if(last < (size_t)sizeof(set->c))				\
-      set->c[last]=1;						\
-CHAROPT(							\
-    else{							\
-      if(set_size &&						\
-	 ((size_t)sp[-1].u.integer) == last-1)			\
-      {								\
-	sp[-1].u.integer++;					\
-      }else{							\
-	check_stack(2);						\
-	push_int64(last);					\
-	push_int64(last);					\
-	set_size++;						\
-      }								\
-    }								\
+    } else {							\
+      last=match[cnt];						\
+MATCH_IS_WIDE(							\
+      if(last < (p_wchar2)sizeof(set->c) && last >= 0)		\
 )								\
+        set->c[last]=1;						\
+MATCH_IS_WIDE(							\
+      else{							\
+        if(set_size &&						\
+	   ((p_wchar2)sp[-1].u.integer) == last-1)		\
+        {							\
+	  sp[-1].u.integer++;					\
+        }else{							\
+	  check_stack(2);					\
+	  push_int(last);					\
+	  push_int(last);					\
+	  set_size++;						\
+        }							\
+      }								\
+  )								\
+    }								\
+    cnt++;							\
+    if(cnt>=match_len)						\
+      Pike_error("Error in sscanf format string.\n");		\
   }								\
 								\
-CHAROPT(							\
+MATCH_IS_WIDE(							\
   if(set_size)							\
   {								\
     INT32 *order;						\
     set->a=aggregate_array(set_size*2);				\
     order=get_switch_order(set->a);				\
-    for(e=0;e<(size_t)set->a->size;e+=2)			\
+    for(e=0;e<(p_wchar2)set->a->size;e+=2)			\
     {								\
       if(order[e]+1 != order[e+1] &&				\
          order[e+1]+1 != order[e]) {				\
         free_array(set->a);					\
         set->a=0;						\
         free((char *)order);					\
-        Pike_error("Overlapping ranges in sscanf not supported.\n");	\
+	Pike_error("Overlapping ranges in sscanf not supported.\n"); \
       }								\
     }								\
 								\
@@ -401,27 +410,8 @@ CHAROPT(							\
 /* Parse binary IEEE strings on a machine which uses a different kind
    of floating point internally */
 
-#ifndef FLOAT_IS_IEEE_BIG
-#ifndef FLOAT_IS_IEEE_LITTLE
-#define NEED_CUSTOM_IEEE
-#endif
-#endif
-#ifndef NEED_CUSTOM_IEEE
-#ifndef DOUBLE_IS_IEEE_BIG
-#ifndef DOUBLE_IS_IEEE_LITTLE
-#define NEED_CUSTOM_IEEE
-#endif
-#endif
-#endif
-
 #ifdef NEED_CUSTOM_IEEE
 
-#if HAVE_LDEXP
-#define LDEXP ldexp
-#else
-extern double LDEXP(double x, int exp); /* defined in encode.c */
-#endif
-
 static INLINE FLOAT_TYPE low_parse_IEEE_float(char *b, int sz)
 {
   unsigned INT32 f, extra_f;
@@ -580,7 +570,7 @@ static INLINE FLOAT_TYPE low_parse_IEEE_float(char *b, int sz)
 
 /* Avoid some warnings about loss of precision */
 #ifdef __ECL
-static inline INT32 TO_INT32(ptrdiff_t x)
+static INLINE INT32 TO_INT32(ptrdiff_t x)
 {
   return DO_NOT_WARN((INT32)x);
 }
@@ -617,7 +607,7 @@ static INT32 PIKE_CONCAT4(very_low_sscanf_,INPUT_SHIFT,_,MATCH_SHIFT)(	 \
 			 PIKE_CONCAT(p_wchar, MATCH_SHIFT) *match,	 \
 			 ptrdiff_t match_len,				 \
 			 ptrdiff_t *chars_matched,			 \
-			 int *success)					 \
+			 int *success, INT32 flags)			 \
 {									 \
   struct svalue sval;							 \
   INT32 matches, arg;							 \
@@ -739,7 +729,7 @@ static INT32 PIKE_CONCAT4(very_low_sscanf_,INPUT_SHIFT,_,MATCH_SHIFT)(	 \
 			 match+cnt+1,					 \
 			 e-cnt-2,					 \
 			 &tmp,						 \
-			 &yes);						 \
+			 &yes, flags);					 \
 	    if(yes && tmp)						 \
 	    {								 \
 	      f_aggregate(TO_INT32(sp-save_sp));			 \
@@ -758,7 +748,7 @@ static INT32 PIKE_CONCAT4(very_low_sscanf_,INPUT_SHIFT,_,MATCH_SHIFT)(	 \
 									 \
 	case 'c':							 \
         {								 \
-CHAROPT2(								 \
+INPUT_IS_WIDE(								 \
           int e;							 \
 )									 \
 	  sval.type=T_INT;						 \
@@ -779,10 +769,10 @@ CHAROPT2(								 \
 	    chars_matched[0]=eye;					 \
 	    return matches;						 \
 	  }								 \
-CHAROPT2(								 \
+INPUT_IS_WIDE(								 \
           for(e=0;e<field_length;e++)					 \
           {								 \
-             if(input[eye+e]>255)					 \
+	    if((unsigned INT32) input[eye+e] > 255)			\
              {								 \
                chars_matched[0]=eye;					 \
                return matches;						 \
@@ -792,7 +782,7 @@ CHAROPT2(								 \
 	  sval.u.integer=0;						 \
 	  if (minus_flag)						 \
 	  {								 \
-	     int x, pos=0;						 \
+	     int pos=0;						 \
 	     if (field_length >= 0) {					 \
 	       pos = (eye += field_length);				 \
 	     }								 \
@@ -854,6 +844,63 @@ CHAROPT2(								 \
 	  }								 \
 	  break;							 \
         }								 \
+ 								         \
+      case 'H':								\
+	{								\
+          int len=0;							\
+          if(field_length == -1)					\
+	    field_length=1;						\
+          if(field_length == 0)                                         \
+            Pike_error("%%H size field is 0.\n");                       \
+          if(eye+field_length > input_len)				\
+	  {								\
+	    chars_matched[0]=eye;					\
+	    return matches;						\
+	  }								\
+	  INPUT_IS_WIDE (						\
+	    for(e=0;e<field_length;e++)					\
+	    {								\
+	      if((unsigned INT32) input[eye+e] > 255)			\
+	      {								\
+		chars_matched[0]=eye;					\
+		return matches;						\
+	      }								\
+	    }								\
+	  );								\
+	  if (minus_flag)						\
+	  {								\
+	    int pos=0;							\
+	    pos = (eye += field_length);				\
+            while(--field_length >= 0)					\
+	    {								\
+	      len<<=8;							\
+	      len |= input[--pos];					\
+	    }								\
+	  } else {							\
+	    while(--field_length >= 0)					\
+	    {								\
+	      len<<=8;							\
+	      len |= input[eye];					\
+	      eye++;							\
+	    }								\
+	  }								\
+	  if(eye+len > input_len)					\
+	  {								\
+	    chars_matched[0]=eye-field_length;				\
+	    return matches;						\
+	  }								\
+	  if (no_assign) {						\
+	    no_assign = 2;						\
+	  } else {							\
+	    sval.type=T_STRING;						\
+	    DO_IF_CHECKER(sval.subtype=0);				\
+	    sval.u.string=PIKE_CONCAT(make_shared_binary_string,	\
+				      INPUT_SHIFT)(input+eye,		\
+						   len);		\
+	  }								\
+	  eye+=len;							\
+	  break;							\
+        }								\
 									 \
         case 'b':							 \
         case 'o':							 \
@@ -950,22 +997,30 @@ CHAROPT2(								 \
 	      return matches;						 \
 	    }								 \
 									 \
-	    sval.type=T_STRING;						 \
-	    DO_IF_CHECKER(sval.subtype=0);				 \
-	    sval.u.string=PIKE_CONCAT(make_shared_binary_string,	 \
-                                      INPUT_SHIFT)(input+eye,		 \
-						   field_length);	 \
+	    if (no_assign) {						 \
+	      no_assign = 2;						 \
+	    } else {							 \
+	      sval.type=T_STRING;					 \
+	      DO_IF_CHECKER(sval.subtype=0);				 \
+	      sval.u.string=PIKE_CONCAT(make_shared_binary_string,	 \
+					INPUT_SHIFT)(input+eye,		 \
+						     field_length);	 \
+	    }								 \
 	    eye+=field_length;						 \
 	    break;							 \
 	  }								 \
 									 \
 	  if(cnt+1>=match_len)						 \
 	  {								 \
-	    sval.type=T_STRING;						 \
-	    DO_IF_CHECKER(sval.subtype=0);				 \
-	    sval.u.string=PIKE_CONCAT(make_shared_binary_string,	 \
-				      INPUT_SHIFT)(input+eye,		 \
-						   input_len-eye);	 \
+	    if (no_assign) {						 \
+	      no_assign = 2;						 \
+	    } else {							 \
+	      sval.type=T_STRING;					 \
+	      DO_IF_CHECKER(sval.subtype=0);				 \
+	      sval.u.string=PIKE_CONCAT(make_shared_binary_string,	 \
+					INPUT_SHIFT)(input+eye,		 \
+						     input_len-eye);	 \
+	    }								 \
 	    eye=input_len;						 \
 	    break;							 \
 	  }else{							 \
@@ -1039,7 +1094,7 @@ CHAROPT2(								 \
 		  PIKE_CONCAT(read_set,MATCH_SHIFT)(match,		 \
 						    s-match+1,		 \
 						    &set,		 \
-						    match_len);		 \
+						    match_len, flags);	 \
 		  set.neg=!set.neg;					 \
 		  goto match_set;					 \
 	      }								 \
@@ -1064,11 +1119,15 @@ CHAROPT2(								 \
 	    end_str_end=match+e;					 \
 									 \
 	    if (end_str_end == end_str_start) {				 \
-	      sval.type=T_STRING;					 \
-	      DO_IF_CHECKER(sval.subtype=0);				 \
-	      sval.u.string=PIKE_CONCAT(make_shared_binary_string,	 \
-	      				INPUT_SHIFT)(input+eye,		 \
-	      					     input_len-eye);	 \
+	      if (no_assign) {						 \
+		no_assign = 2;						 \
+	      } else {							 \
+		sval.type=T_STRING;					 \
+		DO_IF_CHECKER(sval.subtype=0);				 \
+		sval.u.string=PIKE_CONCAT(make_shared_binary_string,	 \
+					  INPUT_SHIFT)(input+eye,	 \
+						       input_len-eye);	 \
+	      }								 \
 	      eye=input_len;						 \
 	      break;							 \
 	    } else if(!contains_percent_percent)			 \
@@ -1109,11 +1168,15 @@ CHAROPT2(								 \
 	      new_eye=p2-input;						 \
 	    }								 \
 									 \
-	    sval.type=T_STRING;						 \
-	    DO_IF_CHECKER(sval.subtype=0);				 \
-	    sval.u.string=PIKE_CONCAT(make_shared_binary_string,	 \
-				      INPUT_SHIFT)(input+start,		 \
-						   eye-start);		 \
+	    if (no_assign) {						 \
+	      no_assign = 2;						 \
+	    } else {							 \
+	      sval.type=T_STRING;					 \
+	      DO_IF_CHECKER(sval.subtype=0);				 \
+	      sval.u.string=PIKE_CONCAT(make_shared_binary_string,	 \
+					INPUT_SHIFT)(input+start,	 \
+						     eye-start);	 \
+	    }								 \
 									 \
 	    cnt=end_str_end-match-1;					 \
 	    eye=new_eye;						 \
@@ -1122,7 +1185,7 @@ CHAROPT2(								 \
 									 \
 	case '[':							 \
 	  cnt=PIKE_CONCAT(read_set,MATCH_SHIFT)(match,cnt+1,		 \
-						&set,match_len);	 \
+						&set,match_len, flags);	 \
 									 \
 	match_set:							 \
 	  {								 \
@@ -1137,19 +1200,20 @@ CHAROPT2(								 \
 	    }								 \
 	    for(e=eye;eye<len;eye++)					 \
 	    {								 \
-CHAROPT2(								 \
-	      if(input[eye]<sizeof(set.c))				 \
+INPUT_IS_WIDE(								 \
+	      if((unsigned INT32) input[eye] < sizeof(set.c))		\
 	      {								 \
 )									 \
 	        if(set.c[input[eye]] == set.neg)			 \
 		  break;						 \
-CHAROPT2(								 \
+INPUT_IS_WIDE(								 \
 	      }else{							 \
 	        if(set.a)						 \
 	        {							 \
 		  INT32 x;						 \
 		  struct svalue tmp;					 \
 		  tmp.type=T_INT;					 \
+		  tmp.subtype = NUMBER_NUMBER;				\
 		  tmp.u.integer=input[eye];				 \
 		  x=switch_lookup(set.a, &tmp);				 \
 		  if( set.neg != (x<0 && (x&1)) ) break;		 \
@@ -1166,10 +1230,14 @@ CHAROPT2(								 \
 	    }								 \
 	  }								 \
           if(set.a) { free_array(set.a); set.a=0; }			 \
-	  sval.type=T_STRING;						 \
-	  DO_IF_CHECKER(sval.subtype=0);				 \
-	  sval.u.string=PIKE_CONCAT(make_shared_binary_string,		 \
-				    INPUT_SHIFT)(input+e,eye-e);	 \
+	  if (no_assign) {						 \
+	    no_assign = 2;						 \
+	  } else {							 \
+	    sval.type=T_STRING;						 \
+	    DO_IF_CHECKER(sval.subtype=0);				 \
+	    sval.u.string=PIKE_CONCAT(make_shared_binary_string,	 \
+				      INPUT_SHIFT)(input+e,eye-e);	 \
+	  }								 \
 	  break;							 \
 									 \
         case 'O':							 \
@@ -1214,8 +1282,9 @@ CHAROPT2(								 \
 									 \
     if(no_assign)							 \
     {									 \
-      free_svalue(&sval);						 \
-    }else{								 \
+      if (no_assign == 1)						 \
+	free_svalue(&sval);						 \
+    } else {								 \
       check_stack(1);							 \
       *sp++=sval;							 \
       dmalloc_touch_svalue(Pike_sp-1);					 \
@@ -1229,36 +1298,37 @@ CHAROPT2(								 \
 
 
 /* Confusing? Yes - Hubbe */
+/* Now slightly less confusing macro names, at least. /mast */
 
-/* CHAROPT(X) is X if the match set is wide.
- * CHAROPT2(X) is X if the input is wide.
+/* MATCH_IS_WIDE(X) is X if the match set is wide.
+ * INPUT_IS_WIDE(X) is X if the input is wide.
  */
-#define CHAROPT(X)
-#define CHAROPT2(X)
+#define MATCH_IS_WIDE(X)
+#define INPUT_IS_WIDE(X)
 
 MKREADSET(0)
 MK_VERY_LOW_SSCANF(0,0)
 
-#undef CHAROPT2
-#define CHAROPT2(X) X
+#undef INPUT_IS_WIDE
+#define INPUT_IS_WIDE(X) X
 
 MK_VERY_LOW_SSCANF(1,0)
 MK_VERY_LOW_SSCANF(2,0)
 
-#undef CHAROPT
-#define CHAROPT(X) X
+#undef MATCH_IS_WIDE
+#define MATCH_IS_WIDE(X) X
 
 MKREADSET(1)
 MKREADSET(2)
 
-#undef CHAROPT2
-#define CHAROPT2(X)
+#undef INPUT_IS_WIDE
+#define INPUT_IS_WIDE(X)
 
 MK_VERY_LOW_SSCANF(0,1)
 MK_VERY_LOW_SSCANF(0,2)
 
-#undef CHAROPT2
-#define CHAROPT2(X) X
+#undef INPUT_IS_WIDE
+#define INPUT_IS_WIDE(X) X
 
 MK_VERY_LOW_SSCANF(1,1)
 MK_VERY_LOW_SSCANF(2,1)
@@ -1266,7 +1336,7 @@ MK_VERY_LOW_SSCANF(1,2)
 MK_VERY_LOW_SSCANF(2,2)
 
 /* Simplified interface to very_low_sscanf_{0,1,2}_{0,1,2}(). */
-INT32 low_sscanf(struct pike_string *data, struct pike_string *format)
+INT32 low_sscanf(struct pike_string *data, struct pike_string *format, INT32 flags)
 {
   ptrdiff_t matched_chars;
   int x;
@@ -1275,64 +1345,66 @@ INT32 low_sscanf(struct pike_string *data, struct pike_string *format)
   check_c_stack(sizeof(struct sscanf_set)*2 + 512);
 
   switch(data->size_shift*3 + format->size_shift) {
+  default:
+#ifdef PIKE_DEBUG
+    Pike_fatal("Unsupported shift-combination to low_sscanf(): %d:%d\n",
+	       data->size_shift, format->size_shift);
+    break;
+#endif
     /* input_shift : match_shift */
   case 0:
     /*      0      :      0 */
     i = very_low_sscanf_0_0(STR0(data), data->len,
 			    STR0(format), format->len,
-			    &matched_chars, &x);
+			    &matched_chars, &x, flags);
     break;
   case 1:
     /*      0      :      1 */
     i = very_low_sscanf_0_1(STR0(data), data->len,
 			    STR1(format), format->len,
-			    &matched_chars, &x);
+			    &matched_chars, &x, flags);
     break;
   case 2:
     /*      0      :      2 */
     i = very_low_sscanf_0_2(STR0(data), data->len,
 			    STR2(format), format->len,
-			    &matched_chars, &x);
+			    &matched_chars, &x, flags);
     break;
   case 3:
     /*      1      :      0 */
     i = very_low_sscanf_1_0(STR1(data), data->len,
 			    STR0(format), format->len,
-			    &matched_chars, &x);
+			    &matched_chars, &x, flags);
     break;
   case 4:
     /*      1      :      1 */
     i = very_low_sscanf_1_1(STR1(data), data->len,
 			    STR1(format), format->len,
-			    &matched_chars, &x);
+			    &matched_chars, &x, flags);
     break;
   case 5:
     /*      1      :      2 */
     i = very_low_sscanf_1_2(STR1(data), data->len,
 			    STR2(format), format->len,
-			    &matched_chars, &x);
+			    &matched_chars, &x, flags);
     break;
   case 6:
     /*      2      :      0 */
     i = very_low_sscanf_2_0(STR2(data), data->len,
 			    STR0(format), format->len,
-			    &matched_chars, &x);
+			    &matched_chars, &x, flags);
     break;
   case 7:
     /*      2      :      1 */
     i = very_low_sscanf_2_1(STR2(data), data->len,
 			    STR1(format), format->len,
-			    &matched_chars, &x);
+			    &matched_chars, &x, flags);
     break;
   case 8:
     /*      2      :      2 */
     i = very_low_sscanf_2_2(STR2(data), data->len,
 			    STR2(format), format->len,
-			    &matched_chars, &x);
-    break;
-  default:
-    Pike_fatal("Unsupported shift-combination to low_sscanf(): %d:%d\n",
-	       data->size_shift, format->size_shift);
+			    &matched_chars, &x, flags);
     break;
   }
   return i;
@@ -1396,15 +1468,32 @@ INT32 low_sscanf(struct pike_string *data, struct pike_string *format)
  *!     characters. If followed by a %[], %s will only read characters not
  *!     present in the set. If followed by normal text, %s will match all
  *!     characters up to but not including the first occurrence of that text.
+ *!   @value "%H"
+ *!     Reads a Hollerith-encoded string, i.e. first reads the length
+ *!     of the string and then that number of characters. The size and
+ *!     byte order of the length descriptor can be modified in the
+ *!     same way as @tt{%c@}. As an example @expr{"%2H"@} first reads
+ *!     @expr{"%2c"@} and then the resulting number of characters.
  *!   @value "%[set]"
  *!     Matches a string containing a given set of characters (those given
- *!     inside the brackets). %[^set] means any character except those inside
- *!     brackets. Ranges of characters can be defined by using a minus
- *!     character between the first and the last character to be included in
- *!     the range. Example: %[0-9H] means any number or 'H'. Note that sets
- *!     that includes the character - must have it first in the brackets to
- *!     avoid having a range defined. Sets including the character ']' must
- *!     list this first (even before -) too, for natural reasons.
+ *!     inside the brackets). Ranges of characters can be defined by using
+ *!     a minus character between the first and the last character to be
+ *!     included in the range. Example: %[0-9H] means any number or 'H'.
+ *!     Note that sets that includes the character '-' must have it first
+ *!     (not possible in complemented sets, see below) or last in the brackets
+ *!     to avoid having a range defined. Sets including the character ']' must
+ *!     list this first too. If both '-' and ']' should be included
+ *!     then put ']' first and '-' last.  It is not possible to make a range
+ *!     that ends with ']'; make the range end with '\' instead and put ']'
+ *!     at the beginning of the set.  Likewise it is generally not possible
+ *!     to have a range start with '-'; make the range start with '.' instead
+ *!     and put '-' at the end of the set.  If the first character after the
+ *!     [ bracket is '^' (%[^set]), and this character does not begin a
+ *!     range, it means that the set is complemented, which is to say that
+ *!     any character except those inside brackets is matched.  To include '-'
+ *!     in a complemented set, it must be put last, not first.  To include '^'
+ *!     in a non-complemented set, it can be put anywhere but first, or be
+ *!     specified as a range ("^-^").
  *!   @value "%{format%}"
  *!     Repeatedly matches 'format' as many times as possible and assigns an
  *!     array of arrays with the results to the lvalue.
@@ -1477,7 +1566,7 @@ INT32 low_sscanf(struct pike_string *data, struct pike_string *format)
  *!   @[sprintf], @[array_sscanf]
  *    @[parse_format]
  */
-void o_sscanf(INT32 args)
+void o_sscanf(INT32 args, INT32 flags)
 {
   INT32 i=0;
   int x;
@@ -1489,7 +1578,7 @@ void o_sscanf(INT32 args)
   if(sp[1-args].type != T_STRING)
     SIMPLE_BAD_ARG_ERROR("sscanf", 2, "string");
 
-  i = low_sscanf(sp[-args].u.string, sp[1-args].u.string);
+  i = low_sscanf(sp[-args].u.string, sp[1-args].u.string, flags);
 
   if(sp-save_sp > args/2-1)
     Pike_error("Too few arguments for sscanf format.\n");
@@ -1530,9 +1619,418 @@ PMOD_EXPORT void f_sscanf(INT32 args)
 
   check_all_args("array_sscanf",args,BIT_STRING, BIT_STRING,0);
 
-  i = low_sscanf(sp[-args].u.string, sp[1-args].u.string);
+  i = low_sscanf(sp[-args].u.string, sp[1-args].u.string, 0);
 
   a = aggregate_array(DO_NOT_WARN(sp - save_sp));
   pop_n_elems(args);
   push_array(a);
 }
+
+void f_sscanf_76(INT32 args)
+{
+  INT32 i;
+  struct svalue *save_sp=sp;
+  struct array *a;
+
+  check_all_args("array_sscanf",args,BIT_STRING, BIT_STRING,0);
+
+  i = low_sscanf(sp[-args].u.string, sp[1-args].u.string, SSCANF_FLAG_76_COMPAT);
+
+  a = aggregate_array(DO_NOT_WARN(sp - save_sp));
+  pop_n_elems(args);
+  push_array(a);
+}
+
+static void push_sscanf_argument_types(PCHARP format, ptrdiff_t format_len,
+				       int cnt, int flags)
+{
+  for(; cnt < format_len; cnt++)
+  {
+    int no_assign=0;
+
+    while((cnt<format_len) && (INDEX_PCHARP(format, cnt) != '%'))
+      cnt++;
+    cnt++;
+    if (cnt >= format_len) break;
+
+    while(1)
+    {
+      switch(INDEX_PCHARP(format, cnt))
+      {
+      case '%':
+	break;
+
+      case '*':
+	no_assign=1;
+	/* FALL_THROUGH */
+
+      case '-':
+      case '+':
+      case '0': case '1': case '2': case '3': case '4':
+      case '5': case '6': case '7': case '8': case '9':
+	cnt++;
+	if(cnt>=format_len) {
+	  yyreport(REPORT_ERROR, type_check_system_string,
+		   0, "Error in sscanf format string.");
+	  break;
+	}
+	continue;
+
+	case '{':
+	{
+	  int e;
+	  int depth = 1;
+	  for(e=cnt+1;depth;e++)
+	  {
+	    if(e>=format_len)
+	    {
+	      yyreport(REPORT_ERROR, type_check_system_string,
+		       0, "Missing %%} in format string.");
+	      break;
+	    }
+	    if(INDEX_PCHARP(format, e)=='%')
+	    {
+	      switch(INDEX_PCHARP(format, e+1))
+	      {
+		case '%': e++; break;
+		case '}': depth--; break;
+		case '{': depth++; break;
+	      }
+	    }
+	  }
+	  if (!no_assign) {
+	    type_stack_mark();
+	    push_sscanf_argument_types(format, e, cnt+1, flags);
+	    if (!(depth = pop_stack_mark())) {
+	      push_type(PIKE_T_ZERO);
+	    } else {
+	      /* Join the argument types. */
+	      while (depth > 1) {
+		push_type(T_OR);
+		depth--;
+	      }
+	    }
+	    push_type(PIKE_T_ARRAY);
+	    push_type(PIKE_T_ARRAY);
+	  }
+	  cnt = e;
+	  break;
+	}
+
+	case 'n':
+        case 'b':
+        case 'o':
+        case 'd':
+        case 'x':
+        case 'D':
+        case 'i':
+	case 'c':
+	  if (!no_assign)
+	    push_finished_type(int_type_string);
+	  break;
+
+	case '[':
+	  {
+	    int ch;
+	    cnt++;
+	    if (cnt >= format_len) {
+	      yyreport(REPORT_ERROR, type_check_system_string,
+		       0, "Error in sscanf format string.");
+	      break;
+	    }
+	    if((INDEX_PCHARP(format, cnt)=='^') &&
+	       (cnt+2>=format_len ||
+		(INDEX_PCHARP(format, cnt+1)!='-') ||
+		(INDEX_PCHARP(format, cnt+2)==']') ||
+		(flags & SSCANF_FLAG_76_COMPAT)))
+	    {
+	      cnt++;
+	      if(cnt >= format_len) {
+		yyreport(REPORT_ERROR, type_check_system_string,
+			 0, "Error in sscanf format string.");
+		break;
+	      }
+	    }
+
+	    if(((ch = INDEX_PCHARP(format, cnt))==']') || (ch=='-'))
+	    {
+	      cnt++;
+	      if(cnt >= format_len) {
+		yyreport(REPORT_ERROR, type_check_system_string,
+			 0, "Error in sscanf format string.");
+		break;
+	      }
+	      ch = INDEX_PCHARP(format, cnt);
+	    }
+
+	    while(ch != ']')
+	    {
+	      if(ch == '-')
+	      {
+		cnt++;
+		if(cnt >= format_len) {
+		  yyreport(REPORT_ERROR, type_check_system_string,
+			   0, "Error in sscanf format string.");
+		  break;
+		}
+
+		if(INDEX_PCHARP(format, cnt)==']')
+		{
+		  break;
+		}
+	      }
+	      cnt++;
+	      if(cnt>=format_len) {
+		yyreport(REPORT_ERROR, type_check_system_string,
+			 0, "Error in sscanf format string.");
+		break;
+	      }
+	      ch = INDEX_PCHARP(format, cnt);
+	    }
+	  }
+	  /* FALL_THROUGH */
+      case 's':
+      case 'H':
+	if (!no_assign)
+	  push_finished_type(string_type_string);
+	break;
+
+	case 'F':
+        case 'f':
+	  if (!no_assign)
+	    push_finished_type(float_type_string);
+	  break;
+
+	  break;
+
+        case 'O':
+	  if (!no_assign)
+	    push_finished_type(mixed_type_string);
+	  break;
+
+	default:
+	  yyreport(REPORT_ERROR, type_check_system_string,
+		   0, "Unknown sscanf token %%%c(0x%02x) at offset %d.",
+		   INDEX_PCHARP(format, cnt), INDEX_PCHARP(format, cnt),
+		   cnt-1);
+	  break;
+      }
+      break;
+    }
+  }
+}
+
+void f___handle_sscanf_format(INT32 args)
+{
+  struct pike_type *res;
+  struct pike_type *tmp;
+  struct pike_string *attr;
+  struct pike_string *fmt;
+  struct pike_string *sscanf_format_string;
+  struct pike_string *sscanf_76_format_string;
+  int flags = 0;
+  int found = 0;
+  int fmt_count;
+
+#if 0
+  fprintf(stderr, "__handle_sprintf_format()\n");
+#endif /* 0 */
+  if (args != 4)
+    SIMPLE_WRONG_NUM_ARGS_ERROR("__handle_sscanf_format", 4);
+  if (Pike_sp[-4].type != PIKE_T_STRING)
+    SIMPLE_ARG_TYPE_ERROR("__handle_sscanf_format", 1, "string");
+  if (Pike_sp[-3].type != PIKE_T_STRING)
+    SIMPLE_ARG_TYPE_ERROR("__handle_sscanf_format", 2, "string");
+  if (Pike_sp[-2].type != PIKE_T_TYPE)
+    SIMPLE_ARG_TYPE_ERROR("__handle_sscanf_format", 3, "type");
+  if (Pike_sp[-1].type != PIKE_T_TYPE)
+    SIMPLE_ARG_TYPE_ERROR("__handle_sscanf_format", 4, "type");
+
+  tmp = Pike_sp[-1].u.type;
+  if ((tmp->type != PIKE_T_FUNCTION) && (tmp->type != T_MANY)) {
+    SIMPLE_ARG_TYPE_ERROR("__handle_sscanf_format", 4, "type(function)");
+  }
+
+  MAKE_CONST_STRING(sscanf_format_string, "sscanf_format");  
+  MAKE_CONST_STRING(sscanf_76_format_string, "sscanf_76_format");  
+
+  if (Pike_sp[-4].u.string != sscanf_format_string) {
+    if (Pike_sp[-4].u.string != sscanf_76_format_string) {
+      pop_n_elems(args);
+      push_undefined();
+      return;
+    }
+    flags = SSCANF_FLAG_76_COMPAT;
+  }
+
+  fmt = Pike_sp[-3].u.string;
+  MAKE_CONST_STRING(attr, "sscanf_args");  
+
+#if 0
+  fprintf(stderr, "Checking sscanf format: \"%s\": ", fmt->str);
+  simple_describe_type(Pike_sp[-1].u.type);
+  fprintf(stderr, "\n");
+#endif /* 0 */
+
+  type_stack_mark();
+  type_stack_mark();
+  for (; tmp; tmp = tmp->cdr) {
+    struct pike_type *arg = tmp->car;
+    int array_cnt = 0;
+    while(arg) {
+      switch(arg->type) {
+      case PIKE_T_ATTRIBUTE:
+	if (arg->car == (struct pike_type *)attr)
+	  break;
+	/* FALL_THROUGH */
+      case PIKE_T_NAME:
+	arg = arg->cdr;
+	continue;
+      case PIKE_T_ARRAY:
+	array_cnt++;
+	arg = arg->car;
+	continue;
+      default:
+	arg = NULL;
+	break;
+      }
+      break;
+    }
+    if (arg) {
+      type_stack_mark();
+      push_sscanf_argument_types(MKPCHARP(fmt->str, fmt->size_shift),
+				 fmt->len, 0, flags);
+      if (!array_cnt) {
+	pop_stack_mark();
+	push_type(T_VOID);	/* No more args */
+	while (tmp->type == PIKE_T_FUNCTION) {
+	  tmp = tmp->cdr;
+	}
+	push_finished_type(tmp->cdr);	/* return type */
+	push_reverse_type(T_MANY);
+	fmt_count = pop_stack_mark();
+	while (fmt_count > 1) {
+	  push_reverse_type(T_FUNCTION);
+	  fmt_count--;
+	}
+	res = pop_unfinished_type();
+	pop_n_elems(args);
+	push_type_value(res);
+#if 0
+	fprintf(stderr, " ==> ");
+	simple_describe_type(res);
+	fprintf(stderr, "\n");
+#endif /* 0 */
+	return;
+      } else {
+	if (!(fmt_count = pop_stack_mark())) {
+	  push_type(PIKE_T_ZERO);
+	} else {
+	  /* Join the argument types into the array. */
+	  while (fmt_count > 1) {
+	    push_type(T_OR);
+	    fmt_count--;
+	  }
+	}
+	while (array_cnt--) {
+	  push_type(PIKE_T_ARRAY);
+	}
+	found = 1;
+      }
+    } else {
+      push_finished_type(tmp->car);
+    }
+    if (tmp->type == T_MANY) {
+      /* Get the return type. */
+      tmp = tmp->cdr;
+      break;
+    }
+  }
+  if (found) {
+    /* Found, but inside an array, so we need to build the function
+     * type here.
+     */
+    push_finished_type(tmp);	/* Return type. */
+    push_reverse_type(T_MANY);
+    fmt_count = pop_stack_mark();
+    while (fmt_count > 1) {
+      push_reverse_type(T_FUNCTION);
+      fmt_count--;
+    }
+    res = pop_unfinished_type();
+    pop_n_elems(args);
+    push_type_value(res);
+#if 0
+    fprintf(stderr, " ==> ");
+    simple_describe_type(res);
+    fprintf(stderr, "\n");
+#endif /* 0 */
+    return;
+  } else if (tmp) {
+    /* Check if it's in the return type. */
+    struct pike_type *arg = tmp;
+    int array_cnt = 0;
+    while(arg) {
+      switch(arg->type) {
+      case PIKE_T_ATTRIBUTE:
+	if (arg->car == (struct pike_type *)attr)
+	  break;
+	/* FALL_THROUGH */
+      case PIKE_T_NAME:
+	arg = arg->cdr;
+	continue;
+      case PIKE_T_ARRAY:
+	array_cnt++;
+	arg = arg->car;
+	continue;
+      default:
+	arg = NULL;
+	break;
+      }
+      break;
+    }
+    if (arg) {
+      type_stack_mark();
+      push_sscanf_argument_types(MKPCHARP(fmt->str, fmt->size_shift),
+				 fmt->len, 0, flags);
+      /* Join the argument types. */
+      if (!(fmt_count = pop_stack_mark())) {
+	push_type(PIKE_T_ZERO);
+      } else while (fmt_count > 2) {
+	push_type(T_OR);
+	fmt_count--;
+      }
+      while (array_cnt--) {
+	push_type(PIKE_T_ARRAY);
+      }
+
+      /* Rebuild the function type backwards. */
+      push_reverse_type(T_MANY);
+      fmt_count = pop_stack_mark();
+      while (fmt_count > 1) {
+	push_reverse_type(T_FUNCTION);
+	fmt_count--;
+      }
+      res = pop_unfinished_type();
+      pop_n_elems(args);
+      push_type_value(res);
+#if 0
+      fprintf(stderr, " ==> ");
+      simple_describe_type(res);
+      fprintf(stderr, "\n");
+#endif /* 0 */
+      return;
+    }
+  }
+
+  /* No marker found. */
+#if 0
+  fprintf(stderr, " ==> No marker found.\n");
+#endif /* 0 */
+  pop_stack_mark();
+  type_stack_pop_to_mark();
+  pop_n_elems(args);
+  push_undefined();
+}
+
+
