@@ -2,12 +2,13 @@
 || This file is part of Pike. For copyright information see COPYRIGHT.
 || Pike is distributed under GPL, LGPL and MPL. See the file COPYING
 || for more information.
-|| $Id: stralloc.h,v 1.82 2004/09/27 21:37:13 mast Exp $
+|| $Id: stralloc.h,v 1.111 2008/10/06 17:00:00 mast Exp $
 */
 
 #ifndef STRALLOC_H
 #define STRALLOC_H
 #include "global.h"
+
 #include "pike_macros.h"
 #include "block_alloc_h.h"
 
@@ -19,7 +20,8 @@
 
 #define PIKE_STRING_CONTENTS						\
   INT32 refs;								\
-  INT32 size_shift; /* 30 bit waste, but good for alignment... */	\
+  INT16 flags;								\
+  INT16 size_shift; /* 14 bit waste, but good for alignment... */	\
   ptrdiff_t len; /* Not counting terminating NUL. */			\
   size_t hval;								\
   struct pike_string *next 
@@ -34,9 +36,42 @@ struct string_builder
 {
   struct pike_string *s;
   ptrdiff_t malloced;
-  size_t known_shift;
+  INT32 known_shift;
 };
 
+/* Flags used in pike_string->flags. */
+#define STRING_NOT_HASHED	1	/* Hash value is invalid. */
+#define STRING_NOT_SHARED	2	/* String not shared. */
+#define STRING_IS_SHORT		4	/* String is blockalloced. */
+#define STRING_CLEAR_ON_EXIT    8       /* Overwrite before free. */
+
+/* Flags used by string_builder_append_integer() */
+#define APPEND_SIGNED		1	/* Value is signed */
+/* Note: The follwoing are NOT true flags. */
+#define APPEND_WIDTH_HALF	2	/* h-flag. */
+#define APPEND_WIDTH_LONG	4	/* l-flag. */
+#define APPEND_WIDTH_LONG_LONG	6	/* ll-flag. */
+#define APPEND_WIDTH_MASK	6	/* Corresponding mask. */
+/* More real flags here. */
+#define APPEND_POSITIVE		8	/* Sign positive values too. */
+#define APPEND_UPPER_CASE	16	/* Use upper case hex. */
+#define APPEND_ZERO_PAD		32	/* Zero pad. */
+#define APPEND_LEFT		64	/* Left align. */
+
+#if SIZEOF_CHAR_P == SIZEOF_INT
+#define APPEND_WIDTH_PTR	0
+#elif SIZEOF_CHAR_P == SIZEOF_LONG
+#define APPEND_WIDTH_PTR	APPEND_WIDTH_LONG
+#elif SIZEOF_CHAR_P == SIZEOF_LONG_LONG
+#define APPEND_WIDTH_PTR	APPEND_WIDTH_LONG_LONG
+#else
+#error "Unknown way to read pointer-wide integers."
+#endif
+
+/* Flags used by string_builder_quote_string() */
+#define QUOTE_NO_STRING_CONCAT	1	/* Don't use string concat in output */
+#define QUOTE_BREAK_AT_LF	2	/* Break after linefeed */
+
 #ifdef PIKE_DEBUG
 PMOD_EXPORT struct pike_string *debug_findstring(const struct pike_string *foo);
 #endif
@@ -51,6 +86,9 @@ PMOD_EXPORT struct pike_string *debug_findstring(const struct pike_string *foo);
 #define my_hash_string(X) PTR_TO_INT(X)
 #define is_same_string(X,Y) ((X)==(Y))
 
+/* NB: This intentionally only works for narrow strings. */
+#define string_has_null(X) (strlen((X)->str)!=(size_t)(X)->len)
+
 #ifdef PIKE_DEBUG
 #define STR0(X) ((p_wchar0 *)debug_check_size_shift((X),0)->str)
 #define STR1(X) ((p_wchar1 *)debug_check_size_shift((X),1)->str)
@@ -65,7 +103,10 @@ PMOD_EXPORT struct pike_string *debug_findstring(const struct pike_string *foo);
   ((SHIFT)==0?((p_wchar0 *)(PTR))[(IND)]:(SHIFT)==1?((p_wchar1 *)(PTR))[(IND)]:((p_wchar2 *)(PTR))[(IND)])
 
 #define SET_INDEX_CHARP(PTR,IND,SHIFT,VAL) \
-  ((SHIFT)==0?(((p_wchar0 *)(PTR))[(IND)]=(VAL)):(SHIFT)==1?(((p_wchar1 *)(PTR))[(IND)]=(VAL)):(((p_wchar2 *)(PTR))[(IND)]=(VAL)))
+  ((SHIFT)==0?								\
+   (((p_wchar0 *)(PTR))[(IND)] = DO_NOT_WARN ((p_wchar0) (VAL))):	\
+   (SHIFT)==1?(((p_wchar1 *)(PTR))[(IND)] = DO_NOT_WARN ((p_wchar1) (VAL))): \
+   (((p_wchar2 *)(PTR))[(IND)] = DO_NOT_WARN ((p_wchar2) (VAL))))
 
 
 #define EXTRACT_CHARP(PTR,SHIFT) INDEX_CHARP((PTR),0,(SHIFT))
@@ -89,7 +130,7 @@ PMOD_EXPORT struct pike_string *debug_findstring(const struct pike_string *foo);
 
 
 
-static INLINE PCHARP MKPCHARP(void *ptr, int shift)
+static INLINE PCHARP MKPCHARP(const void *ptr, int shift)
 {
   PCHARP tmp;
   tmp.ptr=(p_wchar0 *)ptr;
@@ -114,7 +155,7 @@ struct shared_string_location
   struct shared_string_location *next;
 };
 
-extern struct shared_string_location *all_shared_string_locations;
+PMOD_EXPORT extern struct shared_string_location *all_shared_string_locations;
 
 #define MAKE_CONST_STRING(var, text) do {		\
   static struct shared_string_location str_;		\
@@ -126,6 +167,22 @@ extern struct shared_string_location *all_shared_string_locations;
   var = str_.s;						\
 }while(0)
 
+#define MAKE_CONST_STRING_CODE(var, code) do {				\
+    static struct shared_string_location str_;				\
+    if (!str_.s) {							\
+      {code;}								\
+      str_.s = var;							\
+      DO_IF_DEBUG (							\
+	if (!str_.s)							\
+	  Pike_fatal ("Code at " __FILE__ ":" DEFINETOSTR (__LINE__)	\
+		      " failed to produce a string in " #var ".\n");	\
+      );								\
+      str_.next=all_shared_string_locations;				\
+      all_shared_string_locations=&str_;				\
+    }									\
+    else var = str_.s;							\
+  } while (0)
+
 #else
 
 #define MAKE_CONST_STRING(var, text)						\
@@ -134,6 +191,20 @@ extern struct shared_string_location *all_shared_string_locations;
     var = str_;									\
  }while(0)
 
+#define MAKE_CONST_STRING_CODE(var, code) do {				\
+    static struct pike_string *str_;					\
+    if (!str_) {							\
+      {code;}								\
+      str_ = var;							\
+      DO_IF_DEBUG (							\
+	if (!str_)							\
+	  Pike_fatal ("Code at " __FILE__ ":" DEFINETOSTR (__LINE__)	\
+		      " failed to produce a string in " #var ".\n");	\
+      );								\
+    }									\
+    else var = str_;							\
+  } while (0)
+
 #endif
 
 #define REF_MAKE_CONST_STRING(var, text) do {				\
@@ -141,6 +212,11 @@ extern struct shared_string_location *all_shared_string_locations;
     reference_shared_string(var);					\
   } while (0)
 
+#define REF_MAKE_CONST_STRING_CODE(var, code) do {			\
+    MAKE_CONST_STRING_CODE(var, code);					\
+    reference_shared_string(var);					\
+  } while (0)
+
 /* Compatibility. */
 #define MAKE_CONSTANT_SHARED_STRING(var, text)				\
   REF_MAKE_CONST_STRING(var, text)
@@ -160,8 +236,8 @@ INT32 PIKE_CONCAT4(compare_,FROM,_to_,TO)(const PIKE_CONCAT(p_wchar,TO) *to, con
 PMOD_EXPORT extern struct pike_string *empty_pike_string;
 
 /* Prototypes begin here */
-PMOD_EXPORT unsigned INT32 index_shared_string(struct pike_string *s,
-					       ptrdiff_t pos);
+PMOD_EXPORT p_wchar2 index_shared_string(struct pike_string *s,
+					 ptrdiff_t pos);
 PMOD_EXPORT void low_set_index(struct pike_string *s, ptrdiff_t pos,
 			       int value);
 PMOD_EXPORT struct pike_string *debug_check_size_shift(struct pike_string *a,int shift);
@@ -182,9 +258,9 @@ PMOD_EXPORT struct pike_string *findstring(const char *foo);
 struct short_pike_string0;
 struct short_pike_string1;
 struct short_pike_string2;
-BLOCK_ALLOC(short_pike_string0, SHORT_STRING_BLOCK)
-BLOCK_ALLOC(short_pike_string1, SHORT_STRING_BLOCK)
-BLOCK_ALLOC(short_pike_string2, SHORT_STRING_BLOCK)
+BLOCK_ALLOC(short_pike_string0, SHORT_STRING_BLOCK);
+BLOCK_ALLOC(short_pike_string1, SHORT_STRING_BLOCK);
+BLOCK_ALLOC(short_pike_string2, SHORT_STRING_BLOCK);
 
 
 
@@ -207,7 +283,7 @@ PMOD_EXPORT struct pike_string *debug_make_shared_string2(const p_wchar2 *str);
 PMOD_EXPORT void unlink_pike_string(struct pike_string *s);
 PMOD_EXPORT void do_free_string(struct pike_string *s);
 PMOD_EXPORT void do_really_free_string(struct pike_string *s);
-PMOD_EXPORT void do_really_free_pike_string(struct pike_string *s);
+PMOD_EXPORT void do_free_unlinked_pike_string(struct pike_string *s);
 PMOD_EXPORT void really_free_string(struct pike_string *s);
 PMOD_EXPORT void debug_free_string(struct pike_string *s);
 struct pike_string *add_string_status(int verbose);
@@ -223,6 +299,10 @@ PMOD_EXPORT ptrdiff_t generic_quick_binary_strcmp(const char *a,
 						  ptrdiff_t alen, int asize,
 						  const char *b,
 						  ptrdiff_t blen, int bsize);
+PMOD_EXPORT ptrdiff_t generic_find_binary_prefix(const char *a,
+						 ptrdiff_t alen, int asize,
+						 const char *b,
+						 ptrdiff_t blen, int bsize);
 PMOD_EXPORT int c_compare_string(struct pike_string *s, char *foo, int len);
 PMOD_EXPORT ptrdiff_t my_quick_strcmp(struct pike_string *a,
 				      struct pike_string *b);
@@ -250,7 +330,9 @@ PMOD_EXPORT struct pike_string *string_replace(struct pike_string *str,
 				   struct pike_string *to);
 void init_shared_string_table(void);
 void cleanup_shared_string_table(void);
-void count_memory_in_strings(INT32 *num, INT32 *size);
+void count_memory_in_strings(size_t *num, size_t *size);
+PMOD_EXPORT void visit_string (struct pike_string *s, int action);
+void gc_mark_string_as_referenced (struct pike_string *s);
 unsigned gc_touch_all_strings(void);
 void gc_mark_all_strings(void);
 struct pike_string *next_pike_string (struct pike_string *s);
@@ -264,7 +346,12 @@ PMOD_EXPORT void string_build_mkspace(struct string_builder *s,
 				      ptrdiff_t chars, int mag);
 PMOD_EXPORT void *string_builder_allocate(struct string_builder *s, ptrdiff_t chars, int mag);
 PMOD_EXPORT void string_builder_putchar(struct string_builder *s, int ch);
-PMOD_EXPORT void string_builder_binary_strcat(struct string_builder *s, char *str, ptrdiff_t len);
+PMOD_EXPORT void string_builder_binary_strcat0(struct string_builder *s,
+					       const p_wchar0 *str, ptrdiff_t len);
+PMOD_EXPORT void string_builder_binary_strcat1(struct string_builder *s,
+					       const p_wchar1 *str, ptrdiff_t len);
+PMOD_EXPORT void string_builder_binary_strcat2(struct string_builder *s,
+					       const p_wchar2 *str, ptrdiff_t len);
 PMOD_EXPORT void string_builder_append(struct string_builder *s,
 				       PCHARP from,
 				       ptrdiff_t len);
@@ -275,6 +362,22 @@ PMOD_EXPORT void string_builder_fill(struct string_builder *s,
 				     ptrdiff_t offset);
 PMOD_EXPORT void string_builder_strcat(struct string_builder *s, char *str);
 PMOD_EXPORT void string_builder_shared_strcat(struct string_builder *s, struct pike_string *str);
+PMOD_EXPORT ptrdiff_t string_builder_quote_string(struct string_builder *buf,
+						  struct pike_string *str,
+						  ptrdiff_t start,
+						  ptrdiff_t max_len,
+						  int flags);
+PMOD_EXPORT void string_builder_append_integer(struct string_builder *s,
+					       LONGEST val,
+					       unsigned int base,
+					       int flags,
+					       size_t min_width,
+					       size_t precision);
+PMOD_EXPORT void string_builder_vsprintf(struct string_builder *s,
+					 const char *fmt,
+					 va_list args);
+PMOD_EXPORT void string_builder_sprintf(struct string_builder *s,
+					const char *fmt, ...);
 PMOD_EXPORT void reset_string_builder(struct string_builder *s);
 PMOD_EXPORT void free_string_builder(struct string_builder *s);
 PMOD_EXPORT struct pike_string *finish_string_builder(struct string_builder *s);
@@ -291,6 +394,12 @@ PMOD_EXPORT int wide_string_to_svalue_inumber(struct svalue *r,
 					      int base,
 					      ptrdiff_t maxlength,
 					      int shift);
+int safe_wide_string_to_svalue_inumber(struct svalue *r,
+				       void * str,
+				       void *ptr,
+				       int base,
+				       ptrdiff_t maxlength,
+				       int shift);
 PMOD_EXPORT int pcharp_to_svalue_inumber(struct svalue *r,
 					 PCHARP str,
 					 PCHARP *ptr,
@@ -306,8 +415,21 @@ PMOD_EXPORT p_wchar2 *require_wstring2(struct pike_string *s,
 			   char **to_free);
 /* Prototypes end here */
 
+/* Compat alias. */
+#define do_really_free_pike_string do_free_unlinked_pike_string
+
+static INLINE void string_builder_binary_strcat(struct string_builder *s,
+						const char *str, ptrdiff_t len)
+{
+  string_builder_binary_strcat0 (s, (const p_wchar0 *) str, len);
+}
+
 #define ISCONSTSTR(X,Y) c_compare_string((X),Y,sizeof(Y)-sizeof(""))
 
+#define visit_string_ref(S, REF_TYPE)				\
+  visit_ref (pass_string (S), (REF_TYPE),			\
+	     (visit_thing_fn *) &visit_string, NULL)
+
 #ifdef DEBUG_MALLOC
 #define make_shared_string(X) \
  ((struct pike_string *)debug_malloc_pass(debug_make_shared_string(X)))
@@ -362,6 +484,7 @@ PMOD_EXPORT p_wchar2 *require_wstring2(struct pike_string *s,
 
 #undef CONVERT
 
-PMOD_PROTO extern void f_sprintf(INT32 num_arg);
+PMOD_EXPORT void f_sprintf(INT32 num_arg);
+void f___handle_sprintf_format(INT32 args);
 
 #endif /* STRALLOC_H */
