@@ -2,21 +2,20 @@
 || This file is part of Pike. For copyright information see COPYRIGHT.
 || Pike is distributed under GPL, LGPL and MPL. See the file COPYING
 || for more information.
-|| $Id: threads.c,v 1.241 2006/05/23 11:29:07 nilsson Exp $
+|| $Id: threads.c,v 1.280 2009/04/21 16:02:29 jonasw Exp $
 */
 
-#ifndef CONFIGURE_TEST
 #include "global.h"
-RCSID("$Id: threads.c,v 1.241 2006/05/23 11:29:07 nilsson Exp $");
 
 PMOD_EXPORT int num_threads = 1;
 PMOD_EXPORT int threads_disabled = 0;
-#endif	/* !CONFIGURE_TEST */
 
 /* #define PICKY_MUTEX */
 
 #ifdef _REENTRANT
 
+#include "pike_error.h"
+
 #ifndef CONFIGURE_TEST
 
 /* #define VERBOSE_THREADS_DEBUG */
@@ -38,6 +37,7 @@ PMOD_EXPORT int threads_disabled = 0;
 #include "operators.h"
 #include "bignum.h"
 #include "signal_handler.h"
+#include "backend.h"
 #include "pike_rusage.h"
 
 #include <errno.h>
@@ -50,25 +50,33 @@ PMOD_EXPORT int threads_disabled = 0;
 #include <sys/time.h>
 #endif
 
-#ifndef PIKE_THREAD_C_STACK_SIZE
-#define PIKE_THREAD_C_STACK_SIZE (256 * 1024)
-#endif
-
-PMOD_EXPORT int live_threads = 0, disallow_live_threads = 0;
+PMOD_EXPORT int live_threads = 0;
 PMOD_EXPORT COND_T live_threads_change;
 PMOD_EXPORT COND_T threads_disabled_change;
-PMOD_EXPORT size_t thread_stack_size=PIKE_THREAD_C_STACK_SIZE;
 
 #else
 #include "pike_threadlib.h"
 #endif	/* !CONFIGURE_TEST */
 
+#ifndef PIKE_THREAD_C_STACK_SIZE
+#define PIKE_THREAD_C_STACK_SIZE (256 * 1024)
+#endif
+
+PMOD_EXPORT size_t thread_stack_size=PIKE_THREAD_C_STACK_SIZE;
+
 PMOD_EXPORT void thread_low_error (int errcode, const char *cmd,
 				   const char *fname, int lineno)
 {
+#ifdef CONFIGURE_TEST
+  fprintf (stderr, "%s:%d: %s\n"
+	   "Unexpected error from thread function: %d\n",
+	   fname, lineno, cmd, errcode);
+  abort();
+#else
   Pike_fatal ("%s:%d: %s\n"
 	      "Unexpected error from thread function: %d\n",
 	      fname, lineno, cmd, errcode);
+#endif
 }
 
 /* SCO magic... */
@@ -164,6 +172,49 @@ PMOD_EXPORT int co_wait(COND_T *c, MUTEX_T *m)
   return 0;
 }
 
+PMOD_EXPORT int co_wait_timeout(COND_T *c, MUTEX_T *m, int s, int nanos)
+{
+  struct cond_t_queue me;
+  event_init(&me.event);
+  me.next=0;
+  mt_lock(& c->lock);
+
+  if(c->tail)
+  {
+    c->tail->next=&me;
+    c->tail=&me;
+  }else{
+    c->head=c->tail=&me;
+  }
+
+  mt_unlock(& c->lock);
+  mt_unlock(m);
+  if (s || nanos) {
+    DWORD msec;
+    /* Check for overflow (0xffffffff/1000). */
+    if (s >= 4294967) {
+      msec = INFINITE;
+    } else {
+      msec = s*1000 + nanos/1000000;
+      if (!msec) msec = 1;	/* Underflow. */
+    }
+    event_wait_msec(&me.event, msec);
+  } else {
+    event_wait(&me.event);
+  }
+  mt_lock(m);
+
+  event_destroy(& me.event);
+  /* Cancellation point?? */
+
+#ifdef PIKE_DEBUG
+  if(me.next)
+    Pike_fatal("Wait on event return prematurely!!\n");
+#endif
+
+  return 0;
+}
+
 PMOD_EXPORT int co_signal(COND_T *c)
 {
   struct cond_t_queue *t;
@@ -209,7 +260,32 @@ PMOD_EXPORT int co_destroy(COND_T *c)
   return 0;
 }
 
-#endif
+#else /* !SIMULATE_COND_WITH_EVENT */
+
+#ifndef CONFIGURE_TEST
+PMOD_EXPORT int co_wait_timeout(COND_T *c, PIKE_MUTEX_T *m, int s, int nanos)
+{
+#ifdef POSIX_THREADS
+  struct timespec timeout;
+#ifdef HAVE_PTHREAD_COND_RELTIMEDWAIT_NP
+  /* Solaris extension: relative timeout. */
+  timeout.tv_sec = s;
+  timeout.tv_nsec = nanos;
+  return pthread_cond_reltimedwait_np(c, m, &timeout);
+#else /* !HAVE_PTHREAD_COND_RELTIMEDWAIT_NP */
+  /* Absolute timeout. */
+  GETTIMEOFDAY(&current_time);
+  timeout.tv_sec = current_time.tv_sec + s;
+  timeout.tv_nsec = current_time.tv_usec * 1000 + nanos;
+  return pthread_cond_timedwait(c, m, &timeout);
+#endif /* HAVE_PTHREAD_COND_RELTIMEDWAIT_NP */
+#else /* !POSIX_THREADS */
+#error co_wait_timeout does not support this thread model.
+#endif /* POSIX_THREADS */
+}
+#endif /* !CONFIGURE_TEST */
+
+#endif /* SIMULATE_COND_WITH_EVENT */
 
 #ifdef POSIX_THREADS
 pthread_attr_t pattr;
@@ -246,6 +322,8 @@ static void really_low_th_init(void)
 #endif
 }
 
+static void cleanup_thread_state (struct thread_state *th);
+
 #ifndef CONFIGURE_TEST
 
 #define THIS_THREAD ((struct thread_state *)CURRENT_STORAGE)
@@ -269,6 +347,7 @@ struct program *thread_local_prog = 0;
 PMOD_EXPORT ptrdiff_t thread_storage_offset;
 #ifdef USE_CLOCK_FOR_SLICES
 PMOD_EXPORT clock_t thread_start_clock = 0;
+PMOD_EXPORT THREAD_T last_clocked_thread = 0;
 #endif
 
 #ifdef PIKE_DEBUG
@@ -277,7 +356,7 @@ PMOD_EXPORT const char msg_ip_not_locked[] =
 PMOD_EXPORT const char msg_ip_not_locked_this_thr[] =
   "Interpreter not locked by this thread.\n";
 PMOD_EXPORT const char msg_thr_swapped_over[] =
-  "Thread %08x swapped in over existing thread %08x.\n";
+  "Thread %"PRINTSIZET"x swapped in over existing thread %"PRINTSIZET"x.\n";
 PMOD_EXPORT const char msg_saved_thread_id[] =
   "Saved thread id: ";
 PMOD_EXPORT const char msg_swap_in_cur_thr_failed[] =
@@ -293,7 +372,8 @@ PMOD_EXPORT const char msg_thr_allow_in_gc[] =
   "Threads allowed during garbage collection (pass %d).\n";
 PMOD_EXPORT const char msg_thr_allow_in_disabled[] =
   "Threads allowed from a different thread "
-  "while threads are disabled.\n";
+  "while threads are disabled. "
+  "(self: %"PRINTSIZET"x, disabler: %"PRINTSIZET"x)\n";
 PMOD_EXPORT const char msg_global_dynbuf_in_use[] =
   "Threads allowed while the global dynamic buffer is in use.\n";
 #endif
@@ -390,7 +470,7 @@ void low_init_threads_disable(void)
  */
 void init_threads_disable(struct object *o)
 {
-  disallow_live_threads = 1;
+  low_init_threads_disable();
 
   if(live_threads) {
     SWAP_OUT_CURRENT_THREAD();
@@ -399,12 +479,11 @@ void init_threads_disable(struct object *o)
 		      (stderr,
 		       "_disable_threads(): Waiting for %d threads to finish\n",
 		       live_threads));
-      low_co_wait_interpreter(&live_threads_change);
+      _do_co_wait_interpreter(&live_threads_change);
     }
+    THREADS_FPRINTF(0, (stderr, "_disable_threads(): threads now disabled\n"));
     SWAP_IN_CURRENT_THREAD();
   }
-
-  low_init_threads_disable();
 }
 
 void exit_threads_disable(struct object *o)
@@ -417,9 +496,8 @@ void exit_threads_disable(struct object *o)
 
       /* Order shouldn't matter for unlock, so no need to do it backwards. */
       while(im) {
-	THREADS_FPRINTF(0,
-			(stderr,
-			 "exit_threads_disable(): Unlocking IM 0x%p\n", im));
+	THREADS_FPRINTF(0, (stderr,
+			    "exit_threads_disable(): Unlocking IM %p\n", im));
 	mt_unlock(&(im->lock));
 	im = im->next;
       }
@@ -427,7 +505,6 @@ void exit_threads_disable(struct object *o)
       mt_unlock(&interleave_lock);
 
       THREADS_FPRINTF(0, (stderr, "_exit_threads_disable(): Wake up!\n"));
-      disallow_live_threads = 0;
       co_broadcast(&threads_disabled_change);
 #ifdef PIKE_DEBUG
       threads_disabled_thread = 0;
@@ -449,8 +526,7 @@ void init_interleave_mutex(IMUTEX_T *im)
 
   init_threads_disable(NULL);
 
-  THREADS_FPRINTF(0, (stderr,
-		      "init_interleave_mutex(): Locking IM 0x%p\n", im));
+  THREADS_FPRINTF(0, (stderr, "init_interleave_mutex(): Locking IM %p\n", im));
 
   /* Lock it so that it can be unlocked by exit_threads_disable() */
   mt_lock(&(im->lock));
@@ -637,6 +713,78 @@ PMOD_EXPORT struct object *thread_for_id(THREAD_T tid)
      by incrementing refcount though. */
 }
 
+PMOD_EXPORT void call_with_interpreter(void (*func)(void *ctx), void *ctx)
+{
+  struct thread_state *state;
+
+  if((state = thread_state_for_id(th_self()))!=NULL) {
+    /* This is a pike thread.  Do we have the interpreter lock? */
+    if(!state->swapped) {
+      /* Yes.  Go for it... */
+#ifdef PIKE_DEBUG
+      /* Ensure that the thread isn't loose. */
+      int is_loose = state->debug_flags & THREAD_DEBUG_LOOSE;
+      state->debug_flags &= ~THREAD_DEBUG_LOOSE;
+#endif
+
+      func(ctx);
+
+#ifdef PIKE_DEBUG
+      /* Restore the looseness property of the thread. */
+      state->debug_flags |= is_loose;
+#endif
+    } else {
+      /* Nope, let's get it... */
+      mt_lock_interpreter();
+      SWAP_IN_THREAD(state);
+      DO_IF_DEBUG(state->debug_flags &= ~THREAD_DEBUG_LOOSE;)
+
+      func(ctx);
+
+      /* Restore */
+      DO_IF_DEBUG(state->debug_flags |= THREAD_DEBUG_LOOSE;)
+      SWAP_OUT_THREAD(state);
+      mt_unlock_interpreter();
+    }
+  } else {
+    /* Not a pike thread.  Create a temporary thread_id... */
+    struct object *thread_obj;
+
+    mt_lock_interpreter();
+    init_interpreter();
+    Pike_interpreter.stack_top=((char *)&state)+ (thread_stack_size-16384) * STACK_DIRECTION;
+    Pike_interpreter.recoveries = NULL;
+    thread_obj = fast_clone_object(thread_id_prog);
+    INIT_THREAD_STATE((struct thread_state *)(thread_obj->storage +
+                                              thread_storage_offset));
+    num_threads++;
+    thread_table_insert(Pike_interpreter.thread_state);
+
+    func(ctx);
+
+    cleanup_interpret();        /* Must be done before EXIT_THREAD_STATE */
+    Pike_interpreter.thread_state->status=THREAD_EXITED;
+    co_signal(&Pike_interpreter.thread_state->status_change);
+    thread_table_delete(Pike_interpreter.thread_state);
+    EXIT_THREAD_STATE(Pike_interpreter.thread_state);
+    Pike_interpreter.thread_state=NULL;
+    free_object(thread_obj);
+    thread_obj = NULL;
+    num_threads--;
+    mt_unlock_interpreter();
+  }
+}
+
+PMOD_EXPORT void enable_external_threads(void)
+{
+  num_threads++;  
+}
+
+PMOD_EXPORT void disable_external_threads(void)
+{
+  num_threads--;
+}
+
 /*! @module Thread
  */
 
@@ -667,7 +815,7 @@ PMOD_EXPORT void f_all_threads(INT32 args)
 }
 
 #ifdef PIKE_DEBUG
-void debug_list_all_threads(void)
+PMOD_EXPORT void debug_list_all_threads(void)
 {
   INT32 x;
   struct thread_state *s;
@@ -710,15 +858,26 @@ PMOD_EXPORT int count_pike_threads(void)
   return num_pike_threads;
 }
 
+/* #define PROFILE_CHECK_THREADS */
+
 static void check_threads(struct callback *cb, void *arg, void * arg2)
 {
+#ifdef PROFILE_CHECK_THREADS
+  static unsigned long calls = 0;
+  calls++;
+#endif
+
 #ifndef HAVE_NO_YIELD
   /* If we have no yield we can't cut calls here since it's possible
    * that a thread switch will take place only occasionally in the
    * window below. */
-  static int div_;
-  if(div_++ & 255)
-    return;
+
+  /* Could consider using get_cpu_time here to get more accurate
+   * measurements on more platforms. But otoh we don't really need
+   * very high accuracy - the fallback to clock(3) should be good
+   * enough in most cases. So let's avoid as much overhead as
+   * possible. */
+
 #ifdef HAVE_GETHRTIME
   {
     static hrtime_t last_ = 0;
@@ -733,37 +892,108 @@ static void check_threads(struct callback *cb, void *arg, void * arg2)
     task_thread_times_info_data_t info;
     mach_msg_type_number_t        info_size = TASK_THREAD_TIMES_INFO_COUNT;
     
-    /* Get user time and test if 50usec has passed since last check. */
+    /* Before making an expensive call to task_info() we perform a
+       preliminary check that at least 35 ms real time has passed. If
+       not yet true we'll postpone the next check a full interval. */
+    struct timeval                tv;
+    if (GETTIMEOFDAY(&tv) == 0) {
+#ifdef INT64
+      static INT64 real_time_last_check = 0;
+      INT64 real_time_now = tv.tv_sec * 1000000 + tv.tv_usec;
+      if (real_time_now - real_time_last_check < 35000)
+	return;
+      real_time_last_check = real_time_now;
+#else
+      static struct timeval real_time_last_check = { 0, 0 };
+      struct timeval diff;
+      timersub(&real_time_now, &real_time_last_check, &diff);
+      if (diff.tv_usec < 35000 && diff.tv_sec == 0)
+	return;
+      real_time_last_check = tv;
+#endif
+    }
+    
+    /* Get user time and test if 50 ms has passed since last check. */
     if (task_info(mach_task_self(), TASK_THREAD_TIMES_INFO,
 		  (task_info_t) &info, &info_size) == 0) {
+#ifdef INT64
+      static INT64 last_check = 0;
+      INT64 now =
+	info.user_time.seconds * 1000000 +
+	info.user_time.microseconds +
+	info.system_time.seconds * 1000000 +
+	info.system_time.microseconds;
+      if (now - last_check < 50000)
+	return;
+      last_check = now;
+#else
       /* Compute difference by converting kernel time_info_t to timeval. */
-      struct timeval now;
+      static struct timeval last_check = { 0, 0 };
+      struct timeval user, sys, now;
       struct timeval diff;
-      now.tv_sec = info.user_time.seconds;
-      now.tv_usec = info.user_time.microseconds;
+      user.tv_sec = info.user_time.seconds;
+      user.tv_usec = info.user_time.microseconds;
+      sys.tv_sec = info.system_time.seconds;
+      sys.tv_usec = info.system_time.microseconds;
+      timeradd (&user, &sys, &now);
       timersub(&now, &last_check, &diff);
       if (diff.tv_usec < 50000 && diff.tv_sec == 0)
 	return;
       last_check = now;
+#endif
     }
   }
 #elif defined (USE_CLOCK_FOR_SLICES)
+#ifdef PIKE_DEBUG
+  if (last_clocked_thread != th_self())
+    Pike_fatal ("Stale thread %08lx in last_clocked_thread (self is %08lx)\n",
+		(unsigned long) last_clocked_thread, (unsigned long) th_self());
+#endif
   if (clock() - thread_start_clock < (clock_t) (CLOCKS_PER_SEC / 20))
     return;
+#else
+  static int div_;
+  if(div_++ & 255)
+    return;
 #endif
 #endif
 
+#ifdef PROFILE_CHECK_THREADS
+  {
+    static long last_time;
+    struct timeval now;
+    gettimeofday (&now, NULL);
+    if (now.tv_sec > last_time) {
+      fprintf (stderr, "check_threads calls in %ld sec(s): %lu\n",
+	       (long) (now.tv_sec - last_time), calls);
+      last_time = (unsigned long) now.tv_sec;
+      calls = 0;
+    }
+  }
+#endif
+
   DEBUG_CHECK_THREAD();
 
   THREADS_ALLOW();
   /* Allow other threads to run */
+  /* FIXME: Ought to use condition vars or something to get another
+   * thread to run. yield functions are notoriously unreliable and
+   * poorly defined. It might not really yield we need it to. It might
+   * make us yield to another process instead of just another thread.
+   * It might even make us sleep for a short while. */
   th_yield();
   THREADS_DISALLOW();
 
 #ifdef USE_CLOCK_FOR_SLICES
-  /* Must set the base time for the slice here since clock() returns
-   * thread local time. */
+  /* If we didn't yield then give ourselves a new time slice. If we
+   * did yield then thread_start_clock is the current clock anyway
+   * after the thread swap in. */
   thread_start_clock = clock();
+#ifdef PIKE_DEBUG
+  if (last_clocked_thread != th_self())
+    Pike_fatal ("Stale thread %08lx in last_clocked_thread (self is %08lx)\n",
+		(unsigned long) last_clocked_thread, (unsigned long) th_self());
+#endif
 #endif
 
   DEBUG_CHECK_THREAD();
@@ -785,7 +1015,6 @@ TH_RETURN_TYPE new_thread_func(void *data)
   struct object *thread_obj;
   struct thread_state *thread_state;
   JMP_BUF back;
-  INT32 tmp;
 
   THREADS_FPRINTF(0, (stderr,"new_thread_func(): Thread %p created...\n",
 		      arg.thread_state));
@@ -817,8 +1046,11 @@ TH_RETURN_TYPE new_thread_func(void *data)
 #endif
   }
 #endif /* HAVE_BROKEN_LINUX_THREAD_EUID */
-  
-  mt_lock_interpreter();
+
+  /* Lock the interpreter now, but don't wait on
+   * threads_disabled_change since the spawning thread might be
+   * holding it. */
+  low_mt_lock_interpreter();
 
 #if defined(PIKE_DEBUG)
   if(d_flag) {
@@ -831,7 +1063,6 @@ TH_RETURN_TYPE new_thread_func(void *data)
 
   arg.thread_state->swapped = 0;
   Pike_interpreter = arg.thread_state->state;
-  init_interpreter();
 
 #ifdef PROFILING
   Pike_interpreter.stack_bottom=((char *)&data);
@@ -845,6 +1076,14 @@ TH_RETURN_TYPE new_thread_func(void *data)
   /* Inform the spawning thread that we are now running. */
   co_broadcast(&thread_state->status_change);
 
+  /* After signalling the status change to the spawning thread we may
+   * now wait if threads are disabled. */
+  if (threads_disabled) {
+    SWAP_OUT_CURRENT_THREAD();
+    threads_disabled_wait();
+    SWAP_IN_CURRENT_THREAD();
+  }
+
   DEBUG_CHECK_THREAD();
 
   Pike_interpreter.trace_level = default_t_flag;
@@ -854,12 +1093,8 @@ TH_RETURN_TYPE new_thread_func(void *data)
 
   if(SETJMP(back))
   {
-    if(throw_severity < THROW_EXIT)
+    if(throw_severity <= THROW_ERROR)
       call_handle_error();
-    if(throw_severity == THROW_EXIT)
-    {
-      pike_do_exit(throw_value.u.integer);
-    }
   } else {
     INT32 args=arg.args->size;
     back.severity=THROW_EXIT;
@@ -867,9 +1102,13 @@ TH_RETURN_TYPE new_thread_func(void *data)
     arg.args=0;
     f_call_function(args);
 
-    /* copy return value to the arg.thread_state here */
-    assign_svalue(&arg.thread_state->result, Pike_sp-1);
+    /* Copy return value to the thread_state here, if the thread
+     * object hasn't been destructed. */
+    if (thread_state->thread_obj)
+      assign_svalue(&thread_state->result, Pike_sp-1);
     pop_stack();
+
+    throw_severity = THROW_N_A;
   }
 
   UNSETJMP(back);
@@ -889,7 +1128,16 @@ TH_RETURN_TYPE new_thread_func(void *data)
 
   /* This thread is now officially dead. */
 
-  cleanup_interpret();
+  while(Pike_fp)
+    POP_PIKE_FRAME();
+
+  reset_evaluator();
+
+  low_cleanup_interpret(&thread_state->state);
+
+  if (!thread_state->thread_obj)
+    /* Do this only if exit_thread_obj already has run. */
+    cleanup_thread_state (thread_state);
 
   thread_table_delete(thread_state);
   EXIT_THREAD_STATE(thread_state);
@@ -914,6 +1162,11 @@ TH_RETURN_TYPE new_thread_func(void *data)
   debug_print_rusage (stderr);
 #endif
 
+  if (throw_severity == THROW_EXIT)
+    /* Do this after all thread cleanup to avoid false alarms if using
+     * DO_PIKE_CLEANUP. */
+    pike_do_exit (throw_value.u.integer);
+
   /* FIXME: What about threads_disable? */
   mt_unlock_interpreter();
   th_exit(0);
@@ -955,7 +1208,6 @@ void f_thread_create(INT32 args)
   struct thread_starter arg;
   struct thread_state *thread_state =
     (struct thread_state *)Pike_fp->current_storage;
-  ONERROR err;
   int tmp;
 
   if (thread_state->status != THREAD_NOT_STARTED) {
@@ -966,6 +1218,11 @@ void f_thread_create(INT32 args)
   arg.args = aggregate_array(args);
   arg.thread_state = thread_state;
 
+  if (low_init_interpreter(&thread_state->state)) {
+    free_array(arg.args);
+    Pike_error("Out of memory allocating stack.\n");
+  }
+
 #ifdef HAVE_BROKEN_LINUX_THREAD_EUID
   arg.euid = geteuid();
   arg.egid = getegid();
@@ -1000,18 +1257,19 @@ void f_thread_create(INT32 args)
      */
     SWAP_OUT_CURRENT_THREAD();
     while (thread_state->status == THREAD_NOT_STARTED) {
-      THREADS_FPRINTF(0, (stderr, "THREAD_CREATE %p waiting...\n",
+      THREADS_FPRINTF(0, (stderr, "f_thread_create %p waiting...\n",
 			  thread_state));
-      low_co_wait_interpreter(&thread_state->status_change);
+      _do_co_wait_interpreter(&thread_state->status_change);
     }
+    THREADS_FPRINTF(0, (stderr, "f_thread_create %p continue\n", thread_state));
     SWAP_IN_CURRENT_THREAD();
   } else {
+    low_cleanup_interpret(&thread_state->state);
     free_array(arg.args);
     Pike_error("Failed to create thread (errno = %d).\n", tmp);
   }
 
-  THREADS_FPRINTF(0, (stderr, "THREAD_CREATE -> t:%08x\n",
-		      (unsigned int)thread_state));
+  THREADS_FPRINTF(0, (stderr, "f_thread_create %p done\n", thread_state));
   push_int(0);
 }
 
@@ -1419,7 +1677,7 @@ void exit_mutex_obj_compat_7_4(struct object *o)
 #define THIS_KEY ((struct key_storage *)(CURRENT_STORAGE))
 void init_mutex_key_obj(struct object *o)
 {
-  THREADS_FPRINTF(1, (stderr, "KEY k:%p, o:%p\n",
+  THREADS_FPRINTF(1, (stderr, "KEY k:%p, t:%p\n",
 		      THIS_KEY, Pike_interpreter.thread_state));
   THIS_KEY->mut=0;
   THIS_KEY->mutex_obj = NULL;
@@ -1470,8 +1728,6 @@ void exit_mutex_key_obj(struct object *o)
 /*! @endclass
  */
 
-#define THIS_COND ((COND_T *)(CURRENT_STORAGE))
-
 /*! @class Condition
  *!
  *! Implementation of condition variables.
@@ -1488,15 +1744,37 @@ void exit_mutex_key_obj(struct object *o)
  *!   @[Mutex]
  */
 
+struct pike_cond {
+  COND_T cond;
+  int wait_count;
+};
+
+#define THIS_COND ((struct pike_cond *)(CURRENT_STORAGE))
+
 /*! @decl void wait(Thread.MutexKey mutex_key)
+ *! @decl void wait(Thread.MutexKey mutex_key, int(0..)|float seconds)
+ *! @decl void wait(Thread.MutexKey mutex_key, int(0..) seconds, @
+ *!                 int(0..999999999) nanos)
  *!
- *! Wait for contition.
+ *! Wait for condition.
  *!
  *! This function makes the current thread sleep until the condition
- *! variable is signalled. The argument should be a @[Thread.MutexKey]
- *! object for a @[Thread.Mutex]. It will be unlocked atomically
- *! before waiting for the signal and then relocked atomically when
- *! the signal is received.
+ *! variable is signalled or the timeout is reached.
+ *!
+ *! @param mutex_key
+ *!   A @[Thread.MutexKey] object for a @[Thread.Mutex]. It will be
+ *!   unlocked atomically before waiting for the signal and then
+ *!   relocked atomically when the signal is received or the timeout
+ *!   is reached.
+ *!
+ *! @param seconds
+ *!   Seconds to wait before the timeout is reached.
+ *!
+ *! @param nanos
+ *!   Nano (1/1000000000) seconds to wait before the timeout is reached.
+ *!   This value is added to the number of seconds specified by @[seconds].
+ *!
+ *! A timeout of zero seconds disables the timeout.
  *!
  *! The thread that sends the signal should have the mutex locked
  *! while sending it. Otherwise it's impossible to avoid races where
@@ -1504,11 +1782,23 @@ void exit_mutex_key_obj(struct object *o)
  *! @[wait] calls yet.
  *!
  *! @note
+ *!   The support for timeouts was added in Pike 7.8.121, which was
+ *!   after the first public release of Pike 7.8.
+ *!
+ *! @note
+ *!   Note that the timeout is approximate (best effort), and may
+ *!   be exceeded if eg the mutex is busy after the timeout.
+ *!
+ *! @note
  *!   In Pike 7.2 and earlier it was possible to call @[wait()]
  *!   without arguments. This possibility was removed in later
  *!   versions since it unavoidably leads to programs with races
  *!   and/or deadlocks.
  *!
+ *! @note
+ *!   Note also that any threads waiting on the condition will be
+ *!   woken up when it gets destructed.
+ *!
  *! @seealso
  *!   @[Mutex->lock()]
  */
@@ -1516,12 +1806,20 @@ void f_cond_wait(INT32 args)
 {
   struct object *key, *mutex_obj;
   struct mutex_storage *mut;
-  COND_T *c;
+  struct pike_cond *c;
+  INT_TYPE seconds = 0, nanos = 0;
 
   if(threads_disabled)
     Pike_error("Cannot wait for conditions when threads are disabled!\n");
 
-  get_all_args("condition->wait", args, "%o", &key);
+  if (args <= 2) {
+    FLOAT_TYPE fsecs = 0.0;
+    get_all_args("condition->wait", args, "%o.%F", &key, &fsecs);
+    seconds = (INT_TYPE) fsecs;
+    nanos = (INT_TYPE)((fsecs - seconds)*1000000000);
+  } else {
+    get_all_args("condition->wait", args, "%o%i%i", &key, &seconds, &nanos);
+  }
       
   if ((key->prog != mutex_key) ||
       (!(OB2KEY(key)->initialized)) ||
@@ -1545,7 +1843,13 @@ void f_cond_wait(INT32 args)
     
   /* Wait and allow mutex operations */
   SWAP_OUT_CURRENT_THREAD();
-  co_wait_interpreter(c);
+  c->wait_count++;
+  if (seconds || nanos) {
+    co_wait_interpreter_timeout(&(c->cond), seconds, nanos);
+  } else {
+    co_wait_interpreter(&(c->cond));
+  }
+  c->wait_count--;
   SWAP_IN_CURRENT_THREAD();
     
   /* Lock mutex */
@@ -1584,7 +1888,11 @@ void f_cond_wait(INT32 args)
  *! @seealso
  *!   @[broadcast()]
  */
-void f_cond_signal(INT32 args) { pop_n_elems(args); co_signal(THIS_COND); }
+void f_cond_signal(INT32 args)
+{
+  pop_n_elems(args);
+  co_signal(&(THIS_COND->cond));
+}
 
 /*! @decl void broadcast()
  *!
@@ -1593,10 +1901,42 @@ void f_cond_signal(INT32 args) { pop_n_elems(args); co_signal(THIS_COND); }
  *! @seealso
  *!   @[signal()]
  */
-void f_cond_broadcast(INT32 args) { pop_n_elems(args); co_broadcast(THIS_COND); }
+void f_cond_broadcast(INT32 args)
+{
+  pop_n_elems(args);
+  co_broadcast(&(THIS_COND->cond));
+}
 
-void init_cond_obj(struct object *o) { co_init(THIS_COND); }
-void exit_cond_obj(struct object *o) { co_destroy(THIS_COND); }
+void init_cond_obj(struct object *o)
+{
+  co_init(&(THIS_COND->cond));
+  THIS_COND->wait_count = 0;
+}
+
+void exit_cond_obj(struct object *o)
+{
+  /* Wake up any threads that might be waiting on this cond.
+   *
+   * Note that we are already destructed (o->prog == NULL),
+   * so wait_count can't increase.
+   *
+   * FIXME: This code wouldn't be needed if exit callbacks were called
+   *        only when the ref count reaches zero.
+   *	/grubba 2006-01-29
+   */
+  while (THIS_COND->wait_count) {
+    co_broadcast(&(THIS_COND->cond));
+
+    THREADS_ALLOW();
+#ifdef HAVE_NO_YIELD
+    sleep(0);
+#else /* HAVE_NO_YIELD */
+    th_yield();
+#endif /* HAVE_NO_YIELD */
+    THREADS_DISALLOW();
+  }
+  co_destroy(&(THIS_COND->cond));
+}
 
 /*! @endclass
  */
@@ -1653,7 +1993,7 @@ void f_thread_id_status(INT32 args)
   push_int(THIS_THREAD->status);
 }
 
-/*! @decl static string _sprintf(int c)
+/*! @decl protected string _sprintf(int c)
  *!
  *! Returns a string identifying the thread.
  */
@@ -1673,7 +2013,7 @@ void f_thread_id__sprintf (INT32 args)
   f_add (3);
 }
 
-/*! @decl static int id_number()
+/*! @decl protected int id_number()
  *!
  *! Returns an id number identifying the thread.
  *!
@@ -1699,23 +2039,111 @@ static void f_thread_id_result(INT32 args)
     Pike_error("Cannot wait for threads when threads are disabled!\n");
   }
 
-  THREADS_FPRINTF(0, (stderr,
-		      "Thread->result(): Waiting for thread %p (state:%d).\n",
-		      th, th->status));
+  th->waiting++;
+
+  THREADS_FPRINTF(0, (stderr, "Thread->wait(): Waiting for thread_state %p "
+		      "(state:%d).\n", th, th->status));
   while(th->status != THREAD_EXITED) {
     SWAP_OUT_CURRENT_THREAD();
     co_wait_interpreter(&th->status_change);
     SWAP_IN_CURRENT_THREAD();
     check_threads_etc();
     THREADS_FPRINTF(0,
-		    (stderr,
-		     "Thread->result(): Waiting for thread %p (state:%d).\n",
-		     th, th->status));
+		    (stderr, "Thread->wait(): Waiting for thread_state %p "
+		     "(state:%d).\n", th, th->status));
   }
 
   assign_svalue_no_free(Pike_sp, &th->result);
   Pike_sp++;
   dmalloc_touch_svalue(Pike_sp-1);
+
+  th->waiting--;
+
+  if (!th->thread_obj)
+    /* Do this only if exit_thread_obj already has run. */
+    cleanup_thread_state (th);
+}
+
+static int num_pending_interrupts = 0;
+static struct callback *thread_interrupt_callback = NULL;
+
+static void check_thread_interrupt(struct callback *foo,
+				   void *bar, void *gazonk)
+{
+  if (Pike_interpreter.thread_state->flags & THREAD_FLAG_SIGNAL_MASK) {
+    if (Pike_interpreter.thread_state->flags & THREAD_FLAG_TERM) {
+      throw_severity = THROW_THREAD_EXIT;
+    } else {
+      throw_severity = THROW_ERROR;
+    }
+    Pike_interpreter.thread_state->flags &= ~THREAD_FLAG_SIGNAL_MASK;
+    if (!--num_pending_interrupts) {
+      remove_callback(foo);
+      thread_interrupt_callback = NULL;
+    }
+    if (throw_severity == THROW_ERROR) {
+      Pike_error("Interrupted.\n");
+    } else {
+      push_int(-1);
+      assign_svalue(&throw_value, Pike_sp-1);
+      pike_throw();
+    }
+  }
+}
+
+/*! @decl void interrupt()
+ *! @decl void interrupt(string msg)
+ *!
+ *! Interrupt the thread with the message @[msg].
+ *!
+ *! @fixme
+ *!   The argument @[msg] is currently ignored.
+ *!
+ *! @note
+ *!   Interrupts are asynchronous, and are currently not queued.
+ */
+static void f_thread_id_interrupt(INT32 args)
+{
+  /* FIXME: The msg argument is not supported yet. */
+  pop_n_elems(args);
+
+  if (!(THIS_THREAD->flags & THREAD_FLAG_SIGNAL_MASK)) {
+    THIS_THREAD->flags |= THREAD_FLAG_INTR;
+    num_pending_interrupts++;
+    if (!thread_interrupt_callback) {
+      thread_interrupt_callback =
+	add_to_callback(&evaluator_callbacks, check_thread_interrupt, 0, 0);
+    }
+    /* FIXME: Actually interrupt the thread. */
+  }
+  push_int(0);
+}
+
+static void low_thread_kill (struct thread_state *th)
+{
+  if (!(th->flags & THREAD_FLAG_SIGNAL_MASK)) {
+    num_pending_interrupts++;
+    if (!thread_interrupt_callback) {
+      thread_interrupt_callback =
+	add_to_callback(&evaluator_callbacks, check_thread_interrupt, 0, 0);
+    }
+    /* FIXME: Actually interrupt the thread. */
+  }
+  th->flags |= THREAD_FLAG_TERM;
+}
+
+/*! @decl void kill()
+ *!
+ *! Interrupt the thread, and terminate it.
+ *!
+ *! @note
+ *!   Interrupts are asynchronous, and are currently not queued.
+ */
+static void f_thread_id_kill(INT32 args)
+{
+  pop_n_elems(args);
+  low_thread_kill (THIS_THREAD);
+  push_int(0);
 }
 
 void init_thread_obj(struct object *o)
@@ -1724,26 +2152,51 @@ void init_thread_obj(struct object *o)
   THIS_THREAD->thread_obj = Pike_fp->current_object;
   THIS_THREAD->swapped = 0;
   THIS_THREAD->status=THREAD_NOT_STARTED;
+  THIS_THREAD->flags = 0;
+  THIS_THREAD->waiting = 0;
   THIS_THREAD->result.type = T_INT;
   THIS_THREAD->result.subtype = NUMBER_UNDEFINED;
   THIS_THREAD->result.u.integer = 0;
   co_init(& THIS_THREAD->status_change);
   THIS_THREAD->thread_local=NULL;
-#if CPU_TIME_IS_THREAD_LOCAL == PIKE_YES
+#ifdef CPU_TIME_MIGHT_BE_THREAD_LOCAL
   THIS_THREAD->auto_gc_time = 0;
 #endif
 }
 
+static void cleanup_thread_state (struct thread_state *th)
+{
+#ifdef PIKE_DEBUG
+  if (th->thread_obj) Pike_fatal ("Thread state still active.\n");
+#endif
+
+  /* Don't clean up if the thread is running or if someone is waiting
+   * on status_change. They should call cleanup_thread_state later. */
+  if (th->status == THREAD_RUNNING || th->waiting)
+    return;
+
+  if (THIS_THREAD->flags & THREAD_FLAG_SIGNAL_MASK) {
+    Pike_interpreter.thread_state->flags &= ~THREAD_FLAG_SIGNAL_MASK;
+    if (!--num_pending_interrupts) {
+      remove_callback(thread_interrupt_callback);
+      thread_interrupt_callback = NULL;
+    }    
+  }
+
+  co_destroy(& THIS_THREAD->status_change);
+  th_destroy(& THIS_THREAD->id);
+}
 
 void exit_thread_obj(struct object *o)
 {
+  THIS_THREAD->thread_obj = NULL;
+
+  cleanup_thread_state (THIS_THREAD);
+
   if(THIS_THREAD->thread_local != NULL) {
     free_mapping(THIS_THREAD->thread_local);
     THIS_THREAD->thread_local = NULL;
   }
-  co_destroy(& THIS_THREAD->status_change);
-  th_destroy(& THIS_THREAD->id);
-  THIS_THREAD->thread_obj = NULL;
 }
 
 /*! @endclass
@@ -1875,7 +2328,6 @@ void gc_check_thread_local (struct object *o)
   /* Only used by with locate_references. */
   if (Pike_in_gc == GC_PASS_LOCATE) {
     struct svalue key, *val;
-    INT32 x;
     struct thread_state *s;
 
     key.u.integer = ((struct thread_local *)CURRENT_STORAGE)->id;
@@ -2072,7 +2524,7 @@ void th_init(void)
 		    mutex_key_offset + OFFSETOF(key_storage, owner_obj),
 		    tObjIs_THREAD_ID, T_OBJECT, 0);
   PIKE_MAP_VARIABLE("_mutex", mutex_key_offset + OFFSETOF(key_storage, mutex_obj),
-		    tObjIs_THREAD_MUTEX, T_OBJECT, ID_STATIC|ID_PRIVATE);
+		    tObjIs_THREAD_MUTEX, T_OBJECT, ID_PROTECTED|ID_PRIVATE);
   set_init_callback(init_mutex_key_obj);
   set_exit_callback(exit_mutex_key_obj);
   mutex_key=Pike_compiler->new_program;
@@ -2109,9 +2561,11 @@ void th_init(void)
   end_class("mutex_compat_7_4", 0);
 
   START_NEW_PROGRAM_ID(THREAD_CONDITION);
-  ADD_STORAGE(COND_T);
+  ADD_STORAGE(struct pike_cond);
   ADD_FUNCTION("wait",f_cond_wait,
-	       tFunc(tObjIs_THREAD_MUTEX_KEY,tVoid),0);
+	       tOr(tFunc(tObjIs_THREAD_MUTEX_KEY tOr3(tVoid, tIntPos, tFloat),
+			 tVoid),
+		   tFunc(tObjIs_THREAD_MUTEX_KEY tIntPos tIntPos, tVoid)),0);
   ADD_FUNCTION("signal",f_cond_signal,tFunc(tNone,tVoid),0);
   ADD_FUNCTION("broadcast",f_cond_broadcast,tFunc(tNone,tVoid),0);
   set_init_callback(init_cond_obj);
@@ -2136,7 +2590,7 @@ void th_init(void)
   ADD_FUNCTION("get",f_thread_local_get,tFunc(tNone,tMix),0);
   ADD_FUNCTION("set",f_thread_local_set,tFunc(tSetvar(1,tMix),tVar(1)),0);
   ADD_FUNCTION("create", f_thread_local_create,
-	       tFunc(tVoid,tVoid), ID_STATIC);
+	       tFunc(tVoid,tVoid), ID_PROTECTED);
 #ifdef PIKE_DEBUG
   set_gc_check_callback(gc_check_thread_local);
 #endif
@@ -2153,12 +2607,15 @@ void th_init(void)
 		    tMix, T_MIXED, 0);
   ADD_FUNCTION("create",f_thread_create,
 	       tFuncV(tNone,tMixed,tVoid),
-	       ID_STATIC);
+	       ID_PROTECTED);
   ADD_FUNCTION("backtrace",f_thread_backtrace,tFunc(tNone,tArray),0);
   ADD_FUNCTION("wait",f_thread_id_result,tFunc(tNone,tMix),0);
   ADD_FUNCTION("status",f_thread_id_status,tFunc(tNone,tInt),0);
   ADD_FUNCTION("_sprintf",f_thread_id__sprintf,tFunc(tNone,tStr),0);
   ADD_FUNCTION("id_number",f_thread_id_id_number,tFunc(tNone,tInt),0);
+  ADD_FUNCTION("interrupt", f_thread_id_interrupt,
+	       tFunc(tOr(tVoid,tStr), tVoid), 0);
+  ADD_FUNCTION("kill", f_thread_id_kill, tFunc(tNone, tVoid), 0);
   set_gc_recurse_callback(thread_was_recursed);
   set_gc_check_callback(thread_was_checked);
   set_init_callback(init_thread_obj);
@@ -2190,6 +2647,47 @@ void th_init(void)
   thread_table_insert(Pike_interpreter.thread_state);
 }
 
+#ifdef DO_PIKE_CLEANUP
+void cleanup_all_other_threads (void)
+{
+  int i, num_kills = num_pending_interrupts;
+  time_t timeout = time (NULL) + 2;
+
+  mt_lock (&thread_table_lock);
+  for (i = 0; i < THREAD_TABLE_SIZE; i++) {
+    struct thread_state *th;
+    for (th = thread_table_chains[i]; th; th = th->hashlink)
+      if (th != Pike_interpreter.thread_state) {
+	low_thread_kill (th);
+	num_kills++;
+      }
+  }
+  mt_unlock (&thread_table_lock);
+
+  while (num_pending_interrupts && time (NULL) < timeout) {
+    THREADS_ALLOW();
+#ifdef __NT__
+    Sleep (1);
+#elif defined (HAVE_USLEEP)
+    usleep (1000);
+#else
+    sleep (1);
+#endif
+    THREADS_DISALLOW();
+  }
+
+#if 0
+  if (num_kills) {
+    fprintf (stderr, "Killed %d thread(s) in exit cleanup",
+	     num_kills - num_pending_interrupts);
+    if (num_pending_interrupts)
+      fprintf (stderr, ", %d more haven't responded", num_pending_interrupts);
+    fputs (".\n", stderr);
+  }
+#endif
+}
+#endif
+
 void th_cleanup(void)
 {
   th_running = 0;
